/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package examples.invoice.applications;

import org.apache.tools.ant.types.DTDLocation;
import org.objectweb.jorm.api.JormConfigurator;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.lib.JormConfiguratorImpl;
import org.objectweb.jorm.lib.MapperJCA;
import org.objectweb.jorm.mapper.rdb.lib.ConnectionSpecJDBC;
import org.objectweb.jorm.mapper.rdb.lib.MapperJDBC;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.jorm.util.io.api.PathExplorer;
import org.objectweb.jorm.util.io.lib.DirJavaExplorer;
import org.objectweb.perseus.connector.ra.fos.FosAttributeControler;
import org.objectweb.perseus.connector.ra.fos.FosManagedConnectionFactory;
import org.objectweb.perseus.cache.api.CacheManager;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Properties;

/**
 * @author	P. Dechamboux
 */
public class JormInit {
	// JormInit related constants and variables
	public static final String LOGGER_FACTORY_CLASS_NAME
			= "loggerFactoryClassName";
	private static JormInit singleton = new JormInit();
	private static CacheManager cacheManager = new TrivialCache();
	protected LoggerFactory loggerFactory = null;
	protected Logger logger = null;
	// Mapper related constants and variables
	private HashMap mapperMap = new HashMap();
	private HashMap urlMap = new HashMap();
	// RA related constants and variables
	public static final String RESOURCE_JDBC_SUBMAPPER
			= "jdbc.submapper";
	public static final String RESOURCE_FOS_MCF_CLASSNAME
			= "fos.resource.mcf.classname";
	public static final String DEFAULT_FOS_URL = "fos.ConnectionURL";
	private static final ArrayList dtdLocs = new ArrayList();
	private static final PathExplorer pdLocs = new DirJavaExplorer();
	private static final String jormConf = "../../etc/jorm.properties";
	static {
		DTDLocation dtdl = new DTDLocation();
    	dtdl.setPublicId("-//ObjectWeb Consortium//DTD JORM DEFINITIONS 2.0//EN");
    	dtdl.setLocation("../../etc/jorm2.dtd");
    	dtdLocs.add(dtdl);
    	pdLocs.addPath("src");
	}

	/**
	 * Retrieves the JormInit singleton.
	 */
	public static JormInit getInstance() {
		return singleton;
	}

	/**
	 * Retrieves the LoggerFactory
	 */
	public LoggerFactory getLoggerFactory() {
		return loggerFactory;
	}

	/**
	 * Initializes the log system.
	 */
	public void initLogSystem(Properties prop) throws Exception {
		loggerFactory = Monolog.initialize();
		logger = loggerFactory.getLogger("examples.invoice.applications.JormInit");
		logger.log(BasicLevel.INFO, "Log system initialized.");
	}

	/**
	 * Initializes the RA environment
	 */
	public void initResourceAdapter(Properties prop) throws Exception {
		initJDBCResourceAdapter(prop);
		initFOSResourceAdapter(prop);
	}

	private void initJDBCResourceAdapter(Properties prop) throws Exception {
		// Initializes JDBC environment
		JormConfigurator jcc = new JormConfiguratorImpl();
		jcc.configure(jormConf);
		jcc.configureLog(jormConf);
		PMapper mapper = new MapperJDBC(jcc);
        mapper.setDTDLocations(dtdLocs);
        mapper.setPathsToPdFiles(pdLocs);
		mapper.setMapperName("rdb." + prop.getProperty(RESOURCE_JDBC_SUBMAPPER));
		mapper.setConnectionFactory(
				new ConnectionSpecJDBC(
						prop.getProperty("ConnectionURL"),
						prop.getProperty("DriverClassName"),
						prop.getProperty("UserName"),
						prop.getProperty("Password")));
		((Loggable) mapper).setLoggerFactory(loggerFactory);
		mapper.start();
		mapperMap.put(mapper.getMapperName(), mapper);
		urlMap.put(mapper.getMapperName(), prop.getProperty("ConnectionURL"));
		logger.log(BasicLevel.INFO, "Mapper " + mapper.getMapperName()
									+ " instancied and initialised.");
	}

	private void initFOSResourceAdapter(Properties prop) throws Exception {
		PMapper mapper;
		// Initializes FOS environment
		String fosMcfCN = prop.getProperty(RESOURCE_FOS_MCF_CLASSNAME);
		if (fosMcfCN == null) {
			logger.log(BasicLevel.INFO,
					   "Cannot initialize FOS RA environment: MCF=" + fosMcfCN);
			return;
		}
		FosAttributeControler ra
				= (FosAttributeControler) Class.forName(fosMcfCN).newInstance();
		logger.log(BasicLevel.DEBUG, "FOS Ra=" + ra);
		// configure it
		ra.setConnectionURL(prop.getProperty(DEFAULT_FOS_URL));
		ra.setLoggerFactory(loggerFactory);
		ra.setInitializeAtStartUp(false);
		ra.setLoggerBaseName("examples.invoice.ra");
		((FosManagedConnectionFactory) ra).start();
		JormConfigurator jcc = new JormConfiguratorImpl();
		jcc.configure(jormConf);
		jcc.configureLog(jormConf);
		mapper = new MapperJCA(jcc);
        mapper.setDTDLocations(dtdLocs);
        mapper.setPathsToPdFiles(pdLocs);
		mapper.setMapperName("fos");
		mapper.setConnectionFactory(((FosManagedConnectionFactory) ra).createConnectionFactory());
		((Loggable) mapper).setLoggerFactory(loggerFactory);
		mapper.start();
		mapperMap.put(mapper.getMapperName(), mapper);
		urlMap.put(mapper.getMapperName(), prop.getProperty(DEFAULT_FOS_URL));
		logger.log(BasicLevel.INFO, "Mapper " + mapper.getMapperName()
									+ " instancied and initialised.");
	}

	public PMapper getMapper(String name) throws PException {
		return (PMapper) mapperMap.get(name);
	}

	public Iterator registeredMapper() {
		return mapperMap.values().iterator();
	}

	public String getURL(PMapper m) {
		return (String) urlMap.get(m.getMapperName());
	}

	public CacheManager getCacheManager() {
		return cacheManager;
	}
}
