/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package examples.invoice.persistclass;

import org.objectweb.jorm.naming.api.PBinder;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PMapCluster;
import org.objectweb.jorm.api.PBinding;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import java.util.HashMap;
import java.util.Iterator;

import examples.invoice.applications.JormInit;

/**
 * @author	P. Dechamboux
 */
public class AddressHelper implements MapHelper {
	public final static String CN = "examples.invoice.persistclass.Address";
	private LoggerFactory loggerFactory;
	private HashMap pobjMap = new HashMap();

	/**
	 * Prepares the mapping of the Address persistent class for the
	 * AddressAppli or other applications.
	 */
	public void map(PMapper m, LoggerFactory lf, byte clact)
		throws PException {
		loggerFactory = lf;
		Logger logger = loggerFactory.getLogger(CN);
		// LongGenIncr initialization
		PClassMapping pcm = m.lookup(CN);
		// Perform mapping only if it has not been already done
		if (pcm == null) {
			PBinder binder = new AddressIdBinder();
            String cn = m.cn2mn(CN) + PMapper.PCLASSMAPPINGAPPENDER;
			try {
				pcm = (PClassMapping) Class.forName(cn).newInstance();
			} catch (Exception e) {
				throw new PException("Cannot create PClassMapping (probably a ClassLoader problem): " + cn);
			}
			pcm.setPBinder(binder);
			binder.setPClassMapping(pcm);
            binder.setCacheManager(JormInit.getInstance().getCacheManager());
			m.map(pcm);
		}
		PMapCluster cl = m.getPMapCluster(CN);
		switch (clact) {
		case PClassMapping.CREATE_STRUCTURE_IF_NEEDED:
			cl.createMappingStructures(false);
			break;
		case PClassMapping.CLEANUP_REMOVEALL:
			cl.deleteMappingStructures();
			cl.createMappingStructures(false);
			break;
		case PClassMapping.CLEANUP_REMOVEDATA:
			cl.deleteData();
			break;
		case PClassMapping.CLEANUP_DONOTHING:
			break;
		}
	}

	/**
	 * Tries to retrieve an Address knowing its name.
	 */
	public Address getAddress(PMapper m, PName pn) throws PException {
		Address res;
		if (pn == null) {
			System.out.println("WARNING: PName is null in getAddress within AddressHelper!!");
			return null;
		}
		PClassMapping pcm = m.lookup(CN);
		PBinding pb = pcm.getPBinder().lookup(pn);
		if (pb == null) {
			Object conn = m.getConnection();
			try {
				pb = pcm.createPBinding();
				pb.bind(pn);
				res = new Address(conn, pb);
				pobjMap.put(pb, res);
				m.closeConnection(conn);
			} catch (PException pe) {
				m.closeConnection(conn);
				throw pe;
			}
		} else {
			res = (Address) pobjMap.get(pb);
		}
		return res;
	}

	/**
	 * Creates a new persistent Address.
	 */
	public Address createAddress(PMapper m, String n, String s, String z,
								 String t, String c) throws PException {
		Address res;
		PBinding pb = m.lookup(CN).createPBinding();
		Object conn = m.getConnection();
		try {
			res = new Address(conn, pb, n, s, z, t, c);
			m.closeConnection(conn);
			pobjMap.put(pb, res);
		} catch (PException pe) {
			m.closeConnection(conn);
			throw pe;
		}
		return res;
	}

	/**
	 * Deletes a persistent Address.
	 */
	public boolean deleteAddress(PMapper m, String id) throws PException {
		PName pn = m.lookup(CN).getPBinder().decodeString(id);
		Address ad = getAddress(m, pn);
		if (ad == null)
			return false;
		Object conn = m.getConnection();
		try {
			PBinding pb = ad.getPBinding();
			ad.delete(conn);
			m.closeConnection(conn);
			pobjMap.remove(pb);
		} catch (PException pe) {
			m.closeConnection(conn);
			throw pe;
		}
		return true;
	}

	/**
	 * Lists all existing Addresses.
	 */
	public Iterator listAddress(PMapper m) throws Exception {
		Iterator res;
		Object conn = m.getConnection();
		try {
			res = m.lookup(CN).getPNameIterator(conn, true, false, null);
			m.closeConnection(conn);
		} catch (PException pe) {
			m.closeConnection(conn);
			throw pe;
		}
		return res;
	}
}
