/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package examples.invoice.persistclass;

import org.objectweb.jorm.api.*;
import org.objectweb.jorm.facility.naming.basidir.BasidBinder;
import org.objectweb.jorm.facility.naming.polymorphid.PolymorphIdMgr;
import org.objectweb.jorm.facility.naming.polymorphid.PolymorphIdMgrRegistry;
import org.objectweb.jorm.facility.naming.polymorphid.PolymorphIdMgrImpl;
import org.objectweb.jorm.naming.api.PBinder;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import java.util.HashMap;
import java.util.Iterator;

import examples.invoice.applications.JormInit;

/**
 * @author	P. Dechamboux
 */
public class ProdUnitsHelper implements MapHelper {
	public final static String CN = "examples.invoice.persistclass.ProdUnits";
	private LoggerFactory loggerFactory;
	private HashMap pobjMap = new HashMap();
	private ProductHelper productHelper;

	public ProdUnitsHelper(ProductHelper ph) {
		productHelper = ph;
	}

	/**
	 * Prepares the mapping of the ProdUnits persistent class for the
	 * ProdUnitsAppli or other applications.
	 */
	public void map(PMapper m, LoggerFactory lf, byte clact)
		throws PException {
		productHelper.map(m, lf, clact);
		loggerFactory = lf;
		Logger logger = loggerFactory.getLogger(CN);
		// LongGenIncr initialization
		PClassMapping pcm = m.lookup(CN);
		// Perform mapping only if it has not been already done
		if (pcm == null) {
			PBinder binder = new BasidBinder(m, CN, clact);
			try {
				pcm = (PClassMapping) Class.forName(m.cn2mn(CN) + PMapper.PCLASSMAPPINGAPPENDER).newInstance();
			} catch (Exception e) {
				throw new PException("Cannot create PClassMapping (probably a ClassLoader problem).");
			}
			pcm.setPBinder(binder);
			binder.setPClassMapping(pcm);
            binder.setCacheManager(JormInit.getInstance().getCacheManager());
			PolymorphIdMgr pim = PolymorphIdMgrRegistry.getPolymorphIdMgr(m);
			if (pim == null) {
				pim = new PolymorphIdMgrImpl();
				pim.init(m, PClassMapping.CREATE_STRUCTURE_IF_NEEDED);
				PolymorphIdMgrRegistry.registerPolymorphIdMgr(pim);
			}
			((PClassMappingCtrl) pcm).setPNameCoder("product",
								  pim.getRefNC(ProductHelper.CN));
			m.map(pcm);
		}
		PMapCluster cl = m.getPMapCluster(CN);
		switch (clact) {
		case PClassMapping.CREATE_STRUCTURE_IF_NEEDED:
			cl.createMappingStructures(false);
			break;
		case PClassMapping.CLEANUP_REMOVEALL:
			cl.deleteMappingStructures();
			cl.createMappingStructures(false);
			break;
		case PClassMapping.CLEANUP_REMOVEDATA:
			cl.deleteData();
			break;
		case PClassMapping.CLEANUP_DONOTHING:
			break;
		}
	}

	public ProductHelper getProductHelper() {
		if (productHelper == null)
			System.out.println("WARNING: productHelper is NULL in invoiceHelper!!");
		return productHelper;
	}

	/**
	 * Tries to retrieve an ProdUnits knowing its name.
	 */
	public ProdUnits getProdUnits(PMapper m, PName pn) throws Exception {
		ProdUnits res;
		if (pn == null) {
			System.out.println("WARNING: PName is null in getProdUnits within ProdUnitsHelper!!");
			return null;
		}
		PClassMapping pcm = m.lookup(CN);
		PBinding pb = pcm.getPBinder().lookup(pn);
		if (pb == null) {
			Object conn = m.getConnection();
			try {
				pb = pcm.createPBinding();
				pb.bind(pn);
				res = new ProdUnits(this, conn, pb);
				m.closeConnection(conn);
				pobjMap.put(pb, res);
			} catch (PException pe) {
				m.closeConnection(conn);
				throw pe;
			}
		} else {
			res = (ProdUnits) pobjMap.get(pb);
		}
		return res;
	}

	/**
	 * Creates a new persistent ProdUnits.
	 */
	public ProdUnits createProdUnits(PMapper m, String pr, int quant) throws PException {
		ProdUnits res;
		PBinding pb = m.lookup(CN).createPBinding();
		Object conn = m.getConnection();
		try {
			PName pn = m.lookup(ProductHelper.CN).getPBinder().decodeString(pr);
			res = new ProdUnits(this, conn, pb, pn, quant);
			m.closeConnection(conn);
			pobjMap.put(pb, res);
		} catch (PException pe) {
			m.closeConnection(conn);
			throw pe;
		}
		return res;
	}

	/**
	 * Deletes a persistent ProdUnits.
	 */
	public boolean deleteProdUnits(PMapper m, long id) throws Exception {
		PName pn = m.lookup(CN).getPBinder().decodeLong(id);
		ProdUnits ad = getProdUnits(m, pn);
		if (ad == null)
			return false;
		Object conn = m.getConnection();
		try {
			PBinding pb = ad.getPBinding();
			ad.delete(conn);
			m.closeConnection(conn);
			pobjMap.remove(pb);
		} catch (PException pe) {
			m.closeConnection(conn);
			throw pe;
		}
		return true;
	}

	/**
	 * Lists all existing ProdUnitses.
	 */
	public Iterator listProdUnits(PMapper m) throws Exception {
		Iterator res;
		Object conn = m.getConnection();
		try {
			res = m.lookup(CN).getPNameIterator(conn, true, false, null);
			m.closeConnection(conn);
		} catch (PException pe) {
			m.closeConnection(conn);
			throw pe;
		}
		return res;
	}
}
