/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.api;

import org.objectweb.jorm.type.api.PExceptionTyping;
import org.objectweb.jorm.naming.api.PName;

import java.io.Serializable;
import java.util.Date;
import java.math.BigDecimal;
import java.math.BigInteger;

/**
 * The <b>PAccessorGen</b> interface defines a generic accessor to be used by
 * persistent bindings to access persistent variable within memory instances.
 * This interface is implemented for each JORM class in order to define
 * specific accessors associated to such a memory instance.
 * @author	R. Basset, P. Dechamboux
 */
public interface PAccessorGen extends PAccessor {
	/**
	 * It retrieves the value of a boolean field.
	 * @param fn	A String containing the name of the field.
	 * @return	The boolean value of the field as stored within the memory
	 * instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	boolean paGetBooleanField(String fn) throws PException;

	/**
	 * It retrieves the value of a Boolean field.
	 * @param fn	A String containing the name of the field.
	 * @return	The Boolean value of the field as stored within the memory
	 * instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	Boolean paGetObooleanField(String fn) throws PException;

	/**
	 * It retrieves the value of a byte field.
	 * @param fn	A String containing the name of the field.
	 * @return	The byte value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	byte paGetByteField(String fn) throws PException;

	/**
	 * It retrieves the value of a Byte field.
	 * @param fn	A String containing the name of the field.
	 * @return	The Byte value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	Byte paGetObyteField(String fn) throws PException;

	/**
	 * It retrieves the value of a char field.
	 * @param fn	A String containing the name of the field.
	 * @return	The char value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	char paGetCharField(String fn) throws PException;

	/**
	 * It retrieves the value of a Character field.
	 * @param fn	A String containing the name of the field.
	 * @return	The Character value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	Character paGetOcharField(String fn) throws PException;

	/**
	 * It retrieves the value of a short field.
	 * @param fn	A String containing the name of the field.
	 * @return	The short value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	short paGetShortField(String fn) throws PException;

	/**
	 * It retrieves the value of a Short field.
	 * @param fn	A String containing the name of the field.
	 * @return	The Short value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	Short paGetOshortField(String fn) throws PException;

	/**
	 * It retrieves the value of an int field.
	 * @param fn	A String containing the name of the field.
	 * @return	The int value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	int paGetIntField(String fn) throws PException;

	/**
	 * It retrieves the value of an Integer field.
	 * @param fn	A String containing the name of the field.
	 * @return	The Integer value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	Integer paGetOintField(String fn) throws PException;

	/**
	 * It retrieves the value of a long field.
	 * @param fn	A String containing the name of the field.
	 * @return	The long value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	long paGetLongField(String fn) throws PException;

	/**
	 * It retrieves the value of a Long field.
	 * @param fn	A String containing the name of the field.
	 * @return	The Long value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	Long paGetOlongField(String fn) throws PException;

	/**
	 * It retrieves the value of a float field.
	 * @param fn	A String containing the name of the field.
	 * @return	The float value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	float paGetFloatField(String fn) throws PException;

	/**
	 * It retrieves the value of a Float field.
	 * @param fn	A String containing the name of the field.
	 * @return	The Float value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	Float paGetOfloatField(String fn) throws PException;

	/**
	 * It retrieves the value of a double field.
	 * @param fn	A String containing the name of the field.
	 * @return	The double value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	double paGetDoubleField(String fn) throws PException;

	/**
	 * It retrieves the value of a Double field.
	 * @param fn	A String containing the name of the field.
	 * @return	The Double value of the field as stored within the memory
	 * 			instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	Double paGetOdoubleField(String fn) throws PException;

	/**
	 * It retrieves the value of a string field.
	 * @param fn	A String containing the name of the field.
	 * @return	The java.lang.String value of the field as stored within the
	 * 			memory instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	String paGetStringField(String fn) throws PException;

	/**
	 * It retrieves the value of a date field.
	 * @param fn	A String containing the name of the field.
	 * @return	The java.sql.Date value of the field as stored within the
	 * 			memory instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	Date paGetDateField(String fn) throws PException;

	/**
	 * It retrieves the value of a serialized field.
	 * @param fn	A String containing the name of the field.
	 * @return	The java.io.Serializable value of the field as stored within
	 * 			the memory instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	Serializable paGetSerializedField(String fn) throws PException;

	/**
	 * It retrieves the value of a byte[] field.
	 * @param fn	A String containing the name of the field.
	 * @return	The byte[] value of the field as stored within
	 * 			the memory instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	byte[] paGetByteArrayField(String fn) throws PException;

	/**
	 * It retrieves the value of a char[] field.
	 * @param fn	A String containing the name of the field.
	 * @return	The char[] value of the field as stored within
	 * 			the memory instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	char[] paGetCharArrayField(String fn) throws PException;

	/**
	 * It retrieves the value of a BigInteger field.
	 * @param fn	A String containing the name of the field.
	 * @return	The java.math.BigDInteger value of the field as stored within
	 * 			the memory instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	BigInteger paGetBigIntegerField(String fn) throws PException;

	/**
	 * It retrieves the value of a BigDecimal field.
	 * @param fn	A String containing the name of the field.
	 * @return	The java.math.BigDecimal value of the field as stored within
	 * 			the memory instance.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	BigDecimal paGetBigDecimalField(String fn) throws PException;

	/**
	 * It retrieves the value of a reference field.
	 * @param fn	A String containing the name of the field.
	 * @param connection The connection to access to support. This connection
	 * can be used to resolve a pname for example.
	 * @return	The name associated with the reference value of the field.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	PName paGetRefField(String fn, Object connection) throws PException;

	/**
	 * It assigns the value to a boolean field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The boolean value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetBooleanField(String fn, boolean value) throws PException;

	/**
	 * It assigns the value to a Boolean field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The Boolean value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetObooleanField(String fn, Boolean value) throws PException;

	/**
	 * It assigns the value to a byte field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The byte value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetByteField(String fn, byte value) throws PException;

	/**
	 * It assigns the value to a Byte field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The Byte value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetObyteField(String fn, Byte value) throws PException;

	/**
	 * It assigns the value to a char field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The char value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetCharField(String fn, char value) throws PException;

	/**
	 * It assigns the value to a Character field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The Character value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetOcharField(String fn, Character value) throws PException;

	/**
	 * It assigns the value to a short field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The short value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetShortField(String fn, short value) throws PException;

	/**
	 * It assigns the value to a Short field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The Short value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetOshortField(String fn, Short value) throws PException;

	/**
	 * It assigns the value to an int field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The int value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetIntField(String fn, int value) throws PException;

	/**
	 * It assigns the value to an Integer field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The Integer value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetOintField(String fn, Integer value) throws PException;

	/**
	 * It assigns the value to a long field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The long value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetLongField(String fn, long value) throws PException;

	/**
	 * It assigns the value to a Long field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The Long value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetOlongField(String fn, Long value) throws PException;

	/**
	 * It assigns the value to a float field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The float value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetFloatField(String fn, float value) throws PException;

	/**
	 * It assigns the value to a Float field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The Float value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetOfloatField(String fn, Float value) throws PException;

	/**
	 * It assigns the value to a double field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The double value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetDoubleField(String fn, double value) throws PException;

	/**
	 * It assigns the value to a Double field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The Double value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetOdoubleField(String fn, Double value) throws PException;

	/**
	 * It assigns the value to a string field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The java.lang.String value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetStringField(String fn, String value) throws PException;

	/**
	 * It assigns the value to a date field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The java.sql.Date value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetDateField(String fn, Date value) throws PException;

	/**
	 * It assigns the value to a serialized field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The java.io.Serializable value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetSerializedField(String fn, Serializable value)
			throws PException;

	/**
	 * It assigns the value to a byte[] field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The byte[] value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetByteArrayField(String fn, byte[] value) throws PException;

	/**
	 * It assigns the value to a char[] field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The char[] value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetCharArrayField(String fn, char[] value) throws PException;

	/**
	 * It assigns the value to a BigInteger field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The java.math.BigInteger value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetBigIntegerField(String fn, BigInteger value)
			throws PException;

	/**
	 * It assigns the value to a BigDecimal field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The java.math.BigDecimal value to assign.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetBigDecimalField(String fn, BigDecimal value)
			throws PException;

	/**
	 * It assigns the value to a reference field.
	 * @param fn	A String containing the name of the field to be set.
	 * @param value	The name associated with the reference value to assign.
	 * @param connection The connection to access to support. This connection
	 * can be used to resolve a pname for example.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this JORM class.
	 */
	void paSetRefField(String fn, PName value, Object connection)
			throws PException;
}
