/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.compiler.lib;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.compiler.api.JormCompilerConfigurator;
import org.objectweb.jorm.compiler.api.PExceptionCompiler;
import org.objectweb.jorm.generator.api.Generator;
import org.objectweb.jorm.generator.api.MOPFactory;
import org.objectweb.jorm.lib.JormConfiguratorImpl;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.jorm.verifier.api.MappingVerifier;
import org.objectweb.jorm.verifier.api.Verifier;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.LoggerFactory;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.util.HashMap;

/**
 * Used to hold the configuration information related to a JORM compiler
 * instance.
 * @author	P. Dechamboux
 */
public class JormCompilerConfiguratorImpl extends JormConfiguratorImpl implements JormCompilerConfigurator {
    
    private HashMap mapperName2mofFactory = new HashMap();
    
	/**
	 * It assignes the logger factory.
	 * @param lf
	 */
	public void setLoggerFactory(LoggerFactory lf) {
		loggerFactory = lf;
		logger = lf.getLogger(JormCompilerConfigurator.LOGGER_NAME);
	}

	/**
	 * Gets the verifier to be used by JORM.
	 * @return	That verifier.
	 */
	public Verifier getVerifier() throws PException {
		if (properties == null) {
			throw new PExceptionCompiler("JORM has not been configured.");
		}
		String cn = properties.getProperty("jorm.verifier");
        if (cn == null) {
            throw new PException(
                    "Impossible to load a verifier: not configured:"
                    + properties);
        }
		try {
			return (Verifier) loader.loadClass(cn).newInstance();
		} catch (InstantiationException e) {
			throw new PExceptionCompiler(e, "Problem while instantiating " + cn);
		} catch (IllegalAccessException e) {
			throw new PExceptionCompiler(e, "Problem while trying to access " + cn);
		} catch (ClassNotFoundException e) {
			throw new PExceptionCompiler(e, "Problem while loading class " + cn);
		}
	}

	/**
	 * Gets the generator to be used by JORM.
	 * @return	That generator.
	 */
	public Generator getGenerator() throws PException {
		if (properties == null) {
			throw new PExceptionCompiler("JORM has not been configured.");
		}
		String cn = properties.getProperty("jorm.generator");
        if (cn == null) {
            throw new PException(
                    "Impossible to load a generator: not configured:"
                    + properties);
        }
		try {
			Generator res = (Generator) loader.loadClass(cn).newInstance();
			return res;
		} catch (InstantiationException e) {
			throw new PExceptionCompiler(e, "Problem while instantiating " + cn);
		} catch (IllegalAccessException e) {
			throw new PExceptionCompiler(e, "Problem while trying to access " + cn);
		} catch (ClassNotFoundException e) {
			throw new PExceptionCompiler(e, "Problem while loading class " + cn);
		}
	}

	/**
	 * Gets the name to be used for the compiler options file.
	 * @return	That name.
	 */
	public String getJormcOptsFile() throws PException {
		if (properties == null) {
			throw new PExceptionCompiler("JORM has not been configured.");
		}
		return properties.getProperty("jorm.compiler.optsfile");
	}

	/**
	 * Gets the MOP factory associated with the given mapper.
	 * @param mappername	The name of the concerned mapper.
	 * @return The MOPFactory object.
	 * @throws PException
	 */
	public MOPFactory getMOPFactory(String mappername) throws PException {
		MOPFactory mopFactory = (MOPFactory) mapperName2mofFactory.get(mappername);
	    if (mopFactory != null) {
		    return mopFactory; 
		}
	    if (properties == null) {
			throw new PExceptionCompiler("JORM has not been configured.");
		}
		String cn = properties.getProperty("jorm.mapper.mopfactory." + mappername);
        if (cn == null) {
            throw new PException(
                    "Impossible to load a meta object factory: not configured:"
                    + properties);
        }
		try {
		    mopFactory = (MOPFactory) loader.loadClass(cn).newInstance();
		    mapperName2mofFactory.put(mappername, mopFactory);
			if (mopFactory instanceof Loggable) {
				if (loggerFactory != null) {
	                ((Loggable) mopFactory).setLoggerFactory(loggerFactory);
	            } else if (logger != null) {
	                ((Loggable) mopFactory).setLogger(logger);
	            }
			}
			return mopFactory;
		} catch (InstantiationException e) {
			throw new PExceptionCompiler(e, "Problem while instantiating " + cn);
		} catch (IllegalAccessException e) {
			throw new PExceptionCompiler(e, "Problem while trying to access " + cn);
		} catch (ClassNotFoundException e) {
			throw new PExceptionCompiler(e, "Problem while loading class " + cn);
		}
	}

	/**
	 * Gets an InputStream for reading the global compiler parameters.
	 * @return  The allocated InputStream.
	 * @throws PException
	 */
	public InputStream getGlobalJormcOptsFile() throws PException {
		InputStream res = null;
		String jormcOptsFile = getJormcOptsFile();
		if (pathToJormcOpts != null) {
			if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
				logger.log(BasicLevel.DEBUG, "JormcOptsFile=" + pathToJormcOpts
											 + File.separator + jormcOptsFile);
			}
			File f = new File(
					pathToJormcOpts + File.separator + jormcOptsFile);
			if (f.exists()) {
				try {
					res = new FileInputStream(f);
				} catch (FileNotFoundException e) {
					throw new PException(e, "Cannot find jorm compiler option file.");
				}
			}
		}
		if (res == null) {
			File f = new File(jormcOptsFile);
			if (f.exists()) {
				try {
					res = new FileInputStream(f);
				} catch (FileNotFoundException e) {
					throw new PException(e, "Cannot find jorm compiler option file.");
				}
			} else {
				res = loader.getResourceAsStream(jormcOptsFile);
			}
			if (res == null) {
				throw new PExceptionCompiler("Problem while accessing global " +
											 "compiler parameters: unable to locate ["
											 + getJormcOptsFile() + "]");
			}
		}
		return res;
	}

	/**
	 * Gets the verifier associated with the given mapper.
	 * @param mappername	The name of the concerned mapper.
	 * @return The MappingVerifier object.
	 * @throws PException
	 */
	public MappingVerifier getMappingVerifier(String mappername) throws PException {
		if (properties == null) {
			throw new PExceptionCompiler("JORM has not been configured.");
		}
		String cn = properties.getProperty("jorm.mapper.verifier." + mappername);
		try {
			MappingVerifier v = (MappingVerifier) loader.loadClass(cn).newInstance();
			return v;
		} catch (InstantiationException e) {
			throw new PExceptionCompiler(e, "Problem while instantiating " + cn);
		} catch (IllegalAccessException e) {
			throw new PExceptionCompiler(e, "Problem while trying to access " + cn);
		} catch (ClassNotFoundException e) {
			throw new PExceptionCompiler(e, "Problem while loading class " + cn);
		}
	}
}
