/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.compiler.lib;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.compiler.api.PExceptionCompiler;
import org.objectweb.jorm.compiler.api.JormCompilerParameter;
import org.objectweb.jorm.util.io.api.PathExplorer;
import org.objectweb.jorm.util.io.lib.DirJavaExplorer;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.Iterator;

/**
 * A JormCompilerParameter object is an object which manages informations
 * read by the command line parser. These informations are needed to
 * the compiler system to parse, generate, and compile.
 * @author	X. Spengler
 */
public class JormCompilerParameterImpl implements JormCompilerParameter {
	private String logConfFile = null;
	/**
	 * classpath is an object which stores a set of paths (or jar files).
	 */
	private PathExplorer classpath = new DirJavaExplorer();

	/**
	 * path where files will be generated
	 */
	private String output = null;

	/**
	 * verbose flag, set by the user
	 */
	private boolean verbose = false;

	/**
	 * a vector of input file names (.pd files)
	 */
	private Collection inputFiles = new ArrayList();

	/**
	 * keep the source files or not.
	 */
	private boolean keepSrc = true;

	/**
	 * javaCompiler is the name of the javac compiler used to compile
	 * the java sources.
	 */
	private String javacName = "javac";

	/**
	 * projectName is the name of the project for which we generate
	 */
	private String projectName = "";

	/**
	 * if the compile flag is set to true, the java sources will be compiled,
	 * if false, there will be no compilation.
	 */
	private boolean javac = false;
	
	private boolean generateBinding = true;

	/**
	 * if the build flag is set to true, files are generated, else false.
	 */
	private boolean parseOnly = false;

	/**
	 * If this flag is set to true, jorm will generate a state
	 * implementation for the persistent class.
	 */
	private boolean stateGenerated = true;

	/**
	 * The class the generated state implementation should extend.
	 */
	private String stateInheritance = null;

	/**
	 *
	 */
	private String bindingInheritance = null;
	
	private String bindingEnhancedClass = null;

	/**
	 *
	 */
	private String classMappingInheritance = null;

	/**
	 * if the bindingAbstract is set to true, the binding object will be
	 * abstract, else false
	 */
	private boolean bindingAbstract = false;

	/**
	 * a simple logger to log
	 */
	private Logger logger = null;

	/**
	 * the list of DTD locations
	 */
	private ArrayList dtdLocations = new ArrayList();

	/**
	 * indicates if the jorm .pd files must be generated
	 */
	private boolean generatedPDFiles = false;

	private boolean generatedWithMapperPackage = true;

	/**
	 * Prints informations managed by the current object.
	 */
	public void print() {
		if (logger.isLoggable(BasicLevel.DEBUG)) {
			logger.log(BasicLevel.DEBUG, "output: " + output);
			logger.log(BasicLevel.DEBUG, "verbose: " + verbose);
			logger.log(BasicLevel.DEBUG, "inputFiles: ");
			Iterator it = inputFiles.iterator();
			while (it.hasNext()) {
				logger.log(BasicLevel.DEBUG, ((String) it.next()));
			}
		}
	}

	/**
	 * Computes the actual class name which the generated PClassMapping class
	 * must extend.
	 * @param cn	The class name of the persistent class for which the
	 * 				code is generated.
	 * @return	The computed class name.
	 */
	public String computePClassMappingInheritance(String cn) {
		return computeInheritanceString(cn, classMappingInheritance);
	}

	/**
	 * Computes the actual class name which the generated PBinding class
	 * must extend.
	 * @param cn	The class name of the persistent class for which the
	 * 				code is generated.
	 * @return	The computed class name.
	 */
	public String computePBindingInheritance(String cn) {
		return computeInheritanceString(cn, bindingInheritance);
	}

	/**
	 * Load a jorm configuration file.
	 * @param file  the name of the jorm configuration file
	 */
	public void loadConfFile(String file, Iterator knownmappers) throws PException {
		try {
			loadConfFile(new FileInputStream(file), knownmappers);
		} catch (IOException e) {
			throw new PExceptionCompiler(e, "Cannot open the jorm configuration file: " + file);
		}
	}

	/**
	 * Load a jorm configuration file.
	 * @param in  the input stream of the jorm configuration file
	 */
	public void loadConfFile(InputStream in, Iterator knownmappers) throws PException {
		if (in == null) {
			throw new PExceptionCompiler("Cannot load jorm configuration file: InputStream is null.");
		}
		Properties p = new Properties();
		try {
			p.load(in);
		} catch (IOException e) {
			throw new PExceptionCompiler(e, "An error occured during the loading of the jorm configuration file.");
		}
		String s = null;
		if ((s = p.getProperty("log.config.file")) != null)
			logConfFile = s;
		if ((s = p.getProperty("keepsrc")) != null)
			keepSrc = new Boolean(s).booleanValue();
		if ((s = p.getProperty("verbose")) != null)
			verbose = new Boolean(s).booleanValue();
		if ((s = p.getProperty("javacompiler")) != null)
			javacName = s;
		if ((s = p.getProperty("projectname")) != null)
			projectName = s;
		if ((s = p.getProperty("stategenerated")) != null)
			stateGenerated = new Boolean(s).booleanValue();
		if ((s = p.getProperty("stateinheritance")) != null)
			stateInheritance = s;
		if ((s = p.getProperty("bindinginheritance")) != null)
			bindingInheritance = s;
		if ((s = p.getProperty("classmappinginheritance")) != null)
			classMappingInheritance = s;
		if ((s = p.getProperty("inputfiles")) != null) {
			StringTokenizer tok = new StringTokenizer(s, ",");
			while (tok.hasMoreTokens()) {
				inputFiles.add(tok.nextToken().trim());
			}
		}
	}

	public PathExplorer getClasspath() {
		return classpath;
	}

	public String getOutput() {
		return output;
	}

	public boolean isVerbose() {
		return verbose;
	}

	public Collection getInputFiles() {
		return inputFiles;
	}

	public boolean isKeepSrc() {
		return keepSrc;
	}

	public String getJavacName() {
		return javacName;
	}

	public String getProjectName() {
		return projectName;
	}

	public boolean isJavac() {
		return javac;
	}

	public boolean isParseOnly() {
		return parseOnly;
	}

	public boolean isGenerateBinding() {
		return generateBinding;
	}
	public void setGenerateBinding(boolean v) {
		this.generateBinding = v;
	}
	public boolean isStateGenerated() {
		return stateGenerated;
	}
	public void setStateGenerated(boolean v) {
		stateGenerated = v;
	}
	public String getStateInheritance() {
		return stateInheritance;
	}
	public String getBindingInheritance() {
		return bindingInheritance;
	}
	

	public String getBindingEnhancedClass() {
        return bindingEnhancedClass;
    }
	
	public String getClassMappingInheritance() {
		return classMappingInheritance;
	}

	public boolean isBindingAbstract() {
		return bindingAbstract;
	}

	public ArrayList getDtdLocations() {
		return dtdLocations;
	}

	public boolean isGeneratedPDFiles() {
		return generatedPDFiles;
	}

	public boolean isGeneratedWithMapperPackage() {
		return generatedWithMapperPackage;
	}

	// IMPLEMENTATION OF METHODS FROM THE Loggable INTERFACE

	public void setLogger(Logger logger) {
		this.logger = logger;
	}

	public void setLoggerFactory(LoggerFactory lf) {
	}

	public Logger getLogger() {
		return logger;
	}

	public LoggerFactory getLoggerFactory() {
		return null;
	}

	// IMPLEMENTATION OF METHODS FROM THE Cloneable INTERFACE

	public Object clone() {
		JormCompilerParameterImpl res = new JormCompilerParameterImpl();
		res.classpath = classpath;
		res.output = output;
		res.verbose = verbose;
		res.inputFiles = inputFiles;
		res.keepSrc = keepSrc;
		res.javacName = javacName;
		res.projectName = projectName;
		res.javac = javac;
		res.parseOnly = parseOnly;
		res.stateGenerated = stateGenerated;
		res.stateInheritance = stateInheritance;
		res.bindingInheritance = bindingInheritance;
		res.classMappingInheritance = classMappingInheritance;
		res.bindingAbstract = bindingAbstract;
		res.logger = logger;
		res.dtdLocations = dtdLocations;
		return res;
	}

	// PRIVATE METHODS

	private static String computeInheritanceString(String cn, String pattern) {
		if (cn == null || cn.length() == 0) {
			return null;
		}
		int pIdx = pattern.indexOf(INHERITANCE_PACKAGE_PATTERN);
		int cIdx = pattern.indexOf(INHERITANCE_CLASSNAME_PATTERN);

		if (pIdx == -1 && cIdx == -1) {
			return pattern;
		} else {
			String[] icn = isolatePackageName(cn);
			StringTokenizer st = new StringTokenizer(pattern, "%cp", true);
			StringBuffer sb = new StringBuffer();
			boolean hasPercent = false;
			while (st.hasMoreTokens()) {
				String token = st.nextToken();
				if (token.equals("%") && !hasPercent) {
					hasPercent = true;
				} else if (hasPercent) {
					if (token.equals("p")) {
						if (icn[0] != null) {
							sb.append(icn[0]);
						}
					} else if (token.equals("c")) {
						sb.append(icn[1]);
					} else {
						//ignore the unknown tokens
					}
					hasPercent = false;
				} else {
					sb.append(token);
				}
			}
			return sb.toString();
		}
	}

	private static String[] isolatePackageName(String fqcn) {
		int idx = fqcn.lastIndexOf(".");
		if (idx == -1)
			return new String[]{null, fqcn};
		else
			return new String[]{
				fqcn.substring(0, idx + 1),
				fqcn.substring(idx + 1, fqcn.length())
			};
	}

	public void setClasspath(PathExplorer classpath) {
		this.classpath = classpath;
	}

	public void setOutput(String output) {
		this.output = output;
	}

	public void setVerbose(boolean verbose) {
		this.verbose = verbose;
	}

	public void setKeepSrc(boolean keepSrc) {
		this.keepSrc = keepSrc;
	}

	public void setProjectName(String projectName) {
		this.projectName = projectName;
	}

	public void setJavac(boolean javac) {
		this.javac = javac;
	}

	public void setStateInheritance(String stateInheritance) {
		this.stateInheritance = stateInheritance;
	}

	public void setBindingInheritance(String bindingInheritance) {
		this.bindingInheritance = bindingInheritance;
	}
	
	public void setBindingEnhancedClass(String bindingEnhancedClass) {
		this.bindingEnhancedClass = bindingEnhancedClass;
	}
	
	public void setClassMappingInheritance(String classMappingInheritance) {
		this.classMappingInheritance = classMappingInheritance;
	}

	public void setBindingAbstract(boolean bindingAbstract) {
		this.bindingAbstract = bindingAbstract;
	}

	public void setDtdLocations(ArrayList dtdLocations) {
		this.dtdLocations = dtdLocations;
	}

	public void setGeneratedPDFiles(boolean generatedPDFiles) {
		this.generatedPDFiles = generatedPDFiles;
	}

	public void setInputFiles(Collection inputFiles) {
		this.inputFiles = inputFiles;
	}

	public void setGeneratedWithMapperPackage(boolean generatedWithMapperPackage) {
		this.generatedWithMapperPackage = generatedWithMapperPackage;
	}

	public void setLogConfFile(String logConfFile) {
		this.logConfFile = logConfFile;
	}
}
