/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.facility.naming.basidir;

import org.objectweb.jorm.api.PBinding;
import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.api.PStateGraph;
import org.objectweb.jorm.api.PBindingCtrl;
import org.objectweb.jorm.facility.naming.generator.LongGen;
import org.objectweb.jorm.facility.naming.generator.LongGenMgr;
import org.objectweb.jorm.facility.naming.generator.LongGenMgrRegistry;
import org.objectweb.jorm.naming.api.PBinder;
import org.objectweb.jorm.naming.api.PExceptionExistingName;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PNamingContext;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.perseus.cache.api.CacheManager;
import org.objectweb.perseus.cache.api.CacheEntry;
import org.objectweb.perseus.cache.api.CacheException;
import org.objectweb.perseus.cache.api.FixableCacheEntry;

import java.util.Date;
import java.math.BigInteger;
import java.math.BigDecimal;

/**
 *
 * @author Sebastien Chassande-Barrioz
 */
public class BasidBinder implements PBinder, PNamingContext {
	// Static binding to LongGenIncrMgr: TO BE REMOVED
	private final static String LONGGENMGRCLASS
			= "org.objectweb.jorm.facility.naming.generator.LongGenIncr";
	protected int codingType = -1;
	protected Object nullOValue = null;
	protected long nullValue = -1;
	PName PNNULL = null;
	PName PNONULL = null;
	protected PType ptype = null;
	protected PClassMapping pcm = null;
	private CacheManager bindingcache = null;
	private LongGen longGen = null;

	public BasidBinder() {
	}

	public BasidBinder(PMapper m, String genname, byte clact) throws PException {
		this();
		LongGenMgr lgm;
		// LongGen initialization
		synchronized (LongGenMgrRegistry.class) {
			lgm = LongGenMgrRegistry.getLongGenMgr(m);
			if (lgm == null) {
				try {
					lgm = (LongGenMgr) Class.forName(
							m.cn2mn(LONGGENMGRCLASS) + PMapper.PCLASSMAPPINGAPPENDER)
							.newInstance();
				} catch (InstantiationException e) {
					throw new PException(e, "Cannot create LongGenMgr: cannot instanciate.");
				} catch (IllegalAccessException e) {
					throw new PException(e, "Cannot create LongGenMgr: cannot access.");
				} catch (ClassNotFoundException e) {
					throw new PException(e, "Cannot create LongGenMgr: cannot load.");
				}
				lgm.init(m, clact);
				LongGenMgrRegistry.registerLonGenMgr(lgm);
			}
		}
		longGen = lgm.getLongGen(genname);
		codingType = PNamingContext.CTLONG;
		nullValue = -1;
	}

	public BasidBinder(int codingtype) {
		this();
		codingType = codingtype;
	}

	public BasidBinder(int codingtype, Object nullValue) {
		this();
		codingType = codingtype;
		this.nullOValue = nullValue;
	}

	public BasidBinder(int codingtype, long nullValue) {
		this();
		codingType = codingtype;
		this.nullValue = nullValue;
	}

	public int getCodingType() {
		return codingType;
	}

	public void setCodingType(int codingType) {
		this.codingType = codingType;
	}

	public void setNulOValue(Object nullvalue) {
		this.nullOValue = nullvalue;
	}

	public void setNulValue(long nullvalue) {
		this.nullValue = nullvalue;
	}
	//IMPLEMENTATION OF THE PNamingContext INTERFACE //
	//-----------------------------------------------//

	public boolean codingSupported(int codingtype) {
		return codingtype == codingType;
	}

	public PName decode(byte[] en) throws PExceptionNaming {
		return new BasidName(en, this);
	}

	public PName decodeAbstract(Object en, Object context) throws PExceptionNaming, UnsupportedOperationException {
		return new BasidName(en, this);
	}

	public PName decodeByte(byte en) throws PExceptionNaming, UnsupportedOperationException {
		return new BasidName(en, this);
	}

	public PName decodeObyte(Byte en) throws PExceptionNaming, UnsupportedOperationException {
		return new BasidName(en, this);
	}

	public PName decodeChar(char en) throws PExceptionNaming, UnsupportedOperationException {
		return new BasidName(en, this);
	}

	public PName decodeOchar(Character en) throws PExceptionNaming, UnsupportedOperationException {
		return new BasidName(en, this);
	}

	public PName decodeInt(int en) throws PExceptionNaming, UnsupportedOperationException {
		return new BasidName(en, this);
	}

	public PName decodeOint(Integer en) throws PExceptionNaming, UnsupportedOperationException {
		return new BasidName(en, this);
	}

	public PName decodeLong(long en) throws PExceptionNaming, UnsupportedOperationException {
		return new BasidName(en, this);
	}

	public PName decodeOlong(Long en) throws PExceptionNaming, UnsupportedOperationException {
		return new BasidName(en, this);
	}

	public PName decodeShort(short en) throws PExceptionNaming, UnsupportedOperationException {
		return new BasidName(en, this);
	}

	public PName decodeOshort(Short en) throws PExceptionNaming, UnsupportedOperationException {
		return new BasidName(en, this);
	}

	public PName decodeString(String en) throws PExceptionNaming {
		switch (codingType) {
		case PNamingContext.CTBYTE:
		case PNamingContext.CTCHAR:
		case PNamingContext.CTSHORT:
		case PNamingContext.CTINT:
		case PNamingContext.CTLONG:
			return new BasidName(Long.parseLong(en), this);
		}
		return new BasidName(en, this);
	}

	public PName decodeCharArray(char[] en) throws PExceptionNaming {
		return new BasidName(en, this);
	}

	public PName decodeDate(Date en) throws PExceptionNaming {
		return new BasidName(en, this);
	}

	public PName decodeBigInteger(BigInteger en) throws PExceptionNaming {
		return new BasidName(en, this);
	}

	public PName decodeBigDecimal(BigDecimal en) throws PExceptionNaming {
		return new BasidName(en, this);
	}

	public byte[] encode(PName pn) throws PExceptionNaming {
		if (pn.isNull())
			return (byte[]) nullOValue;
		if (codingType == PNamingContext.CTBYTEARRAY)
			return (byte[]) ((BasidName) pn).o;
		throw new PExceptionNaming("Unsupported coding type");
	}

	public Object encodeAbstract(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		return (BasidName) pn;
	}

	public byte encodeByte(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		if (pn.isNull())
			return (byte) nullValue;
		if (codingType == CTBYTE)
			return (byte) ((BasidName) pn).p;
		throw new PExceptionNaming("Unsupported coding type, expected: " + codingType);
	}

	public Byte encodeObyte(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		if (pn.isNull())
			return (Byte) nullOValue;
		if (codingType == CTOBYTE)
			return (Byte) ((BasidName) pn).o;
		throw new PExceptionNaming("Unsupported coding type, expected: " + codingType);
	}

	public char encodeChar(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		if (pn.isNull())
			return (char) nullValue;
		if (codingType == CTCHAR)
			return (char) ((BasidName) pn).p;
		throw new PExceptionNaming("Unsupported coding type, expected: " + codingType);
	}

	public Character encodeOchar(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		if (pn.isNull())
			return (Character) nullOValue;
		if (codingType == CTOCHAR)
			return (Character) ((BasidName) pn).o;
		throw new PExceptionNaming("Unsupported coding type, expected: " + codingType);
	}

	public int encodeInt(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		if (pn.isNull())
			return (int) nullValue;
		if (codingType == CTINT)
			return (int) ((BasidName) pn).p;
		throw new PExceptionNaming("Unsupported coding type, expected: " + codingType);
	}

	public Integer encodeOint(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		if (pn.isNull())
			return (Integer) nullOValue;
		if (codingType == CTOINT)
			return (Integer) ((BasidName) pn).o;
		throw new PExceptionNaming("Unsupported coding type, expected: " + codingType);
	}

	public long encodeLong(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		if (pn.isNull())
			return nullValue;
		if (codingType == CTLONG)
			return ((BasidName) pn).p;
		System.err.println("CT=" + codingType);
		throw new PExceptionNaming("Unsupported coding type, expected: " + codingType);
	}

	public Long encodeOlong(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		if (pn.isNull())
			return (Long) nullOValue;
		if (codingType == CTOLONG)
			return (Long) ((BasidName) pn).o;
		throw new PExceptionNaming("Unsupported coding type, expected: " + codingType);
	}

	public short encodeShort(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		if (pn.isNull())
			return (short) nullValue;
		if (codingType == CTSHORT)
			return (short) ((BasidName) pn).p;
		throw new PExceptionNaming("Unsupported coding type, expected: " + codingType);
	}

	public Short encodeOshort(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		if (pn.isNull())
			return (Short) nullOValue;
		if (codingType == CTOSHORT)
			return (Short) ((BasidName) pn).o;
		throw new PExceptionNaming("Unsupported coding type, expected: " + codingType);
	}

	public String encodeString(PName pn) throws PExceptionNaming {
		if (pn.isNull())
			return null;
		switch (codingType) {
		case CTBYTE:
		case CTCHAR:
		case CTINT:
		case CTLONG:
		case CTSHORT:
			return Long.toString(((BasidName) pn).p);
		case CTOBYTE:
		case CTOCHAR:
		case CTOINT:
		case CTOLONG:
		case CTOSHORT:
		case CTSTRING:
		case CTBIGINTEGER:
		case CTBIGDECIMAL:
			return ((BasidName) pn).o.toString();
		case CTDATE:
			return "" + ((Date) ((BasidName) pn).o).getTime();
		}
		return "?";
	}

	public char[] encodeCharArray(PName pn) throws PExceptionNaming {
		if (pn.isNull())
			return (char[]) nullOValue;
		if (codingType == CTCHARARRAY)
			return (char[]) ((BasidName) pn).o;
		throw new PExceptionNaming("Unsupported coding type, expected: " + codingType);
	}

	public Date encodeDate(PName pn) throws PExceptionNaming {
		if (pn.isNull())
			return (Date) nullOValue;
		if (codingType == CTDATE)
			return (Date) ((BasidName) pn).o;
		throw new PExceptionNaming("[" + getClassName() + "]: unsupported coding type");
	}

	public BigInteger encodeBigInteger(PName pn) throws PExceptionNaming {
		if (pn.isNull())
			return (BigInteger) nullOValue;
		if (codingType == CTBIGINTEGER)
			return (BigInteger) ((BasidName) pn).o;
		throw new PExceptionNaming("[" + getClassName() + "]: unsupported coding type");
	}

	public BigDecimal encodeBigDecimal(PName pn) throws PExceptionNaming {
		if (pn.isNull())
			return (BigDecimal) nullOValue;
		if (codingType == CTBIGDECIMAL)
			return (BigDecimal) ((BasidName) pn).o;
		throw new PExceptionNaming("[" + getClassName() + "]: unsupported coding type");
	}

	public PName export(Object conn, Object en) throws PException {
		PName pn = null;
		if (en == null) {
			throw new PExceptionNaming("[" + getClassName() + "]: cannot export null!");
		}
		if (en instanceof BasidName) {
			if (((PName) en).getPNameManager() == this) {
				return (PName) en;
			} else {
				return new BasidName(en, this);
			}
		}
		PBindingCtrl pb = (PBindingCtrl) en;
		byte nextstate = PStateGraph.nextStatePBinding(pb.getStatus(),
													   PBinding.ACTION_EXPORT);
		if (nextstate == PBinding.LIFECYCLE_ERROR)
			throw new PExceptionProtocol("Export unauthorized operation, current status:" + pb.getStatus());
		// Performs the transition
		if (longGen == null) {
			throw new PExceptionNaming("[" + getClassName() + "]: export not supported by this binder with no generator.");
		}
		pn = new BasidName(longGen.genId(conn), this);
		if (bindingcache != null) {
			synchronized (bindingcache) {
				if (bindingcache.lookup(pn) != null) {
					throw new PExceptionExistingName(
							"[" + getClassName() + "]: an object has been already export with the same identifier");
				}
				try {
					bindingcache.fix(bindingcache.bind(pn, en));
				} catch (CacheException e) {
					throw new PException(e, "[" + getClassName() + "]: problem with cache management");
				}
			}
		}
		pb.setPName(pn);
		pb.setStatus(nextstate);
		return pn;
	}

	public PName export(Object conn, Object en, Object hints) throws PException {
		PName pn = null;
		if (en == null) {
			throw new PExceptionNaming("[" + getClassName() + "]: cannot export null!");
		}
		if (en instanceof PName && ((PName) en).getPNameManager() == this)
			return (PName) en;
		PBindingCtrl pb = (PBindingCtrl) en;
		byte nextstate = PStateGraph.nextStatePBinding(pb.getStatus(),
													   PBinding.ACTION_EXPORT);
		if (nextstate == PBinding.LIFECYCLE_ERROR)
			throw new PExceptionProtocol("Export unauthorized operation, current status:" + pb.getStatus());
		// Performs the transition
		if (longGen != null) {
			throw new PExceptionNaming("[" + getClassName() + "]: export with hints not support by this generator binder.");
		}
		if (en instanceof BasidName
				&& ((PName) en).getPNameManager().equals(this)) {
			return (PName) en;
		}
		if (!(en instanceof PBinding)) {
			throw new PExceptionNaming("[" + getClassName() + "]: the item cannot be exported by this binder (PName or PBinding only): " + en);
		}
		if (hints != null) {
			pn = new BasidName(hints, this);
		} else {
			throw new PExceptionNaming(
					"[" + getClassName() + "]: this item cannot be exported by this binder");
		}
		if (bindingcache != null) {
			synchronized (bindingcache) {
				if (bindingcache.lookup(pn) != null) {
					throw new PExceptionExistingName(
							"[" + getClassName() + "]: an object has been already export with the same identifier");
				}
				try {
					bindingcache.fix(bindingcache.bind(pn, en));
				} catch (CacheException e) {
					throw new PException(e, "[" + getClassName() + "]: problem with cache management");
				}
			}
		}
		pb.setPName(pn);
		pb.setStatus(nextstate);
		return pn;
	}

	public PName getNull() {
		switch (codingType) {
		case CTBYTE:
		case CTINT:
		case CTLONG:
		case CTSHORT:
			return PNNULL;
		case CTOBYTE:
		case CTOCHAR:
		case CTOINT:
		case CTOLONG:
		case CTOSHORT:
		case CTSTRING:
		case CTDATE:
			return PNONULL;
		case CTCHAR:
			return PNNULL;
		default:
			return PNONULL;
		}
	}

	/**
	 * It defines the null PName since a BasidName or an object value
	 * If the encoding is an object format the given value will be the null
	 * encoded valued. Otherwise if the encoded format is a primitive value
	 * then a object value matching to the primitive type is expected:
	 * byte ==> a java.lang.Byte is expected
	 * short ==> a java.lang.Short is expected
	 * ...
	 * @param o
	 */
	public void setNullPName(Object o) {
		if (o == null) {
			nullOValue = o;
		} else if (o instanceof BasidName) {
			nullOValue = ((BasidName) o).o;
			nullValue = ((BasidName) o).p;
		} else {
			nullOValue = o;
			if (o instanceof Byte) {
				nullValue = ((Byte) o).byteValue();
			} else if (o instanceof Short) {
				nullValue = ((Short) o).shortValue();
			} else if (o instanceof Integer) {
				nullValue = ((Integer) o).intValue();
			} else if (o instanceof Long) {
				nullValue = ((Long) o).longValue();
			} else if (o instanceof Character) {
				nullValue = ((Character) o).charValue();
			}
		}
	}

	public PType getPType() {
		return ptype;
	}

	public String getClassName() {
		return ptype.getJormName();
	}

	public PName resolve(Object conn, PName pn) throws PException {
		if (pn == null) {
			throw new PExceptionNaming("[" + getClassName() + "]: this pname is null");
		}
		if (!pn.getPNameManager().equals(this))
			throw new PExceptionNaming("[" + getClassName() + "]: this pname is not valid in this binder");
		return pn;
	}

	public void setPType(PType pt) {
		ptype = pt;
	}

	public boolean supportDynamicComposite() {
		return false;
	}

	public boolean supportCompositeField(String fn, PType ft) {
		return false;
	}

	public boolean supportStaticComposite() {
		return false;
	}

	public void unexport(Object conn, PName pn) throws PException {
		if (pn.isNull()) {
			throw new PExceptionProtocol("[" + getClassName() + "]: cannot unexport with a null pname.");
		}
		if (bindingcache != null) {
			PBindingCtrl pb = null;
			CacheEntry ce = bindingcache.lookup(pn);
			if (ce != null) {
				pb = (PBindingCtrl) ce.getCeObject();
			}
			if (pb == null) {
				throw new PExceptionProtocol("Unauthorized operation: No Pbinding found");
			}
			byte nextstate = PStateGraph.nextStatePBinding(
					pb.getStatus(), PBinding.ACTION_UNEXPORT);
			if (nextstate == PBinding.LIFECYCLE_ERROR) {
				throw new PExceptionProtocol("Unauthorized operation: status =" + pb.getStatus());
			}
			pb.setStatus(nextstate);
		} else {
			throw new PExceptionProtocol("Unauthorized operation: No Pbinding can be found without a CacheManager");
		}
	}

	public void unexport(Object conn, PName pn, Object hints) throws PException {
		if (pn.isNull()) {
			throw new PExceptionProtocol("[" + getClassName() + "]: cannot unexport with a null pname.");
		}
		PBindingCtrl pb = null;
		if (hints instanceof PBindingCtrl) {
			pb = (PBindingCtrl) hints;
		} else if (bindingcache != null) {
			CacheEntry ce = bindingcache.lookup(pn);
			if (ce != null) {
				pb = (PBindingCtrl) ce.getCeObject();
			}
		}
		if (pb == null) {
			throw new PExceptionProtocol("Unauthorized operation: No Pbinding found");
		}
		byte nextstate = PStateGraph.nextStatePBinding(
				pb.getStatus(), PBinding.ACTION_UNEXPORT);
		if (nextstate == PBinding.LIFECYCLE_ERROR) {
			throw new PExceptionProtocol("Unauthorized operation: status =" + pb.getStatus());
		}
		pb.setStatus(nextstate);
	}

	// IMPLEMENTATION OF METHODS FROM THE PBinder INTERFACE

	public PBinding lookup(PName pn) throws PException {
		if (pn == null) {
			throw new PExceptionNaming("[" + getClassName() + "]: this pname is null");
		}
		if (pn.isNull()) {
			return null;
		}
		if (bindingcache == null) {
			throw new PExceptionProtocol("No internal cache: should be managed externally!");
		}
		CacheEntry ce = bindingcache.lookup(pn);
		if (ce != null) {
			return (PBinding) ce.getCeObject();
		}
		return null;
	}

	public void bind(PName pn, PBindingCtrl pb) throws PException {
		if (pn == null)
			throw new PExceptionNaming("[" + getClassName() + "]: this pname is null");
		if (pb == null)
			throw new PExceptionNaming("[" + getClassName() + "]: this pbinding is null");
		if (!(pn instanceof BasidName))
			throw new PExceptionNaming("Can only bind CustomerPkPName");
		if (!pn.getPNameManager().equals(this))
			throw new PExceptionNaming("[" + getClassName() + "]: this pname is not valid in this binder ("
									   + ", \nthis:" + this
									   + ", \nthis.type:" + this.getPType().getJormName()
									   + ", \nthis.pcm:" + this.getBinderClassMapping()
									   + ", \npn.pnc: " + pn.getPNameManager()
									   + ", \npn.pnc.type: " + pn.getPNameManager().getPType().getJormName()
									   + ", \npn.pnc.pcm: " + ((PBinder) pn.getPNameManager()).getBinderClassMapping()
									   + ", \npn:" + pn + ")");
		if (!pb.getPClassMapping().equals(pcm))
			throw new PExceptionNaming("[" + getClassName() + "]: this pbinding is not valid in this binder");
		if (pn.isNull())
			throw new PExceptionNaming("[" + getClassName() + "]: this PName represent the null value");
		byte nextstate = PStateGraph.nextStatePBinding(pb.getStatus(),
													   PBinding.ACTION_BIND);
		// Performs the transition
		if (bindingcache != null) {
			synchronized (bindingcache) {
				try {
					bindingcache.fix(bindingcache.bind(pn, pb));
				} catch (CacheException e) {
					throw new PException(e, "[" + getClassName() + "]: problem with cache management");
				}
			}
		}
		pb.setPName(pn);
		pb.setStatus(nextstate);
	}

	public PClassMapping getBinderClassMapping() {
		return pcm;
	}

	public void unbind(PBindingCtrl pb) throws PException {
		byte nextstate = PStateGraph.nextStatePBinding(pb.getStatus(),
													   PBinding.ACTION_UNBIND);
		// If the state does not change, do nothing
		if (nextstate == pb.getStatus())
			return;
		// Perform the transition
		if (pb == null)
			throw new PExceptionNaming("[" + getClassName() + "]: this pbinding is null");
		PName pn = pb.getPName();
		if (pn == null)
			throw new PExceptionNaming("[" + getClassName() + "]: this PBinding is not bounded with a pname");
		if (!pb.getPClassMapping().equals(pcm))
			throw new PExceptionNaming("[" + getClassName() + "]: this pbinding is not valid in this binder");
		if (bindingcache != null) {
			synchronized (bindingcache) {
				try {
					if (((FixableCacheEntry) bindingcache).getCeFixCount() == 0) {
						// the cache replacement has decided to remove this entry
						pb.setPName(null);
						pb.setStatus(nextstate);
					} else {
						// this entry becomes eligeable for replacement
						CacheEntry ce = bindingcache.lookup(pn);
						if (ce != null) {
							bindingcache.unfix(ce);
						}
					}
				} catch (CacheException e) {
					throw new PException(e, "[" + getClassName() + "]: problem with cache management");
				}
			}
		} else {
			pb.setPName(null);
			pb.setStatus(nextstate);
		}
	}

	/**
	 * Returns the cache manager associated with this binder.
	 * @return  The cache manager.
	 */
	public CacheManager getCacheManager() {
		return bindingcache;
	}

	/**
	 * Assigns the cache manager associated with this binder.
	 * @param cm    The cache manager.
	 * @throws PException
	 */
	public void setCacheManager(CacheManager cm) throws PException {
		bindingcache = cm;
	}

	// IMPLEMENTATION OF METHODS FROM THE PBinder INTERFACE

	public void setPClassMapping(PClassMapping pcm) {
		this.pcm = pcm;
		try {
			switch (codingType) {
			case CTBYTE:
			case CTCHAR:
			case CTINT:
			case CTLONG:
			case CTSHORT:
				PNNULL = new BasidName(nullValue, this);
				break;
			default:
				PNONULL = new BasidName(nullOValue, this);
			}
		} catch (PExceptionNaming pn) {
			// There is no chance having this exception when creation NULL PName
			System.out.println("Very strange problem: this exception should really never happen!!");
			pn.printStackTrace();
		}
	}
}
