/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.jorm.facility.naming.olongid;

import org.objectweb.jorm.api.PAccessor;
import org.objectweb.jorm.api.PBinding;
import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PExceptionNoDSI;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.api.PMapCluster;
import org.objectweb.jorm.facility.naming.basidir.BasidBinder;
import org.objectweb.jorm.facility.naming.generator.LongGen;
import org.objectweb.jorm.facility.naming.generator.LongGenIncrMgr;
import org.objectweb.jorm.facility.naming.generator.LongGenMgrRegistry;
import org.objectweb.jorm.facility.naming.generator.PLongGen;
import org.objectweb.jorm.facility.naming.polymorphid.ClassIdAccessor;
import org.objectweb.jorm.facility.naming.polymorphid.IdClassAccessor;
import org.objectweb.jorm.naming.api.PBinder;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.naming.api.PNameCoder;
import org.objectweb.jorm.naming.api.PNamingContext;

/**
 * Is a manager helping the used of LongId identifier. A LongId is
 * composed of two parts. The high bits describe the class identifier, whereas
 * the low bits describes the object instance identifier. The limit between both
 * parts are defined are building time (constructor).
 *
 * The allocation of identifiers is done through long generator.
 *
 * The manager permits to allocates PBinder and PNamingContext for normal class
 * and generic class.
 *
 * @author S.Chassande-Barrioz
 */
public class LongIdManager {

	/**
	 * Is the default size of the class identifier in term of number of bits
	 */
	public final static int DEFAULT_CID_SIZE = 20;

	/**
	 * is the number of bits of an identifier. It is equal to the number of
	 * bits of a long value (64)
	 */
	public final static int LONG_SIZE = 64;

	/**
	 * Is the name of the persistent class used to load/store the association
	 * between a class name and its class identifier
	 */
	public static final String CLASS_ID_NAME =
		"org.objectweb.jorm.facility.naming.polymorphid.ClassId";

	/**
	 * Is the name of the persistent class used to load/store the association
	 * between a class identifier and its class name.
	 */
	public static final String ID_CLASS_NAME =
		"org.objectweb.jorm.facility.naming.polymorphid.IdClass";

	/**
	 * The the manager of generator. It used to allocate object id generator
	 * for each binder and also to allocate the unique class identifier
	 * generator.
	 */
	private LongGenIncrMgr idGenMgr;

	/**
	 * Is the class identifier generator.
	 */
	private LongGen classIdGen;

	/**
	 * Is the mapper where the generator are persistent.
	 */
	private PMapper mapper;

	/**
	 * is the number of bits allocated for the class identifier among the
	 * number of bits of a long value (64).
	 */
	private int cidSize = DEFAULT_CID_SIZE;

	/**
	 * Builds a LongIdManager with the default class identifier size.
	 * IMMPORTANT: a mapper must be assigned later.
	 */
	public LongIdManager() {
		setCidSize(DEFAULT_CID_SIZE);
	}

	/**
	 * Builds and intializes a LongIdManager with the default class identifier
	 * size.
	 * @param mapper is the mapper where the generator are persistent.
	 */
	public LongIdManager(PMapper mapper) throws PException {
		this(mapper, DEFAULT_CID_SIZE);
	}

	/**
	 * Builds and intializes a LongIdManager with a particular class identifier
	 * size.
	 * @param mapper is the mapper where the generator are persistent.
	 * @param cidSize is the number of bits allocated for the class identifier among the
	 * number of bits of a long value (64).
	 */
	public LongIdManager(PMapper mapper, int cidSize) throws PException {
		setCidSize(cidSize);
		setPMapper(mapper);
	}

	public int getCidSize() {
		return cidSize;
	}

	public void setCidSize(int cidSize) {
		this.cidSize = cidSize;
	}

	public PMapper getMapper() {
		return mapper;
	}

	/**
	 * Instanciates and initializes a LongGenMgr
	 * Allocates a class identifier generator
	 * @param mapper the mapper where the id generator are stored
	 * @throws PException if no LongGenMgr has been found in the specified
	 * mapper.
	 */
	public void setPMapper(PMapper mapper) throws PException {
		if (mapper == null) {
			throw new PExceptionNaming("No mapper specified");
		}
		this.mapper = mapper;
		//try to find the long gen manager in the registry
		synchronized (LongGenMgrRegistry.class) {
			idGenMgr = (LongGenIncrMgr) LongGenMgrRegistry.getLongGenMgr(mapper);
			if (idGenMgr == null) {
				//if not already registered, create it
				try {
					idGenMgr = (LongGenIncrMgr) Class.forName(getLongGenMgr()).newInstance();
				} catch (Exception e) {
					throw new PException(e, "Cannot create LongGenMgr (probably a ClassLoader problem): " + getLongGenMgr());
				}
				idGenMgr.init(mapper, PClassMapping.CREATE_STRUCTURE_IF_NEEDED);
				//and register it
				LongGenMgrRegistry.registerLonGenMgr(idGenMgr, mapper);
			}
		}
		classIdGen = idGenMgr.getLongGen(CLASS_ID_NAME, null);

		synchronized(mapper) {
		    if (mapper.lookup(CLASS_ID_NAME) == null) {
				PBinder binder = new BasidBinder(PNamingContext.CTSTRING);
				String cn = mapper.cn2mn(CLASS_ID_NAME) + PMapper.PCLASSMAPPINGAPPENDER;
				PClassMapping pcm;
				try {
					pcm = (PClassMapping) Class.forName(cn).newInstance();
				} catch (Exception e) {
					throw new PException(e);
				}
				pcm.setPBinder(binder);
				binder.setPClassMapping(pcm);
				mapper.map(pcm);
				PMapCluster cl = mapper.getPMapCluster(pcm.getClassName());
				cl.createMappingStructures(false);
		    }
		}
	}
	
	/**
	 * Calculates the class name of the LongGenMgr from the mapper name. Indeed
	 * the class is the generated PClassMapping of the persistent class
	 * org.objectweb.jorm.facility.naming.generator.LongGenIncr.
	 */
	private String getLongGenMgr() {
		String mn = mapper.getMapperName();
		int idx = mn.indexOf('.');
		if (idx != -1) {
			mn = mn.substring(0, idx);
		}
		return "org.objectweb.jorm.facility.naming.generator."
			+ mn + ".LongGenIncrMapping";
	}

	/**
	 * Allocates a new PBinder for gen class. This PBinder can be used as
	 * PNameCoder too.
	 */
	public PBinder newGenClassPBinder() throws PException {
		return new BasidBinder(PNameCoder.CTOLONG, -1);
	}

	/**
	 * Allocates a new PNamingcontext for a normal persistent class  (not a
	 * generic class).
	 */
	public PNamingContext newClassPNamingContext() throws PException {
		return new LongIdPNC(cidSize);
	}

	/**
	 * Allocates a new PBinder for a normal persistent class (not a generic
	 * class).
	 * @param className is the name of the persistent class
	 * @param conn is a way to access the data support. If the value is null,
	 * a connection is allocated from the mapper.
	 * @throws PException
	 */
	public PBinder newClassPBinder(String className, Object conn) throws PException {
		if (mapper == null) {
			throw new PException("The LongIdManager has not been configured, a PMapper is required");
		}
		Object connection = conn;
		if (conn == null) {
			connection = mapper.getConnection();
		}
		try {
			LongGen lg = new CompositePLongGen(
				(PLongGen) idGenMgr.getLongGen(className, connection),
				getCID(className, connection),
				cidSize);
			LongIdBinder lib = new LongIdPBinder();
			lib.setACFLid(lg);
			return lib;
		} finally {
			if (conn == null) {
				mapper.closeConnection(connection);
			}
		}
	}


	/**
	 * Fetches the identifier (long value) of a normal persistent class.
	 * If no value has been already allocated (nothing store in database), then
	 * a new value is allocated and saved through the class identifier
	 * generator.
	 *
	 * @param className is the name of the persistent class.
	 * @param conn is a way to access the data support. If the value is null,
	 * a connection is allocated from the mapper.
	 * @return the long value representing the persistent class.
	 * @throws PException
	 */
	private long getCID(String className, Object conn) throws PException {
		PClassMapping pcm = mapper.lookup(CLASS_ID_NAME);
		PBinding binding = pcm.createPBinding();
		binding.bind(pcm.getPBinder().decodeString(className));
		Object connection = conn;
		if (conn == null)
			connection = mapper.getConnection();
		CIDAccessor acc = new CIDAccessor(className);
		try {
			binding.read(connection, acc);
		} catch (PExceptionNoDSI e) {
			binding.export(connection, className);
			acc.cid = classIdGen.genId(connection);
			binding.write(connection, acc);
		} finally {
			if (conn == null) {
				mapper.closeConnection(connection);
			}
		}
		binding.unbind();
		return acc.cid;
	}

	/**
	 * Is an accessor permitting to store/load the assocaition between a
	 * persistent class name (String) and the class identifier (long).
	 *
	 */
	private class CIDAccessor implements PAccessor, ClassIdAccessor, IdClassAccessor {
		public long cid;
		public String className;

		public CIDAccessor(String className) {
			this.className = className;
		}

		public CIDAccessor(long cid) {
			this.cid = cid;
		}

		public Object getMemoryInstance() {
			return this;
		}

		//Accessors to the classId field
		public void paSetClassId(long val) throws PException {
			cid = val;
		}

		public long paGetClassId() throws PException {
			return cid;
		}

		//Accessors to the className field
		public void paSetClassName(String val) throws PException {
			className = val;
		}

		public String paGetClassName() throws PException {
			return className;
		}
	}
}
