/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.jorm.facility.naming.olongid;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.naming.api.PNameGetter;
import org.objectweb.jorm.naming.lib.AbstractInheritKeyFilteredPNamingContext;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.type.api.PTypeSpace;

/**
 * Is a PNamingContext managing the LongIdPName and LongIdBinder. It is based
 * on the AbstractInheritKeyFilteredPNamingContext class. The key is the class
 * identifier. The filter (evaluate method) is in charge to extract the class
 * identifier from the identifier (long value). Indeed the long value is
 * composed of two parts the high bits describe the class identifier, whereas
 * the low bits describes the object instance identifier. The limit between both
 * parts are defined are building time (constructor).
 *
 * @author S.Chassande-Barrioz
 */
public class LongIdPNC extends AbstractInheritKeyFilteredPNamingContext {

	Long lid;

	/**
	 * is the number of bits allocated for the class identifier among the
	 * number of bits of a long value (64).
	 */
	private int cidSize;

	/**
	 * Builds a LongIdPNC with a particular class identifier size
	 * @param cidsize is the number of bits allocated for the class identifier
	 * among the number of bits of a long value (64).
	 */
	public LongIdPNC(int cidsize) throws PException {
		super(PTypeSpace.LONG, CTLONG);
		setNullPName(new LongIdPName(null, new LongIdBinder()));
		((LongIdBinder) getNull().getPNameManager()).setNullPName(getNull());
		setCidSize(cidsize);
		init();
	}

	/**
	 * @see this.cidSize
	 */
	public int getCidSize() {
		return cidSize;
	}

	/**
	 * @see this.cidSize
	 */
	public void setCidSize(int cidSize) {
		this.cidSize = cidSize;
	}

	/**
	 * Calculates the class identifier from the long identifier value
	 * @param png is a PNameGetter or a Long instance
	 * @param ctx is the context used for evaluate the PNameGetter
	 * @return a Long instance containing the class identifier
	 * @throws PException if it is not possible to fetch the long value, or if
	 * the png parameter is not managed.
	 */
	public Object evaluate(Object png, Object ctx) throws PException {
		if (png instanceof LongIdPNG) {
			Long id = ((LongIdPNG) png).pnGetLid(ctx);
			if (id == null) {
			    return new Long(-1);
			} else {
			    return new Long(id.longValue() >> (LongIdManager.LONG_SIZE - cidSize));
			}
		} else if (png instanceof PNameGetter) {
			Long id = ((PNameGetter) png).pngetOlongField("lid", ctx);
			if (id == null) {
			    return new Long(-1);
			} else {
			    return new Long(id.longValue() >> (LongIdManager.LONG_SIZE - cidSize));
			}
		} else if (png instanceof Long) {
			Long id = (Long) png;
			if (id == null) {
			    return new Long(-1);
			} else {
			    return new Long(id.longValue() >> (LongIdManager.LONG_SIZE - cidSize));
			}
		}
		throw new PExceptionNaming(
			"Impossible to evaluate without a PNameGetter: " + png);
	}

	public void setNullPName(Object o) throws PException {
		super.setNullPName(o);
		//Set the field values representing the null PName
		if (o == null) {
		    lid = null;
		} else {
			PNameGetter png = (PNameGetter) nullPName.encodeAbstract();
			lid = png.pngetOlongField("lid", null);
		}
	}

	public boolean isNull(Object _png, Object ctx) throws PException {
	    Long l;
	    if (_png instanceof org.objectweb.jorm.facility.naming.olongid.LongIdPNG) {
			l = ((org.objectweb.jorm.facility.naming.olongid.LongIdPNG) _png).pnGetLid(ctx);
		} else if (_png instanceof PNameGetter) {
			l = ((PNameGetter) _png).pngetOlongField("lid", ctx);
		} else {
			throw new PException("The specified pname getter is not supported: " + _png);
		}
	    if (lid == null) {
	        return l == null;
	    } else {
	        return lid.equals(l);
	    }
	}
	
	public void setPType(PType pt) {
		super.setPType(pt);
		nullPName.getPNameManager().setPType(pt);
	}

}
