/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.facility.naming.rdbsequence;

import org.objectweb.jorm.api.PBinding;
import org.objectweb.jorm.api.PBindingCtrl;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.api.PStateGraph;
import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.naming.api.NamingFilterKeyProvider;
import org.objectweb.jorm.naming.api.PExceptionExistingName;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.lib.BasicPBinder;
import org.objectweb.jorm.mapper.rdb.lib.PMapperRdb;
import org.objectweb.perseus.cache.api.CacheEntry;
import org.objectweb.perseus.cache.api.CacheException;
import org.objectweb.perseus.persistence.api.ConnectionHolder;
import org.objectweb.perseus.persistence.api.PersistenceException;


/**
 * This binder is specific to the rdb mapper.
 * @author S.Chassande-Barrioz
 */
public class RdbSequenceBinder extends BasicPBinder {

	private RdbSequenceHelper sequenceHelper;
	private SequenceIdAllocator allocator;
	private Object key;

	public RdbSequenceBinder() {
		super();
		nullPName = new RdbSequencePName(this, RdbSequencePName.NULL_VALUE);
		sequenceHelper = new RdbSequenceHelper();
	}

	public void setSequenceIdAllocator(SequenceIdAllocator sia) {
		allocator = sia;
	}

	public SequenceIdAllocator getSequenceIdAllocator() {
		return allocator;
	}

	public RdbSequenceHelper getSequenceHelper() {
		return sequenceHelper;
	}

	public void setSequenceHelper (RdbSequenceHelper sequenceHelper) {
		this.sequenceHelper = sequenceHelper;
	}
	
	protected void initSequenceHelper(){
		if (sequenceHelper.getAdapter() != null) {
			return;
		}
		PClassMapping pcm = getBinderClassMapping();
		if (pcm == null) {
			throw new IllegalStateException("No PClassMapping bound to the PBinder");
		}
		PMapperRdb mapper = (PMapperRdb) pcm.getPMapper();
		if (mapper == null) {
			throw new IllegalStateException(
				"No PMapper bound to the PClassMapping of the class "
				+ pcm.getClassName());
		}
		if (pcm instanceof NamingFilterKeyProvider) {
			key = ((NamingFilterKeyProvider) pcm).getNamingFilterKey();
		}
		sequenceHelper.setAdapter(mapper.getRdbAdapter());
	}

	// IMPLEMENTATION OF THE METHOD OF BasicPNamingContext INTERFACE

	public PName resolve(Object conn, PName pn) throws PException {
		if (pn == null) {
			throw new PExceptionNaming("[" + getClassName()
									   + "]: this pname is null");
		}
		if (!pn.getPNameManager().equals(this))
			throw new PExceptionNaming("[" + getClassName()
									   + "]: this pname is not valid in this binder");
		return pn;
	}

	public PName export(Object conn, Object infoitem, Object hints)
			throws PException {
		if (hints instanceof RdbSequencePName) {
			//Binder role
			PBindingCtrl pb = (PBindingCtrl) infoitem;
			byte nextstate = PStateGraph.nextStatePBinding(pb.getStatus(),
														   PBinding.ACTION_EXPORT);
			if (nextstate == PBinding.LIFECYCLE_ERROR)
				throw new PExceptionProtocol("Unauthorized operation");

			PName pn = new RdbSequencePName(this, ((RdbSequencePName) hints).value);
			if (cache != null) {
				synchronized (cache) {
					if (cache.lookup(pn) != null) {
						throw new PExceptionExistingName("[" + getClassName()
														 + "]: an object has been already export with the same identifier");
					}
					try {
						cache.fix(cache.bind(pn, pb));
					} catch (CacheException e) {
						throw new PException(e, "[" + getClassName()
												+ "]: problem with cache management");
					}
				}
			}
			pb.setPName(pn);
			pb.setStatus(nextstate);
			return pn;
		} else {
			if (sequenceHelper.getSequenceName() == null
				&& (hints instanceof String)) {
				sequenceHelper.setSequenceName((String) hints);
			}
			return export(conn, infoitem);
		}
	}

	public PName export(Object c, Object en) throws PException {
		if (en == null) {
			throw new PExceptionNaming("[" + getClassName() + "]: cannot export null!");
		}
		if (en instanceof PName) {
			//Naming context role
			if (((PName) en).getPNameManager() == this) {
				return (PName) en;
			} else {
				return new RdbSequencePName(this, en);
			}
		}

		//Binder role
		PBindingCtrl pb = (PBindingCtrl) en;
		byte nextstate = PStateGraph.nextStatePBinding(pb.getStatus(),
													   PBinding.ACTION_EXPORT);
		if (nextstate == PBinding.LIFECYCLE_ERROR)
			throw new PExceptionProtocol("Unauthorized operation");

		PName pn = null;
		boolean connectionAllocatedLocaly = false;
		try {
			if (c == null) {
				c = getBinderClassMapping().getPMapper().getConnection();
				connectionAllocatedLocaly = true;
			}
			initSequenceHelper();
			if (!sequenceHelper.isSequenceCreated()) {
				sequenceHelper.createSequence(c);
			}
			long lid = sequenceHelper.allocateId(c);
			if (allocator != null) {
				lid = allocator.allocateId(lid, key);
			}
			pn = new RdbSequencePName(this, new Long(lid));
		} finally {
			if (connectionAllocatedLocaly) {
				getBinderClassMapping().getPMapper().closeConnection(c);
			} else if (c instanceof ConnectionHolder) {
				try {
					((ConnectionHolder) c).releaseCHConnection();
				} catch (PersistenceException e) {
					throw new PException(e, "Problem with Perseus connection releasing.");
				}
			}
		}

		if (cache != null) {
			synchronized (cache) {
				if (cache.lookup(pn) != null) {
					throw new PExceptionExistingName("[" + getClassName()
													 + "]: an object has been already export with the same identifier");
				}
				try {
					cache.fix(cache.bind(pn, pb));
				} catch (CacheException e) {
					throw new PException(e, "[" + getClassName()
											+ "]: problem with cache management");
				}
			}
		}
		pb.setPName(pn);
		pb.setStatus(nextstate);
		return pn;
	}

	public void unexport(Object conn, PName pn) throws PException {
		if (pn.isNull()) {
			throw new PExceptionProtocol("[" + getClassName() + "]: cannot unexport with a null pname.");
		}
		if (cache != null) {
			PBindingCtrl pb = null;
			CacheEntry ce = cache.lookup(pn);
			if (ce != null) {
				pb = (PBindingCtrl) ce.getCeObject();
			}
			if (pb == null) {
				throw new PExceptionProtocol("Unauthorized operation: No Pbinding found");
			}
			byte nextstate = PStateGraph.nextStatePBinding(
					pb.getStatus(), PBinding.ACTION_UNEXPORT);
			if (nextstate == PBinding.LIFECYCLE_ERROR) {
				throw new PExceptionProtocol("Unauthorized operation: status =" + pb.getStatus());
			}
			pb.setStatus(nextstate);
		} else {
			throw new PExceptionProtocol("Unauthorized operation: No Pbinding can be found without a CacheManager");
		}
	}

	public void unexport(Object conn, PName pn, Object hints) throws PException {
		if (pn.isNull()) {
			throw new PExceptionProtocol("[" + getClassName() + "]: cannot unexport with a null pname.");
		}
		PBindingCtrl pb = null;
		if (hints instanceof PBindingCtrl) {
			pb = (PBindingCtrl) hints;
		} else if (cache != null) {
			CacheEntry ce = cache.lookup(pn);
			if (ce != null) {
				pb = (PBindingCtrl) ce.getCeObject();
			}
		}
		if (pb == null) {
			throw new PExceptionProtocol("Unauthorized operation: No Pbinding found");
		}
		byte nextstate = PStateGraph.nextStatePBinding(
				pb.getStatus(), PBinding.ACTION_UNEXPORT);
		if (nextstate == PBinding.LIFECYCLE_ERROR) {
			throw new PExceptionProtocol("Unauthorized operation: status =" + pb.getStatus());
		}
		pb.setStatus(nextstate);
	}

	public long encodeLong(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		Long val = ((RdbSequencePName) pn).value;
		return (val == null ? -1 : val.longValue());
	}

	public Long encodeOlong(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		return ((RdbSequencePName) pn).value;
	}

	public PName decodeLong(long en) throws PExceptionNaming, UnsupportedOperationException {
		return decodeOlong(new Long(en));
	}
	
	public PName decodeOlong(Long en) throws PExceptionNaming, UnsupportedOperationException {
		if (en == RdbSequencePName.NULL_VALUE || (en != null && en.longValue() <0)) {
			return nullPName;
		} else {
			return new RdbSequencePName(this, en);
		}
	}

	public PName decodeString(String en) throws PExceptionNaming {
		Long val;
		if (en == null || en.equals("null")) {
			val = null;
		} else {
			val = new Long(Long.parseLong(en));
		}
		try {
			return decodeOlong(val);
		} catch (NumberFormatException e) {
			throw new PExceptionNaming(e, "[" + getClassName()
				+ "]: Impossible to decode the String to long: " + en);
		}
	}

	public boolean codingSupported(int codingtype) {
		switch (codingtype) {
		case CTLONG:
		case CTOLONG:
		case CTSTRING:
			return true;
		default:
			return false;
		}
	}

	public String encodeString(PName pn) throws PExceptionNaming {
		return "" + ((RdbSequencePName) pn).value;
	}
}
