/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.genclass.lib;

import org.objectweb.jorm.api.PAccessor;
import org.objectweb.jorm.api.PBinding;
import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.api.PMappingCallback;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PClassMappingCtrl;
import org.objectweb.jorm.api.PNameIterator;
import org.objectweb.jorm.metainfo.api.GenClass;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.naming.api.PBinder;
import org.objectweb.jorm.naming.api.PNameCoder;
import org.objectweb.jorm.genclass.api.FieldDesc;
import org.objectweb.jorm.genclass.api.GenClassProp;
import org.objectweb.jorm.type.api.PExceptionTyping;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import java.util.Iterator;
import java.util.Collections;

/**
 * This abstract class provides all the generic code for the PClassMapping
 * associated with JORM generic classes. It implements all methods except the
 * following ones: createPBinding, and isConform.
 * @author	P. Dechamboux
 */
public abstract class GenClassMapping
		implements PClassMapping, PClassMappingCtrl, GenClassProp, Loggable {
	/**
	 * This field references the mapper. The mapper is used to know the mapper
	 * name and the mapping name.
	 */
	protected PMappingCallback mapper = null;

	/**
	 * It references the meta-information associated with this generic class.
	 */
	protected GenClass genClassMetaInfo = null;

	/**
	 * This field is the PType of the involved generic class.
	 */
	protected PType genClassPType = null;

	/** Fully quailied name of the persitent class */
	protected String jormClassName = "";

	/**
	 * This array contains the ScalarField meta object elements which
	 * represent the index fields of the generic class instance.
	 */
	protected FieldDesc[] indexFields = null;

	/**
	 * This array contains the ScalarField meta object elements which
	 * represent the identifier fields of the generic class instance.
	 */
	protected FieldDesc[] identifierFields = null;

	/**
	 * This array contains the ScalarField meta object elements which
	 * represent the element fields of the generic class instance. If it is
	 * a reference, there must be at least two fields: one describing the
	 * reference field, and one describing its structural projection. This
	 * structural projection may be a composite name in which case it should
	 * describes all the fields of this composite name.
	 */
	protected FieldDesc[] elemFields = null;

	/**
	 * The actual PBinder.
	 */
	protected PBinder binder = null;

	protected PNameCoder classPNC;

	/**
	 * The PClassMapping of the generic class instance representing the element
	 * stored within this generic class object. This is valid only if the
	 * elements of this generic class object are also generic class objects.
	 */
	protected PClassMapping elemClassMapping = null;

	/**
	 * The naming context associated to the reference elements of this generic
	 * class instance.
	 */
	protected PNameCoder elemNameCoder = null;

	protected String projectName = null;

	/**
	 * The logger to log all information related to the mapping of this class.
	 */
	protected Logger logger = null;

	/**
	 * The loggerFactory.
	 */
	protected LoggerFactory loggerFactory = null;


	public FieldDesc[] getElemFields() {
		return elemFields;
	}

	public FieldDesc[] getIdentifierFields() {
		return identifierFields;
	}

	public FieldDesc[] getIndexFields() {
		return indexFields;
	}


	public String getProjectName() {
		return projectName;
	}

	public void setProjectName(String projectName) {
		this.projectName = projectName;
	}

	/**
	 * It retrieves the name of the class mapped by this PClassMapping. It
	 * should be the same as the one get when calling
	 * "getPType().getJormName()" if the PType exists.
	 */
	public String getClassName() {
		return jormClassName;
	}

	/**
	 * It gives access to the PClassMapping associated to the element of a
	 * generic class whose type is a generic class.
	 * @return	The PClassMapping of the generic class instance.
	 */
	public PClassMapping getGenClassMapping() {
		return elemClassMapping;
	}

	/**
	 * It gives access to the PClassMapping associated to a field whose type is
	 * a generic class. It has no meaning within this context.
	 * @param fn	The name of the generic class field.
	 * @return	The PClassMapping of the generic class instance: always null
	 * 			here.
	 */
	public PClassMapping getGenClassMapping(String fn) {
		return null;
	}

	/**
	 * It returns the Class/GenClass object representing the meta-information
	 * associated to the JORM class/generic class whose mapping is defined by
	 * this PClassMapping. It may be null if not requested at mapping time.
	 * @return	The Class/GenClass meta-object or null.
	 * @see	org.objectweb.jorm.api.PMapper#map
	 */
	public MetaObject getMetaInfo() {
		return genClassMetaInfo;
	}

	/**
	 * It yields the PBinder associated to this PClassMapping. As a PBinder
	 * itself, this PClassMapping delegates this behaviour to this PBinder.
	 * @return	The PBinder associated to this PClassMapping.
	 */
	public PBinder getPBinder() {
		return binder;
	}

	public PNameCoder getClassPNameCoder() {
		return classPNC;
	}

	/**
	 * It gives access to the mapper that has mapped the associated JORM class
	 * as defined by this PClassMapping.
	 * @return	The associated PMapper.
	 */
	public PMapper getPMapper() {
		return mapper;
	}

	/**
	 * It returns the naming context that manages the references contained by
	 * the generic class instance whose mapping is defined by this
	 * PClassMapping.
	 * @return	The naming context that manages references contained by a
	 * 			generic class instance.
	 * @exception	UnsupportedOperationException	It is raised when this
	 * 												PClassMapping does not map
	 * 												a generic class instance, or
	 * 												when the elements of this
	 * 												generic class instance are
	 * 												not references.
	 */
	public PNameCoder getPNameCoder()
			throws UnsupportedOperationException {
		return elemNameCoder;
	}

	/**
	 * It returns the naming context that manages the reference associated to
	 * a field whose type is a JORM class.
	 * @return	The naming context that manages the reference defined by a
	 * 			field.
	 * @exception	UnsupportedOperationException	It is raised when no naming
	 * 												context is associated to
	 * 												the corresponding field.
	 */
	public PNameCoder getPNameCoder(String fn)
			throws UnsupportedOperationException {
		throw new UnsupportedOperationException();
	}

	/**
	 * It intialises the class mapping. It especially assigns a mapper as well
	 * as a meta-object describing either a class or a generic class.
	 * @param mapper	The mapper to be assigned to this class mapping.
	 * @param metaclass	The meta-object describing the associated class. It may
	 * 					be null.
	 */
	public void init(PMappingCallback mapper, MetaObject metaclass) throws PException {
		this.mapper = mapper;
		genClassMetaInfo = (GenClass) metaclass;
	}

	public Iterator getPNameIterator(Object conn) throws PException {
		return getPNameIterator(conn, false, false, null);
	}

	/**
	 * It assigns the PBinder associated to this PClassMapping. As a PBinder
	 * itself, this PClassMapping delegates this behaviour to this PBinder.
	 * @param pb	The PBinder associated to this PClassMapping.
	 * @exception	PExceptionTyping	It is raised when the structure of the
	 * 									the name that identifies the DSI for
	 * 									the PClassMapping is not supported by
	 * 									this PBinder that should be connected
	 * 									to this PClassMapping.
	 */
	public void setPBinder(PBinder pb) throws PExceptionTyping {
		binder = pb;
	}

	public void setClassPNameCoder(PNameCoder pnc) throws PException {
		this.classPNC = pnc;
	}

	public void configureRefFields(PClassMapping.ReferenceConfigurator rc)
			throws PException, UnsupportedOperationException {
		throw new UnsupportedOperationException();
	}

	/**
	 * It assigns the PNameCoder used to manage the references hold by a
	 * generic class instance whose mapping is described by this PClassMapping.
	 * It should verify that the name structure for this reference is supported
	 * by the corresponding naming context.
	 * @param pnc	The PNameCoder to be associated to this generic class
	 * 				PClassMapping for managing its reference elements.
	 * @exception	PExceptionTyping	It is raised when the structure of the
	 * 									the names that reference the DSI stored
	 * 									within the generic class instance is not
	 * 									supported by this naming context.
	 * @exception	UnsupportedOperationException	It is raised when this
	 * 												PClassMapping does not map
	 * 												a generic class instance, or
	 * 												when the elements of this
	 * 												generic class instance are
	 * 												not references.
	 */
	public void setPNameCoder(PNameCoder pnc)
			throws PException, UnsupportedOperationException {
		elemNameCoder = pnc;
	}

	/**
	 * It assigns the PNameCoder used to manage the reference hold by a
	 * field whose type is a JORM class.
	 * It should verify that the name structure for this reference is supported
	 * by the corresponding naming context.
	 * @param fn	The name of the involved field.
	 * @param pnc	The PNameCoder to be associated to this field for
	 * 				managing the reference it may hold.
	 * @exception	PExceptionTyping	It is raised when the structure of the
	 * 									the names that reference the DSI stored
	 * 									within the generic class instance is not
	 * 									supported by this naming context.
	 * @exception	UnsupportedOperationException	It is raised when no naming
	 * 												context is associated to
	 * 												the corresponding field.
	 */
	public void setPNameCoder(String fn, PNameCoder pnc)
			throws PException, UnsupportedOperationException {
		throw new UnsupportedOperationException();
	}

	/**
	 * It assigns the GenClassMapping used to manage the reference hold by a
	 * field whose type is a JORM gen class.
	 */
	public void setGenClassMapping(PClassMapping pcm)
			throws UnsupportedOperationException {
		// TODO: verifies with the PType that the elements are gen class
		// themselves
		elemClassMapping = pcm;
	}

	/**
	 * It assigns the GenClassMapping used to manage the reference hold by a
	 * field whose type is a JORM class.
	 */
	public void setGenClassMapping(String fn, PClassMapping pcm)
			throws UnsupportedOperationException {
		throw new UnsupportedOperationException();
	}

	/**
	 * Retrieves the PType associated to the JORM class which is the type of
	 * the name managed by this naming context.
	 * @return	The related PType.
	 */
	public PType getPType() {
		return genClassPType;
	}

	/**
	 * It assigns the PType associated to the JORM class which is the type of
	 * the reference managed by this naming context.
	 * @param pt	The PType of the JORM class reference managed by this naming
	 * 				context.
	 */
	public void setPType(PType pt) {
		if (genClassPType == null) {
			genClassPType = pt;
			jormClassName += pt.getJormName();
			binder.setPType(pt);
			if (pt.getNestedPType() != null) {
				if (elemClassMapping != null) {
					// element is a genclass
					elemClassMapping.getPBinder().setPType(pt.getNestedPType());
				}
				if (pt.getNestedPType().getTypeCode() == PType.TYPECODE_REFERENCE) {
					elemNameCoder.setPType(pt.getNestedPType());
				}
			}
		}
	}
    public void read(PBinding pb, Object conn, PAccessor pa, Object txctx) throws PException {
        read(pb, conn, pa);
    }

    public void read(PBinding pb, Object conn, PAccessor pa, Object txctx, boolean forUpdate) throws PException {
        read(pb, conn, pa);
    }

	// IMPLMENTATION OF THE GenClassProp INTERFACE

	/**
	 * Defines the fields property specifying the fields composing the
	 * identifier structure of a generic class instance.
	 * @param fds	The array of descriptions of the associated fields.
	 */
	public void setIdentifierDesc(FieldDesc[] fds) {
		identifierFields = fds;
	}

	/**
	 * Defines the fields property specifying the fields composing the
	 * index structure of a generic class instance elements.
	 * @param fds	The array of descriptions of the associated fields.
	 */
	public void setIndexDesc(FieldDesc[] fds) {
		indexFields = fds;
	}

	/**
	 * Defines the fields property specifying the fields composing the
	 * element value structure of a generic class instance elements.
	 * @param fds	The array of descriptions of the associated fields.
	 */
	public void setElemDesc(FieldDesc[] fds) {
		elemFields = fds;
	}

	/**
	 * It assigns a logger factory that allows the creation of new loggers.
	 * @param lf	The LoggerFactory object to obtain new loggers.
	 */
	public void setLoggerFactory(LoggerFactory lf) {
		loggerFactory = lf;
	}

	/**
	 * It assigns a logger to an component that wants to log things.
	 * @param logger	The logger object.
	 */
	public void setLogger(Logger logger) {
		this.logger = logger;
	}

	public Logger getLogger() {
		return logger;
	}

	public LoggerFactory getLoggerFactory() {
		return loggerFactory;
	}
}
