/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2005 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.generator.lib;

import java.io.File;
import java.io.FileOutputStream;
import java.util.ArrayList;
import java.util.Iterator;

import org.objectweb.asm.ClassWriter;
import org.objectweb.jorm.api.PAccessor;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.compiler.api.JormCompilerParameter;
import org.objectweb.jorm.compiler.api.PExceptionCompiler;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.Reference;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.util.io.api.TargetHolder;
import org.objectweb.jorm.util.lib.StringReplace;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class is a generator of XXXAccessor objects. It is generic either the
 * mapper type. This generator use the velocity tools. The used template is
 * Accessor.vm . Associated to this template this generator builds a velocity
 * context which contains the following information:<br/>
 * <table border=1>
 * 	<tr><td>Key</td><td>Value</td></tr>
 * 	<tr><td>"class"</td><td>The reference to the Class meta object</td></tr>
 * 	<tr>
 * 		<td>"tools"</td>
 * 		<td>The reference to the <a href="CommonHelper.html">CommonHelper<a/></td>
 * </tr>
 * 	<tr>
 * 		<td>"header"</td>
 * 		<td>The file name of the common template which contains the header of
 * 			the generate files.</td>
 * </tr>
 * </table>
 */
public class AccessorGenerator extends CommonGenerator {

	/**
	 * This method generates a XAccessor file corresponding to the pod
	 * parameter in the directory parameter.
	 * @param co		The meta object interface.
	 * @param holder	The target holder which allows to create files.
	 * @param cp		This parameter permits to reach the compilation parameters.
	 */
	public void generate(Class clazz,
						 TargetHolder holder,
						 JormCompilerParameter cp) throws PException {
        boolean debug = logger != null && logger.isLoggable(BasicLevel.DEBUG);
        ClassWriter cw = new ClassWriter(true);
		String itfname = getAccessorItfName(clazz);

		ArrayList superInterfaces = new ArrayList();
		superInterfaces.add(getJVMClassName(PAccessor.class));
		Iterator it = clazz.getSuperClasses().iterator();
		if (it.hasNext()) {
			while(it.hasNext()) {
				superInterfaces.add(getAccessorItfName((Class) it.next()));
			}
		}
		if (debug) {
		    logger.log(BasicLevel.DEBUG, "Interface " + itfname 
		            + " extends " + superInterfaces);
		}
		
		cw.visit(V1_1, ACC_PUBLIC + ACC_ABSTRACT + ACC_INTERFACE, //access
		        itfname, //name
		        getJVMClassName(Object.class), //superName
				(String[]) superInterfaces.toArray(new String[superInterfaces.size()]), //interfaces
				itfname // sourcefile
				);
		for (Iterator fieldIt = clazz.getFields().iterator(); fieldIt.hasNext();) {
            TypedElement te = (TypedElement) fieldIt.next();
            StringBuffer setParams = new StringBuffer();
            StringBuffer getParams = new StringBuffer();
            String returnType;
            if (te instanceof Reference) {
                returnType = getJVMType(PName.class);
                setParams.append(returnType);
                getParams.append(getJVMType(Object.class));
                setParams.append(getParams);
            } else {
                returnType = getJVMType(te.getType());
                setParams.append(returnType);
            }
            String fn = upperFL(te.getName());
            //void paSetF1(String val)throws PException;
            //or void paSetRefToB(PName pn, Object connection)throws PException;
            cw.visitMethod(
                    ACC_PUBLIC + ACC_ABSTRACT, 
                    "paSet" + fn, 
                    "(" + setParams.toString() + ")V", 
                    new String[]{getJVMClassName(PException.class)},
                    null);
            //String paGetF1() throws PException;
            //or PName paGetRefToB(Object connection)throws PException;
            cw.visitMethod(
                    ACC_PUBLIC + ACC_ABSTRACT, 
                    "paGet" + fn, 
                    "(" + getParams.toString() + ")" + returnType, 
                    new String[]{getJVMClassName(PException.class)},
                    null);
        }        
		try {
			// Fetch a block with the fileName
		    String fileName = itfname;
		    fileName = StringReplace.replaceChar('/', File.separatorChar, fileName) + ".class";
		    FileOutputStream fos = holder.getFileOutputStream(fileName);
	        fos.write(cw.toByteArray());
	        fos.close();
		} catch (Exception e) {
			throw new PExceptionCompiler(e, "Problem while generating PAccessor.");
		}
    }
    public String getAccessorItfName(java.lang.Class clazz) {
		return getAccessorItfName(clazz.getName());
    }
    public String getAccessorItfName(Class clazz) {
		return getAccessorItfName(clazz.getFQName());
    }
    public String getAccessorItfName(String className) {
		return getJVMClassName(className + "Accessor");
    }    
}
