/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2005 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.generator.lib;

import java.io.File;
import java.io.FileOutputStream;
import java.util.Arrays;
import java.util.Iterator;

import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Label;
import org.objectweb.asm.Type;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.compiler.api.JormCompilerParameter;
import org.objectweb.jorm.compiler.api.PExceptionCompiler;
import org.objectweb.jorm.metainfo.api.CompositeName;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.naming.api.PExceptionNameCoding;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.naming.api.PNCStringCoder;
import org.objectweb.jorm.naming.api.PNameCoder;
import org.objectweb.jorm.naming.lib.BasicStringCoder;
import org.objectweb.jorm.naming.lib.CTHelper;
import org.objectweb.jorm.naming.lib.CompositePBinder;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.util.io.api.TargetHolder;
import org.objectweb.jorm.util.lib.StringReplace;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class is a generator of XXXBinder classes. It is generic either the
 * mapper type. This generator the ASM  tools.
 *  
 * @author S.Chassande-Barrioz
 */
public class BinderGenerator extends CommonGenerator {
    
    public final static String ACF_TYPE = "org.objectweb.jorm.facility.naming.generator.LongGen";

	/**
	 * This method generates a XPBinder file corresponding to the co
	 * parameter in the directory parameter.
	 * @param co		The meta object interface.
	 * @param holder	The target holder which allows to create files.
	 * @param cp		This parameter permits to reach the compilation parameters.
	 */
	public void generate(CompositeName co,
			 TargetHolder holder,
			 JormCompilerParameter cp) throws PException {
        final ClassWriter cv = new ClassWriter(true);
        final BinderGenContext gc = new BinderGenContext(co, cv, this);
        if (debug) {
            logger.log(BasicLevel.DEBUG, "Generate the class '"
                + gc.classToWrite + "'.");
        }
        //create the class
        cv.visit(V1_1, ACC_PUBLIC + ACC_SUPER,
                gc.classToWrite, gc.compositeBinderJCN, 
                new String[]{}, gc.classToWrite);
        //generate fields
        generateNullFields(gc);
        if (gc.hasAutoCalculatedField) {
            generateACFieldsNAccessors(gc);
        }
        //Constructors
        generateNoArgConstructor(gc);
        generatePCMConstructor(gc);
        //Methods
		if (gc.allAutoCalculated) {
		    generateIsAllAutoCalculatedMethod(gc);
		}
        generateGetPNameFromMethod(gc);
        generateIsNullMethod(gc);
        generateIsManagedPNameMethod(gc);
        generateNewPNameMethod(gc);
        generateCodingSupportedMethod(gc);
        generateSupportCompositeFieldMethod(gc);
        generateGetNullMethod(gc);
        if (gc.uniqueField != null) {
            generateEncodeXXXMethod(gc);            
            generateDecodeXXXMethod(gc);            
        }
        if (!gc.uniqueFieldIsString) {
            generateDecodeString(gc);
        } //else the decodeString method has been generated previously (generateDecodeXXXMethod)
        generateFillStringCoderMethod(gc);
        cv.visitEnd();
        
        try {
            // Fetch a block with the fileName
            String fileName = gc.classToWrite;
            fileName = StringReplace.replaceChar('/', File.separatorChar,
                    fileName)
                    + ".class";
            FileOutputStream fos = holder.getFileOutputStream(fileName);
            fos.write(cv.toByteArray());
            fos.close();
        } catch (Exception e) {
            throw new PExceptionCompiler(e,
                    "Problem while generating PAccessor.");
        }
	}

	private void generateNullFields(final BinderGenContext gc) {
		for (Iterator it = gc.coAllFields.iterator(); it.hasNext();) {
			final PrimitiveElement pe = (PrimitiveElement) it.next();
			//Generate field containing the null value
			gc.clav.visitField(ACC_PRIVATE, getNullFieldName(pe), 
			        getJVMType(pe.getType()), null, null);
		}
	}
	
	private void generateACFieldsNAccessors(final BinderGenContext gc) {
	    final String acfJVMType = getJVMType(ACF_TYPE);
		for (Iterator it = gc.coAllFields.iterator(); it.hasNext();) {
			final PrimitiveElement pe = (PrimitiveElement) it.next();
			if (pe.isAutoCalculated()) {
			    final String fn = getACFName(pe);
				//field declaration
			    gc.clav.visitField(ACC_PRIVATE, fn, acfJVMType, null, null);

			    final String upperfl = upperFL(pe.getName());
			    //Field getter
	            CodeVisitor cv = gc.clav.visitMethod(ACC_PUBLIC, 
	                    "getACF" + upperfl, "()" + acfJVMType, null, null);
	            cv.visitVarInsn(ALOAD, 0);
	            cv.visitFieldInsn(GETFIELD, gc.classToWrite, fn, acfJVMType);
	            cv.visitInsn(ARETURN);
	            cv.visitMaxs(0, 0);
				
	            //Field setter
	            cv = gc.clav.visitMethod(ACC_PUBLIC, "setACF" + upperfl, 
	                    "("  + acfJVMType + ")V", null, null);
	            cv.visitVarInsn(ALOAD, 0);
	            cv.visitVarInsn(ALOAD, 1);
	            cv.visitFieldInsn(PUTFIELD, gc.classToWrite, fn, acfJVMType);
	            cv.visitInsn(RETURN);
	            cv.visitMaxs(0, 0);
			}
		}
	}

	private void generateIsAllAutoCalculatedMethod(final BinderGenContext gc) {
	    final CodeVisitor cv = gc.clav.visitMethod(ACC_PROTECTED, 
	            "isAllAutoCalculated", "()Z", null, null);
	    cv.visitInsn(ICONST_1);
	    cv.visitInsn(IRETURN);
	    cv.visitMaxs(0, 0);
	}
	
	private void generateNoArgConstructor(final BinderGenContext gc) {
	    //public TotoBinder() {
        final CodeVisitor cv = gc.clav.visitMethod(ACC_PUBLIC, "<init>", "()V", null, null);
        //super()
        cv.visitVarInsn(ALOAD, 0);
        cv.visitMethodInsn(INVOKESPECIAL, 
                gc.compositeBinderJCN, "<init>", "()V");
		//initialisation of null fields
        generateNullFieldInit(gc, cv);
        cv.visitInsn(RETURN);
        cv.visitMaxs(0, 0);
	}
	
	private void generatePCMConstructor(final BinderGenContext gc) {
	    //public TotoBinder(PClassMapping pcm) {
        final CodeVisitor cv = gc.clav.visitMethod(ACC_PUBLIC, "<init>", 
                "(" + gc.pcmJT + ")V", null, null);
        //super(pcm)
        cv.visitVarInsn(ALOAD, 0);
        cv.visitVarInsn(ALOAD, 1);
        cv.visitMethodInsn(INVOKESPECIAL, 
                gc.compositeBinderJCN, "<init>", "(" + gc.pcmJT + ")V");
		//initialisation of null fields
        generateNullFieldInit(gc, cv);
        cv.visitInsn(RETURN);
        cv.visitMaxs(0, 0);
	}
	
	private void generateGetPNameFromMethod(final BinderGenContext gc)throws PException {
	    //protected PName getPNameFrom(Object png, Object ctx, Object connection) throws PException {
        CodeVisitor cv = gc.clav.visitMethod(ACC_PROTECTED, 
                "getPNameFrom", 
                "(" + gc.objectJT 	//png 1
                	+ gc.objectJT 	//ctx 2
                	+ gc.objectJT 	//connection 3
                	+ "Z" 			//computeACF 4
                	+ ")" + gc.pnameJT, 
                new String[]{gc.pexceptionJCN}, null);
        final int ctxRegisterIndex = 2;
        final int connRegisterIndex = 3;
        final int computeACFRegisterIndex = 4;
        if (gc.allAutoCalculated) {
            //When all fields are auto calculated and if the boolean flag
            // computeACF is true then no png is required.
            // So we generate here code only for this case
            Label labelElsePNGS = new Label();
            //if (computeACF) {
            cv.visitVarInsn(ILOAD, computeACFRegisterIndex);
	        cv.visitJumpInsn(IFEQ, labelElsePNGS);
	        {
	    	    //new TotoPName
	            cv.visitTypeInsn(NEW, gc.xpnameJCN);
	            cv.visitInsn(DUP);
	            for (Iterator it = gc.coAllFields.iterator(); it.hasNext();) {
	    			final PrimitiveElement pe = (PrimitiveElement) it.next();
	    			computeACFValue(gc, cv, connRegisterIndex, pe);
	            }
	            cv.visitVarInsn(ALOAD, 0);
	            cv.visitMethodInsn(INVOKESPECIAL, gc.xpnameJCN, "<init>", "(" 
	                    + gc.xconstructorFieldsPnc + ")V");
	            cv.visitInsn(ARETURN);
	        }
            cv.visitLabel(labelElsePNGS);
        }
        
        final int pngRegisterIndex = 5;
        //if (png instanceof  TotoPNG) {
        Label labelElsePNGG = new Label();
        cv.visitVarInsn(ALOAD, 1);
        cv.visitTypeInsn(INSTANCEOF, gc.xpngJCN);
        cv.visitJumpInsn(IFEQ, labelElsePNGG);
        {
            //generate code for building PName from a specific pname getter
            //TotoPNG spng = (TotoPNG) png;
            cv.visitVarInsn(ALOAD, 1);
	        cv.visitTypeInsn(CHECKCAST, gc.xpngJCN);
            cv.visitVarInsn(ASTORE, pngRegisterIndex);
            //return new TotoPName(..., this);
            cv.visitTypeInsn(NEW, gc.xpnameJCN);
            cv.visitInsn(DUP);
            fillStackWithPNConstParam(gc, cv, true, pngRegisterIndex, 
                    ctxRegisterIndex, connRegisterIndex, 
                    computeACFRegisterIndex);
            cv.visitMethodInsn(INVOKESPECIAL, gc.xpnameJCN, "<init>", "(" 
                    + gc.xconstructorFieldsPnc + ")V");
            cv.visitInsn(ARETURN);
        }
        
        //} else if (png instanceof  PNameGetter) {
        cv.visitLabel(labelElsePNGG);
        cv.visitVarInsn(ALOAD, 1);
        cv.visitTypeInsn(INSTANCEOF, gc.pngJCN);
        Label labelElse = new Label();
        cv.visitJumpInsn(IFEQ, labelElse);
        {
            //generate code for building PName from a generic pname getter
            //PNameGetter gpng = (PNameGetter) png;
            cv.visitVarInsn(ALOAD, 1);
	        cv.visitTypeInsn(CHECKCAST, gc.pngJCN);
            cv.visitVarInsn(ASTORE, pngRegisterIndex);
            //return new TotoPName(..., this);
            cv.visitTypeInsn(NEW, gc.xpnameJCN);
            cv.visitInsn(DUP);
            fillStackWithPNConstParam(gc, cv, false, pngRegisterIndex, 
                    ctxRegisterIndex, connRegisterIndex, 
                    computeACFRegisterIndex);
            cv.visitMethodInsn(INVOKESPECIAL, gc.xpnameJCN, "<init>", "(" 
                    + gc.xconstructorFieldsPnc + ")V");
            cv.visitInsn(ARETURN);
        }
        //} else {
        cv.visitLabel(labelElse);
        {
            //generate code rerturning null because the specified png cannot
            // be used.
            cv.visitInsn(ACONST_NULL);
            cv.visitInsn(ARETURN);
        }
        cv.visitMaxs(0, 0);
	}

	private void generateIsNullMethod(final BinderGenContext gc)throws PException {
	    //protected boolean isNull(Object png, Object ctx) throws PException
        final CodeVisitor cv = gc.clav.visitMethod(ACC_PROTECTED, 
                "isNull", 
                "(" + gc.objectJT //png
                	+ gc.objectJT //ctx
                	+ ")Z", 
                new String[]{gc.pexceptionJCN}, null);
        final Label labelFalse = new Label();
        //if (png instanceof  TotoPNG) {
        final Label labelElsePNG = new Label();
        cv.visitVarInsn(ALOAD, 1);
        cv.visitTypeInsn(INSTANCEOF, gc.xpngJCN);
        cv.visitJumpInsn(IFEQ, labelElsePNG);
        {
            //TotoPNG spng = (TotoPNG) png;
            cv.visitVarInsn(ALOAD, 1);
	        cv.visitTypeInsn(CHECKCAST, gc.xpngJCN);
            cv.visitVarInsn(ASTORE, 3);
            
            //return spng.pnGetF1(ctx) == jormf1;
            comparePNGwithNullValue(gc, cv, true, 3, 2, labelFalse);
            cv.visitInsn(ICONST_1);
            cv.visitInsn(IRETURN);
        }
        //} else if (png instanceof  PNameGetter) {
        cv.visitLabel(labelElsePNG);
        cv.visitVarInsn(ALOAD, 1);
        cv.visitTypeInsn(INSTANCEOF, gc.pngJCN);
        final Label labelElse = new Label();
        cv.visitJumpInsn(IFEQ, labelElse);
        {
            //PNameGetter gpng = (PNameGetter) png;
            cv.visitVarInsn(ALOAD, 1);
	        cv.visitTypeInsn(CHECKCAST, gc.pngJCN);
            cv.visitVarInsn(ASTORE, 3);

            //return gpng.pnGetChar("f1", ctx) == jormf1;
            comparePNGwithNullValue(gc, cv, false, 3, 2, labelFalse);
            cv.visitInsn(ICONST_1);
            cv.visitInsn(IRETURN);
        }
        //} else {
        cv.visitLabel(labelElse);
        {
            final String exceptionName = getJVMClassName(PExceptionNaming.class);
            //throw new PExceptionNaming();
            cv.visitTypeInsn(NEW, exceptionName);
            cv.visitInsn(DUP);
            cv.visitTypeInsn(NEW, gc.stringBufferJCN);
            cv.visitInsn(DUP);
            cv.visitMethodInsn(INVOKESPECIAL, gc.stringBufferJCN, "<init>", "()V");
            cv.visitLdcInsn("A PNamegetter (generic of specific) is required to decode this object:");
            cv.visitMethodInsn(INVOKEVIRTUAL, gc.stringBufferJCN, "append", 
                    "(" + gc.stringJT + ")" 
                    + gc.stringBufferJT);
            cv.visitVarInsn(ALOAD, 1);
            cv.visitMethodInsn(INVOKEVIRTUAL, gc.stringBufferJCN, "append", 
                    "(" + gc.objectJT + ")" + gc.stringBufferJT);
            cv.visitMethodInsn(INVOKEVIRTUAL, gc.stringBufferJCN, "toString", "()" 
                    + gc.stringJT);
            
            cv.visitMethodInsn(INVOKESPECIAL, exceptionName,
                    "<init>", "(" + gc.stringJT+ ")V");
            cv.visitInsn(ATHROW);
        }
        cv.visitLabel(labelFalse);
        cv.visitInsn(ICONST_0);
        cv.visitInsn(IRETURN);
        cv.visitMaxs(0, 0);
	}

	private void generateIsManagedPNameMethod(final BinderGenContext gc) {
	    //protected boolean isManagedPName(Object pn) {
	    final CodeVisitor cv = gc.clav.visitMethod(ACC_PROTECTED, "isManagedPName", 
	            "(" + gc.objectJT + ")Z", null, null);
	    //return pn instanceof TotoPName;
	    cv.visitVarInsn(ALOAD, 1);
	    cv.visitTypeInsn(INSTANCEOF, gc.xpnameJCN);
	    cv.visitInsn(IRETURN);
	    cv.visitMaxs(0, 0);
	}

	private void generateNewPNameMethod(final BinderGenContext gc) {
	    //protected PName newPName(PName oldpn) throws PException {
	    final CodeVisitor cv = gc.clav.visitMethod(ACC_PROTECTED, "newPName", 
	            "(" + gc.pnameJT + ")" + gc.pnameJT, 
	            new String[] {gc.pexceptionJCN}, null);
	    //return new TotoPName(this, (PName) oldpn);
	    cv.visitTypeInsn(NEW, gc.xpnameJCN);
	    cv.visitInsn(DUP);
	    cv.visitVarInsn(ALOAD, 0);
	    cv.visitVarInsn(ALOAD, 1);
	    cv.visitMethodInsn(INVOKESPECIAL, gc.xpnameJCN, "<init>", 
	            "(" + gc.pnameManagerJT + gc.pnameJT + ")V");
	    cv.visitInsn(ARETURN);
	    cv.visitMaxs(0, 0);    
	}
	
	private void generateCodingSupportedMethod(final BinderGenContext gc) {
	    //public boolean codingSupported(int codingtype) {
	    final CodeVisitor cv = gc.clav.visitMethod(ACC_PUBLIC, "codingSupported", "(I)Z", null, null);

	    //switch(tc) {
	    cv.visitVarInsn(ILOAD, 1);
	    final Label labelTrue = new Label();
	    final Label labelFalse = new Label();
	    final int[] values; // the array of the values used in the switch
	    if (gc.uniqueField == null || gc.uniqueFieldIsString) {
	        values = new int[2];
	    } else {
	        // the pname is composed of an unique field ==> allow the particular
	        // coding (be careful, String is always allowed)
	        values = new int[3];
	        //case PNameCoder.CTXXX
	        values[2] = CTHelper.ptc2ct(gc.uniqueField.getType().getTypeCode());
	    }
	    //case PNameCoder.CTSTRING
        values[0] = PNameCoder.CTSTRING;
	    //case PNameCoder.CTCOMPOSITE
        values[1] = PNameCoder.CTCOMPOSITE;
        Arrays.sort(values);
        // labels: the label corresponding to the values
        //if the value is values[i] goto to labels[i].
        //otherwise goto the default label
        //Here there are two case only:
        // - allowed coding ==> goto labelTrue
        // - otherwise ==> goto labelFalse
	    final Label[] labels = new Label[values.length]; 
	    Arrays.fill(labels, labelTrue);
	    cv.visitLookupSwitchInsn(labelFalse, values, labels);
	    //return true
	    cv.visitLabel(labelTrue);
	    cv.visitInsn(ICONST_1);
	    cv.visitInsn(IRETURN);
	    
	    //return false
	    cv.visitLabel(labelFalse);
	    cv.visitInsn(ICONST_0);
	    cv.visitInsn(IRETURN);

	    cv.visitMaxs(0, 0);
	}

	private void generateSupportCompositeFieldMethod(final BinderGenContext gc) {
	    //public boolean supportCompositeField(String fn, PType ft) {
	    final CodeVisitor cv = gc.clav.visitMethod(ACC_PUBLIC, 
	            "supportCompositeField", "(" + gc.stringJT + gc.ptypeJT + ")Z", null, null);
		for (Iterator it = gc.coAllFields.iterator(); it.hasNext();) {
			final PrimitiveElement pe = (PrimitiveElement) it.next();
		    final Label labelNextTest = new Label();
		    //if ("f1".equals(fn)) {
		    cv.visitLdcInsn(pe.getName());
		    cv.visitVarInsn(ALOAD, 1);
		    cv.visitMethodInsn(INVOKEVIRTUAL, gc.stringJCN, 
		            "equals", "(" + gc.objectJT + ")Z");
		    cv.visitJumpInsn(IFNE, labelNextTest);
		    {
			    // return ft.isa(PTypeSpace.CHAR);
		        cv.visitVarInsn(ALOAD, 2);
		        cv.visitFieldInsn(GETSTATIC, gc.ptypespaceJCN, 
		                getPTypeSpaceFieldName(pe.getType()), gc.ptypeJT);
		        cv.visitMethodInsn(INVOKEVIRTUAL, gc.ptypeJCN, "isa", 
		            "(" + gc.ptypeJT + ")Z");
		        cv.visitInsn(IRETURN);
		    }
		    cv.visitLabel(labelNextTest);
		}
		//return false;
	    cv.visitInsn(ICONST_0);
	    cv.visitInsn(IRETURN);
	    cv.visitMaxs(0, 0);
	}
	
	private void generateGetNullMethod(final BinderGenContext gc) {
	    //public PName getNull() {
	    final CodeVisitor cv = gc.clav.visitMethod(ACC_PUBLIC, "getNull", 
	            "()" + gc.pnameJT, null, null);
	    cv.visitVarInsn(ALOAD, 0);
	    cv.visitFieldInsn(GETFIELD, gc.classToWrite, "nullPName", gc.pnameJT);
	    final Label l1 = new Label();
	    cv.visitJumpInsn(IFNONNULL, l1);
	    cv.visitVarInsn(ALOAD, 0);
	    cv.visitVarInsn(ALOAD, 0);
	    cv.visitMethodInsn(INVOKESTATIC, gc.xpnameJCN, 
	            "nullPName", "(" + gc.pnameManagerJT + ")" + gc.pnameJT);
	    cv.visitFieldInsn(PUTFIELD, gc.classToWrite, "nullPName", gc.pnameJT);
	    cv.visitLabel(l1);
	    cv.visitVarInsn(ALOAD, 0);
	    cv.visitFieldInsn(GETFIELD, gc.classToWrite, "nullPName", gc.pnameJT);
	    cv.visitInsn(ARETURN);
	    cv.visitMaxs(0, 0);
	}
	
	private void generateEncodeXXXMethod(final BinderGenContext gc)throws PException {
	    //public long encodeLong(PName pn) throws PExceptionNaming, UnsupportedOperationException {
	    final PType type = gc.uniqueField.getType();
	    final String jvmFieldType = getJVMType(type);
		final Type asmType = Type.getType(type.getJavaClass());
	    final CodeVisitor cv = gc.clav.visitMethod(ACC_PUBLIC, 
	            getPNameEncodeFunction(type), 
	            "(" +  gc.pnameJT + ")" + jvmFieldType, 
	            new String[] { 
	            getJVMClassName(PExceptionNaming.class), 
	            getJVMClassName(UnsupportedOperationException.class)}, null);
	    //return ((TotoPName) pn).f1;
	    cv.visitVarInsn(ALOAD, 1);
	    cv.visitTypeInsn(CHECKCAST, gc.xpnameJCN);
	    cv.visitFieldInsn(GETFIELD, gc.xpnameJCN, gc.uniqueField.getName(), 
	            jvmFieldType);
	    cv.visitInsn(asmType.getOpcode(IRETURN));
	    cv.visitMaxs(0, 0);
	}

	private void generateDecodeXXXMethod(final BinderGenContext gc)throws PException {
	    //public PName decodeLong(long val) throws PExceptionNaming, UnsupportedOperationException {
	    final PType type = gc.uniqueField.getType();
	    final String jvmFieldType = getJVMType(type);
		final Type asmType = Type.getType(type.getJavaClass());
	    final int loadInstr = asmType.getOpcode(ILOAD);
	    final CodeVisitor cv = gc.clav.visitMethod(ACC_PUBLIC, 
	            getPNameDecodeFunction(type), 
	            "(" +  jvmFieldType + ")" + gc.pnameJT, 
	            new String[] { 
	            getJVMClassName(PExceptionNaming.class), 
	            getJVMClassName(UnsupportedOperationException.class)}, null);
	    int nullvalRegister = 2;
	    final boolean isObjectType = isObjectType(type);
	    if (type.getTypeCode() == PType.TYPECODE_LONG 
	            || type.getTypeCode() == PType.TYPECODE_DOUBLE) {
	        nullvalRegister ++;
	    }
	    // char nullval = getNull().encodeChar();
	    cv.visitVarInsn(ALOAD, 0);
	    cv.visitMethodInsn(INVOKEVIRTUAL, gc.classToWrite, "getNull", "()" + gc.pnameJT);
	    cv.visitMethodInsn(INVOKEINTERFACE, gc.pnameJCN, 
	            getPNameEncodeFunction(type), "()" + jvmFieldType);
	    cv.visitVarInsn(asmType.getOpcode(ISTORE), nullvalRegister);
	    
	    //compare both values (param and the local variable (nullvalRegister))
	    final Label labelNoNulPName = new Label();
        if (isObjectType) {
    	    final Label labelNulPName = new Label();
            //if (nullval == null)
            Label labelNullCase = new Label(); 
    	    cv.visitVarInsn(loadInstr, nullvalRegister);
            cv.visitJumpInsn(IFNULL, labelNullCase);
            {
                // nullval.equals(param)
        	    cv.visitVarInsn(loadInstr, nullvalRegister);
        	    cv.visitVarInsn(loadInstr, 1);
    	        cv.visitMethodInsn(INVOKEVIRTUAL, 
    	                getJVMClassName(type.getJavaClass()), "equals", 
    	                "(" + getJVMType(Object.class)+ ")Z");
	            cv.visitJumpInsn(IFNE, labelNulPName);
		        cv.visitJumpInsn(GOTO, labelNoNulPName);
            }
            // else
            cv.visitLabel(labelNullCase);
            {
                // param == null
        	    cv.visitVarInsn(loadInstr, 1);
	            cv.visitJumpInsn(IFNULL, labelNulPName);
		        cv.visitJumpInsn(GOTO, labelNoNulPName);
            }
    	    cv.visitLabel(labelNulPName);
        } else {
            //nullval == param
    	    cv.visitVarInsn(loadInstr, nullvalRegister);
    	    cv.visitVarInsn(loadInstr, 1);
			switch (type.getTypeCode()) {
			case PType.TYPECODE_BYTE:
			case PType.TYPECODE_CHAR:
			case PType.TYPECODE_SHORT:
			case PType.TYPECODE_INT:
    	        cv.visitJumpInsn(IF_ICMPNE, labelNoNulPName);
				break;
			case PType.TYPECODE_LONG:
			case PType.TYPECODE_DOUBLE:
    	        cv.visitInsn(LCMP);
    	        cv.visitJumpInsn(IFNE, labelNoNulPName);
				break;
			}                
        }
        
        //return getNull();
        cv.visitVarInsn(ALOAD, 0);
        cv.visitMethodInsn(INVOKEVIRTUAL, gc.classToWrite, "getNull", "()" + gc.pnameJT);
        cv.visitInsn(ARETURN);

	    // return new TotoPName(param, this);
	    cv.visitLabel(labelNoNulPName);
	    cv.visitTypeInsn(NEW, gc.xpnameJCN);
	    cv.visitInsn(DUP);
	    
	    cv.visitVarInsn(loadInstr, 1);
	    cv.visitVarInsn(ALOAD, 0);
	    cv.visitMethodInsn(INVOKESPECIAL, gc.xpnameJCN, "<init>", 
	            "(" + jvmFieldType + gc.pnameManagerJT + ")V");
	    cv.visitInsn(ARETURN);
	    cv.visitMaxs(0, 0);
	}

	private void generateDecodeString(final BinderGenContext gc) {
	    //public PName decodeString(String en) throws PExceptionNaming {
	    final CodeVisitor cv = gc.clav.visitMethod(ACC_PUBLIC, "decodeString", 
	            "("  + gc.stringJT + ")" + gc.pnameJT, 
	            new String[] { getJVMClassName(PExceptionNaming.class)}, 
	            null);
	    final Label beginTry = new Label();
	    //try {
	    cv.visitLabel(beginTry);
	    String bsc = getJVMClassName(BasicStringCoder.class);
	    //		PNCStringCoder cnsc = new BasicStringCoder(param);
	    cv.visitTypeInsn(NEW, bsc);
	    cv.visitInsn(DUP);
	    cv.visitVarInsn(ALOAD, 1);
	    cv.visitMethodInsn(INVOKESPECIAL, bsc, "<init>", "(" + gc.stringJT + ")V");
	    cv.visitVarInsn(ASTORE, 2);
	    
	    // 		TotoPName pn = new TotoPName(this);
	    cv.visitTypeInsn(NEW, gc.xpnameJCN);
	    cv.visitInsn(DUP);
	    cv.visitVarInsn(ALOAD, 0);
	    cv.visitMethodInsn(INVOKESPECIAL, gc.xpnameJCN, "<init>", 
	            "(" + gc.pnameManagerJT + ")V");
	    cv.visitVarInsn(ASTORE, 3);
	    
	    final String pncsc = getJVMClassName(PNCStringCoder.class);
        for (final Iterator it = gc.coAllFields.iterator(); it.hasNext();) {
            //	pn.f1 = cnsc.getChar();
			final PrimitiveElement pe = (PrimitiveElement) it.next();
		    final PType type = pe.getType();
		    final String jvmFieldType = getJVMType(type);
            cv.visitVarInsn(ALOAD, 3);
            cv.visitVarInsn(ALOAD, 2);
            cv.visitMethodInsn(INVOKEINTERFACE, pncsc, 
                    "get" + getCoderName(type), "()" + jvmFieldType);
            cv.visitFieldInsn(PUTFIELD, gc.xpnameJCN, pe.getName(), jvmFieldType);
        }
	    final Label endTry = new Label();
	    cv.visitLabel(endTry);
	    final Label endCatch = new Label();
	    cv.visitJumpInsn(GOTO, endCatch);
	    // } catch (PExceptionNaming) {
	    final Label beginCatch = new Label();
	    {
		    cv.visitLabel(beginCatch);
		    cv.visitVarInsn(ASTORE, 4);
		    final String pen = getJVMClassName(PExceptionNaming.class);
		    cv.visitTypeInsn(NEW, pen);
		    cv.visitInsn(DUP);
		    cv.visitVarInsn(ALOAD, 4);
		    cv.visitLdcInsn("Wrong name structure!!");
		    cv.visitMethodInsn(INVOKESPECIAL, pen, "<init>", 
		            "(" + getJVMType(Exception.class) + gc.stringJT + ")V");
		    cv.visitInsn(ATHROW);
	    }
	    // } //end catch
	    cv.visitLabel(endCatch);
        //		return pn;
	    cv.visitVarInsn(ALOAD, 3);
	    cv.visitInsn(ARETURN);
	    cv.visitTryCatchBlock(beginTry, endTry, beginCatch, gc.pexceptionJCN);
	    cv.visitMaxs(0, 0);

	}
	
	private void generateFillStringCoderMethod(final BinderGenContext gc) {
	    //protected void fillStringCoder(PNCStringCoder cnsc, PName pn) throws PExceptionNameCoding {
	    final CodeVisitor cv = gc.clav.visitMethod(ACC_PROTECTED, 
	            "fillStringCoder", 
	            "(" + getJVMType(PNCStringCoder.class) + gc.pnameJT + ")V", 
	            new String[] {getJVMClassName(PExceptionNameCoding.class)}, 
	            null);
	    if (!gc.uniqueFieldIsString) {
	        //TotoPName rpn = (TotoPName) pn;
		    cv.visitVarInsn(ALOAD, 2);
		    cv.visitTypeInsn(CHECKCAST, gc.xpnameJCN);
		    cv.visitVarInsn(ASTORE, 3);
		    final String pncsc = getJVMClassName(PNCStringCoder.class);
	        for (Iterator it = gc.coAllFields.iterator(); it.hasNext();) {
				final PrimitiveElement pe = (PrimitiveElement) it.next();
			    final PType type = pe.getType();
			    final String jvmFieldType = getJVMType(type);
			    //cnsc.putLong(rpn.f1);
				cv.visitVarInsn(ALOAD, 1);
				cv.visitVarInsn(ALOAD, 3);
				cv.visitFieldInsn(GETFIELD, gc.xpnameJCN, pe.getName(), jvmFieldType);
				cv.visitMethodInsn(INVOKEINTERFACE, pncsc, 
				        "put" + getCoderName(type), "(" + jvmFieldType + ")V");
	        }
        }
	    cv.visitInsn(RETURN);
	    cv.visitMaxs(0, 0);
	}

	/**
	 * Generates byte initializing the null field of the generated Binder.
	 *  
	 * @param cv is the visitor of the method where the code has to be generated
	 */
	private void generateNullFieldInit(final BinderGenContext gc, final CodeVisitor  cv) {
        for (Iterator it = gc.coAllFields.iterator(); it.hasNext();) {
			final PrimitiveElement pe = (PrimitiveElement) it.next();
	        //this.jormf1 = -1;
			cv.visitVarInsn(ALOAD, 0);
	        loadNullValue(pe, cv);
	        cv.visitFieldInsn(PUTFIELD, gc.classToWrite, getNullFieldName(pe), 
	                getJVMType(pe.getType()));
		}
	}
	
	/**
	 * This method generates byte code to put on the stack the values required
	 * for the XXXPname constrcutor based on null values and the PNameManger. 
	 *  
	 * @param cv is the visitor of the method where the code has to be generated
	 * @param specificPNG indicates the value loading has to be done with the
	 * specific/generated pname getter (XXXPNG). A false value indicates that
	 * the generic pname getter (PNameGetter) has to be used.
	 * @param pngRegisterIndex is the register number where the reference to 
	 * the pname getter (generic or specific) is stored. 
	 * @param ctxRegisterIndex is the register numbere where the reference to 
	 * the context is stored. The context is a parameter of the pname getter 
	 * methods (generic or specific).
	 * @param connRegisterIndex is the register number where the reference to
	 * the connection is stored. This connection is used for auto calculated
	 * fields.
	 * @param computeACFRegisterIndex is the register number where the boolean
	 * flag indicates if the ACF (autoCalculatedField) must be computed from 
	 * generator (true value) or if the value must be fetch from the pname 
	 * getter (false value).
	 */
	private void fillStackWithPNConstParam(final BinderGenContext gc,
	        final CodeVisitor cv, 
	        final boolean specificPNG,
	        final int pngRegisterIndex,
	        final int ctxRegisterIndex,
	        final int connRegisterIndex,
	        final int computeACFRegisterIndex) throws PException {
        for (Iterator it = gc.coAllFields.iterator(); it.hasNext();) {
			final PrimitiveElement pe = (PrimitiveElement) it.next();
			if (pe.isAutoCalculated()) {
	            Label labelElse = new Label();
	            Label labelEnd = new Label();
	            cv.visitVarInsn(ILOAD, computeACFRegisterIndex);
    	        cv.visitJumpInsn(IFEQ, labelElse);

    	        computeACFValue(gc, cv, connRegisterIndex, pe);
    	        cv.visitJumpInsn(GOTO, labelEnd);
    	        
    	        //Else: the value is pick up from the png
		        cv.visitLabel(labelElse);
    	        loadValuesFromPNG(gc, cv, specificPNG, pngRegisterIndex, 
    	                ctxRegisterIndex, connRegisterIndex, pe);
		        
	    	    cv.visitLabel(labelEnd);
			} else {
			    loadValuesFromPNG(gc, cv, specificPNG, pngRegisterIndex, 
			            ctxRegisterIndex, connRegisterIndex, pe);
			}
		}
        cv.visitVarInsn(ALOAD, 0);
	}
	
	/**
	 * generate byte code for computing an autocalculted field. This method 
     * supports int, long, Integer and Long auto calculated field only.
     * 
	 * @param cv is the codeVisitor to use
	 * @param connRegisterIndex is the register number where the reference to
	 * the connection is stored. This connection is used for auto calculated
	 * @param pe is the JORM meta object representing the auto calculated field
	 * of a composite name.
	 */
   	private void computeACFValue(final BinderGenContext gc, 
   	        final CodeVisitor cv, 
   	        final int connRegisterIndex,
   	        final PrimitiveElement pe) throws PException {
        int tc = pe.getType().getTypeCode();
        String classToInstanciate = null;
        switch (tc) {
        case PType.TYPECODE_OBJINT: 
        case PType.TYPECODE_OBJLONG:
            // in case of the field is an object like, we have to create it before
            // compute the primitive value.
            // new Integer(...)
            classToInstanciate = getJVMClassName(pe.getType().getJavaClass());
            cv.visitTypeInsn(NEW, classToInstanciate);
            cv.visitInsn(DUP);
            break;
        }

        //f1JormLongGen.genId(connection)
        cv.visitVarInsn(ALOAD, 0);
        cv.visitFieldInsn(GETFIELD, gc.classToWrite, 
                getACFName(pe), getJVMType(ACF_TYPE));
        cv.visitVarInsn(ALOAD, connRegisterIndex);
        cv.visitMethodInsn(INVOKEINTERFACE, getJVMClassName(ACF_TYPE), "genId", 
                "(" + gc.objectJT + ")J");
        
        //finish to convert the long into the asked type
        switch (tc) {
        case PType.TYPECODE_INT: // long to int
            cv.visitInsn(L2I);
            break;
        case PType.TYPECODE_LONG:
            //nothing more to do
            break;
        case PType.TYPECODE_OBJINT: //long to java.lang.Integer
            cv.visitInsn(L2I);
            cv.visitMethodInsn(INVOKESPECIAL, classToInstanciate, "<init>", "(I)V");
            break;
        case PType.TYPECODE_OBJLONG: //lonog to java.lang.Long
            cv.visitMethodInsn(INVOKESPECIAL, classToInstanciate, "<init>", "(J)V");
            break;
        default:
            throw new PException(
                "Auto calculated field not supported for the field type: "
                    + pe.getType().getJavaName());
        }
   	}
	
   	/**
   	 * generate byte code for loading a value from a pname getter (generic or 
   	 * specific).
   	 * 
	 * @param cv is the visitor of the method where the code has to be generated
	 * @param specificPNG indicates the value loading has to be done with the
	 * specific/generated pname getter (XXXPNG). A false value indicates that
	 * the generic pname getter (PNameGetter) has to be used.
	 * @param pngRegisterIndex is the register number where the reference to 
	 * the pname getter (generic or specific) is stored. 
	 * @param ctxRegisterIndex is the register numbere where the reference to 
	 * the context is stored. The context is a parameter of the pname getter 
	 * methods (generic or specific).
	 * @param connRegisterIndex is the register number where the reference to
	 * the connection is stored. This connection is used for auto calculated
	 * fields.
	 * @param pe is the JORM meta object representing the field of a composite 
	 * name.
   	 * @throws PException
   	 */
	private void loadValuesFromPNG(final BinderGenContext gc,
	        final CodeVisitor cv, 
	        final boolean specificPNG,
	        final int pngRegisterIndex,
	        final int ctxRegisterIndex,
	        final int connRegisterIndex,
	        final PrimitiveElement pe) throws PException {
		final String jvmFieldType = getJVMType(pe.getType());
	    if (specificPNG) {
		    //spng.pnGetF1(ctx)
            cv.visitVarInsn(ALOAD, pngRegisterIndex);
            cv.visitVarInsn(ALOAD, ctxRegisterIndex);
            cv.visitMethodInsn(INVOKEINTERFACE, gc.xpngJCN,
                    "pnGet" + upperFL(pe.getName()),
                    "(" + gc.objectJT + ")" + jvmFieldType);
		} else {
		    //gpng.pnGetChar("f1", ctx)
            cv.visitVarInsn(ALOAD, pngRegisterIndex);
            cv.visitLdcInsn(pe.getName());
            cv.visitVarInsn(ALOAD, ctxRegisterIndex);
            cv.visitMethodInsn(INVOKEINTERFACE,gc.pngJCN,
                    getPNameGetterGetFunction(pe.getType()),
                    "(" + gc.stringJT + gc.objectJT + ")" + jvmFieldType);
		}	    
	}
	
	/**
	 * This method generates byte code to compare the fields of the generated
	 * binder to a pname getter. The fields of the generated binder represents
	 * the null values. This method manages specific or generic pname getter.
	 * If the test passes the next instruction is immaediatly after. If the test
	 * fails, this method generates a jump a given label.
	 *  
	 * @param cv is the visitor of the method where the code has to be generated
	 * @param specificPNG indicates the value loading has to be done with the
	 * specific/generated pname getter (XXXPNG). A false value indicates that
	 * the generic pname getter (PNameGetter) has to be used.
	 * @param pngRegisterIndex is the register number where the reference to 
	 * the pname getter (generic or specific) is stored. 
	 * @param ctxRegisterIndex is the register numbere where the reference to 
	 * the context is stored. The context is a parameter of the pname getter 
	 * methods (generic or specific).
	 * @param labelFalse is the label to jump in case of the values are not the
	 * same.
	 * @throws PException
	 */
	private void comparePNGwithNullValue(final BinderGenContext gc,
	        final CodeVisitor cv, 
	        final boolean specificPNG,
	        final int pngRegisterIndex,
	        final int ctxRegisterIndex,
	        final Label labelFalse
	        ) throws PException {
        for (Iterator it = gc.coAllFields.iterator(); it.hasNext();) {
			final PrimitiveElement pe = (PrimitiveElement) it.next();
			final String fn = getNullFieldName(pe);
    		final String jvmFieldType = getJVMType(pe.getType());
            final Label labelNextTest = new Label(); 
            if (isObjectType(pe.getType())) {
	            //if (jormF1 == null)
	            Label labelNullCase = new Label(); 
	            cv.visitVarInsn(ALOAD, 0);
	            cv.visitFieldInsn(GETFIELD, gc.classToWrite, fn, jvmFieldType);
	            cv.visitJumpInsn(IFNULL, labelNullCase);
	            {
	                //the field jormf1 is NOT null
	                // && jormf1.equals(spng.pnGetF1(ctx))
	                cv.visitVarInsn(ALOAD, 0);
	                cv.visitFieldInsn(GETFIELD, gc.classToWrite, fn, jvmFieldType);
	                loadValueFromPNG(gc, pe, cv, specificPNG, pngRegisterIndex, ctxRegisterIndex);
	    	        cv.visitMethodInsn(INVOKEVIRTUAL, getJVMClassName(pe.getType().getJavaClass()), 
	    	                "equals", 
	    	                "(" + getJVMType(Object.class)+ ")Z");
		            cv.visitJumpInsn(IFEQ, labelFalse);
			        cv.visitJumpInsn(GOTO, labelNextTest);
	            }
	            // else
	            cv.visitLabel(labelNullCase);
	            {
	                //the field jormf1 is null
	                // && spng.pnGetF1(ctx) == null
	                loadValueFromPNG(gc, pe, cv, specificPNG, pngRegisterIndex, ctxRegisterIndex);
		            cv.visitJumpInsn(IFNONNULL, labelFalse);
	            }
            } else {
                //jormf1 == spng.pnGetF1(ctx)
                cv.visitVarInsn(ALOAD, 0);
                cv.visitFieldInsn(GETFIELD, gc.classToWrite, fn, jvmFieldType);
                loadValueFromPNG(gc, pe, cv, specificPNG, pngRegisterIndex, ctxRegisterIndex);
    			switch (pe.getType().getTypeCode()) {
    			case PType.TYPECODE_BYTE:
    			case PType.TYPECODE_CHAR:
    			case PType.TYPECODE_SHORT:
    			case PType.TYPECODE_INT:
        	        cv.visitJumpInsn(IF_ICMPNE, labelFalse);
    				break;
    			case PType.TYPECODE_LONG:
    			case PType.TYPECODE_DOUBLE:
        	        cv.visitInsn(LCMP);
        	        cv.visitJumpInsn(IFNE, labelFalse);
    				break;
    			}                
            }
            cv.visitLabel(labelNextTest);
		}
	}
	
	/**
	 * This helper method loads on the stack a value from a pname getter.
	 * @param pe is the JORM meta object representing the field (to load) of 
	 * the composite name.
	 * @param cv is the visitor of the method where the code has to be generated
	 * @param specificPNG indicates the value loading has to be done with the
	 * specific/generated pname getter (XXXPNG). A false value indicates that
	 * the generic pname getter (PNameGetter) has to be used.
	 * @param pngRegisterIndex is the register number where the reference to 
	 * the pname getter (generic or specific) is stored. 
	 * @param ctxRegisterIndex is the register numbere where the reference to 
	 * the context is stored. The context is a parameter of the pname getter 
	 * methods (generic or specific).
	 * @throws PException if it is not possible to compute a specific pname
	 * getter from the given composite name field (param pe).
	 */
	private void loadValueFromPNG(final BinderGenContext gc, 
	        final PrimitiveElement pe,
	        final CodeVisitor cv, 
	        final boolean specificPNG, 
	        final int pngRegisterIndex, 
	        final int ctxRegisterIndex) throws PException {
		final String jvmFieldType = getJVMType(pe.getType());
		if (specificPNG) {
		    //spng.pnGetF1(ctx)
            cv.visitVarInsn(ALOAD, pngRegisterIndex);
            cv.visitVarInsn(ALOAD, ctxRegisterIndex);
            cv.visitMethodInsn(INVOKEINTERFACE, gc.xpngJCN,
                    "pnGet" + upperFL(pe.getName()),
                    "(" + gc.objectJT + ")" + jvmFieldType);
		} else {
		    //gpng.pnGetChar("f1", ctx)
            cv.visitVarInsn(ALOAD, pngRegisterIndex);
            cv.visitLdcInsn(pe.getName());
            cv.visitVarInsn(ALOAD, ctxRegisterIndex);
            cv.visitMethodInsn(INVOKEINTERFACE,gc.pngJCN,
                    getPNameGetterGetFunction(pe.getType()),
                    "(" + gc.stringJT + gc.objectJT + ")" + jvmFieldType);
		}	    
	}
	
	/**
	 * Gets the name of the auto calculted field in this generated Binder for
	 * a given field
	 * @param pe is the meta object representing the field
	 */
	private String getACFName(final PrimitiveElement pe) {
	    return pe.getName() + "JormLongGen";
	}

	/**
	 * Gets the name of the field containing the null in this generated Binder 
	 * for a given field
	 * @param pe is the meta object representing the field
	 */
	private String getNullFieldName(final PrimitiveElement pe) {
	    return "jorm" + pe.getName();
	}
	
	/**
	 * Load in the stack the null value for a given field
	 * @param pe is the meta object representing the field
	 * @param cv is the CodeVisitor to use for loading the null value
	 */
	private void loadNullValue(final PrimitiveElement pe, final CodeVisitor cv) {
            loadNull(cv, pe.getType());
	}
	
	/**
	 * This class is a generation context for binder generation with ASM.
	 *
	 * @author S.Chassande-Barrioz
	 */
	private static final class BinderGenContext extends CNGenContext {
	    /**
	     * Is the name of the class inherited from the generated PBinder.
	     */
	    public final String compositeBinderJCN;
	    
	    /**
	     * Is the name of the class to generates
	     */
	    public final String classToWrite;

	    public BinderGenContext(final CompositeName _co, 
	            final ClassVisitor _clav, 
	            final CommonJVMHelper h) {
	        super(_co, _clav, h);
	        classToWrite = xbinderJCN;
	        this.compositeBinderJCN = helper.getJVMClassName(CompositePBinder.class); 
	    }
	}
}
