/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2005 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.generator.lib;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.Collections;
import java.util.HashSet;

import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Constants;
import org.objectweb.asm.attrs.Attributes;
import org.objectweb.jorm.api.PBinding;
import org.objectweb.jorm.api.PBindingCtrl;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.compiler.api.JormCompilerConfigurator;
import org.objectweb.jorm.compiler.api.JormCompilerParameter;
import org.objectweb.jorm.compiler.api.PExceptionCompiler;
import org.objectweb.jorm.lib.PBindingImpl;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.util.io.api.TargetHolder;
import org.objectweb.jorm.util.lib.StringReplace;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class is a generator of XXXBinding objects. This generator use the
 * ASM tools. 
 * 
 * @author S.Chassande-Barrioz
 */
public class BindingGenerator extends CommonGenerator {

    /**
     * This method generates a XPBinding file corresponding to the pod parameter
     * in the directory parameter
     * 
     * @param clazz
     *            the class meta object interface
     * @param holder
     *            The target holder which allows to create files
     * @param cp
     *            This parameter permits to reach the compilation parameters
     * @param jcc
     *            the JormCompilerConfigurator
     */
    public void generate(Class clazz, 
            TargetHolder holder,
            JormCompilerParameter cp, 
            JormCompilerConfigurator jcc) throws PException {
        String extend = cp.getBindingInheritance();
        if (extend != null) {
            //convert the class name into the JVM format
            extend = getJVMClassName(extend);
        }
        boolean injectPBindingImpl = true;
        boolean newClass = true;
        String classToWrite = getJVMClassName(clazz.getFQName() + "Binding");
        if (extend == null) {
            if (cp.getBindingEnhancedClass() == null) {
                //
                extend = getJVMClassName(PBindingImpl.class);
                injectPBindingImpl = false;
                if (debug) {
                    logger.log(BasicLevel.DEBUG, "Generate the class '"
                            + classToWrite + "' extending PBindingImpl.");
                }
            } else {
                //Inject PBindingImpl code and PNG code into the enhanced class
                newClass = false;
                extend = getJVMClassName(Object.class);
                classToWrite = getJVMClassName(cp.getBindingEnhancedClass());
                if (debug) {
                    logger.log(BasicLevel.DEBUG, "Enhance the class '"
                            + classToWrite + "'.");
                }
            }
        } else {
            //generate a class extending the user class containing
            //Inject PBindingImpl code and PNG code into
            if (debug) {
                logger.log(BasicLevel.DEBUG, "Generate the class '"
                        + classToWrite + "' extending '" + extend + "'.");
            }
        }
        ClassWriter cw = new ClassWriter(true);
        RedundencyRemover rr = new RedundencyRemover(cw,logger);
        HashSet interfaces = new HashSet();
        interfaces.add(getJVMClassName(PBinding.class));
        interfaces.add(getJVMClassName(PBindingCtrl.class));
        
        if (debug) {
            logger.log(BasicLevel.DEBUG, "interfaces: " + interfaces);
        }
        if (newClass) {
            //Create the class
            cw.visit(V1_1, ACC_PUBLIC
                    + (cp.isBindingAbstract() ? ACC_ABSTRACT : 0),
                    classToWrite, extend, 
                    (String[]) interfaces.toArray(new String[interfaces.size()]), 
                    classToWrite);
			CodeVisitor _cv = cw.visitMethod(
					Constants.ACC_PUBLIC, "<init>", "()V", null, null);
				_cv.visitVarInsn(Constants.ALOAD, 0);
				_cv.visitMethodInsn(Constants.INVOKESPECIAL, extend, "<init>", "()V");
				_cv.visitInsn(Constants.RETURN);
				_cv.visitMaxs(0, 0);
        } else {
            //inject the code of the enhanced class 
            ClassInterfaceAdder cia = new ClassInterfaceAdder(rr, interfaces, logger);
            ClassMergerVisitor cmv = new ClassMergerVisitor(cia, logger);
            ClassReader cr = loadJavaClass(classToWrite, holder, true);
            cr.accept(cmv, Attributes.getDefaultAttributes(), false);
        }
        if (injectPBindingImpl) {
            //Inject the byte code from the PBindingImpl class
            ClassMergerVisitor cmv = new ClassMergerVisitor(rr, logger);
            ClassReader cr = null;
            try {
                cr = new ClassReader(getClass().getClassLoader().getResourceAsStream(
                        getJVMClassName(PBindingImpl.class) + ".class"));
            } catch (Exception e) {
                logger.log(BasicLevel.ERROR, e.getMessage(), e);
                throw new PException("Impossible to load the PBinding class, check your classpath");
            }
            HashSet oldNames = new HashSet();
            oldNames.add(getJVMClassName(PBindingImpl.class));
            ClassRenamer cren = new ClassRenamer(cmv, oldNames, classToWrite, logger);
            cr.accept(cren, Attributes.getDefaultAttributes(), false);
        }
        cw.visitEnd();
        
        try {
            // Fetch a block with the fileName
            String fileName = classToWrite;
            fileName = StringReplace.replaceChar('/', File.separatorChar,
                    fileName)
                    + ".class";
            FileOutputStream fos = holder.getFileOutputStream(fileName);
            fos.write(cw.toByteArray());
            fos.close();
        } catch (Exception e) {
            throw new PExceptionCompiler(e,
                    "Problem while generating PAccessor.");
        }
    }

    public ClassReader loadJavaClass(final String completeName,
            final TargetHolder holder, final boolean remove) throws PException {
        ClassReader jclass = null;
        File f = holder.getFile(
                completeName.replace('.', File.separatorChar) + ".class");
        try {
            FileInputStream fis = new FileInputStream(f);
            jclass = new ClassReader(fis);
            fis.close();
            if (remove) {
                f.delete();
            }
        } catch (Exception e) {
            throw new PException(e);
        }
        return jclass;
    }

}