/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2005 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.generator.lib;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.Package;
import org.objectweb.jorm.metainfo.api.ClassRef;
import org.objectweb.jorm.metainfo.api.CompositeName;
import org.objectweb.jorm.metainfo.api.FieldName;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.NameRef;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.api.Reference;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.jorm.metainfo.api.Mapping;
import org.objectweb.jorm.metainfo.api.ReferenceMapping;
import org.objectweb.jorm.metainfo.api.PrimitiveElementMapping;
import org.objectweb.jorm.type.api.PExceptionTyping;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.jorm.type.lib.TypeHelper;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.jorm.compiler.api.PExceptionCompiler;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;
import java.util.Collection;
import java.util.HashMap;

/**
 * This class is a helper for the generation. All methods provided in this
 * class are shortcut or facilties which simplify the generation.
 * @author Sebastien Chassande-Barrioz
 */
public class CommonHelper extends TypeHelper implements Loggable {

	protected LoggerFactory loggerFactory = null;
	protected Logger logger = null;
	protected boolean debug = false;

	public final static PType[] PN_GET_TYPES = {
		PTypeSpace.BYTE,
		PTypeSpace.CHAR,
		PTypeSpace.DATE,
		PTypeSpace.INT,
		PTypeSpace.LONG,
		PTypeSpace.OBJBYTE,
		PTypeSpace.OBJCHAR,
		PTypeSpace.OBJINT,
		PTypeSpace.OBJLONG,
		PTypeSpace.OBJSHORT,
		PTypeSpace.SHORT,
		PTypeSpace.STRING,
		PTypeSpace.BYTEARRAY,
		PTypeSpace.CHARARRAY,
		PTypeSpace.BIGDECIMAL,
		PTypeSpace.BIGINTEGER
	};

	/**
	 * This methods returns the word specified in parameter with the first
	 * character in upper-case. This method is used for the generation of
	 * accessor methods.
	 * @param word	The String which the first character must be set in
	 * 				upper-case
	 * @return		The String whith the first character in upper-case
	 */
	public String upperFL(String word) {
		if (word != null && word.length() > 0)
			return word.substring(0, 1).toUpperCase() + word.substring(1);
		else
			return word;
	}

	/**
	 * This method allow to retrieve a primitive element from th meta object mo.
	 * if this metaobject is a primitive element, it is returned itself.
	 * if this meta object is a single namedef the primitive element composing the namedef is returned.
	 * This method does not work with composite pname.
	 * @param mo the meta object embeding the primitive element
	 * @param c the class embeding the meta object mo
	 * @return the primitive element
	 */

	public PrimitiveElement getPrimitiveElement(Object mo,
												Class c,
												boolean canBeFieldName)
			throws PException {
		if (mo instanceof PrimitiveElement)
			return (PrimitiveElement) mo;
		else if (mo instanceof NameDef) {
			NameDef nd = ((NameDef) mo);
			if (nd.isFieldName())
				return (PrimitiveElement) (c.getTypedElement(nd.getFieldName()));
			else
				throw new PException("This is composite name !" + mo);
		} else
			throw new PException(
					"impossible to obtain a PrimitiveElement with " + mo);
	}

	/**
	 * This method tests if a class contains a reference. A reference can be a
	 * designs a generic class or a class.
	 * @param co	The meta object Class
	 * @return true if the class contains a reference.
	 */
	public boolean containsReference(Class co) {
		Iterator it = co.getAllFields().iterator();
		while (it.hasNext()) {
			if (it.next() instanceof Reference)
				return true;
		}
		return false;
	}

	/**
	 * This method tests if a class contains a composite reference. A reference can be a
	 * designs a generic class or a class.
	 * @param mo	The meta object Mapping
	 * @return true if the class contains the namedef for the composite reference.
	 *
	 public boolean containsCompositeReference(Mapping mo) throws Exception {
	 Iterator it = mo.getClassMapping().getReferenceMappings().iterator();
	 while (it.hasNext()) {
	 ReferenceMapping rm = (ReferenceMapping) it.next();
	 if (((NameDef) rm.getLinkedMO()).isNameRef()) {
	 return true;
	 }
	 }
	 return false;
	 }
	 */

	/**
	 * This method return a list containing the nameref implied in composite reference field
	 * @param mo	The meta object Mapping to be analysed.
	 * @return an ArrayList which contains all nameref related to reference fields.
	 */

	public ArrayList getCompositeReferences(Mapping mo) throws Exception {
		ArrayList res = new ArrayList();
		Iterator it = mo.getClassMapping().getReferenceMappings().iterator();
		while (it.hasNext()) {
			ReferenceMapping rm = (ReferenceMapping) it.next();
			if (((NameDef) rm.getLinkedMO()).isNameRef()) {
				res.add(((NameDef) rm.getLinkedMO()).getNameRef());
			}
		}
		return res;
	}

	/**
	 * This method return a SET containing the full name of compositename implied
	 * in the reference field of the class.
	 * @param moc	The meta object Class
	 * @param cp the compiler parameter
	 * @return an iterator which contains all pname related to the Reference.
	 *
	 public Set getCompositeNames(Class moc, JormCompilerParameter cp) throws Exception {

	 Set res = new HashSet();
	 NameDef nd = moc.getNameDef(cp.projectName);
	 if (nd.isNameRef()) {
	 String cn = nd.getNameRef().getCompositeName().getFQName();
	 if (res.add(cn)) {
	 res.add(cn + "PNG");
	 }
	 }
	 Iterator it = moc.iterateAllField();
	 while (it.hasNext()) {
	 TypedElement te = (TypedElement) it.next();
	 if (te instanceof Reference) {
	 nd = ((Reference) te).getRefNameDef(cp.projectName);
	 if (nd.isNameRef()) {
	 res.add(nd.getNameRef().getCompositeName().getFQName()
	 + "PNG");
	 }
	 }
	 }
	 return res;
	 }
	 */

	/**
	 * This method retrieve the package name for the meta object passed as parameter.
	 * The meta object must be a child of a Package MetaObject.
	 * @param mo	The meta object
	 * @return the name of the package if it exist, null otherwise.
	 */
	public String getCNPackageName(MetaObject mo) {
		if (mo instanceof Manager) return null;
		while (!(mo instanceof Package)) {
			mo = mo.getParent();
		}
		if (mo instanceof Package)
			return ((Package) mo).getName();
		else
			return null;
	}

	/**
	 * This method tests if a class contains a serialized field.
	 * @param co	The meta object Class
	 * @return true if the class contains a serialized field.
	 */
	public boolean containsSerializedField(Class co) {
		Iterator it = co.getAllFields().iterator();
		while (it.hasNext()) {
			if (((TypedElement) it.next()).getType().getTypeCode()
					== PType.TYPECODE_SERIALIZED)
				return true;
		}
		return false;
	}

	/**
	 * This method tests if a class contains a char[] field.
	 * @param m	The meta object Mapping
	 * @return true if the class contains a serialized field.
	 */
	public boolean containsCharArrayField(Mapping m) {
		Class co = (Class) m.getClassMapping().getLinkedMO();
		Iterator it = co.getAllFields().iterator();
		while (it.hasNext()) {
			TypedElement te = (TypedElement) it.next();
			if (te == null) {
				throw new NullPointerException(
						"A field of the class '" + co.getFQName() + "' is null.");
			}
			if (te.getType() == null) {
				throw new NullPointerException(
						"the type of field '" + te.getName()
						+ "' of the class '" + co.getFQName() + "' is null.");
			}
			if (te.getType().getTypeCode() == PType.TYPECODE_CHARARRAY)
				return true;
		}
		it = co.getAllHiddenFields().iterator();
		while (it.hasNext()) {
			if (((TypedElement) it.next()).getType().getTypeCode()
					== PType.TYPECODE_CHARARRAY)
				return true;
		}
		NameDef nd = (NameDef)
				m.getClassMapping().getIdentifierMapping().getLinkedMO();
		if (nd.isNameRef()) {
			it = nd.getNameRef().getCompositeName().getAllFields().iterator();
			while (it.hasNext()) {
				if (((TypedElement) it.next()).getType().getTypeCode()
						== PType.TYPECODE_CHARARRAY)
					return true;
			}
		}
		return false;
	}

	/**
	 * This method tests if a class contains a serialized field.
	 * @param co	The meta object Class
	 * @return true if the class contains a serialized field.
	 */
	public boolean containsFieldWhichCanBeNull(Class co) throws PException {
		for (Iterator it = co.getAllFields().iterator(); it.hasNext();) {
			if (canBeNullValue(((TypedElement) it.next()).getType()))
				return true;
		}
		for (Iterator it = co.getAllHiddenFields().iterator(); it.hasNext();) {
			if (canBeNullValue(((TypedElement) it.next()).getType()))
				return true;
		}
		return false;
	}

	/**
	 * This method tests if a String is valid. A String is valid if it is not
	 * null and if the length is greater than 0.
	 * @param str   The string which must be checked
	 * @return true if the string is valid
	 */
	public boolean isValidString(String str) {
		return str != null && str.length() > 0;
	}

	/**
	 * This method tests if a class references to a generic class.
	 * @param co	The meta object Class
	 * @return true if the class references a generic class.
	 */
	public boolean containsGenClassRef(Class co) {
		Iterator it = co.getAllFields().iterator();
		while (it.hasNext()) {
			if (it.next() instanceof GenClassRef)
				return true;
		}
		return false;
	}

	/**
	 * This method checks if the NameDef is composed by several fields.
	 * @param nd the meta object NameDef which must be tested
	 * @return true if the NameDef is composed by several fields.
	 */
	public boolean isCompositeName(NameDef nd) {
		return nd.isNameRef();
	}

	/**
	 * This method provides an iterator over the map entries which associated
	 * a composite field name (key) to the class field name (value). These map
	 * elements describe the projection of a name definition.
	 *
	 * @param mo the meta object NameDef which must be used
	 */

	public Collection getCompositeNameDefEntries(NameDef mo) throws Exception {
		if (mo == null)
			throw new PException("Null NameDef Parameter");
		else if (mo.isNameRef())
			return mo.getNameRef().getProjection().entrySet();
		else if (mo.isFieldName())
			throw new PException("The NameDef is a FieldName ("
								 + mo.getFieldName() + ")");
		else if (mo.isSystem())
			throw new PException("System identifier unmanaged");
		else
			throw new PException("Unknown NameDef: " + mo);
	}

	public String getFQNOfCompositeName(Object o) {
		CompositeName cn = null;
		if (o instanceof CompositeName)
			cn = (CompositeName) o;
		else if (o instanceof NameDef)
			cn = ((NameDef) o).getNameRef().getCompositeName();
		else
			return "";

		String packName = ((Package) cn.getParent()).getName();
		if (packName == null || packName.length() == 0) {
			return cn.getName();
		} else {
			return packName + "." + cn.getName();
		}
	}

	/**
	 * This method allows to know if a class represented by its Class meta
	 * object, contains a composite PName. This method analyzes the NameDef of
	 * the @c class parameter and the NameDef of all reference fields.
	 * @param mo the Mapping which must analyzed
	 */
	public boolean containsCompositeName(Mapping mo) {
		if (mo == null) {
			return false;
		}
		NameDef nd = (NameDef)
				mo.getClassMapping().getIdentifierMapping().getLinkedMO();
		if (nd.isNameRef())
			return true;
		if (logger != null)
			logger.log(BasicLevel.DEBUG, "PName class has a single field");
		Iterator it = mo.getClassMapping().getReferenceMappings().iterator();
		while (it.hasNext()) {
			ReferenceMapping rm = (ReferenceMapping) it.next();
			nd = (NameDef) rm.getLinkedMO();
			if (nd.isNameRef()) {
				return true;
			}
		}
		return false;
	}

	/**
	 * This method checks if the parameter is an instance of
	 * org.objectweb.jorm.metainfo.api.Class
	 * @param o the object which the type must be checked
	 * @return true the parameter is an instance of Class
	 */
	public boolean isClass(Object o) {
		return o instanceof Class;
	}

	/**
	 * This method checks if the parameter is an instance of
	 * org.objectweb.jorm.metainfo.api.Reference
	 * @param te the object which the type must be checked
	 * @return true the parameter is an instance of Reference
	 */
	public boolean isReference(TypedElement te) {
		return te instanceof Reference;
	}

	/**
	 * This method checks if the parameter is an instance of
	 * org.objectweb.jorm.metainfo.api.ClassRef
	 * @param te the object which the type must be checked
	 * @return true the parameter is an instance of ClassRef
	 */
	public boolean isClassRef(TypedElement te) {
		return te instanceof ClassRef;
	}

	/**
	 * This method checks if the parameter is an instance of
	 * org.objectweb.jorm.metainfo.api.GenClassRef
	 * @param te the object which the type must be checked
	 * @return true the parameter is an instance of GenClassRef
	 */
	public boolean isGenClassRef(TypedElement te) {
		return te instanceof GenClassRef;
	}

	/**
	 * This method checks if the parameter is an instance of
	 * org.objectweb.jorm.metainfo.api.PrimitiveElement
	 * @param o the object which the type must be checked
	 * @return true the parameter is an instance of PrimitiveElement
	 */
	public boolean isPrimitiveElement(Object o) {
		return o instanceof PrimitiveElement;
	}

	/**
	 * @param val1 the variable name of the first element
	 * @param val2 the variable name of the second element
	 * @param type the type of the elements
	 * @return a String comparing two primitive element
	 */
	public String comparePE(String val1, String val2, PType type) {
		if (isObjectType(type)) {
			return "((" + val1 + " == null && " + val2 + " == null) || ("
					+ val1 + " !=null && " + val1 + ".equals(" + val2 + ")))";
		} else {
			return val1 + " == " + val2;
		}
	}

	public boolean isNull(Object o) {
		return o == null;
	}

	/**
	 * This method checks if the parameter is an instance of
	 * org.objectweb.jorm.metainfo.api.PrimitiveElement
	 * @param pe the object which the type must be checked
	 * @return true the parameter is an instance of PrimitiveElement
	 */
	public boolean isHiddenField(PrimitiveElement pe) {
		return ((Class) pe.getParent()).getHiddenField(pe.getName()) != null;
	}

	/**
	 * This method checks if the parameter is an instance of
	 * org.objectweb.jorm.metainfo.api.FieldName
	 * @param o the object which the type must be checked
	 * @return true the parameter is an instance of FieldName
	 */
	public boolean isFieldName(Object o) {
		return o instanceof FieldName;
	}

	/**
	 * This method checks if the parameter designs a Serialized field.
	 * @param o the object which the type must be checked
	 * @return true the parameter is an instance of PrimitiveElement and if its
	 * PType is SERIALIZED.
	 */
	public boolean isSerializedField(Object o) {
		return (o instanceof PrimitiveElement)
				&& (((PrimitiveElement) o).getType().getTypeCode()
				== PType.TYPECODE_SERIALIZED);
	}

	/**
	 * This method checks if the PType parameter designs a Serialized type.
	 * @param t the PType object which the type must be checked
	 * @return true the parameter is a PType.SERIALIZED.
	 */
	public boolean isSerializedType(PType t) {
		return t.getTypeCode() == PType.TYPECODE_SERIALIZED;
	}

	/**
	 * This method checks if the PType parameter designs a Serialized type.
	 * @param t the PType object which the type must be checked
	 * @return true the parameter is a PType.SERIALIZED.
	 */
	public boolean isCharArrayType(PType t) {
		return t.getTypeCode() == PType.TYPECODE_CHARARRAY;
	}

	public PrimitiveElementMapping[] buildPemList(Collection pemcoll) {
		return (PrimitiveElementMapping[]) pemcoll.toArray(new PrimitiveElementMapping[0]);
	}

	public int getPemPos(PrimitiveElementMapping[] pemlist, PrimitiveElement pe) throws PExceptionCompiler {
		for (int i = 0; i < pemlist.length; i++) {
			if (pemlist[i].getLinkedMO() == pe) {
				return i + 1;
			}
		}
		throw new PExceptionCompiler("Cannot find a mapping for field: " + pe.getName());
	}

	public PrimitiveElement getPrimitiveElement(String fn, Class cl) throws PExceptionCompiler {
		TypedElement res = cl.getTypedElement(fn);
		if (res == null) {
			throw new PExceptionCompiler("NameDef with field name does not correspond to any field - searched field: " + fn);
		}
		if (!(res instanceof PrimitiveElement)) {
			throw new PExceptionCompiler("NameDef with field name does not correspond to any primitive field - searched field: " + fn);
		}
		return (PrimitiveElement) res;
	}

	public NameDef getRefNameDef(Mapping mo, Reference field) throws PException {
		if (field == null) {
			return null;
		}
		if (mo == null) {
			return null;
		}
		ReferenceMapping rm = mo.getClassMapping().getReferenceMapping(field.getName());
		NameDef nd = null;
		if (rm != null) {
			nd = (NameDef) rm.getLinkedMO();
		}
		if (nd == null) {
			throw new PException("No NameDef found for the reference field '"
				+ field.getName() +"' of the persistent class '"
				+ ((Class) mo.getClassMapping().getLinkedMO()).getFQName()
				+ "'.");
		}
		return nd;
	}

	public String getTupleDecoding(String fn, PType pt, String value)
		throws PException {
		if (pt == null) {
			throw new PException("Unauthorized null PType");
		}
		StringBuffer sb = new StringBuffer();
		sb.append(fn);
		sb.append("Pnc.");
		sb.append(getPNameDecodeFunction(pt));
		sb.append("((");
		sb.append(pt.getJavaName());
		sb.append(") ");
		sb.append(value);
		sb.append(")");
		return sb.toString();
	}

	/**
	 * Builds the string representing the expression to decode a PName wrt the
	 * relevant tuple field having the right type.
	 * @param tmpv      The temporary variable that holds the tuple field value.
	 * @param fn        The name of the involved application field.
	 * @param tuplevn   The variable name that holds the corresponding tuple.
	 * @param pt        The PType of the associated field.
	 * @param pos       The position of the tuple field into this tuple.
	 * @return  The string representing the getter expression.
	 */
	public String getTupleDecoding(String tmpv, String fn, String tuplevn, PType pt, int pos)
			throws PException {
		if (pt == null)
			throw new PException("Unauthorized null PType");
		StringBuffer sb = new StringBuffer();
		if (isObjectType(pt)) {
			sb.append("(((");
			sb.append(tmpv);
			sb.append(" = ");
			sb.append(getTupleGetter(tuplevn, pt, pos) );
			sb.append(") == null) ? ");
			sb.append(fn);
			sb.append("Pnc.getNull() : ");
			sb.append(fn);
			sb.append("Pnc.");
			sb.append(getPNameDecodeFunction(pt));
			sb.append("((");
			sb.append(pt.getJavaName());
			sb.append(") ");
			sb.append(tmpv);
			sb.append("))");
		} else {
			sb.append(fn);
			sb.append("Pnc.");
			sb.append(getPNameDecodeFunction(pt));
			sb.append("(");
			sb.append(getTupleGetter(tuplevn, pt, pos) );
			sb.append(")");
		}
		return  sb.toString();
	}

	/**
	 * Builds the string representing the expression to access the relevant
	 * tuple field having the right type.
	 * @param tuplevn   The variable name that holds the corresponding tuple.
	 * @param pt        The PType of the associated field.
	 * @param pos       The position of the tuple field into this tuple.
	 * @return  The string representing the getter expression.
	 */
	public String getTupleGetter(String tuplevn, PType pt, int pos)
			throws PException {
		if (pt == null)
			throw new PException("Unauthorized null PType");
	    StringBuffer sb = new StringBuffer();
		switch (pt.getTypeCode()) {
		case PType.TYPECODE_OBJBOOLEAN:
		case PType.TYPECODE_OBJBYTE:
		case PType.TYPECODE_OBJCHAR:
		case PType.TYPECODE_OBJSHORT:
		case PType.TYPECODE_OBJINT:
		case PType.TYPECODE_OBJLONG:
		case PType.TYPECODE_OBJFLOAT:
		case PType.TYPECODE_OBJDOUBLE:
			sb.append("(new ").append(pt.getJavaName()).append("("); 
		    sb.append(getTupleGetter(tuplevn, toPrimitive(pt), pos));
		    sb.append("))");
		    break;

		case PType.TYPECODE_SERIALIZED:
		    sb.append("(Serializable) ");
		default:
		    sb.append(tuplevn).append(".");
			sb.append(getTupleGetterAsString(pt));
			sb.append("(").append(pos).append(")");
		}
	    return sb.toString();
	}

	/**
	 * This method does the conversion between a PType @pt and the getter
	 * function to use on a PAccessorGen (generic accessor).
	 * @param pt the ptype which helps to find the right paccessor method
	 * @return 	The string which represents the paccessor method associated
	 * 			to pt parameter
	 */
	public String getPaGetMethod(PType pt) throws PException {
		if (pt == null)
			throw new PException("Unauthorized null PType");
		return "paGet" + getGenPaMethodPart(pt) + "Field";
	}

	/**
	 * This method does the conversion between a PType @pt and the setter
	 * function to use on a PAccessorGen (generic accessor).
	 * @param pt the ptype which helps to find the right paccessor method
	 * @return 	The string which represents the paccessor method associated
	 * 			to pt parameter
	 */
	public String getPaSetMethod(PType pt) throws PException {
		if (pt == null)
			throw new PException("Unauthorized null PType");
		return "paSet" + getGenPaMethodPart(pt) + "Field";
	}

	/**
	 * This method permits to find the right decode function either the
	 * PType specified by the parameter @pt. These functions name are available
	 * on a PName interface.
	 * Be carefull the only jorm types support for the naming are the scalar
	 * types (char, byte, short, int, long and string) !
	 * @param pt the ptype which helps to find the right decode method
	 * @return 	The string which represents the decode method associated
	 * 			to pt parameter
	 */
	public String getPNameDecodeFunction(PType pt) throws PException {
		if (pt == null) {
			throw new PExceptionTyping("The parameter must be not null");
		}
	    return "decode" + getPNameCodingMethodPart(pt);
	}

	public String getCoderName(PType t) {
	    return t.getCodingName();
	}

	/**
	 * This method permits to find the right pngetXField function either the
	 * PType specified by the parameter @pt. These functions name are available
	 * on a PNameGetter interface.
	 * Be carefull the only jorm types support for the naming are the scalar
	 * types (char, byte, short, int, long and string) !
	 * @param pt the ptype which helps to find the right pname getter method
	 * @return 	The string which represents the pname getter method associated
	 * 			to pt parameter
	 */
	public String getPNameGetterGetFunction(PType pt) throws PException {
		if (pt == null) {
			throw new PExceptionTyping("The parameter must be not null");
		}
		return "pnget" + pt.getCodingName() + "Field";
	}

	/**
	 * This method permits to find the right encode function either the
	 * PType specified by the parameter @pt. These functions name are available
	 * on a PName interface.
	 * Be carefull the only jorm types support for the naming are the scalar
	 * types (char, byte, short, int, long and string) !
	 * @param pt the ptype which helps to find the right encode method
	 * @return 	The string which represents the encode method associated
	 * 			to pt parameter
	 */
	public String getPNameEncodeFunction(PType pt) throws PException {
		if (pt == null) {
			throw new PExceptionTyping("The parameter must be not null");
		}
	    return "encode" + getPNameCodingMethodPart(pt);
	}

	public String getCTDeclaration(PType t) {
	    return getCodingTypeAsString(t.getTypeCode());
	}

	public String getCTDeclaration(int typecode) {
	    return getCodingTypeAsString(typecode);
	}

	/**
	 * This method provides the projection of the field name belonging to the class
	 * over a field name belonging to the composite name corresponding definition.
	 * @param nd the meta object NameDef embeding a nameref with a projection.
	 * @param cfn the field name belonging to the pname in the class.
	 * @return the field name of the projection. this field name is a field belonging to the class.
	 */

	public String getProjectionField(NameDef nd, String cfn) throws Exception {
		if (nd.isNameRef()) {
			NameRef nr = nd.getNameRef();
			// get the projection of this composite name onto the field of the class
			// classfields represents the name of the field related to the name structure.
			// classfields are either primitive field or hiddenfield of the class.
			Iterator it = nr.getProjection().entrySet().iterator();
			while (it.hasNext()) {
				Map.Entry me = (Map.Entry) (it.next());
				if (me.getValue().equals(cfn))
					return (String) (me.getKey());
			}
		}
		throw new PException("impossible to obtain field " + cfn + " of " + nd);
	}

	public PType getPType(String ptn) {
		if (ptn.equals("boolean")) return PTypeSpace.BOOLEAN;
		if (ptn.equals("Boolean")) return PTypeSpace.OBJBOOLEAN;
		if (ptn.equals("char")) return PTypeSpace.CHAR;
		if (ptn.equals("Character")) return PTypeSpace.OBJCHAR;
		if (ptn.equals("byte")) return PTypeSpace.BYTE;
		if (ptn.equals("Byte")) return PTypeSpace.OBJBYTE;
		if (ptn.equals("short")) return PTypeSpace.SHORT;
		if (ptn.equals("Short")) return PTypeSpace.OBJSHORT;
		if (ptn.equals("int")) return PTypeSpace.INT;
		if (ptn.equals("Integer")) return PTypeSpace.OBJINT;
		if (ptn.equals("long")) return PTypeSpace.LONG;
		if (ptn.equals("Long")) return PTypeSpace.OBJLONG;
		if (ptn.equals("float")) return PTypeSpace.FLOAT;
		if (ptn.equals("Float")) return PTypeSpace.OBJFLOAT;
		if (ptn.equals("double")) return PTypeSpace.DOUBLE;
		if (ptn.equals("Double")) return PTypeSpace.OBJDOUBLE;
		if (ptn.equals("string")) return PTypeSpace.STRING;
		if (ptn.equals("date")) return PTypeSpace.DATE;
		if (ptn.equals("serialized")) return PTypeSpace.SERIALIZED;
		if (ptn.equals("BigInteger")) return PTypeSpace.BIGINTEGER;
		if (ptn.equals("BigDecimal")) return PTypeSpace.BIGDECIMAL;
		if (ptn.equals("byte[]")) return PTypeSpace.BYTEARRAY;
		if (ptn.equals("char[]")) return PTypeSpace.CHARARRAY;
		if (ptn.equals("reference")) return PTypeSpace.REFTOP;
		return null;
	}

	/**
	 * Its retrieves the PType object which represents the boolean jorm type.
	 */
	public PType getPTypeBoolean() {
		return PTypeSpace.BOOLEAN;
	}

	/**
	 * Its retrieves the PType object which represents the Boolean jorm type.
	 */
	public PType getPTypeOboolean() {
		return PTypeSpace.OBJBOOLEAN;
	}

	/**
	 * Its retrieves the PType object which represents the char jorm type.
	 */
	public PType getPTypeChar() {
		return PTypeSpace.CHAR;
	}

	/**
	 * Its retrieves the PType object which represents the Character jorm type.
	 */
	public PType getPTypeOchar() {
		return PTypeSpace.OBJCHAR;
	}

	/**
	 * Its retrieves the PType object which represents the byte jorm type.
	 */
	public PType getPTypeByte() {
		return PTypeSpace.BYTE;
	}

	/**
	 * Its retrieves the PType object which represents the Byte jorm type.
	 */
	public PType getPTypeObyte() {
		return PTypeSpace.OBJBYTE;
	}

	/**
	 * Its retrieves the PType object which represents the short jorm type.
	 */
	public PType getPTypeShort() {
		return PTypeSpace.SHORT;
	}

	/**
	 * Its retrieves the PType object which represents the Short jorm type.
	 */
	public PType getPTypeOshort() {
		return PTypeSpace.OBJSHORT;
	}

	/**
	 * Its retrieves the PType object which represents the int jorm type.
	 */
	public PType getPTypeInt() {
		return PTypeSpace.INT;
	}

	/**
	 * Its retrieves the PType object which represents the Integer jorm type.
	 */
	public PType getPTypeOint() {
		return PTypeSpace.OBJINT;
	}

	/**
	 * Its retrieves the PType object which represents the long jorm type.
	 */
	public PType getPTypeLong() {
		return PTypeSpace.LONG;
	}

	/**
	 * Its retrieves the PType object which represents the Long jorm type.
	 */
	public PType getPTypeOlong() {
		return PTypeSpace.OBJLONG;
	}

	/**
	 * Its retrieves the PType object which represents the float jorm type.
	 */
	public PType getPTypeFloat() {
		return PTypeSpace.FLOAT;
	}

	/**
	 * Its retrieves the PType object which represents the Float jorm type.
	 */
	public PType getPTypeOfloat() {
		return PTypeSpace.OBJFLOAT;
	}

	/**
	 * Its retrieves the PType object which represents the double jorm type.
	 */
	public PType getPTypeDouble() {
		return PTypeSpace.DOUBLE;
	}

	/**
	 * Its retrieves the PType object which represents the Double jorm type.
	 */
	public PType getPTypeOdouble() {
		return PTypeSpace.OBJDOUBLE;
	}

	/**
	 * Its retrieves the PType object which represents the string jorm type.
	 */
	public PType getPTypeString() {
		return PTypeSpace.STRING;
	}

	/**
	 * Its retrieves the PType object which represents the date jorm type.
	 */
	public PType getPTypeDate() {
		return PTypeSpace.DATE;
	}

	/**
	 * Its retrieves the PType object which represents the chararray jorm type.
	 */
	public PType getPTypeCharArray() {
		return PTypeSpace.CHARARRAY;
	}

	/**
	 * Its retrieves the PType object which represents the bytearray jorm type.
	 */
	public PType getPTypeByteArray() {
		return PTypeSpace.BYTEARRAY;
	}

	/**
	 * Its retrieves the PType object which represents the serialized jorm type.
	 */
	public PType getPTypeSerialized() {
		return PTypeSpace.SERIALIZED;
	}

	/**
	 * Its retrieves the PType object which represents the BigInteger jorm type.
	 */
	public PType getPTypeBigInteger() {
		return PTypeSpace.BIGINTEGER;
	}

	/**
	 * Its retrieves the PType object which represents the BigDecimal jorm type.
	 */
	public PType getPTypeBigDecimal() {
		return PTypeSpace.BIGDECIMAL;
	}

	/**
	 * Its retrieves the PType object which represents a reference.
	 */
	public PType getPTypeReference() {
		return PTypeSpace.REFTOP;
	}

	public PType[] getPNGTypes() {
		return PN_GET_TYPES;
	}

	public boolean not(boolean b) {
		return !b;
	}

	public boolean canBeNullValue(PType type) throws PException {
	    return isObjectType(type);
	}

	/**
	 * This method is a helper to reach the getInheritedClasses(Class, HashMap)
	 * method. Indeed in the velocity template is not possible to instanciate
	 * an object.
	 */
	public String getInheritedClasses(Class c) {
		Map m = getInheritedClasses(c, new HashMap(), false);
		StringBuffer sb = new StringBuffer();
		sb.append('{');
		Iterator it = m.entrySet().iterator();
		String sep = "";
		while(it.hasNext()) {
			Map.Entry me = (Map.Entry) it.next();
			String child = (String) me.getKey();
			String[] parents = (String[]) me.getValue();
			for(int i=0; i<parents.length; i++) {
				sb.append(sep);
				sb.append("{\"");
				sb.append(child);
				sb.append("\", \"");
				sb.append(parents[i]);
				sb.append("\"}");
				sep =", ";
			}
		}
		sb.append('}');
		return sb.toString();
	}

	/**
	 * This recursive method add in the HashMap @calculated all inherited class
	 * of the @current class parameter.
	 * @return an hashmap where keys are class names and their values are the
	 * direct inhertied class.
	 */
	private HashMap getInheritedClasses(Class current,
	                                    HashMap calculated,
	                                    boolean addCurrent) {
		int parentNumber = current.getInheritedClassNumber();
		// Base case : No inherited classes
		if (parentNumber == 0)
			return calculated;

		int addcurrent = (addCurrent ? 1 : 0);
		String[] list = new String[parentNumber + addcurrent];
		if (addCurrent) {
			// Put the current class in first position into the list
			list[0] = current.getFQName();
		}
		// Put the list into the structure
		calculated.put(current.getFQName(), list);
		//Full the list of inherited classes for the current class
		Iterator it = current.getSuperClasses().iterator();
		int i;
		for (i = addcurrent; it.hasNext() && i < list.length; i++) {
			Class parent = (Class) it.next();
			list[i] = parent.getFQName();
			if (!calculated.containsKey(parent.getFQName()))
				getInheritedClasses(parent, calculated, false);
		}
		if (it.hasNext() || i < list.length) {
			throw new RuntimeException("MalFormed JMI");
		}
		return calculated;
	}

	public String log(String msg) {
		System.out.println(msg);
		return msg;
	}
	
	public String debug(String msg) {
	    if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
	        logger.log(BasicLevel.DEBUG, msg);
	        return "//" + msg;
	    }
	    return "";
	}
	/**
	 * Retrieves the name of the static field of the interface PTypeSpace
	 * corresponding to a PType.
	 */
	public String getPTypeSpaceFieldName(PType pt) {
	    return pt.getProgName().substring("PTypeSpace.".length());
	}
	

	//------------------------------------------//
	// IMPLEMENTATION OF THE Loggable INTERFACE //
	//------------------------------------------//

	/**
	 * Its assigns a logger factory.
	 */
	public void setLoggerFactory(LoggerFactory lf) {
		if (logger == null && lf != null)
			setLogger(lf.getLogger(getClass().getName()));
		loggerFactory = lf;
	}

	/**
	 * Its assigns a logger.
	 */
	public void setLogger(Logger l) {
		logger = l;
		debug = logger != null && logger.isLoggable(BasicLevel.DEBUG);
	}

	/**
	 * Its retrives the current logger.
	 */
	public Logger getLogger() {
		return logger;
	}

	/**
	 * Its retrives the current logger factory.
	 */
	public LoggerFactory getLoggerFactory() {
		return loggerFactory;
	}
}
