/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2005 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Authors: S.Chassande-Barrioz.
 * Created on 23 juin 2005
 *
 */
package org.objectweb.jorm.generator.lib;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Iterator;

import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Constants;
import org.objectweb.asm.Type;
import org.objectweb.jorm.compiler.api.PExceptionCompiler;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.CompositeName;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.jorm.naming.api.PNameManager;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.util.io.api.TargetHolder;
import org.objectweb.jorm.util.lib.StringReplace;

/**
 * This class is a helper for the byte code enhancement.
 * 
 * @author S.Chassande-Barrioz
 */
public class CommonJVMHelper extends CommonHelper implements Constants {

    public static final String getJVMClassName(String className) {
        return StringReplace.replaceChar('.', '/', className);
    }

    public static final String getJVMClassName(java.lang.Class clazz) {
        return getJVMClassName(clazz.getName());
    }

    public static final String getJVMClassName(Class clazz) {
        return getJVMClassName(clazz.getFQName());
    }

    public static final String getJVMType(java.lang.Class clazz) {
        return getJVMType(clazz.getName());
    }

    public static final String getJVMType(String className) {
        return "L" + getJVMClassName(className) + ";";
    }

    public static final String getJVMType(PType pt) {
        return Type.getType(pt.getJavaClass()).getDescriptor();
    }

    public static void write(ClassWriter cw, String name,
                             TargetHolder holder) throws PExceptionCompiler {
        try {
            // Fetch a block with the fileName
            String fileName = StringReplace.replaceChar
                ('/', File.separatorChar, name) + ".class";
            FileOutputStream fos = holder.getFileOutputStream(fileName);
            fos.write(cw.toByteArray());
            fos.close();
        } catch (Exception e) {
            throw new PExceptionCompiler
                (e, "Problem while generating PAccessor.");
        }
    }

    public static void writeJavaClass(
            final String name,
            final ClassWriter jclass, 
            final String srcFiles) throws IOException {
        int p = name.lastIndexOf(File.separatorChar);
        String pkg;
        String clas;
        if (p == -1) {
            pkg = "";
            clas = name;
        } else {
            pkg = name.substring(0, p);
            clas = name.substring(p + 1);
        }
        File outputDir;
        if (srcFiles == null) {
            outputDir = new File(pkg);
        } else {
            outputDir = new File(srcFiles + File.separatorChar + pkg);
        }

        outputDir.mkdirs();
        File outputFile = new File(outputDir, clas + ".class");
        outputFile.createNewFile();
        FileOutputStream fos = new FileOutputStream(outputFile);
        fos.write(jclass.toByteArray());
        fos.close();
    }

    public final String getConstructorFieldsPncParam(CompositeName co) {
        StringBuffer constructParam = new StringBuffer();
        for (Iterator it = co.getAllFields().iterator(); it.hasNext();) {
            TypedElement te = (TypedElement) it.next();
            constructParam.append(getJVMType(te.getType()));
        }
        constructParam.append(getJVMType(PNameManager.class));
        return constructParam.toString();
    }

    public static final void loadNull(CodeVisitor cv, PType type) {
        switch(type.getTypeCode()){
        case PType.TYPECODE_BOOLEAN:
            cv.visitInsn(ICONST_0); //false
            break;
        case PType.TYPECODE_CHAR:
            cv.visitIntInsn(BIPUSH, 48); //'0'
            break;
        case PType.TYPECODE_BYTE: //(byte) -1
            cv.visitInsn(ICONST_M1);
            cv.visitInsn(I2B);
            break;
        case PType.TYPECODE_SHORT: //(short) -1
            cv.visitInsn(ICONST_M1);
            cv.visitInsn(I2S);
            break;
        case PType.TYPECODE_INT: // -1
            cv.visitInsn(ICONST_M1);
            break;
        case PType.TYPECODE_LONG: //(long) -1
            cv.visitInsn(ICONST_M1);
            cv.visitInsn(I2L);
            break;
        case PType.TYPECODE_FLOAT: //(float) -1
            cv.visitInsn(ICONST_M1);
            cv.visitInsn(I2F);
            break;
        case PType.TYPECODE_DOUBLE: //(double) -1
            cv.visitInsn(ICONST_M1);
            cv.visitInsn(I2D);
            break;
        default:
            cv.visitInsn(ACONST_NULL);
            break;
        }
    }

    public static final void any2I(CodeVisitor cv, String type) {
        any2I(cv, Type.getType(type));
    }

    public static final void any2I(CodeVisitor cv, Type type) {
        switch (type.getSort()) {
        case Type.DOUBLE:
            cv.visitInsn(D2I);
            break;
        case Type.FLOAT:
            cv.visitInsn(F2I);
            break;
        case Type.LONG:
            cv.visitInsn(L2I);
            break;
        default:
            // noop
            break;
        }
    }

}
