/**
 * Copyright (C) 2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.jorm.generator.lib;

import org.apache.velocity.VelocityContext;
import org.apache.velocity.context.Context;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.compiler.api.JormCompilerParameter;
import org.objectweb.jorm.compiler.api.PExceptionCompiler;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.Package;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.util.io.api.TargetHolder;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.api.Operand;
import org.objectweb.medor.expression.api.Operator;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.lib.Equal;
import org.objectweb.medor.expression.lib.NotEqual;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.io.File;
import java.io.FileWriter;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 *
 * @author S.Chassande-Barrioz
 */
public class FPNCGenerator extends CommonGenerator {

	public CommonHelper helper;



	public void generate(Class clazz,
	                     TargetHolder holder,
	                     JormCompilerParameter cp) throws PException {

		Collection nds = clazz.getNameDefs();
		Class classWithNd = clazz;
		while(nds.size() < 1 && classWithNd.getInheritedClassNumber() > 0) {
			classWithNd = (Class) classWithNd.getSuperClasses().iterator().next();
			nds = classWithNd.getNameDefs();
		}
		if (nds.size() < 1) {
			logger.log(BasicLevel.WARN, "FPNC: No NameDef defined for the class " + clazz.getFQName());
			return;
		}
		if (nds.size() == 1) {
			generate(clazz, (NameDef) nds.iterator().next(), null, holder);
		} else {
			Iterator it = nds.iterator();
			while(it.hasNext()) {
				NameDef nd = (NameDef) it.next();
				String ndName = (nd.getName() != null
				    && nd.getName().length() == 0
				    ? null
				    : nd.getName());
				generate(clazz, nd, ndName, holder);
			}
		}
	}

	/**
	 * This method generates a XFPNC file corresponding to the clazz
	 * parameter in the directory parameter.
	 * @param clazz		The clazz requiring the generation of a PNamingContext.
	 * @param nd		The NameDef to use
	 * @param ndName    The name def name to use fot the package name
	 * (can be null)
	 * @param holder	The target holder which allows to create files.
	 */
	private void generate(Class clazz,
	                      NameDef nd,
	                      String ndName,
	                      TargetHolder holder) throws PException {
		Expression exp;
		try {
			exp = clazz.getInheritanceFilter(nd);
		} catch (ExpressionException e) {
			throw new PException(e);
		}
		if (exp == null) {
			return;
		}
		PType expType = exp.getType();
		if (expType.getTypeCode() != PType.TYPECODE_BOOLEAN
			&& expType.getTypeCode() != PType.TYPECODE_OBJBOOLEAN) {
			return;
		}
		

		// Calculate fileName
		String packName = ((Package) clazz.getParent()).getName();
		if (ndName != null) {
			if (packName != null && packName.length() > 0) {
				packName += "." + ndName;
			} else {
				packName = ndName;
			}
		}
		String className = clazz.getName() + "FPNC";
		String fileName = className;
		if (packName != null && packName.length() > 0) {
			fileName = packName + "." + fileName;
			fileName = fileName.replace('.', File.separatorChar);
		}
		logger.log(BasicLevel.DEBUG, "Generate the " + fileName);


		Context ctx = new VelocityContext();
		ctx.put("class", clazz);
		ctx.put("helper", this);
		ctx.put("header", GEN_TEMPLATE_DIR + "Header.vm");
		ctx.put("className", className);
		ctx.put("packName", packName);
		boolean isComposite = false;
		if (nd.isNameRef()) {
			isComposite = true;
			ctx.put("isComposite", Boolean.TRUE);
			ctx.put("codingType", "PNameCoder.CTCOMPOSITE");
			ctx.put("compositeName", nd.getNameRef().getCompositeName());
		} else if (nd.isFieldName()) {
			isComposite = false;
			ctx.put("isComposite", Boolean.FALSE);
			TypedElement te = clazz.getTypedElement(nd.getFieldName());
			if (te == null) {
				throw new PException("The field '" + nd.getFieldName()
				    + "' used in the namedef of the class '"
				    + clazz.getFQName() + "' is not defined.");
			} else if (!(te instanceof PrimitiveElement)) {
				throw new PException("MalFormed Meta information: the field '"
				    + nd.getFieldName() + "' used in the namedef of the class '"
				    + clazz.getFQName() + "' is not a primitive element.");
			}
			ctx.put("codingType", getCTDeclaration(te.getType()));
		}

		FPNCMatchInfo mi;
		try {
			mi = new FPNCMatchInfo(exp, isComposite, logger);
		} catch (ExpressionException e) {
			throw new PException(e, "Generation of " + fileName);
		}
		if (mi.filter.length() == 0) {
			mi.filter.append("true");
		}
		ctx.put("mi", mi);

		//Generate the file
		try {
			FileWriter fw = holder.getFileWriter(fileName + ".java");
			if (template == null) {
				template = velocityEngine.getTemplate(GEN_TEMPLATE_DIR + "FPNC.vm");
			}
			template.merge(ctx, fw);
			fw.flush();
			fw.close();
		} catch (Exception e) {
			throw new PExceptionCompiler(e, "Problem while generating " + fileName + ".java");
		}
	}

	public class FPNCMatchInfo extends NamingFilterExpressionHelper {

		/**
		 * Indicates if the naming is based on a composite name.
		 */
		public boolean isComposite;

		/**
		 * When the naming is based on a single field (not a composite). This field
		 * is the type of the naming field.
		 */
		public PType singleFieldPType = null;

		/**
		 * When the naming is based on a single field (not a composite). This field
		 * is the name of the naming field. By default this field is equals to 'o'.
		 * this field is used as parameter name in match methods.
		 */
		public String singleFieldName = "o";

		/**
		 * Builds a FPNCMatchInfo fro an expression.
		 * @param e is the filter expression
		 * @param isComposite indicates if the naming is composite or not. If yes a
		 * composite name will be used to reach naming field.
		 */
		public FPNCMatchInfo(Expression e,
							 boolean isComposite,
							 Logger logger)
			throws PException, ExpressionException {
			super(logger);
			this.isComposite = isComposite;
			fillMatchInfo(e);
		}

		public PType getSingleFieldPType() {
			return singleFieldPType;
		}

		public String getSingleFieldName() {
			return singleFieldName;
		}

		public String toString() {
			return super.toString()
				+ " / singleFieldName=" + singleFieldName
				+ " / singleFieldPType=" + singleFieldPType
				;
		}

		protected void fillMatchInfo(ParameterOperand po) throws PException, ExpressionException {
			logger.log(BasicLevel.DEBUG, "ParameterOperand " + po.getName());
			if (isComposite) {
				if (field2declarations.get(po.getName()) == null) {
					if (field2declarations.size() == 0) {
						field2declarations.put("", "PNameGetter png = (PNameGetter) o;");
					}
					StringBuffer declaration = new StringBuffer();
					declaration.append(po.getType().getJavaName());
					declaration.append(' ');
					declaration.append(po.getName());
					declaration.append(" = png.pnGet");
					declaration.append(getCoderName(po.getType()));
					declaration.append("Field(\"");
					declaration.append(po.getName());
					declaration.append("\");");
					field2declarations.put(po.getName(), declaration);
					if (logger.isLoggable(BasicLevel.DEBUG)) {
						logger.log(BasicLevel.DEBUG, "Add declaration: " + declaration);
					}
				}
				filter.append(po.getName());
			} else {
				if (singleFieldPType != null && singleFieldPType != po.getType()) {
					throw new ExpressionException("Several parameters with a non composite name");
				}
				singleFieldPType = po.getType();
				singleFieldName = po.getName();
				filter.append(singleFieldName);
				logger.log(BasicLevel.DEBUG, "singleFieldPType " + singleFieldPType);
				logger.log(BasicLevel.DEBUG, "singleFieldName " + singleFieldName);
			}
		}
	}
}
