/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.generator.lib;

import org.apache.velocity.VelocityContext;
import org.apache.velocity.context.Context;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.compiler.api.JormCompilerConfigurator;
import org.objectweb.jorm.compiler.api.JormCompilerParameter;
import org.objectweb.jorm.compiler.api.PExceptionCompiler;
import org.objectweb.jorm.generator.api.MOP;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ClassMapping;
import org.objectweb.jorm.metainfo.api.ClassProject;
import org.objectweb.jorm.metainfo.api.CompositeName;
import org.objectweb.jorm.metainfo.api.GenClass;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.Mapping;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.Package;
import org.objectweb.jorm.metainfo.api.ParentClassMapping;
import org.objectweb.jorm.metainfo.api.PrimitiveElementMapping;
import org.objectweb.jorm.metainfo.api.ReferenceMapping;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.jorm.type.api.PExceptionTyping;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.jorm.util.io.api.TargetHolder;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.util.monolog.api.BasicLevel;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

/**
 * This class is a generator of XPMapping objects. It is generic either the
 * mapper type. This generator use the velocity tools. The used template is
 * Mapping.vm . Associated to this template, this
 * generator builds a velocity context which contains the following
 * information:<br/>
 * <table border=1>
 * 	<tr><td>Key</td><td>Value</td></tr>
 * 	<tr><td>"class"</td><td>The reference to the Class meta object</td></tr>
 * 	<tr>
 * 		<td>"tools"</td>
 * 		<td>The reference to the <a href="CommonHelper.html">CommonHelper<a/></td>
 * </tr>
 * 	<tr>
 * 		<td>"header"</td>
 * 		<td>The file name of the common template which contains the header of
 * 			the generate files.</td>
 * </tr>
 * 	<tr>
 * 		<td>"mappingTools"</td>
 * 		<td>The reference to current instance of MappingGenerator.</td>
 * </tr>
 * 	<tr>
 * 		<td>"cparam"</td>
 * 		<td>The reference to a CompilerParameter instance. This instance
 * 		describes the user compilation parameter.</td>
 * </tr>
 * </table>
 *
 * As a XXXMapping class is specific to a mapper, the Mapping.vm template uses
 * externals macros which must be defined as an external template library in
 * the velocity tools.
 */
public class MappingGenerator extends CommonGenerator {
	/**
	 * The mapper generation specializations (some MOP objects)
	 * key = mapper name
	 * value = a MOP
	 */
	protected HashMap mmops = new HashMap();

	/**
	 * This constrcutor permits to specify which BindingMOP must be used
	 */
	public MappingGenerator(JormCompilerConfigurator jcc, JormCompilerParameter cp) throws PException {
		super();
		Iterator it = jcc.knownMappers();
		while (it.hasNext()) {
			String mn = (String) it.next();
			mmops.put(mn, jcc.getMOPFactory(mn).createMappingMOP(cp, mn));
		}
	}

	public String[][] getTemplateLibraries() {
		String[][] res = new String[mmops.size()][];
		Iterator it = mmops.values().iterator();
		int i = 0;
		while (it.hasNext()) {
			res[i] = ((MOP) it.next()).getTemplateLibraries();
			i++;
		}
		return res;
	}

    /**
     * Generates java code mapping for a jorm class. 
     * Code in put in a file named XMapping.java whre X is the name of the class
     * Iterates java code generation for each mapping defined for this project
     * @param clazz     the meta object interface
     * @param holder    The target holder which allows to create files
     * @param cp        the compilation parameters
     * @param jcc       the compiler configurator (not used at the moment)
     */
    public void generate(Class clazz,
                         TargetHolder holder,
                         JormCompilerParameter cp,
                         JormCompilerConfigurator jcc) throws PException {
            concreteGenerate(clazz, holder, cp);
    }

	/**
     * Generates java code mapping for a jorm CONCRETE class. 
     * @see generate
	 * @param clazz		the meta object interface
	 * @param holder	The target holder which allows to create files
	 * @param cp		the compilation parameters
	 */
	private void concreteGenerate(Class clazz,
						 TargetHolder holder,
						 JormCompilerParameter cp) throws PException {

        // find the class project corresponding to the project name
		ClassProject cpr = null;
		Iterator it = clazz.getClassProjects().iterator();
		while (it.hasNext()) {
			ClassProject tmpCp = (ClassProject) it.next();
			if (tmpCp.getProjectName().equals(cp.getProjectName())) {
                cpr = tmpCp;
				break;
			}
		}
		if (cpr == null) {
			throw new PExceptionCompiler("Cannot find the relevant project!");
		}

        // for all its mapping, generate the corresponding code
		it = cpr.getMappings().iterator();
		while (it.hasNext()) {
			Mapping mo = (Mapping) it.next();
			concreteGenerateWithMapper(mo.getMapperName(), clazz, holder, cp, mo);
			logger.log(BasicLevel.INFO, "     -> \"" + mo.getMapperName()
					+ "\" mapping generated.");
		}
	}

    /**
     * Generate java code mapping for a concrete jorm class with a known mapper
     * (in a known project)
     * @param mn name of the mapper
     * @param co jorm class
     * @param holder holder of the destination file
     * @param cp compiler parameters
     * @param mo mapping from which generating java code
     * @throws PException
     */

	private void concreteGenerateWithMapper(String mn, Class co,
                                            TargetHolder holder,
                                            JormCompilerParameter cp,
                                            Mapping mo) throws PException {
		String pack = ((Package) co.getParent()).getName();
		String packroot = pack;
		if (cp.isGeneratedWithMapperPackage()) {
			pack = ("".equals(pack) ? mn : pack + "." + mn);
		}
		// File name
		String fileName = co.getName() + PMapper.PCLASSMAPPINGAPPENDER;
		if (pack.length() > 0) {
			fileName = (pack + '.').replace('.', File.separatorChar) + fileName;
		}
		logger.log(BasicLevel.DEBUG, "Generate the " + fileName + " class");
		MOP mmop = (MOP) mmops.get(mo.getMapperName());

		// Fetch a block with the fileName
		Context ctx = new VelocityContext();
		ctx.put("abstractClass", new Boolean(co.isAbstract()));
		ctx.put("velocityengine", velocityEngine);
		ctx.put("holder", holder);
		ctx.put("class", co);
		ctx.put("mapping", mo);
		ctx.put("dependencies", mo.getClassMapping().getDependencies());
		if (pack.length() > 0) {
			ctx.put("package", pack);
			ctx.put("packageroot", packroot);
		}
		ctx.put("mappername", mn);
		ctx.put("noBindingCreation", new Boolean(cp.isBindingAbstract()));
		ctx.put("cparam", cp);
		CommonHelper ch = new CommonHelper();
		ch.setLogger(logger);
		ctx.put("containsCharArrayField", new Boolean(ch.containsCharArrayField(mo)));
		ctx.put("tools", ch);
		ctx.put("header", GEN_TEMPLATE_DIR + "Header.vm");
		ctx.put("mappingTools", this);
		ctx.put("bindingTools", this);
        ((Loggable) mmop).setLogger(logger);

        List pems = mo.getClassMapping().getAllPrimitiveElementMappings();
		ctx.put("pemlist",
			pems.toArray(new PrimitiveElementMapping[pems.size()]));

		mmop.initContext(ctx, co, cp.getProjectName(), mn);

		ctx.put("useNamingKey", Boolean.FALSE);
		NameDef nd = mo.getClassMapping().getIdentifierMapping().getNameDef();
		Class ancestorClass = getAncestor(co);
		Expression namingfilter;
		try {
			namingfilter = ancestorClass.getInheritanceFilter(nd);
		} catch (ExpressionException e) {
			throw new PException(e);
		}
		Object namingKey = co.getInheritanceNamingKey(nd);
		if (namingfilter != null && namingKey !=null) {
			ctx.put("useNamingKey", Boolean.TRUE);
			//calculate the type of the expression
			PType namingKeyType = namingfilter.getType();
			ctx.put("namingKeyTypeStr", ptype2String(namingKeyType));
			String namingKeyStr = null;;
			switch(namingKeyType.getTypeCode()) {
			case PType.TYPECODE_BOOLEAN:
				if (namingKey instanceof Boolean) {
					namingKeyStr = namingKey.toString();
				}
				break;
			case PType.TYPECODE_OBJBOOLEAN:
				if (namingKey instanceof Boolean) {
					namingKeyStr = (((Boolean) namingKey).booleanValue()
						? "Boolean.TRUE" : "Boolean.FALSE");
				}
				break;
			case PType.TYPECODE_BYTE:
			case PType.TYPECODE_SHORT:
			case PType.TYPECODE_INT:
			case PType.TYPECODE_LONG:
				PType jlpt = getJavaLangPrimitive(namingKeyType);
				if (namingKey instanceof String) {
					namingKeyStr = jlpt.getJavaName() + ".valueOf"
						+"(\"" + namingKey.toString() + "\")";
				} else {
					namingKeyStr = "new " + jlpt.getJavaName()
						+ "((" + namingKeyType.getJavaName()
						+ ") " + namingKey.toString() + ")";
				}
				break;
			case PType.TYPECODE_OBJBYTE:
			case PType.TYPECODE_OBJSHORT:
			case PType.TYPECODE_OBJINT:
			case PType.TYPECODE_OBJLONG:
				if (namingKey instanceof String) {
					namingKeyStr = "new " + namingKeyType.getJavaName()
						+ "(\"" + namingKey.toString() + "\")";
				} else {
					namingKeyStr = "new " + namingKeyType.getJavaName()
						+ "((" + getPrimitive(namingKeyType).getJavaName()
						+ ") " + namingKey.toString() + ")";
				}
				break;
			case PType.TYPECODE_STRING:
				namingKeyStr = "\"" + namingKey.toString() + "\"";
				break;
			}

			if (namingKey == null) {
				throw new PExceptionTyping("Naming key value no corresponding to the type of the filter: "
					+ "filter type: " + namingKeyType.getJavaName()
					+ " / naming key value: " + namingKey);
			}
			ctx.put("namingKey", namingKeyStr);
		}

		ArrayList gcms = new ArrayList();
		getGenClassMapping(mo, cp.getProjectName(), gcms);
		ctx.put("gcms", gcms);

		List crms = getClassReferenceMapping(mo, cp.getProjectName());
		ctx.put("crms", crms);
		ctx.put("classND", nd);

		try {
			FileWriter fw = holder.getFileWriter(fileName + ".java");
			if (template == null) {
				template = velocityEngine.getTemplate(GEN_TEMPLATE_DIR + "Mapping.vm");
			}
			template.merge(ctx, fw);
			fw.flush();
			fw.close();
		} catch (Exception e) {
			throw new PExceptionCompiler(e, "Problem while writing Mapping java file: " + fileName + ".java");
		}
	}


    /**
     * Generates java code mapping for an abstract jorm class
     * @param clazz the jorm class
     * @param holder holder of the destination file
     * @param cp jorm compilation parameters
     * @throws PException
     */
    private void abstractGenerate(Class clazz,
                     TargetHolder holder,
                     JormCompilerParameter cp) throws PException {
        HashSet mapperNames = new HashSet();
        getMapperNames(clazz, cp.getProjectName(), mapperNames);
        for (Iterator itMapperName = mapperNames.iterator(); itMapperName.hasNext();) {
            String mapperName = (String) itMapperName.next();
            Collection nds = clazz.getNameDefs();
            Class classWithNd = clazz;
            while(nds.size() < 1 && classWithNd.getInheritedClassNumber() > 0) {
                // get first super class
                classWithNd = (Class) classWithNd.getSuperClasses().iterator().next();
                nds = classWithNd.getNameDefs();
            }
            if (nds.size() < 1) {
                logger.log(BasicLevel.WARN, "AbstractClass: No NameDef defined for the class " + clazz.getFQName());
                return;
            }
            Iterator it = nds.iterator();
            boolean useNdName = nds.size() > 1; 
            while(it.hasNext()) {
                NameDef nd = (NameDef) it.next();
                abstractGenerateWithNameDef(mapperName, clazz, holder, cp,nd, useNdName);
            }
        }
}

    private void getMapperNames(Class clazz, String projectName, Set res) {
        if (clazz.isAbstract()) {
            for (Iterator itClass = clazz.getSubClasses().iterator(); itClass.hasNext();) {
                Class subClass = (Class) itClass.next();
                getMapperNames(subClass, projectName, res);
            }
        } else {
            res.addAll(clazz.getClassProject(projectName).getMappers());
        }
    }

    private void abstractGenerateWithNameDef(String mapperName,
                                             Class clazz,
                                             TargetHolder holder,
                                             JormCompilerParameter cp,
                                             NameDef nd,
                                             boolean useNdName) throws PException {
        String packName = ((Package) clazz.getParent()).getName();
        if (cp.isGeneratedWithMapperPackage()) {
            packName = ("".equals(packName) ? mapperName : packName + "." + mapperName);
        }

         // File name
         String fileName = clazz.getName() + PMapper.PCLASSMAPPINGAPPENDER;
         if (useNdName) {
             if (packName != null && packName.length() > 0) {
                 packName += "." + nd.getName();
             } else {
                 packName = nd.getName();
             }
         }
         if (packName.length() > 0) {
             fileName = (packName + '.').replace('.', File.separatorChar) + fileName;
         }
         logger.log(BasicLevel.DEBUG, "Generate the " + fileName + " class");
         MOP mmop = (MOP) mmops.get(mapperName);
         Context ctx = new VelocityContext();
         ctx.put("class", clazz);
         if (packName.length() > 0) {
             ctx.put("package", packName);
         }
         ctx.put("cparam", cp);
         CommonHelper ch = new CommonHelper();
         ch.setLogger(logger);
         ctx.put("tools", ch);
         ctx.put("mappername", mapperName);
         ctx.put("noBindingCreation", Boolean.TRUE);
         ctx.put("abstractClass", Boolean.TRUE);
         ctx.put("header", GEN_TEMPLATE_DIR + "Header.vm");
         ctx.put("mappingTools", this);
 		 ctx.put("classND", nd);

         ClassProject classproject = clazz.getClassProject(cp.getProjectName());
         ArrayList dependencies = new ArrayList();
         List pems = null;
         if (classproject != null) {
             Collection ms = classproject.getMappings();
             if (!ms.isEmpty()) {
                 Iterator it = ms.iterator();
                 while(it.hasNext()) {
                     Mapping m = (Mapping) it.next();
                     pems = m.getClassMapping().getAllPrimitiveElementMappings();
                     ClassMapping cm = m.getClassMapping();
                     if (cm != null) {
                         dependencies.addAll(cm.getDependencies());
                     }
                 }
             }
         }
         if (pems != null) {
         	ctx.put("pemlist",
 				pems.toArray(new PrimitiveElementMapping[pems.size()]));
         }
        ((Loggable) mmop).setLogger(logger);
         mmop.initContext(ctx, clazz, cp.getProjectName(), mapperName);
         ctx.put("dependencies", dependencies);
         FileWriter fw = null;
         try {
             // Fetch a block with the fileName
             fw = holder.getFileWriter(fileName + ".java");
             velocityEngine.getTemplate(GEN_TEMPLATE_DIR + "Mapping.vm").merge(ctx, fw);
         } catch (Exception e) {
             throw new PExceptionCompiler(e, "Problem while writing Mapping java file: " + fileName + ".java");
         } finally {
            try {
            	if (fw != null) {
            		fw.flush();
            		fw.close();
            	}
			} catch (IOException e1) {
				e1.printStackTrace();
			}
         }

     }

	private void getGenClassMapping(Mapping mo, String projname, ArrayList gcms) {
		gcms.addAll(mo.getGenClassMappings());
		for(Iterator it = mo.getClassMapping().getParentClassMappings().iterator();
			it.hasNext();) {
			getGenClassMapping(((ParentClassMapping)  it.next()).getMOClass()
				.getClassProject(projname)
				.getMapping(mo.getMapperName()),
				projname,
				gcms);
		}
	}

	private List getClassReferenceMapping(Mapping mapping, String projname) {
	    HashMap ref2refmap = new HashMap();
	    ArrayList ms = new ArrayList();
	    ms.add(mapping);
	    while(!ms.isEmpty()) {
	        //Current Mapping to treat
	        Mapping mo = (Mapping) ms.remove(0);
	        //get ReferenceMapping instances defined in the current Mapping
	        //avoid to register several reference mapping for the same reference
	        for(Iterator it = mo.getClassMapping().getReferenceMappings().iterator();it.hasNext();) {
	            ReferenceMapping rm = (ReferenceMapping) it.next(); 
	            Object old = ref2refmap.put(rm.getLinkedMO(), rm);
	            if (old != null) {
	                //keep the first ReferenceMapping found
	                ref2refmap.put(rm.getLinkedMO(), old);
	            }
	        }
			for(Iterator it = mo.getClassMapping().getParentClassMappings().iterator();
				it.hasNext();) {
			    ms.add(((ParentClassMapping)  it.next()).getMOClass()
					.getClassProject(projname)
					.getMapping(mo.getMapperName()));
			}
	    }
	    return new ArrayList(ref2refmap.values());
	}

	private Class getAncestor(Class clazz) throws PException {
		Collection superClasses = clazz.getSuperClasses();
		if (superClasses.isEmpty()) {
			return clazz;
		}
		if (superClasses.size() == 1) {
			return getAncestor((Class) superClasses.iterator().next());
		}
		throw new PException("Multiple inheritance not supported in the current implementation");
	}

	// METHOD USED BY THE TEMPLATE //
	//-----------------------------//

	public String getOwner(TypedElement te, Class owner, String ownerName) {
		return (te.getParent() == owner
			? ownerName
			: "\"" + ((Class) te.getParent()).getFQName() + "\"");
	}

	/**
	 * This generation helper method retrieves the string corresponding
	 * to the codding type (short value) associated to a PType.
	 */
	public String ptype2CTString(PType pt) throws PExceptionTyping {
		if (pt == null) {
			throw new PExceptionTyping("null parameter forbidden");
		}
		switch (pt.getTypeCode()) {
		case PType.TYPECODE_CHAR:
			return "PNamingContext.CTCHAR";
		case PType.TYPECODE_OBJCHAR:
			return "PNamingContext.CTOCHAR";
		case PType.TYPECODE_BYTE:
			return "PNamingContext.CTBYTE";
		case PType.TYPECODE_OBJBYTE:
			return "PNamingContext.CTOBYTE";
		case PType.TYPECODE_SHORT:
			return "PNamingContext.CTSHORT";
		case PType.TYPECODE_OBJSHORT:
			return "PNamingContext.CTOSHORT";
		case PType.TYPECODE_INT:
			return "PNamingContext.CTINT";
		case PType.TYPECODE_OBJINT:
			return "PNamingContext.CTOINT";
		case PType.TYPECODE_LONG:
			return "PNamingContext.CTLONG";
		case PType.TYPECODE_OBJLONG:
			return "PNamingContext.CTOLONG";
		case PType.TYPECODE_STRING:
			return "PNamingContext.CTSTRING";
		case PType.TYPECODE_DATE:
			return "PNamingContext.CTDATE";
		case PType.TYPECODE_CHARARRAY:
			return "PNamingContext.CTCHARARRAY";
		case PType.TYPECODE_BYTEARRAY:
			return "PNamingContext.CTBYTEARRAY";
		default:
			throw new PExceptionTyping("The following " + pt.getJormName()
									   + " type has'nt been allowed to be used as a PName field");
		}

	}

	public PType getJavaLangPrimitive(PType pt) {
		switch (pt.getTypeCode()) {
		case PType.TYPECODE_BOOLEAN:
		case PType.TYPECODE_OBJBOOLEAN:
			return PTypeSpace.OBJBOOLEAN;
		case PType.TYPECODE_CHAR:
		case PType.TYPECODE_OBJCHAR:
			return PTypeSpace.OBJCHAR;
		case PType.TYPECODE_BYTE:
		case PType.TYPECODE_OBJBYTE:
			return PTypeSpace.OBJBYTE;
		case PType.TYPECODE_SHORT:
		case PType.TYPECODE_OBJSHORT:
			return PTypeSpace.OBJSHORT;
		case PType.TYPECODE_INT:
		case PType.TYPECODE_OBJINT:
			return PTypeSpace.OBJINT;
		case PType.TYPECODE_LONG:
		case PType.TYPECODE_OBJLONG:
			return PTypeSpace.OBJLONG;
		default:
			return null;
		}
	}

	public PType getPrimitive(PType pt) {
		switch (pt.getTypeCode()) {
		case PType.TYPECODE_BOOLEAN:
		case PType.TYPECODE_OBJBOOLEAN:
			return PTypeSpace.BOOLEAN;
		case PType.TYPECODE_CHAR:
		case PType.TYPECODE_OBJCHAR:
			return PTypeSpace.CHAR;
		case PType.TYPECODE_BYTE:
		case PType.TYPECODE_OBJBYTE:
			return PTypeSpace.BYTE;
		case PType.TYPECODE_SHORT:
		case PType.TYPECODE_OBJSHORT:
			return PTypeSpace.SHORT;
		case PType.TYPECODE_INT:
		case PType.TYPECODE_OBJINT:
			return PTypeSpace.INT;
		case PType.TYPECODE_LONG:
		case PType.TYPECODE_OBJLONG:
			return PTypeSpace.LONG;
		default:
			return null;
		}
	}

	/**
	 * This generation helper method retrieves the string corresponding
	 * to the instance of the PType.
	 */
	public String ptype2String(PType pt) throws PExceptionTyping {
		if (pt == null) {
			throw new PExceptionTyping("null parameter forbidden");
		}
		return pt.getProgName();
	}

	/**
	 * This method is a generation helper. It returns a string like the
	 * following regular expression: (".getGenClassMapping()")*
	 * where the number of repetition of the string is equals to the level
	 * parameter.
	 */
	public String getGenclass(int level) {
		String res = "";
		for (int i = 0; i < level; i++) {
			res += ".getGenClassMapping()";
		}
		return res;
	}

	/**
	 * This method calculates the number of indexes in a GenClass. This method
	 * should be moved into the meta object GenClass.
	 */
	public int getIndexNumber(GenClass gc) {
		Iterator it = gc.iterateIndexField();
		int res = 0;
		while (it.hasNext()) {
			res++;
			it.next();
		}
		return res;
	}

	public TypedElement getLastElement(GenClassRef gcr) {
		TypedElement o = gcr;
		while (o instanceof GenClassRef) {
			GenClassRef g = ((GenClassRef) o);
			if (g.isClassRef())
				o = g.getClassRef();
			else if (g.isGenClassRef())
				o = g.getClassRef();
			else if (g.isPrimitive())
				o = g.getPrimitiveElement();
		}
		return o;
	}

	public ArrayList getGenClassNames(GenClassRef gcr) {
		ArrayList gcns = new ArrayList(5);
		TypedElement o = gcr;
		while (o instanceof GenClassRef) {
			GenClassRef g = ((GenClassRef) o);
			gcns.add(g.getGenClassName());
			if (g.isClassRef())
				o = g.getClassRef();
			else if (g.isGenClassRef())
				o = g.getGenClassRef();
			else if (g.isPrimitive())
				o = g.getPrimitiveElement();
		}
		ArrayList res = new ArrayList(gcns.size());
		for (int i = 0; i < gcns.size(); i++) {
			String s = "new String[]{";
			for (int j = 0; j <= i; j++) {
				String elem = "\"" + gcns.get(j) + "\"";
				s += elem + (i == j ? "}" : ",");
			}
			res.add(s);
		}
		return res;
	}
}
