/**
 * Copyright (C) 2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.jorm.generator.lib;

import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.api.PException;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.api.Operator;
import org.objectweb.medor.expression.api.Operand;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.lib.Equal;
import org.objectweb.medor.expression.lib.NotEqual;
import org.objectweb.medor.expression.lib.Mod;
import org.objectweb.medor.expression.lib.Round;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.util.Map;
import java.util.HashMap;

/**
 * This class is a helper for managing the evaluation of medor expression in
 * naming.
 *
 * @author S.Chassande-Barrioz
 */
public abstract class NamingFilterExpressionHelper extends CommonHelper {

	/**
	 * A map containing a list of field declaration/initialisation. The key is
	 * the field name.
	 * key = String, field name
	 * value = String, declaration of the field
	 */
	public Map field2declarations = new HashMap();

	/**
	 * the expression evaluating the filter
	 */
	public StringBuffer filter = new StringBuffer();

	public Logger logger;


	public NamingFilterExpressionHelper(Logger logger)
		throws PException, ExpressionException {
		this.logger = logger;
	}

	public Map getField2declarations() {
		return field2declarations;
	}

	public String getFilter() {
		return filter.toString();
	}

	public String toString() {
		return super.toString()
			+ " / filter=" + filter.toString()
			+ " / field2declarations=" + field2declarations
			;
	}

	protected void fillMatchInfo(Expression e) throws PException, ExpressionException {
		if (e instanceof Round) {
		    PType type = ((Operator) e).getExpression(0).getType();
		    if (type != null && (
		            type.getTypeCode() == PType.TYPECODE_FLOAT
		            || type.getTypeCode() == PType.TYPECODE_DOUBLE)) {
		        filter.append("java.lang.Math.round(");
		        fillMatchInfo(((Operator) e).getExpression(0));
		        filter.append(')');
		    } else {
		        fillMatchInfo(((Operator) e).getExpression(0));
		    }
		    
		} else if (e instanceof Operator) {
			Operator operator = (Operator) e;
			logger.log(BasicLevel.DEBUG, "Operator " + operator.getOperatorString());
			int nb = operator.getOperandNumber();
			if (nb == 1) {
				filter.append(operator.getOperatorString());
				fillMatchInfo(operator.getExpression(0));
			} else if (nb == 2) {
				String opPreString = "";
				String opString = " " + operator2String(operator) + " ";
				String opPostString = "";
				//particular case of the equal operator which is translated
				// - to the equals method on java object,
				// - to a equality (==) on primitive type,
				// - to the static method java.util.Arrays.equals(o1,o2) on arrays
				if (((e instanceof Equal) || (e instanceof NotEqual))
					&& (isObjectOperand(operator.getExpression(0))
						|| isObjectOperand(operator.getExpression(1)))) {
					opString = ".equals(";
					opPostString = ")";
					if (e instanceof NotEqual) {
						opPreString = "!";
					}
					if (isArrayOperand(operator.getExpression(0))
						|| isArrayOperand(operator.getExpression(1))) {
						opPreString += "java.util.Arrays.equals(";
						opString = ",";
					}
				}

				filter.append('(');
				filter.append(opPreString);
				fillMatchInfo(operator.getExpression(0));
				filter.append(opString);
				fillMatchInfo(operator.getExpression(1));
				filter.append(opPostString);
				filter.append(')');
			} else {
				throw new PException("Operators with more than 2 operands are not supported: " + e);
			}
		} else if (e instanceof ParameterOperand) {
			fillMatchInfo((ParameterOperand) e);
		} else if ((e instanceof Operand) && ((Operand) e).isDefined()) {
			Operand operand = (Operand) e;
			if (isObjectType(operand.getType())) {
				Object o = operand.getObject();
				if (o == null) {
					filter.append("null");
				} else {
					filterAppendOperand(operand);
				}
			} else {
				filterAppendOperand(operand);
			}
		}
	}

	private void filterAppendOperand(Operand operand) throws PException, ExpressionException{
		switch (operand.getType().getTypeCode()) {
		case PType.TYPECODE_CHAR:
		case PType.TYPECODE_OBJCHAR:
			filter.append("'");
			filter.append(operand.getChar());
			filter.append("'");
			break;

		case PType.TYPECODE_BYTE:
		case PType.TYPECODE_OBJBYTE:
			filter.append(operand.getByte());
			break;

		case PType.TYPECODE_SHORT:
		case PType.TYPECODE_OBJSHORT:
			filter.append(operand.getShort());
			break;

		case PType.TYPECODE_INT:
		case PType.TYPECODE_OBJINT:
			filter.append(operand.getInt());
			break;

		case PType.TYPECODE_LONG:
		case PType.TYPECODE_OBJLONG:
			filter.append(operand.getLong());
			filter.append('l');
			break;

		case PType.TYPECODE_STRING:
			filter.append('"');
			filter.append(operand.getString());
			filter.append('"');
			break;

		case PType.TYPECODE_BYTEARRAY: {
			filter.append("new byte[]{");
			String sep = "";
			byte[] arrays = operand.getByteArray();
			for(int i=0; i<arrays.length; i++) {
				filter.append(sep);
				sep = ",";
				filter.append(arrays[i]);
			}
			filter.append("}");
			break;
		}
		case PType.TYPECODE_CHARARRAY: {
			char[] arrays = operand.getCharArray();
				filter.append("new char[]{");
				String sep = "";
				for(int i=0; i<arrays.length; i++) {
					filter.append(sep);
					sep = ",";
					filter.append(arrays[i]);
				}
				filter.append("}");
			break;
		}
		case PType.TYPECODE_DATE:
			filter.append("new java.util.Date(");
			filter.append(operand.getDate().getTime());
			filter.append(')');
			break;
		case PType.TYPECODE_BIGDECIMAL:
			filter.append("new java.math.BigDecimal(\"");
			filter.append(operand.getBigDecimal().toString());
			filter.append("\")");
			break;

		case PType.TYPECODE_BIGINTEGER:
			filter.append("new java.math.BigInteger(\"");
			filter.append(operand.getBigDecimal().toString());
			filter.append("\")");
			break;

		default:
			
		}
	}
	
	private boolean isObjectOperand(Expression e) {
		return (e instanceof Operand) && isObjectType(e.getType());
	}

	private boolean isArrayOperand(Expression e) {
		return (e instanceof Operand)
		    && (e.getType().getTypeCode() == PType.TYPECODE_CHARARRAY
		           || e.getType().getTypeCode() == PType.TYPECODE_BYTEARRAY);
	}

	protected String operator2String(Operator op) {
		if (op instanceof Mod) {
			return "%";
		}
		return op.getOperatorString();
	}

	protected abstract void fillMatchInfo(ParameterOperand po)
		throws PException, ExpressionException;
	
}

