/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.generator.lib;

import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.velocity.VelocityContext;
import org.apache.velocity.context.Context;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Constants;
import org.objectweb.asm.Label;
import org.objectweb.asm.Type;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.compiler.api.JormCompilerParameter;
import org.objectweb.jorm.compiler.api.PExceptionCompiler;
import org.objectweb.jorm.metainfo.api.CompositeName;
import org.objectweb.jorm.metainfo.api.ScalarField;
import org.objectweb.jorm.metainfo.api.Package;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PNameGetter;
import org.objectweb.jorm.naming.api.PNameManager;
import org.objectweb.jorm.naming.lib.BasicCompositePName;
import org.objectweb.jorm.naming.lib.BasicPName;
import org.objectweb.jorm.naming.lib.CompositePBinder;
import org.objectweb.jorm.type.api.PExceptionTyping;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.util.io.api.TargetHolder;
import org.objectweb.jorm.util.lib.StringReplace;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class is a generator of XXXBinder objects. It is generic either the
 * mapper type. This generator use the velocity tools. The used template is
 * Binder.vm . Associated to this template this generator builds a velocity
 * context which contains the following information:<br/>
 * <table border=1>
 * 	<tr><td>Key</td><td>Value</td></tr>
 * 	<tr><td>"compositename"</td><td>The reference to the compositename meta object</td></tr>
 * 	<tr>
 * 		<td>"tools"</td>
 * 		<td>The reference to the <a href="CommonHelper.html">CommonHelper<a/></td>
 * </tr>
 * 	<tr>
 * 		<td>"header"</td>
 * 		<td>The file name of the common template which contains the header of
 * 			the generate files.</td>
 * </tr>
 * </table>
 */
public class PNameGenerator extends CommonGenerator {

    final String PARENT_CLASS_NAME = getJVMClassName(BasicCompositePName.class);
    
    /**
     * Defines the name of the field containing the null value for primitive 
     * field.
     */
    final static String NULLVALUE = "NULLVALUE";

    /**
     * Defines the name of the field containing the null value for Object 
     * field.
     */
    final static String NULLOVALUE = "NULLOVALUE";

    /**
     * Defines the name of the field containing the null value for char 
     * field.
     */
    final static String NULL_CHAR_VALUE = "NULL_CHAR_VALUE";
    	
	/**
	 * This method generates a XPName file corresponding to the pod
	 * parameter in the directory parameter.
	 * @param co		The meta object interface.
	 * @param holder	The target holder which allows to create files.
	 * @param cp		This parameter permits to reach the compilation parameters.
	 */
	public void generate(CompositeName co,
						 TargetHolder holder,
						 JormCompilerParameter cp) throws PException {
        ClassWriter cv = new ClassWriter(true);
	    PNameGenContext gc = new PNameGenContext(co, cv, this);
        if (debug) {
            logger.log(BasicLevel.DEBUG, "Generate the class '"
                + gc.classToWrite + "'.");
        }
        //create the class
        cv.visit(V1_1, ACC_PUBLIC + ACC_SUPER,
                gc.classToWrite, getJVMClassName(BasicCompositePName.class), 
                new String[]{getJVMClassName(co.getFQName() + "PNG")}, 
                gc.classToWrite);
        
        generateFields(gc);
		generateConstructorPnc(gc);
		generateConstructorPncPn(gc);
		generateConstructorFieldsPnc(gc);
		generateInit(gc);
		generateToString(gc);
		generateCheckPName(gc);
		generateValuesEqual(gc);
		generateHashCode(gc);
		generateNullPName(gc);
		generatePnGetSpecific(gc);
		generatePnGetGeneric(gc);
		
        try {
            // Fetch a block with the fileName
            String fileName = gc.classToWrite;
            fileName = StringReplace.replaceChar('/', File.separatorChar,
                    fileName)
                    + ".class";
            FileOutputStream fos = holder.getFileOutputStream(fileName);
            fos.write(cv.toByteArray());
            fos.close();
        } catch (Exception e) {
            throw new PExceptionCompiler(e,
                    "Problem while generating PAccessor.");
        }
	    
	}

	private void generateFields(PNameGenContext gc) {
        gc.clav.visitField(ACC_PUBLIC + ACC_STATIC, NULLVALUE, 
                Type.LONG_TYPE.getDescriptor(), new Long(-1), null);
        gc.clav.visitField(ACC_PUBLIC + ACC_STATIC, NULL_CHAR_VALUE, 
                Type.CHAR_TYPE.getDescriptor(), new Character('0'), null);
        gc.clav.visitField(ACC_PUBLIC + ACC_STATIC, NULLOVALUE, 
                getJVMType(Object.class), null, null);
        
        for (Iterator it = gc.coAllFields.iterator(); it.hasNext();) {
            TypedElement te = (TypedElement) it.next();
            String jvmFieldType = getJVMType(te.getType());
            //declare the field
            gc.clav.visitField(ACC_PROTECTED, te.getName(), jvmFieldType, null, null);
        }	    
	}
	
	private void generateConstructorPnc(PNameGenContext gc) {
        
        //public TotoPName(PNameManager pnc) { this.pnc = pnc; }
        CodeVisitor cv = gc.clav.visitMethod(ACC_PUBLIC, 
                "<init>", "(" + gc.pnameManagerJT + ")V", null, null);
	    //super();
        cv.visitVarInsn(ALOAD, 0);
        cv.visitMethodInsn(INVOKESPECIAL, PARENT_CLASS_NAME, "<init>", "()V");
        assignPNC(cv, gc, 1);
        cv.visitInsn(RETURN);
        cv.visitMaxs(0, 0);
	}
	
	private void generateConstructorPncPn(PNameGenContext gc) {
        //public TotoPName(PNameManager pnc, PName pn) {
	    CodeVisitor cv = gc.clav.visitMethod(ACC_PUBLIC, 
                "<init>", 
                "(" + gc.pnameManagerJT + gc.pnameJT + ")V", 
                null, null);
	    //super();
        cv.visitVarInsn(ALOAD, 0);
        cv.visitMethodInsn(INVOKESPECIAL, PARENT_CLASS_NAME, "<init>", "()V");
        
        // if (pn instanceof TotoPName) {
        cv.visitVarInsn(ALOAD, 2);
        cv.visitTypeInsn(INSTANCEOF, gc.classToWrite);
        Label labelElse = new Label();
        Label labelEnd = new Label();
        cv.visitJumpInsn(IFEQ, labelElse);
        {
	        for (Iterator it = gc.coAllFields.iterator(); it.hasNext();) {
	            //this.f1 = ((TotoPName) pn).f1;
	            TypedElement te = (TypedElement) it.next();
		        String jvmFieldType = getJVMType(te.getType());
		        cv.visitVarInsn(ALOAD, 0);
		        cv.visitVarInsn(ALOAD, 2);
		        cv.visitTypeInsn(CHECKCAST, gc.classToWrite);
		        cv.visitFieldInsn(GETFIELD, gc.classToWrite, te.getName(), jvmFieldType);
		        cv.visitFieldInsn(PUTFIELD, gc.classToWrite, te.getName(), jvmFieldType);
	        }
	        cv.visitJumpInsn(GOTO, labelEnd);
        }
        // } else {
        {
            cv.visitLabel(labelElse);
	        //PNameGetter png = (PNameGetter) pn.encodeAbstract();
	        cv.visitVarInsn(ALOAD, 2);
	        cv.visitMethodInsn(INVOKEINTERFACE, gc.pnameJCN, 
	                "encodeAbstract", "()" + gc.objectJT);
	        cv.visitTypeInsn(CHECKCAST, gc.pngJCN);
	        cv.visitVarInsn(ASTORE, 3);
	        //if (png instanceof ${compositename.getName()}PNG) {
	        cv.visitVarInsn(ALOAD, 3);
	        cv.visitTypeInsn(INSTANCEOF, gc.xpngJCN);
	        Label labelElse2 = new Label();
	        cv.visitJumpInsn(IFEQ, labelElse2);
	        {
	            for (Iterator it = gc.coAllFields.iterator(); it.hasNext();) {
	                //this.f1 = ((TotoPNG) png).pnGetF1(null);
	                TypedElement te = (TypedElement) it.next();
	                String jvmFieldType = getJVMType(te.getType());
	                cv.visitVarInsn(ALOAD, 0);
	                cv.visitVarInsn(ALOAD, 3);
	    	        cv.visitTypeInsn(CHECKCAST, gc.xpngJCN);
	    	        cv.visitMethodInsn(INVOKEINTERFACE, gc.xpngJCN, 
	    	                "pnGet" + upperFL(te.getName()), 
	    	                "()" + jvmFieldType);
	                cv.visitFieldInsn(PUTFIELD, gc.classToWrite, te.getName(), 
	                        jvmFieldType);
	            }
	            cv.visitJumpInsn(GOTO, labelEnd); //go directly after the global if
	        }
            cv.visitLabel(labelElse2);
            // } else {
            {
                //throw new PException();
                cv.visitTypeInsn(NEW, gc.pexceptionJCN);
                cv.visitInsn(DUP);
                cv.visitMethodInsn(INVOKESPECIAL, gc.pexceptionJCN, "<init>", "()V");
                cv.visitInsn(ATHROW);
            }
        }
        //}
        cv.visitLabel(labelEnd);
        assignPNC(cv, gc, 1);
        cv.visitInsn(RETURN);
        cv.visitMaxs(0, 0);
	}
	
	private void generateConstructorFieldsPnc(PNameGenContext gc) {

	    //public TotoPName(int f1, long f3, PNameManager pnc) { this.pnc = pnc; }
        CodeVisitor cv = gc.clav.visitMethod(ACC_PUBLIC, 
                "<init>", "(" + gc.xconstructorFieldsPnc + ")V", 
                null, null);
	    //super();
        cv.visitVarInsn(ALOAD, 0);
        cv.visitMethodInsn(INVOKESPECIAL, PARENT_CLASS_NAME, "<init>", "()V");
        int idx = assignFields(gc, cv);
        assignPNC(cv, gc, idx);
        cv.visitInsn(RETURN);
        cv.visitMaxs(0, 0);
	}
	
	private void generateInit(PNameGenContext gc) {
        CodeVisitor cv = gc.clav.visitMethod(ACC_PUBLIC, 
                "init", "(" + gc.xconstructorFieldsPnc + ")V", null, null);
        int idx = assignFields(gc, cv);
        assignPNC(cv, gc, idx);
        cv.visitInsn(RETURN);
        cv.visitMaxs(0, 0);
	}
	
	private void generateNullPName(PNameGenContext gc) {
	    //public static PName nullPName(PNameManager _pnm) {
        CodeVisitor cv = gc.clav.visitMethod(ACC_PUBLIC + ACC_STATIC, 
                "nullPName", 
                "(" + gc.pnameManagerJT + ")" + gc.pnameJT, null, null);
        //return new TotoPName(f1, pncParam});
        cv.visitTypeInsn(NEW, gc.classToWrite);
        cv.visitInsn(DUP);
        for (Iterator it = gc.coAllFields.iterator(); it.hasNext();) {
			TypedElement te = (TypedElement) it.next();
			PType type = te.getType();
            if (isObjectType(type)) {
		        cv.visitFieldInsn(GETSTATIC, gc.classToWrite, NULLOVALUE, 
		                gc.objectJT);
		        cv.visitTypeInsn(CHECKCAST, 
		                getJVMClassName(te.getType().getJavaClass()));
			} else {
	    	    switch(type.getTypeCode()){
	    	    case PType.TYPECODE_CHAR:
			        cv.visitFieldInsn(GETSTATIC, gc.classToWrite, NULL_CHAR_VALUE, 
			                Type.CHAR_TYPE.getDescriptor());
	    	    	break;
	    	    case PType.TYPECODE_BYTE:
				    cv.visitFieldInsn(GETSTATIC, gc.classToWrite, NULLVALUE, 
				            Type.LONG_TYPE.getDescriptor());
	    	    	cv.visitInsn(L2I);
	    	    	cv.visitInsn(I2B);
	    	    	break;
	    	    case PType.TYPECODE_SHORT:
				    cv.visitFieldInsn(GETSTATIC, gc.classToWrite, NULLVALUE, 
				            Type.LONG_TYPE.getDescriptor());
	    	    	cv.visitInsn(L2I);
	    	    	cv.visitInsn(I2B);
	    	    	break;
	    	    case PType.TYPECODE_INT:
				    cv.visitFieldInsn(GETSTATIC, gc.classToWrite, NULLVALUE, 
				            Type.LONG_TYPE.getDescriptor());
	    	    	cv.visitInsn(L2I);
	    	    	break;
	    	    case PType.TYPECODE_LONG: 
				    cv.visitFieldInsn(GETSTATIC, gc.classToWrite, NULLVALUE, 
				            Type.LONG_TYPE.getDescriptor());
	    	    	break;
	    	    }
			}
        }
        cv.visitVarInsn(ALOAD, 0);
        cv.visitMethodInsn(INVOKESPECIAL, gc.classToWrite, "<init>", 
                "(" + gc.xconstructorFieldsPnc + ")V");
        
        cv.visitInsn(ARETURN);
        cv.visitMaxs(0, 0);
	}

	private void generatePnGetSpecific(PNameGenContext gc) {
        for (Iterator it = gc.coAllFields.iterator(); it.hasNext();) {
            TypedElement te = (TypedElement) it.next();
            String jvmFieldType = getJVMType(te.getType());            
            //implement the specific PName getter: 
            // public char pnGetF1(Object ctx) throws PException { return f1;}
            CodeVisitor cv = gc.clav.visitMethod(ACC_PUBLIC, 
                    "pnGet" + upperFL(te.getName()), 
                    "(" + gc.objectJT + ")" + jvmFieldType, 
                    new String[]{gc.pexceptionJCN}, null);
            cv.visitVarInsn(ALOAD, 0);
            cv.visitFieldInsn(GETFIELD, gc.classToWrite, 
                    te.getName(), jvmFieldType);
            cv.visitInsn(Type.getType(te.getType().getJavaClass()).getOpcode(IRETURN));
            cv.visitMaxs(0, 0);
        }
	}
	
	private void generatePnGetGeneric(PNameGenContext gc) throws PException {
		Map pt2fieldlist = new HashMap();
        for (Iterator it = gc.coAllFields.iterator(); it.hasNext();) {
			TypedElement te = (TypedElement) it.next();
			int tc = te.getType().getTypeCode();
			List fs = (List) pt2fieldlist.get(te.getType());
			if (fs == null) {
			    fs = new ArrayList();
			    pt2fieldlist.put(te.getType(), fs);
			}
			fs.add(te);
		}
        for (Iterator mapit = pt2fieldlist.entrySet().iterator(); mapit.hasNext();) {
            Map.Entry me = (Map.Entry) mapit.next();
            PType type = (PType) me.getKey();
            String jvmFieldType = getJVMType(type);
            String methodName = getPNameGetterGetFunction(type); 
            String methodDesc = "(" + gc.stringJT + gc.objectJT + ")" + jvmFieldType;
            int methodReturnType = Type.getType(type.getJavaClass()).getOpcode(IRETURN);
            //char pnGetCharField(String fn, Object ctx) throws PException
            CodeVisitor cv = gc.clav.visitMethod(ACC_PUBLIC,
                    methodName, methodDesc, 
                    new String[]{gc.pexceptionJCN} , null);

            for (Iterator fieldIt = ((List) me.getValue()).iterator(); fieldIt.hasNext();) {
                TypedElement te = (TypedElement) fieldIt.next();
                
                //if ("lid".equals(fn)) {
                cv.visitLdcInsn(te.getName());
                cv.visitVarInsn(ALOAD, 1);
    	        cv.visitMethodInsn(INVOKEVIRTUAL, gc.stringJCN, 
    	                "equals", "(" + gc.objectJT+ ")Z");
    	        Label labelElse = new Label();
    	        cv.visitJumpInsn(IFEQ, labelElse);
    	        //    return this.f1;
                cv.visitVarInsn(ALOAD, 0);
		        cv.visitFieldInsn(GETFIELD, gc.classToWrite, te.getName(), jvmFieldType);
		        cv.visitInsn(methodReturnType);
                // }
		        cv.visitLabel(labelElse);
            }
            //return super.pnGetLongField(fn, ctx);
            cv.visitVarInsn(ALOAD, 0);
            cv.visitVarInsn(ALOAD, 1);
            cv.visitVarInsn(ALOAD, 2);
            cv.visitMethodInsn(INVOKESPECIAL, PARENT_CLASS_NAME, methodName, methodDesc);
	        cv.visitInsn(methodReturnType);
            cv.visitMaxs(0, 0);
        }
	}	

	private void generateCheckPName(PNameGenContext gc) {
	    //protected boolean checkPName(Object obj) {
        CodeVisitor cv = gc.clav.visitMethod(ACC_PUBLIC, 
                "checkPName", "(" + gc.objectJT + ")Z", null, null);
        //return obj instanceof LongIdPName;
        cv.visitVarInsn(ALOAD, 1);
        cv.visitTypeInsn(INSTANCEOF, gc.classToWrite);
        cv.visitInsn(IRETURN);
        cv.visitMaxs(0, 0);
	}
	
	private void generateToString(PNameGenContext gc) {
	    String toStringMDesc = "()" + gc.stringJT;
	    //protected boolean checkPName(Object obj) {
        CodeVisitor cv = gc.clav.visitMethod(ACC_PUBLIC, 
                "toString", toStringMDesc, null, null);
        //sb = new StringBuffer();
        cv.visitTypeInsn(NEW, gc.stringBufferJCN);
        cv.visitInsn(DUP);
        //sb.append(super.toString());
        cv.visitMethodInsn(INVOKESPECIAL, gc.stringBufferJCN, "<init>", "()V");
        cv.visitVarInsn(ALOAD, 0);
        cv.visitMethodInsn(INVOKESPECIAL, PARENT_CLASS_NAME, "toString", toStringMDesc);
        String mas = "(" + gc.stringJT + ")" + gc.stringBufferJT;
        cv.visitMethodInsn(INVOKEVIRTUAL, gc.stringBufferJCN, "append", mas);

        for (Iterator it = gc.coAllFields.iterator(); it.hasNext();) {
            //sb.append(" / f1=");
			TypedElement te = (TypedElement) it.next();
			PType type = te.getType();
			cv.visitLdcInsn(" / " + te.getName() + "=");
	        cv.visitMethodInsn(INVOKEVIRTUAL, gc.stringBufferJCN, "append", mas);
            //sb.append(f1);
            cv.visitVarInsn(ALOAD, 0);
	        cv.visitFieldInsn(GETFIELD, gc.classToWrite, te.getName(), getJVMType(type));
	        String amd = "(";
			if (isObjectType(type)) {
			     amd += gc.objectJT;
			} else {
			     amd += getJVMType(type);
			}
			amd += ")" + getJVMType(StringBuffer.class);
	        cv.visitMethodInsn(INVOKEVIRTUAL, gc.stringBufferJCN, "append", amd);
			
        }
        // return sb.toString();
        cv.visitMethodInsn(INVOKEVIRTUAL, gc.stringBufferJCN, "toString", toStringMDesc);
        cv.visitInsn(ARETURN);
        cv.visitMaxs(0, 0);
	}

	private void generateHashCode(PNameGenContext gc) {
        CodeVisitor cv = gc.clav.visitMethod(ACC_PUBLIC, "hashCode", "()I", null, null);
        cv.visitInsn(ICONST_0);
        for (Iterator it = gc.coAllFields.iterator(); it.hasNext();) {
			TypedElement te = (TypedElement) it.next();
			PType type = te.getType();
			switch (te.getType().getTypeCode()) {
			case PType.TYPECODE_BYTE:
			case PType.TYPECODE_CHAR:
			case PType.TYPECODE_SHORT:
			case PType.TYPECODE_INT:
			    // f1
	            cv.visitVarInsn(ALOAD, 0);
				cv.visitFieldInsn(GETFIELD, gc.classToWrite, te.getName(), getJVMType(type));
				break;
			case PType.TYPECODE_LONG:
			    //(long) f1
	            cv.visitVarInsn(ALOAD, 0);
				cv.visitFieldInsn(GETFIELD, gc.classToWrite, te.getName(), getJVMType(type));
		        cv.visitInsn(L2I);
				break;
			case PType.TYPECODE_OBJCHAR:
			case PType.TYPECODE_OBJBYTE:
			case PType.TYPECODE_OBJSHORT:
			case PType.TYPECODE_OBJINT:
			case PType.TYPECODE_OBJLONG:
			case PType.TYPECODE_DATE:
			case PType.TYPECODE_BIGDECIMAL:
			case PType.TYPECODE_BIGINTEGER:
			case PType.TYPECODE_STRING:
				//(f1 == null ? 0 : f1.hashCode())
	            cv.visitVarInsn(ALOAD, 0);
				cv.visitFieldInsn(GETFIELD, gc.classToWrite, te.getName(), getJVMType(type));
				Label l1 = new Label();
				Label l2 = new Label();
				cv.visitJumpInsn(IFNULL, l1);
	            cv.visitVarInsn(ALOAD, 0);
				cv.visitFieldInsn(GETFIELD, gc.classToWrite, te.getName(), getJVMType(type));
		        cv.visitMethodInsn(INVOKEVIRTUAL, 
		                getJVMClassName(te.getType().getJavaName()), 
		                "hashCode", "()I");
				cv.visitJumpInsn(GOTO, l2);
				cv.visitLabel(l1);
		        cv.visitInsn(ICONST_0);
				cv.visitLabel(l2);
				break;
			case PType.TYPECODE_BYTEARRAY:
			case PType.TYPECODE_CHARARRAY:
			default:
		        cv.visitInsn(ICONST_0);
			}
	        cv.visitInsn(IADD);
        }
        cv.visitInsn(IRETURN);
        cv.visitMaxs(0, 0);
	}

	private void generateValuesEqual(PNameGenContext gc) {
        // protected boolean valuesEqual(PName obj) {
        CodeVisitor cv = gc.clav.visitMethod(ACC_PROTECTED, 
                "valuesEqual", "(" + gc.pnameJT + ")Z", null, null);
        
        //TotoPName pn = (TotoPName) obj;
        cv.visitVarInsn(ALOAD, 1);
        cv.visitTypeInsn(CHECKCAST, gc.classToWrite);
        cv.visitVarInsn(ASTORE, 2);
        Label labelFalse = new Label();
	    
        for (Iterator it = gc.coAllFields.iterator(); it.hasNext();) {
			TypedElement te = (TypedElement) it.next();
			PType type = te.getType();
			String jvmFieldType = getJVMType(type);
			Type asmType = Type.getType(type.getJavaClass());
			switch (type.getTypeCode()) {
			case PType.TYPECODE_BYTE:
			case PType.TYPECODE_CHAR:
			case PType.TYPECODE_SHORT:
			case PType.TYPECODE_INT:
				//&&  f1 == pn.f1
			    cv.visitVarInsn(ALOAD, 0);
				cv.visitFieldInsn(GETFIELD, gc.classToWrite, te.getName(), jvmFieldType);
				cv.visitVarInsn(ALOAD, 2);
				cv.visitFieldInsn(GETFIELD, gc.classToWrite, te.getName(), jvmFieldType);
    	        cv.visitJumpInsn(IF_ICMPNE, labelFalse);
				break;
			case PType.TYPECODE_LONG:
				//&&  f1 == pn.f1
			    cv.visitVarInsn(ALOAD, 0);
				cv.visitFieldInsn(GETFIELD, gc.classToWrite, te.getName(), jvmFieldType);
				cv.visitVarInsn(ALOAD, 2);
				cv.visitFieldInsn(GETFIELD, gc.classToWrite, te.getName(), jvmFieldType);
				
    	        cv.visitInsn(LCMP);
    	        cv.visitJumpInsn(IFNE, labelFalse);
				break;

			case PType.TYPECODE_OBJBYTE:
			case PType.TYPECODE_OBJCHAR:
			case PType.TYPECODE_OBJSHORT:
			case PType.TYPECODE_OBJINT:
			case PType.TYPECODE_OBJLONG:
			case PType.TYPECODE_DATE:
			case PType.TYPECODE_BIGDECIMAL:
			case PType.TYPECODE_BIGINTEGER:
			case PType.TYPECODE_STRING:
				//&&  (f1 == null ? pn.f1 == null : f1.equals(pn.f1))
	            
			    // f1 == null
			    cv.visitVarInsn(ALOAD, 0);
				cv.visitFieldInsn(GETFIELD, gc.classToWrite, te.getName(), jvmFieldType);
		        Label labelNull = new Label();
		        Label labelNextTest = new Label();
				cv.visitJumpInsn(IFNULL, labelNull); 
	            
				// f1.equals(pn.f1)
	            cv.visitVarInsn(ALOAD, 0);
				cv.visitFieldInsn(GETFIELD, gc.classToWrite, te.getName(), jvmFieldType);
	            cv.visitVarInsn(ALOAD, 2);
				cv.visitFieldInsn(GETFIELD, gc.classToWrite, te.getName(), jvmFieldType);
    	        cv.visitMethodInsn(INVOKEVIRTUAL, getJVMClassName(Object.class), 
    	                "equals", "(" + gc.objectJT + ")Z");
		        cv.visitJumpInsn(IFEQ, labelFalse); 
		        cv.visitJumpInsn(GOTO, labelNextTest);
		        
		        cv.visitLabel(labelNull);
		        // pn.f1 == null
	            cv.visitVarInsn(ALOAD, 2);
				cv.visitFieldInsn(GETFIELD, gc.classToWrite, te.getName(), jvmFieldType);
				cv.visitJumpInsn(IFNULL, labelNextTest);
		        cv.visitJumpInsn(GOTO, labelFalse);

		        cv.visitLabel(labelNextTest);
				break;
			case PType.TYPECODE_BYTEARRAY:
			case PType.TYPECODE_CHARARRAY:
				//&& this.equal(f1, pn.f1)
	            cv.visitVarInsn(ALOAD, 0);
	            cv.visitVarInsn(ALOAD, 0);
				cv.visitFieldInsn(GETFIELD, gc.classToWrite, te.getName(), jvmFieldType);
	            cv.visitVarInsn(ALOAD, 2);
				cv.visitFieldInsn(GETFIELD, gc.classToWrite, te.getName(), jvmFieldType);
    	        cv.visitMethodInsn(INVOKEVIRTUAL, PARENT_CLASS_NAME, 
    	                "equal", "(" + jvmFieldType + jvmFieldType + ")Z");
		        cv.visitJumpInsn(IFEQ, labelFalse);
				break;
			default:
				logger.log(BasicLevel.WARN, "Type '" + type.getJavaName() 
				        + "' occurs a problem with this visitor");
			}
        }
        
        cv.visitInsn(ICONST_1);
        Label labelEnd = new Label();
        cv.visitJumpInsn(GOTO, labelEnd);
        cv.visitLabel(labelFalse);
        cv.visitInsn(ICONST_0);
        cv.visitLabel(labelEnd);
        cv.visitInsn(Type.BOOLEAN_TYPE.getOpcode(IRETURN));
        cv.visitMaxs(0, 0);
	}

	private int assignFields(PNameGenContext gc, CodeVisitor cv) {
        int i = 1;
        for (Iterator it = gc.coAllFields.iterator(); it.hasNext();) {
            TypedElement te = (TypedElement) it.next();
            cv.visitVarInsn(ALOAD, 0);
            cv.visitVarInsn(Type.getType(te.getType().getJavaClass()).getOpcode(ILOAD), i);
            cv.visitFieldInsn(PUTFIELD, gc.classToWrite, te.getName(), 
                    getJVMType(te.getType()));
            switch(te.getType().getTypeCode()) {
            case PType.TYPECODE_LONG:
            case PType.TYPECODE_DOUBLE:
                i +=2;
            	break;
            default:
                i++;
            	break;
            }
        }
        return i;
	}
	
	private void assignPNC(CodeVisitor cv, PNameGenContext gc, int paramIdx) {
        cv.visitVarInsn(ALOAD, 0);
        cv.visitVarInsn(ALOAD, paramIdx);
        cv.visitFieldInsn(PUTFIELD, gc.classToWrite, "pnc", gc.pnameManagerJT);
	}
	
	/**
	 * This class is a generation context for pname generation with ASM.
	 *
	 * @author S.Chassande-Barrioz
	 */
	private static final class PNameGenContext extends CNGenContext {
	    
	    /**
	     * Is the name of the class to generates
	     */
	    public final String classToWrite;

	    public PNameGenContext(final CompositeName _co, 
	            final ClassVisitor _clav, 
	            final CommonJVMHelper h) {
	        super(_co, _clav, h);
	        classToWrite = xpnameJCN;
	    }
	}

}
