/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.generator.lib;

import java.io.File;
import java.io.FileOutputStream;
import java.util.Iterator;

import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Label;
import org.objectweb.asm.Type;
import org.objectweb.jorm.api.PAccessor;
import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.compiler.api.JormCompilerParameter;
import org.objectweb.jorm.compiler.api.PExceptionCompiler;
import org.objectweb.jorm.lib.PSerializableImpl;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.Reference;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.util.io.api.TargetHolder;
import org.objectweb.jorm.util.lib.StringReplace;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class generates a XXXSeriazable class for each persistent class.
 * This generator is based on ASM.
 * 
 * @author yoann (initial version with velocity)
 * @author S.Chassande-Barrioz (Migrate to ASM)
 */
public class SerializableGenerator extends CommonGenerator {
	/**
	 * This method generates a XSerializableData file corresponding to the pod
	 * parameter in the directory parameter.
	 * @param co		The meta object interface.
	 * @param holder	The target holder which allows to create files.
	 * @param cp		This parameter permits to reach the compilation parameters.
	 */
	public void generate(Class co,
			 TargetHolder holder,
			 JormCompilerParameter cp) throws PException {
		String classToWrite = getJVMClassName(co.getFQName() + "Serializable");
        if (debug) {
            logger.log(BasicLevel.DEBUG, "Generate the class '"
                + classToWrite + "'.");
        }
        ClassWriter cv = new ClassWriter(true);
        //create the class
        cv.visit(V1_1, ACC_PUBLIC + ACC_SUPER
                + (cp.isBindingAbstract() ? ACC_ABSTRACT : 0),
                classToWrite, getJVMClassName(PSerializableImpl.class), 
                new String[]{getJVMClassName(co.getFQName() + "Accessor")}, 
                classToWrite);        
        generateGetClassName(co, cv);
        generateFieldDeclaration(co, cv);
        generatePAccessors(co, classToWrite, cv);
        generateToString(co, classToWrite, cv);
        generateRead(co, classToWrite, cv);
        generateWrite(co, classToWrite, cv);
        try {
            // Fetch a block with the fileName
            String fileName = classToWrite;
            fileName = StringReplace.replaceChar('/', File.separatorChar,
                    fileName)
                    + ".class";
            FileOutputStream fos = holder.getFileOutputStream(fileName);
            fos.write(cv.toByteArray());
            fos.close();
        } catch (Exception e) {
            throw new PExceptionCompiler(e,
                    "Problem while generating PAccessor.");
        }
	}

	private void generateFieldDeclaration(Class co, ClassVisitor cv) {
        for (Iterator it = co.getAllFields().iterator(); it.hasNext();) {
            TypedElement te = (TypedElement) it.next();
            if (te instanceof Reference) {
                cv.visitField(ACC_PROTECTED, te.getName(), 
                        getJVMType(String.class), null, null);
            } else {
                cv.visitField(ACC_PROTECTED, te.getName(), 
                        getJVMType(te.getType()), null, null);
            }
        }
	}
	
	private void generateGetClassName(Class co, 
	        ClassVisitor clav) {
		//public String getClassName(){ return "${class.getFQName()}"; }
        CodeVisitor cv = clav.visitMethod(ACC_PUBLIC, 
                    "getClassName", "()" + getJVMType(String.class), null, null);
        cv.visitLdcInsn(co.getFQName());
        cv.visitInsn(ARETURN);
        cv.visitMaxs(0, 0);
	}

	private void generatePAccessors(Class co, String classToWrite, ClassVisitor clav) {
        for (Iterator it = co.getAllFields().iterator(); it.hasNext();) {
            TypedElement te = (TypedElement) it.next();
            PType type = te.getType();
            if (te instanceof Reference) {
                //public void paSetF1(PName pn, Object hints) {
                CodeVisitor cv = clav.visitMethod(ACC_PUBLIC, 
                        "paSet" + upperFL(te.getName()),	
                        "(" + getJVMType(PName.class) 
                        + getJVMType(Object.class) + ")" 
                        + getJVMType(String.class), null, null);
                //this.f1 = this.encodeString(pn);
                cv.visitVarInsn(ALOAD, 0);
		        cv.visitMethodInsn(INVOKEVIRTUAL, 
		                getJVMClassName(PSerializableImpl.class), 
		                "encodeString", "(" + getJVMType(PName.class) +")" 
		                + getJVMType(String.class));
                cv.visitVarInsn(ALOAD, 0);
	            cv.visitFieldInsn(PUTFIELD, classToWrite, te.getName(), 
	                    getJVMType(PName.class));
                //}
                cv.visitInsn(RETURN);
                cv.visitMaxs(0, 0);

                //public PName paGetF1(Object hints) {
                cv = clav.visitMethod(ACC_PUBLIC, 
                        "paGet" + upperFL(te.getName()),	
                        "(" + getJVMType(Object.class) + ")" 
                        + getJVMType(PName.class), null, null);
                //return this.decodeString(f1, "f1", hints);
                cv.visitVarInsn(ALOAD, 0);
                cv.visitVarInsn(ALOAD, 0);
                cv.visitFieldInsn(GETFIELD, classToWrite, te.getName(), 
                        getJVMType(PName.class));
                cv.visitVarInsn(ALOAD, 1);
		        cv.visitMethodInsn(INVOKEVIRTUAL, 
		                getJVMClassName(PSerializableImpl.class), 
		                "decodeString", "(" + getJVMType(Object.class) +")" 
		                + getJVMType(String.class));
                //}
                cv.visitInsn(ARETURN);
                cv.visitMaxs(0, 0);
            } else {
                String jvmFieldType = getJVMType(te.getType());            
                //public void paSetF1(char val) {
                CodeVisitor cv = clav.visitMethod(ACC_PUBLIC, 
                        "paSet" + upperFL(te.getName()),	
                        "(" + jvmFieldType 
                        + getJVMType(Object.class) + ")V", null, null);
                //this.f1 = val;
                cv.visitVarInsn(ALOAD, 0);
                cv.visitVarInsn(Type.getType(type.getJavaClass()).getOpcode(ILOAD), 1);
	            cv.visitFieldInsn(PUTFIELD, classToWrite, te.getName(), 
	                    jvmFieldType);
                //}
                cv.visitInsn(RETURN);
                cv.visitMaxs(0, 0);

                //public char paGetF1() {
                cv = clav.visitMethod(ACC_PUBLIC, 
                        "paGet" + upperFL(te.getName()),	
                        "()" + jvmFieldType, null, null);
                //return this.f1;
                cv.visitVarInsn(ALOAD, 0);
	            cv.visitFieldInsn(GETFIELD, classToWrite, te.getName(), 
	                    jvmFieldType);
                //}
                cv.visitInsn(Type.getType(type.getJavaClass()).getOpcode(IRETURN));
                cv.visitMaxs(0, 0);
            }
        }
	}

	private void generateToString(Class co, String classToWrite, ClassVisitor clav) {
	    String sbClass = getJVMClassName(StringBuffer.class);
	    //public String toString(){
	    CodeVisitor cv = clav.visitMethod(ACC_PUBLIC, "toString", 
	            "()" + getJVMType(String.class), null, null);
		//StringBuffer s = new StringBuffer();
	    cv.visitTypeInsn(NEW, sbClass);
	    cv.visitInsn(DUP);
	    cv.visitMethodInsn(INVOKESPECIAL, getJVMClassName(StringBuffer.class), "<init>", "()V");
		//s.append("data(");
	    cv.visitLdcInsn("data(");
	    cv.visitMethodInsn(INVOKEVIRTUAL, sbClass, "append", 
	            "(Ljava/lang/String;)Ljava/lang/StringBuffer;");

	    for (Iterator it = co.getAllFields().iterator(); it.hasNext();) {
			TypedElement te = (TypedElement) it.next();
			PType type = te.getType();
            //sb.append(f1);
            cv.visitVarInsn(ALOAD, 0);
	        String amd = "(";
            if (te instanceof Reference) {
		        cv.visitFieldInsn(GETFIELD, classToWrite, te.getName(), 
		                getJVMType(PName.class));
			    amd += getJVMType(Object.class);
            } else {
		        cv.visitFieldInsn(GETFIELD, classToWrite, te.getName(), getJVMType(type));
				if (isObjectType(type)) {
				     amd += getJVMType(Object.class);
				} else {
				     amd += getJVMType(type);
				}
            }
			amd += ")" + getJVMType(StringBuffer.class);
	        cv.visitMethodInsn(INVOKEVIRTUAL, sbClass, "append", amd);
			
        }
	    //s.append(", ");
	    cv.visitLdcInsn(", ");
	    cv.visitMethodInsn(INVOKEVIRTUAL, sbClass, "append", 
	            "(Ljava/lang/String;)Ljava/lang/StringBuffer;");
		//s.append(this.pName)
        cv.visitVarInsn(ALOAD, 0);
        cv.visitFieldInsn(GETFIELD, classToWrite, "pName", getJVMType(String.class));
	    cv.visitMethodInsn(INVOKEVIRTUAL, sbClass, "append", 
        "(Ljava/lang/String;)Ljava/lang/StringBuffer;");
	    
	    //s.append(")");
	    cv.visitLdcInsn(")");
	    cv.visitMethodInsn(INVOKEVIRTUAL, sbClass, "append", 
	            "(Ljava/lang/String;)Ljava/lang/StringBuffer;");
	    // return sb.toString();
        cv.visitMethodInsn(INVOKEVIRTUAL, sbClass, "toString", "()" + getJVMType(String.class));
        cv.visitInsn(ARETURN);
        cv.visitMaxs(0, 0);
	}

	private void generateRead(Class co, String classToWrite, ClassVisitor clav) {
	    //public void read(PClassMapping pcm, PAccessor pa) throws PException {
	    CodeVisitor cv = clav.visitMethod(ACC_PUBLIC, "read", 
	            "(" + getJVMType(PClassMapping.class) 
	            + getJVMType(PAccessor.class) + ")V", 
	            new String[] {getJVMClassName(PException.class)}, null);
	    Label labelEnd = new Label();
	    
	    //if (pa instanceof ACompositeBigDecimalIdAccessor)
	    cv.visitVarInsn(ALOAD, 2);
	    cv.visitTypeInsn(INSTANCEOF, getJVMClassName(co.getFQName() + "Accessor"));
	    //ACompositeBigDecimalIdAccessor paccessor = null;
	    cv.visitVarInsn(ALOAD, 2);
	    cv.visitTypeInsn(CHECKCAST, getJVMClassName(co.getFQName() + "Accessor"));
	    cv.visitVarInsn(ASTORE, 3);
	    
	    for (Iterator it = co.getAllFields().iterator(); it.hasNext();) {
			TypedElement te = (TypedElement) it.next();
			PType type = te.getType();
            if (te instanceof Reference) {
                //paccessor.paSetMyself1(paGetMyself1(pcm), null);
                cv.visitVarInsn(ALOAD, 3);
                cv.visitVarInsn(ALOAD, 0);
                cv.visitVarInsn(ALOAD, 1);
                cv.visitMethodInsn(INVOKEVIRTUAL, classToWrite, 
                        "paGet" + upperFL(te.getName()), 
                        "(" + getJVMClassName(String.class) 
                        + ")" + getJVMClassName(PName.class));
                cv.visitInsn(ACONST_NULL);
				cv.visitMethodInsn(INVOKEINTERFACE, 
				        getJVMClassName(co.getFQName() + "Accessor"), 
				        "paSet" + upperFL(te.getName()), 
                		"(" + getJVMClassName(PName.class) 
                		+ getJVMClassName(Object.class) + ")V");
            } else {
                String jvmFieldType = getJVMType(te.getType());            
    			//paccessor.paSetF1(paGetF1());
				cv.visitVarInsn(ALOAD, 3);
				cv.visitVarInsn(ALOAD, 0);
				cv.visitMethodInsn(INVOKEVIRTUAL, classToWrite, 
				        "paGet" + upperFL(te.getName()), 
				        "()" + jvmFieldType);
				cv.visitMethodInsn(INVOKEINTERFACE, 
				        getJVMClassName(co.getFQName() + "Accessor"), 
				        "paSet" + upperFL(te.getName()), 
				        "(" + jvmFieldType + ")V");
            }
	    }
	    cv.visitLabel(labelEnd);
        cv.visitInsn(RETURN);
        cv.visitMaxs(0, 0);
	}

	private void generateWrite(Class co, String classToWrite, ClassVisitor clav) {
	    //public void read(PClassMapping pcm, PAccessor pa) throws PException {
	    CodeVisitor cv = clav.visitMethod(ACC_PUBLIC, "write", 
	            "(" + getJVMType(PClassMapping.class) 
	            + getJVMType(PAccessor.class) + ")V", 
	            new String[] {getJVMClassName(PException.class)}, null);
	    Label labelEnd = new Label();
	    
	    //if (pa instanceof ACompositeBigDecimalIdAccessor)
	    cv.visitVarInsn(ALOAD, 2);
	    cv.visitTypeInsn(INSTANCEOF, getJVMClassName(co.getFQName() + "Accessor"));
	    //ACompositeBigDecimalIdAccessor paccessor = null;
	    cv.visitVarInsn(ALOAD, 2);
	    cv.visitTypeInsn(CHECKCAST, getJVMClassName(co.getFQName() + "Accessor"));
	    cv.visitVarInsn(ASTORE, 3);
	    
	    for (Iterator it = co.getAllFields().iterator(); it.hasNext();) {
			TypedElement te = (TypedElement) it.next();
			PType type = te.getType();
            if (te instanceof Reference) {
                //paccessor.paSetMyself1(paGetMyself1(pcm), null);
                cv.visitVarInsn(ALOAD, 0);
                cv.visitVarInsn(ALOAD, 3);
                cv.visitVarInsn(ALOAD, 1);
                cv.visitMethodInsn(INVOKEINTERFACE, 
                        getJVMClassName(co.getFQName() + "Accessor"), 
                        "paGet" + upperFL(te.getName()), 
                        "(" + getJVMClassName(String.class) 
                        + ")" + getJVMClassName(PName.class));
                cv.visitInsn(ACONST_NULL);
				cv.visitMethodInsn(INVOKEVIRTUAL, 
				        classToWrite, 
				        "paSet" + upperFL(te.getName()), 
                		"(" + getJVMClassName(PName.class) 
                		+ getJVMClassName(Object.class) + ")V");
            } else {
                String jvmFieldType = getJVMType(te.getType());            
    			//paSetF1(paccessor.paGetF1());
				cv.visitVarInsn(ALOAD, 0);
				cv.visitVarInsn(ALOAD, 3);
				cv.visitMethodInsn(INVOKEINTERFACE, 
				        getJVMClassName(co.getFQName() + "Accessor"), 
				        "paGet" + upperFL(te.getName()), 
				        "()" + jvmFieldType);
				cv.visitMethodInsn(INVOKEVIRTUAL, 
				        classToWrite, 
				        "paSet" + upperFL(te.getName()), 
				        "(" + jvmFieldType + ")V");
            }
	    }
	    cv.visitLabel(labelEnd);
        cv.visitInsn(RETURN);
        cv.visitMaxs(0, 0);
	}
}
