package org.objectweb.jorm.generator.lib;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;
import java.util.Set;

import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Label;
import org.objectweb.asm.Type;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.compiler.api.JormCompilerParameter;
import org.objectweb.jorm.compiler.api.PExceptionCompiler;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbClassMapping;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbPrimitiveElementMapping;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ClassProject;
import org.objectweb.jorm.metainfo.api.ClassRef;
import org.objectweb.jorm.metainfo.api.CompositeName;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.Mapping;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.NameRef;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.api.Reference;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.jorm.naming.api.FieldCoder;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PNameCoder;
import org.objectweb.jorm.naming.api.PNameGetter;
import org.objectweb.jorm.naming.lib.MapPNameGetter;
import org.objectweb.jorm.util.io.api.TargetHolder;
import org.objectweb.jorm.type.api.PType;


/**
 * The StateGenerator class uses ASM to generate an abstract state
 * implementation for a JORM persistent class. The generated class
 * will implement the specialized PAccessor interface for the JORM
 * class in question. The generated state implementation will
 * automatically manage coherence between multiple memory fields
 * mapped onto a single persistent field. A JORM user wishing to take
 * advantage of this generated state implementation must extend the
 * generated class and provide an implementation of the {@link
 * FieldCoder} interface.
 *
 * @author Rafael H. Schloming &lt;rhs@mit.edu&gt;
 **/

public class StateGenerator extends CommonGenerator {

    public StateGenerator() {}

    /**
     * Public entry point for invoking this generator.
     *
     * @param clazz The persistent class for which to provide a state
     * implementation.
     *
     * @param holder The {@link TargetHolder} used to create the
     * output .class file.
     *
     * @param pc A {@link JormCompilerParameter} object containing
     * parameters for this generator.
     */
    public void generate(Class clazz, TargetHolder holder,
                         JormCompilerParameter pc) throws PException {
        StateGenerator gen = new StateGenerator(clazz, holder, pc);
        gen.emit();
    }


    // Constants for various method and type signatures needed by ASM.
    private static final String OBJECT_T = getJVMType(Object.class);
    private static final String STRING_T = getJVMType(String.class);
    private static final String PNAME_T = getJVMType(PName.class);
    private static final String PNAME_CODER_T = getJVMType(PNameCoder.class);
    private static final String PNAME_GETTER_T = getJVMType(PNameGetter.class);
    private static final String PNAME_CODER_N =
        getJVMClassName(PNameCoder.class);
    private static final String PNAME_GETTER_N = getJVMClassName(PNameGetter.class);
    private static final String MAP_GETTER_N =
        getJVMClassName(MapPNameGetter.class);
    private static final String FIELD_CODER = getJVMClassName(FieldCoder.class);
    private static final String ENCODE = "encode";
    private static final String ENCODE_PARAMS =
        "(" + STRING_T + PNAME_T + ")" + PNAME_GETTER_T;
    private static final String DECODE = "decode";
    private static final String DECODE_PARAMS =
        "(" + STRING_T + PNAME_GETTER_T + OBJECT_T + ")" + PNAME_T;

    // the persistent class
    private Class clazz;
    // the TargetHolder used to create the output .class fileo
    private TargetHolder holder;
    // the compiler params for this invocation
    private JormCompilerParameter params;
    // the RdbClassMapping object for clazz
    private RdbClassMapping rcm;
    // the pname columns for clazz
    private Collection pncolumns;
    // the name of the generated class
    private String name;
    // the persistent class that the generated class will extend
    private String sup;
    // the PAccessor interface for the specific persistent class
    private String acc;
    private ClassWriter cw;
    // field name -> field type
    private Map types = new HashMap();
    // emitted field -> column name
    private Map emitted = new HashMap();

    /**
     * Private constructor used by the {@link #generate(Class,
     * TargetHolder, JormCompilerParameter)} method.
     */
    private StateGenerator(Class clazz, TargetHolder holder,
                           JormCompilerParameter params) {
        this.clazz = clazz;
        this.holder = holder;
        this.params = params;
        rcm = getRdbClassMapping();
        pncolumns = getPNameColumns(clazz);
        name = getJVMClassName(clazz.getFQName() + "State");
        if (params.getStateInheritance() != null) {
            sup = getJVMClassName(params.getStateInheritance());
        }
        acc = getJVMClassName(clazz.getFQName() + "Accessor");
        cw = new ClassWriter(true);
        cw.visit(V1_1,
                 ACC_ABSTRACT + ACC_PUBLIC + ACC_SUPER, // accessor modifiers
                 name,                                  // class name
                 sup,                                   // superclass
                 new String[]{acc, FIELD_CODER},        // interfaces
                 name);                                 // source file
        CodeVisitor cd = cw.visitMethod(ACC_PUBLIC, "<init>", "()V", null, null);
        if (sup != null) {
            cd.visitVarInsn(ALOAD, 0);
            cd.visitMethodInsn(INVOKESPECIAL, sup, "<init>", "()V");
        }
        cd.visitInsn(RETURN);
        cd.visitMaxs(0, 0);
    }

    /**
     * Extracts the RdbClassMapping for the persistent class.
     *
     * @return the RdbClassMapping for the persistent class
     */
    private RdbClassMapping getRdbClassMapping() {
        ClassProject cp = clazz.getClassProject(params.getProjectName());
        Mapping rdb = cp.getMapping("rdb");
        return (RdbClassMapping) rdb.getClassMapping();
    }

    /**
     * Utility method for turning a singleton or empty collection of
     * {@link NameDef} objects into a {@link NameDef} or null.
     *
     * @param ndefs a collection containing zero or one {@link
     * NameDef} objects
     *
     * @return the {@link NameDef} object or null if the collection is
     * empty
     *
     * @throws IllegalArgumentException if ndefs contains more than
     * one object
     */
    private NameDef name(Collection ndefs) {
        // XXX: I have no idea why there can be multiple name defs,
        // but to my knowledge this feature is never used, and this
        // hack seems to work.
        if (ndefs.isEmpty()) {
            return null;
        } else if (ndefs.size() > 1) {
            throw new IllegalArgumentException
                ("too many names: " + ndefs);
        } else {
            return (NameDef) ndefs.iterator().next();
        }
    }

    /**
     * Extracts the PName columns for the given persistent class.
     *
     * @param clazz the persistent class
     * @return a {@link String} collection of column names
     */
    private Collection getPNameColumns(Class clazz) {
        NameDef nd = name(clazz.getNameDefs());
        if (nd != null) {
            return getColumns(nd);
        } else {
            Collection sups = clazz.getSuperClasses();
            for (Iterator it = sups.iterator(); it.hasNext(); ) {
                Class sup = (Class) it.next();
                Collection result = getPNameColumns(sup);
                if (result != null) {
                    return result;
                }
            }

            return null;
        }
    }

    /**
     * Extracts the column names for the given {@link NameDef}.
     *
     * @param nd the {@link NameDef}
     * @return a {@link String} collection of column names
     */
    private Collection getColumns(NameDef nd) {
        Collection result = new ArrayList();
        for (Iterator it = nd.iterateField(); it.hasNext(); ) {
            String field = (String) it.next();
            result.add(getColumn(field));
        }
        return result;
    }

    /**
     * Creates or returns an existing field in the class produced by
     * this generator.
     *
     * @param field the field name
     * @param el the type for the field
     *
     * @return the ASM type descriptor for the field
     */
    private String field(String field, TypedElement el) throws PExceptionCompiler {
        return field(field, el.getType());
    }

    /**
     * Creates or returns an existing field in the class produced by
     * this generator.
     *
     * @param field the field name
     * @param el the type for the field
     *
     * @return the ASM type descriptor for the field
     */
    private String field(String field, PType type) throws PExceptionCompiler {
        return field(field, getJVMType(type));
    }

    /**
     * Creates or returns an existing field in the class produced by
     * this generator.
     *
     * @param field the field name
     * @param el the type for the field
     *
     * @return the ASM type descriptor for the field
     */
    private String field(String field, java.lang.Class type) throws PExceptionCompiler {
        return field(field, getJVMType(type));
    }

    /**
     * Creates or returns an existing field in the class produced by
     * this generator.
     *
     * @param field the field name
     * @param el the type for the field
     *
     * @return the ASM type descriptor for the field
     */
    private String field(String field, String type) throws PExceptionCompiler {
        String result = (String) types.get(field);
        if (result == null) {
            cw.visitField(ACC_PRIVATE, field, type, null, null);
            types.put(field, type);
            result = type;
        } else if (!type.equals(result)) {
            throw new PExceptionCompiler
                ("incompatible types for field '" + field + "': " + type + ", " + result);
        }
        return result;
    }

    /**
     * Converts an integer opcode into the corresponding opcode for
     * the given type.
     *
     * @param type the desired instruction type
     * @param code the integer version of the opcode
     *
     * @return the type specific version of the opcode
     */
    private int getOpcode(String type, int code) {
        return Type.getType(type).getOpcode(code);
    }

    private static final String[] ACCESSOR_EX = {getJVMClassName(PException.class)};

    /**
     * Generate a pair of persistent field accessors (getters and
     * setters).
     *
     * @param el the type of the field
     * @param field the name of the field
     * @param ref true if the field is a reference, false otherwise
     */
    private void accessors(TypedElement el, String field, boolean ref) {
        String fn = upperFL(el.getName());
        String type = (String) types.get(field);

        CodeVisitor cd = cw.visitMethod
            (ACC_PUBLIC, "paSet" + fn,
             "(" + type + (ref ? OBJECT_T : "") + ")V",
             ref ? ACCESSOR_EX : null, null);
        cd.visitVarInsn(ALOAD, 0);
        cd.visitVarInsn(getOpcode(type, ILOAD), 1);
        cd.visitFieldInsn(PUTFIELD, name, field, type);
        cd.visitInsn(RETURN);
        cd.visitMaxs(0, 0);

        cd = cw.visitMethod
            (ACC_PUBLIC, "paGet" + fn,
             "(" + (ref ? OBJECT_T : "") + ")" + type,
             ref ? ACCESSOR_EX : null, null);
        cd.visitVarInsn(ALOAD, 0);
        cd.visitFieldInsn(GETFIELD, name, field, type);
        cd.visitInsn(getOpcode(type, IRETURN));
        cd.visitMaxs(0, 0);

        emitted.put(el, field);
    }

    /**
     * Returns true if and only if the given reference is mapped onto
     * part or all of the pname for this persistent class. When this
     * is true that means the reference is implemented by a back
     * pointer in the database.
     *
     * @param ref the reference
     * @return true iff the reference is mapped onto part or all of
     * the pname for this class
     */
    private boolean isMappedOnPName(Reference ref) {
        return pncolumns.containsAll(getColumns(name(ref.getRefNameDef())));
    }

    /**
     * Entry point for the ASM generation code of this class. This is
     * invoked by the {@link #generate(Class, TargetHolder,
     * JormCompilerParameter)} method after the {@link
     * #StateGenerator(Class, TargetHolder, JormCompilerParameter)}
     * has been used to set up instance state.
     */
    private void emit() throws PException {
        Collection mappings = rcm.getAllPrimitiveElementMappings();
        for (Iterator it = mappings.iterator(); it.hasNext(); ) {
            RdbPrimitiveElementMapping pem =
                (RdbPrimitiveElementMapping) it.next();
            TypedElement el = (TypedElement) pem.getLinkedMO();
            String col = pem.getName();
            field(col, el.getType());
            accessors(el, col, false);
        }

        Collection fields = clazz.getAllFields();

        for (Iterator it = fields.iterator(); it.hasNext(); ) {
            TypedElement el = (TypedElement) it.next();
            if (emitted.containsKey(el)) {
                continue;
            } else if (el instanceof Reference) {
                Reference ref = (Reference) el;
                if (isMappedOnPName(ref)) {
                    String fn = ref.getName() + "PName";
                    field(fn, PName.class);
                    accessors(ref, fn, true);
                } else {
                    reference(ref);
                }
            } else {
                field(el.getName(), el.getType());
                accessors(el, el.getName(), false);
            }
        }

        write(cw, name, holder);
    }

    /**
     * Generates code that reads the column value out of an instance
     * field in the state class and places it in a MapPNameGetter.
     *
     * @param cv the {@link CodeVisitor} object used to generate the
     * code
     * @param obj the variable index of the MapPNameGetter
     * @param field the name of the persistence field
     * @param col the name of the column
     */
    private void readColumn(CodeVisitor cv, int obj, String field, String col) {
        // $<obj>.set("<field>", this.<col>);
        cv.visitVarInsn(ALOAD, obj);
        if (field == null) {
            cv.visitInsn(ACONST_NULL);
        } else {
            cv.visitLdcInsn(field);
        }
        cv.visitVarInsn(ALOAD, 0);
        String ft = (String) types.get(col);
        cv.visitFieldInsn(GETFIELD, name, col, ft);
        String param;
        if (Type.getType(ft).getSort() == Type.OBJECT) {
            param = OBJECT_T;
        } else {
            param = ft;
        }
        cv.visitMethodInsn(INVOKEVIRTUAL, MAP_GETTER_N, "set",
                           "(" + STRING_T + param + ")V");
    }

    /**
     * Generates code that extracts the column value from a
     * PNameGetter and writes it into an instance field.
     *
     * @param cv the {@link CodeVisitor} used to generate the code
     * @param col the name of the instance field to write to
     * @param obj the variable index for the PNameGetter
     * @param getter the PNameGetter function
     * @param field the field name
     * @param ctx the context object for the PNameGetter
     */
    private void writeColumn(CodeVisitor cv, String col, int obj,
                             String getter, String field, int ctx) {
        // paSet: this.<col> = $<obj>.pnget<blah>("<field>", $<ctx>)
        cv.visitVarInsn(ALOAD, 0);
        cv.visitVarInsn(ALOAD, obj);
        if (field == null) {
            cv.visitInsn(ACONST_NULL);
        } else {
            cv.visitLdcInsn(field);
        }
        cv.visitVarInsn(ALOAD, ctx);
        String ft = (String) types.get(col);
        cv.visitMethodInsn(INVOKEINTERFACE, PNAME_GETTER_N, getter,
                           "(" + STRING_T + OBJECT_T + ")" + ft);
        cv.visitFieldInsn(PUTFIELD, name, col, ft);
    }

    /**
     * Look up a field for the persistent class.
     *
     * @param field the field name
     * @return the TypedElement from the class metadata
     */
    private TypedElement getField(String field) {
        TypedElement result = clazz.getTypedElement(field);
        if (result == null) {
            throw new IllegalArgumentException
                ("class " + clazz.getFQName() + " has no such field: " + field);
        }
        return result;
    }

    /**
     * Lookup the column for a given field.
     *
     * @param field the field name
     * @return the column name
     */
    private String getColumn(String field) {
        RdbPrimitiveElementMapping pem = (RdbPrimitiveElementMapping)
            rcm.getPrimitiveElementMapping(field, true);
        return pem.getName();
    }

    /**
     * Generate code accessors (setters and getters) for a reference
     * field. The generated code must encode/decode between the PName
     * object and the database values stored by the state instance.
     * The encoding/decoding process is delegated to subclass of the
     * generated class via the {@link FieldCoder} interface.
     *
     * @param ref the reference field
     */
    private void reference(Reference ref) throws PException {
        String pn = upperFL(ref.getName());
        CodeVisitor getter = cw.visitMethod
            (ACC_PUBLIC, "paGet" + pn, "(" + OBJECT_T + ")" + PNAME_T,
             ACCESSOR_EX, null);
        CodeVisitor setter = cw.visitMethod
            (ACC_PUBLIC, "paSet" + pn, "(" + PNAME_T + OBJECT_T + ")V",
             ACCESSOR_EX, null);

        Label end = new Label();
        NameDef ndef = name(ref.getRefNameDef());
        if (ndef.isNameRef()) {
            Label nonNull = new Label();
            for (Iterator it = ndef.iterateField(); it.hasNext(); ) {
                String field = (String) it.next();
                String col = getColumn(field);
                if (pncolumns.contains(col)) {
                    continue;
                }
                PType type = getField(field).getType();
                String jvmType = field(col, type);
                getter.visitVarInsn(ALOAD, 0);
                getter.visitFieldInsn(GETFIELD, name, col, jvmType);
                any2I(getter, jvmType);
                loadNull(getter, type);
                any2I(getter, jvmType);
                if (Type.getType(jvmType).getSort() == Type.OBJECT) {
                    getter.visitJumpInsn(IF_ACMPNE, nonNull);
                } else {
                    getter.visitJumpInsn(IF_ICMPNE, nonNull);
                }
            }

            // paSet: $2 = null
            getter.visitInsn(ACONST_NULL);
            getter.visitVarInsn(ASTORE, 2);
            getter.visitJumpInsn(GOTO, end);
            getter.visitLabel(nonNull);
        }

        // paGet: $2 = new MapPNameGetter();
        getter.visitTypeInsn(NEW, MAP_GETTER_N);
        getter.visitInsn(DUP);
        getter.visitMethodInsn(INVOKESPECIAL, MAP_GETTER_N, "<init>", "()V");
        getter.visitVarInsn(ASTORE, 2);

        // paSet: $3 = encode("<field>", $1);
        setter.visitVarInsn(ALOAD, 0);
        setter.visitLdcInsn(ref.getName());
        setter.visitVarInsn(ALOAD, 1);
        setter.visitMethodInsn(INVOKEVIRTUAL, name, ENCODE, ENCODE_PARAMS);
        setter.visitVarInsn(ASTORE, 3);

        if (ndef.isFieldName()) {
            String field = ndef.getFieldName();
            String col = getColumn(field);
            // paGet: $2.set(null, this.<col>);
            readColumn(getter, 2, null, col);
            // paSet: this.<col> = $3.pnget<blah>(null, $2);
            writeColumn
                (setter, col, 3,
                 getPNameGetterGetFunction(getField(field).getType()), null,
                 2);
        } else if (ndef.isNameRef()) {
            NameRef nref = ndef.getNameRef();
            CompositeName cn = nref.getCompositeName();
            Collection fields = cn.getAllFields();
            for (Iterator it = fields.iterator(); it.hasNext(); ) {
                PrimitiveElement pe = (PrimitiveElement) it.next();
                String cnfield = pe.getName();
                String nrfield = nref.getClassFieldName(cnfield);
                String col = getColumn(nrfield);
                // paGet: $2.set("<cnfield>", this.<col>);
                readColumn(getter, 2, cnfield, col);
                // paSet: this.<col> = $3.pnget<blah>("<cnfield>", $2);
                writeColumn
                    (setter, col, 3,
                     getPNameGetterGetFunction(pe.getType()), cnfield, 2);
            }
        } else {
            throw new IllegalStateException
                ("unknown kind of name definition: " + ndef);
        }

        getter.visitLabel(end);
        // paGet: return decode("<field>", $2, $1);
        getter.visitVarInsn(ALOAD, 0);
        getter.visitLdcInsn(ref.getName());
        getter.visitVarInsn(ALOAD, 2);
        getter.visitVarInsn(ALOAD, 1);
        getter.visitMethodInsn(INVOKEVIRTUAL, name, DECODE, DECODE_PARAMS);
        getter.visitInsn(ARETURN);
        getter.visitMaxs(0, 0);

        // paSet: return;
        setter.visitInsn(RETURN);
        setter.visitMaxs(0, 0);
    }

}
