/**
 * Copyright (C) 2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.jorm.lib;

import org.objectweb.jorm.api.PMapCluster;
import org.objectweb.jorm.api.PException;

import java.util.Set;
import java.util.Collection;
import java.util.HashSet;

/**
 * Abstract common implementation of a PMapCluster. This implementation
 * manages the dependencies (the list of classes of the cluster).
 *
 * @author S.Chassande-Barrioz
 */
public abstract class AbstractPMapcluster implements PMapCluster {

	/**
	 * The names of the Jorm classes defined and managed  by this PMapCluster
	 */
	protected Set jcNames;

	/**
	 * The names of the Jorm classes not yet defined but to manage by this
	 * PMapCluster. An empty list mean the cluster is complete and can be
	 * started.
	 */
	protected Set unresolvedDependencies;

	/**
	 * Indicates if the cluster is active:
	 * - data structure created
	 * - cluster started
	 */
	protected boolean structuresActive;

	public AbstractPMapcluster() {
		jcNames = new HashSet(1);
		unresolvedDependencies = new HashSet(1);
		structuresActive = false;
	}

	/**
	 * Retrieves an Iterator over the collection of all JORM classes whose
	 * mapping structures have to be controlled altogether.
	 * @return  The Iterator over this collection.
	 */
	public Collection getClusterClasses() {
		if (unresolvedDependencies.isEmpty()) {
			return jcNames;
		} else {
			Set res = new HashSet(jcNames);
			res.addAll(unresolvedDependencies);
			return res;
		}
	}

	public boolean isDefined() {
		return unresolvedDependencies.isEmpty();
	}

	public Set getUnResolvedDependencies() {
		return unresolvedDependencies;
	}

	public void addDependency(String jcname) {
		if (!jcNames.contains(jcname) && !unresolvedDependencies.contains(jcname)) {
			unresolvedDependencies.add(jcname);
		}
	}

	public void classDefined(String jcname) {
		if (!jcNames.contains(jcname)) {
			jcNames.add(jcname);
			unresolvedDependencies.remove(jcname);
		}
	}

	/**
	 * Looks for this JORM class name into this map cluster.
	 * @param jcname	The JORM class name.
	 * @return true if this JORM class name is already assigned to this cluster.
	 */
	public boolean containClass(String jcname) {
		return jcNames.contains(jcname);
	}

	/**
	 * Starts this map cluster. The following operation cannot be executed
	 * anymore: createMappingStructures, deleteMappingStructures, and
	 * updateMappingStructures.
	 * @throws org.objectweb.jorm.api.PException
	 */
	public void start() throws PException {
		if (!structuresActive) {
			throw new PException("Cannot start map cluster: not yet inactive");
		}
		if (!isDefined()) {
			throw new PException("Cannot start map cluster: the following " +
								 "dependencies have not been resolved: "
								 + unresolvedDependencies);
		}
		structuresActive = true;
	}

	/**
	 * Stops this map cluster.
	 * @throws org.objectweb.jorm.api.PException
	 */
	public void stop() throws PException {
		if (structuresActive) {
			throw new PException("Cannot stop map cluster: not yet active");
		}
		structuresActive = false;
	}

	/**
	 * Aligns existing mapping structures to the ones define by this map
	 * cluster. This operation allows schema evolution on existing data store.
	 * It is thus usually difficult to support in the general case.
	 * @throws org.objectweb.jorm.api.PException   Thrown when the data store cannot perform this
	 * operation.
	 * @throws UnsupportedOperationException    Thrown if this operation is not
	 * supported by the mapping.
	 */
	public void updateMappingStructures()
			throws PException, UnsupportedOperationException {
		if (structuresActive) {
			throw new PException("Cannot change mapping structures while they are under use");
		}
		throw new UnsupportedOperationException();
	}
}
