/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */
package org.objectweb.jorm.lib;

import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ClassRef;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.naming.api.PNameCoder;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.genclass.lib.GenClassMapping;

/**
 * This class is a Helper for the Jorm user (Speedo, JOnAS, Medor, ...) which
 * uses pathes. A Path is a string which designates a persistent object class.
 * A path always begins with the name of a jorm persistent class.
 *
 * Some path examples:
 * - '/org.toto.titi' represents the persistent class 'org.toto.titi'
 * - '/org.toto.titi/f1' represents the persistent class referenced by the
 * field 'f1' of the persistent class 'org.toto.titi'.
 * @author S.Chassande-Barrioz
 */
public class JormPathHelper {

	public final static char SEP = '/';
	public final static String ELEMENT = SEP + "element";

	/**
	 * Retrieves the path of a GenClass.
	 * @param gcr is the Jorm meta object repesenting the generic class.
	 * @param element indicates if the returned path must represent the element
	 * of the generic class or the generic class itself.
	 */
	public static String getPath(GenClassRef gcr, boolean element) {
		MetaObject parent = gcr.getParent();
		if (parent instanceof Class) {
			return getPath((Class) parent) + SEP + gcr.getName()
					+ (element ? ELEMENT : "");
		} else if (parent instanceof GenClassRef) {
			GenClassRef p = (GenClassRef) parent;
			return getPath(p, false) + SEP + p.getGenClassName()
					+ (element ? ELEMENT : "");
		} else {
			return null;
		}
	}

	/**
	 * Retrieves the path of the reference to a class. This reference can be
	 * hosted by a class or generic class.
	 * @param cr is the Jorm meta object representing the reference.
	 */
	public static String getPath(ClassRef cr) {
		MetaObject parent = cr.getParent();
		if (parent instanceof Class) {
			return getPath((Class) parent) + SEP + cr.getName();
		} else if (parent instanceof GenClassRef) {
			GenClassRef p = (GenClassRef) parent;
			return getPath(p, false) + SEP + p.getGenClassName() + ELEMENT;
		} else {
			return null;
		}
	}

	public static String replaceFieldOwner(String path, String newOwner) {
		if (path == null) {
			return null;
		}
		int idx = path.indexOf(SEP);
		if (idx == -1) {
			return path;
		}
		if (idx == 0) {
			idx = path.indexOf(SEP, 1);
		}
		if (idx == -1) {
			return path;
		}
		return getPath(newOwner) + path.substring(idx);
	}

	/**
	 * Retrieves the path of the class. This is the fully qualied class name of
	 * the class.
	 * @param clazz is the Jorm meta object representing the persistent class.
	 */
	public static String getPath(Class clazz) {
		return getPath(clazz.getFQName());
	}

	public static String getPath(String fqclassname) {
		return SEP + fqclassname;
	}

	public static String getPath(String sourceclassName,
								 String refFieldName,
								 String[] genClassNames) {
		StringBuffer cn = new StringBuffer();
		cn.append(getPath(sourceclassName));
		cn.append(SEP);
		cn.append(refFieldName);
		for (int i = 1; i < genClassNames.length; i++) {
			cn.append(SEP);
			cn.append(genClassNames[i - 1]);
		}
		return cn.toString();
	}

	/**
	 * Retrieves the PNameCoder instance corresponding to a path
	 * @param path is th path to reach the PNameCoder.
	 * @param mapper is the PMapper to find the first class name.
	 */
	public static PNameCoder getPNameCoder(String path, PMapper mapper) {
		if (path == null || path.length() == 0 || mapper == null) {
			return null;
		}
		int idx = path.indexOf(SEP);
		if (idx == -1) {
			// PName of a class
			PClassMapping pcm = mapper.lookup(path);
			if (pcm == null) {
				return null;
			} else {
				PNameCoder pnc = pcm.getClassPNameCoder();
				if (pnc == null) {
					pnc = pcm.getPBinder();
				}
				return pnc;
			}
		} else if (idx == 0) {
			return getPNameCoder(path.substring(1), mapper);
		} else {
			//the first part of the pnamingcontext id is a class name
			return getPNameCoder(path.substring(idx + 1),
								 mapper.lookup(path.substring(0, idx)));
		}
	}

	/**
	 * Retrieves the PNameCoder instance corresponding to a path
	 * @param path is th path to reach the PNameCoder.
	 * @param pcm is the which is the start of the path
	 */
	private static PNameCoder getPNameCoder(String path, PClassMapping pcm) {
		if (pcm == null) {
			return null;
		}
		int idx = path.indexOf(SEP);
		if (pcm instanceof GenClassMapping) {
			if (idx == -1) {
				return pcm.getPBinder();
			} else {
				String subpath = path.substring(idx + 1);
				if ((SEP + subpath).startsWith(ELEMENT)) {
					return pcm.getPNameCoder();
				}
				return getPNameCoder(
						subpath,
						pcm.getGenClassMapping());
			}
		} else {
			if (idx == -1) {
				return pcm.getPNameCoder(path);
			} else {
				return getPNameCoder(
						path.substring(idx + 1),
						pcm.getGenClassMapping(path.substring(0, idx)));
			}
		}
	}


	/**
	 * Retrieves the GenClassMapping instance corresponding to a path
	 * @param path is th path to reach the GenClassMapping
	 * (PClassMapping implementation).
	 * @param mapper is the PMapper to find the first class name.
	 */
	public static PClassMapping getPClassMapping(String path,
												 PMapper mapper) {
		if (path == null || path.length() == 0 || mapper == null) {
			return null;
		}
		int idx = path.indexOf(SEP);
		if (idx == -1) {
			//The path represents a persistent class and not a generic
			//persistent class
			return mapper.lookup(path);
		} else if (idx == 0) {
			return getPClassMapping(path.substring(1), mapper);
		}
		return getPClassMapping(
				path.substring(idx + 1),
				mapper.lookup(path.substring(0, idx)));
	}

	private static PClassMapping getPClassMapping(String path,
												  PClassMapping pcm) {
		if (pcm == null) {
			return null;
		}
		int idx = path.indexOf(SEP);
		if (pcm instanceof GenClassMapping) {
			if (idx == -1) {
				return pcm.getGenClassMapping();
			} else {
				String subpath = path.substring(idx + 1);
				if ((SEP + subpath).startsWith(ELEMENT)) {
					return null;
				}
				return getPClassMapping(
						subpath,
						pcm.getGenClassMapping());
			}
		} else {
			if (idx == -1) {
				return pcm.getGenClassMapping(path);
			} else {
				return getPClassMapping(
						path.substring(idx + 1),
						pcm.getGenClassMapping(path.substring(0, idx)));
			}
		}
	}

	public static String getOriginClass(String path) {
		if (path == null || path.length() == 0) {
			return null;
		}
		int idx = path.indexOf(SEP);
		if (idx == 0) {
			idx = path.indexOf(SEP, idx);
		}
		if (idx == -1) {
			return path;
		} else {
			int idx2 = path.indexOf(SEP, idx + 1);
			if (idx2 == -1) {
				return path.substring(idx + 1);
			} else {
				return path.substring(idx + 1, idx2);
			}
		}
	}
}
