/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.fos.genclass;

import org.objectweb.jorm.api.PAccessor;
import org.objectweb.jorm.api.PBinding;
import org.objectweb.jorm.api.PBindingCtrl;
import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.api.PGenClassAccessor;
import org.objectweb.jorm.api.PIndexedElem;
import org.objectweb.jorm.api.PMappingCallback;
import org.objectweb.jorm.api.PMappingStructuresManager;
import org.objectweb.jorm.api.PNameIterator;
import org.objectweb.jorm.api.PStateGraph;
import org.objectweb.jorm.genclass.api.FieldDesc;
import org.objectweb.jorm.genclass.lib.GenClassMapping;
import org.objectweb.jorm.lib.PBindingImpl;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.mapper.fos.lib.FosPMappingStructuresManager;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.perseus.connector.ra.fos.FosConnection;
import org.objectweb.perseus.fos.api.FosAccess;
import org.objectweb.perseus.fos.api.FosException;
import org.objectweb.perseus.fos.api.FosStructure;

import java.io.File;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.math.BigDecimal;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;

public class FosGenClassMapping
		extends GenClassMapping
		implements FosGenClassProp, FosStructure {
	protected String className = null;
	protected String fieldName = null;
	protected String genClassLevel = null;
	protected String dirName = null;

	private void initDirName() {
		if (dirName == null) {
			dirName = className + File.separator + fieldName + File.separator
					+ genClassLevel;
		}
	}

	// IMPLEMENTATION OF METHODS FROM THE PClassMapping INTERFACE

	/**
	 * It intialises the class mapping. It especially assigns a mapper as well
	 * as a meta-object describing either a class or a generic class.
	 * @param mapper	The mapper to be assigned to this class mapping.
	 * @param metaclass	The meta-object describing the associated class. It may
	 * 					be null.
	 */
	public void init(PMappingCallback mapper, MetaObject metaclass) throws PException {
		super.init(mapper, metaclass);
		initDirName();
		((FosPMappingStructuresManager) mapper.getPMappingStructuresManager())
				.addDirName(genClassPType.getJormName(), dirName);
	}

	/**
	 * It creates a new PBinding.
	 * @return	The new PBinding with lifecycle state set to
	 * 			LIFECYCLE_NOTBOUND.
	 */
	public PBinding createPBinding() throws PException {
		return new PBindingImpl(this);
	}


	/**
	 * It returns an iterator over all pname of persitent objects availlable in
	 * this class.
	 * @param conn is the connection which permits to access to the support
	 * @return the iterator over PName objects
	 * @exception PExceptionProtocol	A problem occured during the request
	 * 									that is not related to I/O.
	 * @exception PExceptionIO	A problem occured during the request.
	 */
	public PNameIterator getPNameIterator(Object conn,
	                                      boolean withSubType,
	                                      boolean prefetching,
										  Object txctx) throws PException {
		throw new PException("Not supported yet");
	}

	/**
	 * It specifies if the given kind of PMapper is supported by this
	 * PClassMapping.
	 * @param mappername	The name defining the PMapper kind.
	 * @return	true if this kind of mapper is supported.
	 */
	public boolean isConform(String mappername) {
		return mappername.equals("fos");
	}

	public boolean exist(PBinding pb, Object conn) throws PException {
		byte status = pb.getStatus();
		byte nextState =
				PStateGraph.nextStatePBinding(status, PBinding.ACTION_EXIST);
		if (nextState == PBinding.LIFECYCLE_ERROR)
			throw new PExceptionProtocol("Unauthorized operation");
		status = nextState;
		FosConnection connection;
		if (conn == null)
			throw new PExceptionProtocol("Cannot check the existence without a connection");
		else if (conn instanceof FosConnection)
			connection = (FosConnection) conn;
		else
			throw new PExceptionProtocol("Cannot check the existence with this kind of connection");
		try {
			return connection.exist(dirName, pb.getPName().encodeString());
		} catch (FosException fe) {
			throw new PExceptionIO(fe, "Problem while testing the existence of the DSI.");
		}
	}

	public void read(PBinding pb, Object conn, PAccessor pa) throws PException {
		byte status = pb.getStatus();
		byte nextState =
				PStateGraph.nextStatePBinding(status, PBinding.ACTION_READ);
		if (nextState == PBinding.LIFECYCLE_ERROR)
			throw new PExceptionProtocol("Unauthorized operation");
		FosConnection connection;
		if (conn == null)
			throw new PExceptionProtocol("Cannot read without a connection");
		else if (conn instanceof FosConnection)
			connection = (FosConnection) conn;
		else
			throw new PExceptionProtocol("Cannot read with this kind of connection");
		try {
			connection.read(dirName, pb.getPName().encodeString(), this, pa);
		} catch (FosException fe) {
			throw new PExceptionIO(fe, "Problem while reading the DSI.");
		}
		((PBindingCtrl) pb).setStatus(nextState);
	}

	public void write(PBinding pb, Object conn, PAccessor pa) throws PException {
		byte status = pb.getStatus();
		byte nextState =
				PStateGraph.nextStatePBinding(status, PBinding.ACTION_WRITE);
		if (nextState == PBinding.LIFECYCLE_ERROR)
			throw new PExceptionProtocol("Unauthorized operation");
		FosConnection connection;
		if (conn == null)
			throw new PExceptionProtocol("Cannot write without a connection");
		else if (conn instanceof FosConnection)
			connection = (FosConnection) conn;
		else
			throw new PExceptionProtocol("Cannot write with this kind of connection");
		try {
			if (status == PBinding.LIFECYCLE_DELTOWRITE) {
				connection.delete(dirName, pb.getPName().encodeString());
			} else if (status == PBinding.LIFECYCLE_NEWTOWRITE) {
				connection.write(dirName, pb.getPName().encodeString(), this, pa);
			} else if (status == PBinding.LIFECYCLE_ACTIVEFORIO) {
				connection.write(dirName, pb.getPName().encodeString(), this, pa);
			}
		} catch (FosException fe) {
			throw new PExceptionIO(fe, "Problem while writing the DSI.");
		}
		((PBindingCtrl) pb).setStatus(nextState);
	}

	/**
	 * Returns null in case of genclass.
	 */
	public PClassMapping[] getSubPCMs() throws PException {
		return null;
	}
	
	// IMPLEMENTATION OF THE METHODS FROM THE FosGenClassProp INTERFACE

	public void setClassName(String cn) {
		className = cn;
	}

	public void setFieldName(String fn) {
		fieldName = fn;
	}

	public void setGenClassLevel(String level) {
		genClassLevel = level;
	}

	// IMPLEMENTATION OF METHODS FROM THE FosStructure INTERFACE

	/**
	 * Reads the content of a persistent object from the file represented by
	 * the given ObjectInputStream.
	 * @param ois	The ObjectInputStream representing the file from which to
	 *				read.
	 * @param conn	The connection used to access the File Object Store.
	 */
	public void readFile(ObjectInputStream ois, FosAccess conn, Object ctxt)
			throws Exception {
		PGenClassAccessor paccessorGenClass = null;
		if (ctxt == null)
			throw new PExceptionProtocol("Accessor must not be null");
		else if (ctxt instanceof PGenClassAccessor)
			paccessorGenClass = (PGenClassAccessor) ctxt;
		else
			throw new PExceptionProtocol("Wrong type of accessor");
		//Read the number of elements
		int nbelement = ois.readInt();
		paccessorGenClass.paSetNbElem(nbelement);
		for (int j = 0; j < nbelement; j++) {
			PIndexedElem pie = paccessorGenClass.createPIndexedElem();
			// Read the index fields
			FieldDesc[] indexes = getIndexFields();
			for (int i = 0; i < indexes.length; i++) {
				switch (indexes[i].type.getTypeCode()) {
				case PType.TYPECODE_BYTE:
					pie.pieSetByteIndexField(indexes[i].name, ois.readByte());
					break;
				case PType.TYPECODE_CHAR:
					pie.pieSetCharIndexField(indexes[i].name, ois.readChar());
					break;
				case PType.TYPECODE_INT:
					pie.pieSetIntIndexField(indexes[i].name, ois.readInt());
					break;
				case PType.TYPECODE_LONG:
					pie.pieSetLongIndexField(indexes[i].name, ois.readLong());
					break;
				case PType.TYPECODE_SHORT:
					pie.pieSetShortIndexField(indexes[i].name, ois.readShort());
					break;
				case PType.TYPECODE_DATE:
				case PType.TYPECODE_OBJBYTE:
				case PType.TYPECODE_OBJCHAR:
				case PType.TYPECODE_OBJINT:
				case PType.TYPECODE_OBJLONG:
				case PType.TYPECODE_OBJSHORT:
				case PType.TYPECODE_STRING:
					pie.pieSetStringIndexField(indexes[i].name,
											   (String) ois.readObject());
					break;
				default:
					throw new PExceptionProtocol("Unauthorized index type: ("
												 + indexes[i].name + ", "
												 + indexes[i].type.getJormName());
				}
			}
			// Write the element fields
			FieldDesc[] elements = getElemFields();
			switch (elements[0].type.getTypeCode()) {
			case PType.TYPECODE_BOOLEAN:
				pie.pieSetBooleanElem(ois.readBoolean());
				break;
			case PType.TYPECODE_OBJBOOLEAN:
				pie.pieSetObooleanElem((Boolean) ois.readObject());
				break;
			case PType.TYPECODE_BYTE:
				pie.pieSetByteElem(ois.readByte());
				break;
			case PType.TYPECODE_OBJBYTE:
				pie.pieSetObyteElem((Byte) ois.readObject());
				break;
			case PType.TYPECODE_CHAR:
				pie.pieSetCharElem(ois.readChar());
				break;
			case PType.TYPECODE_OBJCHAR:
				pie.pieSetOcharElem((Character) ois.readObject());
				break;
			case PType.TYPECODE_SHORT:
				pie.pieSetBooleanElem(ois.readBoolean());
				break;
			case PType.TYPECODE_OBJSHORT:
				pie.pieSetOshortElem((Short) ois.readObject());
				break;
			case PType.TYPECODE_INT:
				pie.pieSetIntElem(ois.readInt());
				break;
			case PType.TYPECODE_OBJINT:
				pie.pieSetOintElem((Integer) ois.readObject());
				break;
			case PType.TYPECODE_LONG:
				pie.pieSetLongElem(ois.readLong());
				break;
			case PType.TYPECODE_OBJLONG:
				pie.pieSetOlongElem((Long) ois.readObject());
				break;
			case PType.TYPECODE_FLOAT:
				pie.pieSetFloatElem(ois.readFloat());
				break;
			case PType.TYPECODE_OBJFLOAT:
				pie.pieSetOfloatElem((Float) ois.readObject());
				break;
			case PType.TYPECODE_DOUBLE:
				pie.pieSetDoubleElem(ois.readDouble());
				break;
			case PType.TYPECODE_OBJDOUBLE:
				pie.pieSetOdoubleElem((Double) ois.readObject());
				break;
			case PType.TYPECODE_DATE:
				pie.pieSetDateElem((Date) ois.readObject());
				break;
			case PType.TYPECODE_CHARARRAY:
				pie.pieSetCharArrayElem((char[]) ois.readObject());
				break;
			case PType.TYPECODE_BYTEARRAY:
				pie.pieSetByteArrayElem((byte[]) ois.readObject());
				break;
			case PType.TYPECODE_SERIALIZED:
				pie.pieSetSerializedElem((Serializable) ois.readObject());
				break;
			case PType.TYPECODE_BIGDECIMAL:
				pie.pieSetBigDecimalElem((BigDecimal) ois.readObject());
				break;
			case PType.TYPECODE_STRING:
				pie.pieSetStringElem((String) ois.readObject());
				break;
			case PType.TYPECODE_REFERENCE:
				pie.pieSetRefElem(getPNameCoder()
						.decodeString((String) ois.readObject()));
				break;
			default:
				throw new PExceptionProtocol("Unauthorized element type: ("
											 + elements[0].name + ", "
											 + elements[0].type.getJormName());
			}
			paccessorGenClass.paAdd(pie, conn);
		}
	}

	/**
	 * Writes the content of a persistent object to the file represented by
	 * the given ObjectOutputStream.
	 * @param oos	The ObjectOutputStream representing the file to which to
	 *				write.
	 * @param conn	The connection used to access the File Object Store.
	 */
	public void writeFile(ObjectOutputStream oos, FosAccess conn, Object ctxt)
			throws Exception {
		PGenClassAccessor paccessorGenClass = null;
		if (ctxt == null)
			throw new PExceptionProtocol("Accessor must not be null");
		else if (ctxt instanceof PGenClassAccessor)
			paccessorGenClass = (PGenClassAccessor) ctxt;
		else
			throw new PExceptionProtocol("Wrong type of accessor");
		//Write the number of element
		oos.writeInt(paccessorGenClass.paGetNbElem());
		Iterator pies = paccessorGenClass.paIterator();
		// For each element of the genclass
		while (pies.hasNext()) {
			PIndexedElem pie = (PIndexedElem) pies.next();
			if (pie.getElemStatus() == PIndexedElem.ELEM_DELETED)
				continue;
			// Write the index fields
			FieldDesc[] indexes = getIndexFields();
			for (int i = 0; i < indexes.length; i++) {
				switch (indexes[i].type.getTypeCode()) {
				case PType.TYPECODE_BYTE:
					oos.writeByte(pie.pieGetByteIndexField(indexes[i].name));
					break;
				case PType.TYPECODE_CHAR:
					oos.writeChar(pie.pieGetCharIndexField(indexes[i].name));
					break;
				case PType.TYPECODE_INT:
					oos.writeInt(pie.pieGetIntIndexField(indexes[i].name));
					break;
				case PType.TYPECODE_LONG:
					oos.writeLong(pie.pieGetLongIndexField(indexes[i].name));
					break;
				case PType.TYPECODE_SHORT:
					oos.writeShort(pie.pieGetShortIndexField(indexes[i].name));
					break;
				case PType.TYPECODE_DATE:
				case PType.TYPECODE_OBJBYTE:
				case PType.TYPECODE_OBJCHAR:
				case PType.TYPECODE_OBJINT:
				case PType.TYPECODE_OBJLONG:
				case PType.TYPECODE_OBJSHORT:
				case PType.TYPECODE_STRING:
					oos.writeObject(pie.pieGetStringIndexField(indexes[i].name));
					break;
				default:
					throw new PExceptionProtocol("Unauthorized index type: ("
												 + indexes[i].name + ", "
												 + indexes[i].type.getJormName() + ")");
				}
			}
			// Write the element fields
			FieldDesc[] elements = getElemFields();
			switch (elements[0].type.getTypeCode()) {
			case PType.TYPECODE_BOOLEAN:
				oos.writeBoolean(pie.pieGetBooleanElem());
				break;
			case PType.TYPECODE_OBJBOOLEAN:
				oos.writeObject(pie.pieGetObooleanElem());
				break;
			case PType.TYPECODE_BYTE:
				oos.writeByte(pie.pieGetByteElem());
				break;
			case PType.TYPECODE_OBJBYTE:
				oos.writeObject(pie.pieGetObyteElem());
				break;
			case PType.TYPECODE_CHAR:
				oos.writeChar(pie.pieGetCharElem());
				break;
			case PType.TYPECODE_OBJCHAR:
				oos.writeObject(pie.pieGetOcharElem());
				break;
			case PType.TYPECODE_SHORT:
				oos.writeShort(pie.pieGetShortElem());
				break;
			case PType.TYPECODE_OBJSHORT:
				oos.writeObject(pie.pieGetOshortElem());
				break;
			case PType.TYPECODE_INT:
				oos.writeInt(pie.pieGetIntElem());
				break;
			case PType.TYPECODE_OBJINT:
				oos.writeObject(pie.pieGetOintElem());
				break;
			case PType.TYPECODE_LONG:
				oos.writeLong(pie.pieGetLongElem());
				break;
			case PType.TYPECODE_OBJLONG:
				oos.writeObject(pie.pieGetOlongElem());
				break;
			case PType.TYPECODE_FLOAT:
				oos.writeFloat(pie.pieGetFloatElem());
				break;
			case PType.TYPECODE_OBJFLOAT:
				oos.writeObject(pie.pieGetOfloatElem());
				break;
			case PType.TYPECODE_DOUBLE:
				oos.writeDouble(pie.pieGetDoubleElem());
				break;
			case PType.TYPECODE_OBJDOUBLE:
				oos.writeObject(pie.pieGetOdoubleElem());
				break;
			case PType.TYPECODE_DATE:
				oos.writeObject(pie.pieGetDateElem());
				break;
			case PType.TYPECODE_CHARARRAY:
				oos.writeObject(pie.pieGetCharArrayElem());
				break;
			case PType.TYPECODE_BYTEARRAY:
				oos.writeObject(pie.pieGetByteArrayElem());
				break;
			case PType.TYPECODE_SERIALIZED:
				oos.writeObject(pie.pieGetSerializedElem());
				break;
			case PType.TYPECODE_BIGDECIMAL:
				oos.writeObject(pie.pieGetBigDecimalElem());
				break;
			case PType.TYPECODE_STRING:
				oos.writeObject(pie.pieGetStringElem());
				break;
			case PType.TYPECODE_REFERENCE:
				PName pn = pie.pieGetRefElem();
				oos.writeObject(pn.encodeString());
				break;
			default:
				throw new PExceptionProtocol("Unauthorized element type: ("
											 + elements[0].name + ", "
											 + elements[0].type.getJormName() + ")");
			}
		}
	}

	/* (non-Javadoc)
	 * @see org.objectweb.jorm.api.PClassMapping#getAssociationTable()
	 */
	public HashMap getAssociationTable() {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see org.objectweb.jorm.api.PClassMapping#addAssociation(org.objectweb.jorm.api.PClassMapping, int[])
	 */
	public void addAssociation(PClassMapping targetClass, int[] indexes) {
		// TODO Auto-generated method stub
		
	}

	/* (non-Javadoc)
	 * @see org.objectweb.jorm.api.PClassMapping#getIndexesTable(org.objectweb.jorm.api.PClassMapping)
	 */
	public int[] getIndexesTable(PClassMapping targetClass) {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see org.objectweb.jorm.api.PClassMapping#resolve(org.objectweb.jorm.naming.api.PName)
	 */
	public PName resolve(Object conn, PName pname) throws PException {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see org.objectweb.jorm.api.PClassMapping#match(org.objectweb.jorm.naming.api.PName)
	 */
	public boolean match(Object obj, boolean b) throws PException{
		// TODO Auto-generated method stub
		return false;
	}

	/* (non-Javadoc)
	 * @see org.objectweb.jorm.api.PClassMapping#getDecodedPName(org.objectweb.medor.tuple.api.TupleCollection, org.objectweb.jorm.naming.api.PName)
	 */
	public PName getDecodedPName(TupleCollection tc, PName pname, boolean b) throws PException {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see org.objectweb.jorm.api.PClassMapping#init(org.objectweb.jorm.api.PMappingStructuresManager)
	 */
	public void init(PMappingStructuresManager pmsm) throws PException {
		// TODO Auto-generated method stub
		
	}

	/* (non-Javadoc)
	 * @see org.objectweb.jorm.api.PClassMapping#classDefined(org.objectweb.jorm.api.PMappingStructuresManager)
	 */
	public void classDefined(PMappingStructuresManager pmsm) throws PException {
		// TODO Auto-generated method stub
		
	}
}
