/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.fos.lib;

import org.objectweb.jorm.api.PMappingStructuresManager;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.api.PMapCluster;
import org.objectweb.util.monolog.api.Logger;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Collection;

/**
 * Defines the way to manage relational schema for the FOS mappers.
 * @author	P. Dechamboux
 */
public class FosPMappingStructuresManager implements PMappingStructuresManager {
	private PMapper pMapper = null;
	private ArrayList clusterList = new ArrayList();
	protected Logger logger = null;


	public synchronized void declareClass(String jcname) {
		Iterator it = clusterList.iterator();
		FosPMapCluster cl;
		while (it.hasNext()) {
			cl = (FosPMapCluster) it.next();
			if (cl.containClass(jcname)) {
				return;
			}
		}
		cl = new FosPMapCluster(jcname, this);
		clusterList.add(cl);
	}

	/**
	 * Closes a mapper connection.
	 * @param conn	The mapper connection.
	 * @throws org.objectweb.jorm.api.PException
	 */
	void closeConnection(Object conn) throws PException {
		pMapper.closeConnection(conn);
	}

	/**
	 * Gets a connection from the mapper.
	 * @return The allocated connection.
	 * @throws org.objectweb.jorm.api.PException
	 */
	Object getConnection() throws PException {
		Object conn = pMapper.getConnection();
		if (conn == null)
			throw new PExceptionProtocol("Impossible to initialize the mapping structures without connection");
		return conn;
	}

	/**
	 * Retrieves or creates an FosPMapCluster associated with this JORM class.
	 * @param jcname The name of this JORM class.
	 * @return The associated FosPMapCluster.
	 */
	private FosPMapCluster getAlwaysPMapCluster(String jcname,
												String dirname)
			throws PException {
		FosPMapCluster cl = null;
		Iterator it = clusterList.iterator();
		while (it.hasNext()) {
			cl = (FosPMapCluster) it.next();
			if (cl.containDirectory(dirname)) {
				break;
			}
			cl = null;
		}
		if (cl == null) {
			// no map cluster use dirname
			cl = (FosPMapCluster) getPMapCluster(jcname);
			if (cl == null) {
				// no map cluster for jcname : creates a new one
				cl = new FosPMapCluster(jcname, dirname, this);
				clusterList.add(cl);
			}
			// there is already a map cluster for jcname
		} else {
			// there is already a map cluster for dirname
			FosPMapCluster cl2 = (FosPMapCluster) getPMapCluster(jcname);
			if ((cl != cl2) && (cl2 != null)) {
				// there is also a map cluster for jcname : merge both into cl
				cl.merge(cl2);
				cl.addDirName(dirname);
				clusterList.remove(cl2);
			} else {
				// no map cluster for jcname : declare jcname into cl
				cl.addDependency(jcname);
			}
		}
		return cl;
	}

	public void addDependency(String jcname1, String jcname2) throws PException {
//TODO:
	}

	public void classDefined(String jcname) throws PException {
//TODO:
	}

	/**
	 * Adds a table column into these RDB mapping structures.
	 * @param jcname	The JORM class name.
	 * @param dirname	The table name.
	 * @throws PException
	 */
	synchronized public void addDirName(String jcname, String dirname)
			throws PException {
		FosPMapCluster cl = getAlwaysPMapCluster(jcname, dirname);
		cl.addDirName(dirname);
	}

	// IMPLEMENTATION OF METHODS FROM THE PMappingStructuresManager INTERFACE

	/**
	 * Asks for the map cluster associated with a JORM class.
	 * @return The relevant map cluster.
	 * @throws PException   This class has not been mapped.
	 */
	public PMapCluster getPMapCluster(String jcname) throws PException {
		Iterator it = clusterList.iterator();
		while (it.hasNext()) {
			FosPMapCluster cl = (FosPMapCluster) it.next();
			if (cl.containClass(jcname)) {
				return cl;
			}
		}
		return null;
	}

	/**
	 * Asks for all the map clusters defined within this mapper.
	 * @return The Iterator over the collection of map clusters.
	 */
	public Collection getPMapClusters() {
		return clusterList;
	}

	/**
	 * Assigns a mapper to this schema manager.
	 * @param pm	The concerned mapper.
	 */
	public void setPMapper(PMapper pm) {
		pMapper = pm;
	}

	/**
	 * Assigns a logger to this schema manager.
	 * @param l		The concerned logger.
	 */
	public void setLogger(Logger l) {
		logger = l;
	}

	public void clear() throws PException {
		// TODO Auto-generated method stub
		
	}
}
