/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.rdb.genclass;

import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapter;
import org.objectweb.jorm.naming.api.PNameGetter;
import org.objectweb.jorm.genclass.api.FieldDesc;
import org.objectweb.jorm.type.api.PExceptionTyping;
import org.objectweb.jorm.type.api.PType;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.PreparedStatement;
import java.util.Date;
import java.math.BigInteger;
import java.math.BigDecimal;

/**
 * This is a class that implements the PNameGetter interface and that provides
 * a generic access based on the description of fields of the element of the
 * relevant generic class instance.
 * @author	P. D?chamboux
 */

class RdbGenClassNameGetter implements PNameGetter {
	/**
	 * The ResultSet pointing to the row from which to fetch the value of the
	 * fields of a composite name.
	 */
	protected ResultSet resultSet;
	/**
	 * The description of the fields of a composite name.
	 */
	protected FieldDesc[] elemFields;

	protected PreparedStatement ps;
	
	protected RdbAdapter adapter;
	
	protected int startIndex = 0;

	/**
	 * Constructs a getter that enables to retrieve the values of the fields
	 * of a composite name.
	 */
	RdbGenClassNameGetter(PreparedStatement ps, 
	        ResultSet rs, 
	        FieldDesc[] rfds, 
	        int startIndex,
	        RdbAdapter adapter) {
		resultSet = rs;
		elemFields = rfds;
		this.ps = ps;
		this.startIndex = startIndex;
		this.adapter = adapter;
	}

	protected void finalize() throws Throwable {
		if (resultSet != null)
			resultSet.close();
		if (ps != null)
			ps.close();
		super.finalize();
	}

	/**
	 * It retrieves the value of a byte field.
	 * @param fn	A String containing the name of the field.
	 * @param ctx A context used by PNameGetter to extract the related data.
	 * @return	The byte value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 */
	public byte pngetByteField(String fn, Object ctx)
			throws PExceptionTyping, PExceptionIO {
		try {
			for (short i = 1; i < elemFields.length; i++)
				if (fn.equals(elemFields[i].compositeName))
					if (elemFields[i].type.getTypeCode() == PType.TYPECODE_BYTE)
						return adapter.getByte(resultSet, startIndex + i, (byte) -1);
					else
						break;
		} catch (SQLException e) {
			throw new PExceptionIO(e, "Problem while retreiving PNameGetter field from ResultSet: field <" + fn + ">");
		}
		throw new PExceptionTyping("Wrong composite name field " + fn + " in a GenClass.");
	}

	/**
	 * It retrieves the value of a Byte field.
	 * @param fn	A String containing the name of the field.
	 * @return	The Byte value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	public Byte pngetObyteField(String fn, Object context) throws PExceptionTyping, PExceptionIO {
		try {
			for (short i = 1; i < elemFields.length; i++)
				if (fn.equals(elemFields[i].compositeName))
					if (elemFields[i].type.getTypeCode()
							== PType.TYPECODE_OBJBYTE) {
						return adapter.getObyte(resultSet, startIndex + i, null);
					} else
						break;
		} catch (SQLException e) {
			throw new PExceptionIO(e, "Problem while retreiving PNameGetter field from ResultSet: field <" + fn + ">");
		}
		throw new PExceptionTyping("Wrong composite name field " + fn + " in a GenClass.");
	}

	/**
	 * It retrieves the value of a char field.
	 * @param fn	A String containing the name of the field.
	 * @param ctx A context used by PNameGetter to extract the related data.
	 * @return	The char value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 */
	public char pngetCharField(String fn, Object ctx)
			throws PExceptionTyping, PExceptionIO {
		try {
			for (short i = 1; i < elemFields.length; i++)
				if (fn.equals(elemFields[i].compositeName))
					if (elemFields[i].type.getTypeCode() == PType.TYPECODE_CHAR)
						return adapter.getChar(resultSet, startIndex + i, (char) -1);
					else
						break;
		} catch (SQLException e) {
			throw new PExceptionIO(e, "Problem while retreiving PNameGetter field from ResultSet: field <" + fn + ">");
		}
		throw new PExceptionTyping("Wrong composite name field " + fn + " in a GenClass.");
	}

	/**
	 * It retrieves the value of a Character field.
	 * @param fn	A String containing the name of the field.
	 * @return	The Character value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	public Character pngetOcharField(String fn, Object context) throws PExceptionTyping, PExceptionIO {
		try {
			for (short i = 1; i < elemFields.length; i++)
				if (fn.equals(elemFields[i].compositeName))
					if (elemFields[i].type.getTypeCode()
							== PType.TYPECODE_OBJCHAR) {
						return adapter.getOchar(resultSet, startIndex + i, null);
					} else
						break;
		} catch (SQLException e) {
			throw new PExceptionIO(e, "Problem while retreiving PNameGetter field from ResultSet: field <" + fn + ">");
		}
		throw new PExceptionTyping("Wrong composite name field " + fn + " in a GenClass.");
	}

	/**
	 * It retrieves the value of a short field.
	 * @param fn	A String containing the name of the field.
	 * @param ctx A context used by PNameGetter to extract the related data.
	 * @return	The short value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 */
	public short pngetShortField(String fn, Object ctx)
			throws PExceptionTyping, PExceptionIO {
		try {
			for (short i = 1; i < elemFields.length; i++)
				if (fn.equals(elemFields[i].compositeName))
					if (elemFields[i].type.getTypeCode()
							== PType.TYPECODE_SHORT)
						return adapter.getShort(resultSet, startIndex + i, (short) -1);
					else
						break;
		} catch (SQLException e) {
			throw new PExceptionIO(e, "Problem while retreiving PNameGetter field from ResultSet: field <" + fn + ">");
		}
		throw new PExceptionTyping("Wrong composite name field " + fn + " in a GenClass.");
	}

	/**
	 * It retrieves the value of a Short field.
	 * @param fn	A String containing the name of the field.
	 * @return	The Short value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	public Short pngetOshortField(String fn, Object context) throws PExceptionTyping, PExceptionIO {
		try {
			for (short i = 1; i < elemFields.length; i++)
				if (fn.equals(elemFields[i].compositeName))
					if (elemFields[i].type.getTypeCode()
							== PType.TYPECODE_OBJSHORT) {
						return adapter.getOshort(resultSet, startIndex + i, null);
					} else
						break;
		} catch (SQLException e) {
			throw new PExceptionIO(e, "Problem while retreiving PNameGetter field from ResultSet: field <" + fn + ">");
		}
		throw new PExceptionTyping("Wrong composite name field " + fn + " in a GenClass.");
	}

	/**
	 * It retrieves the value of an int field.
	 * @param fn	A String containing the name of the field.
	 * @param ctx A context used by PNameGetter to extract the related data.
	 * @return	The int value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 */
	public int pngetIntField(String fn, Object ctx)
			throws PExceptionTyping, PExceptionIO {
		try {
			for (short i = 1; i < elemFields.length; i++)
				if (fn.equals(elemFields[i].compositeName))
					if (elemFields[i].type.getTypeCode() == PType.TYPECODE_INT)
						return adapter.getInt(resultSet, startIndex + i, -1);
					else
						break;
		} catch (SQLException e) {
			throw new PExceptionIO(e, "Problem while retreiving PNameGetter field from ResultSet: field <" + fn + ">");
		}
		throw new PExceptionTyping("Wrong composite name field " + fn + " in a GenClass.");
	}

	/**
	 * It retrieves the value of an Integer field.
	 * @param fn	A String containing the name of the field.
	 * @return	The Integer value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	public Integer pngetOintField(String fn, Object ctxt) throws PExceptionTyping, PExceptionIO {
		try {
			for (short i = 1; i < elemFields.length; i++)
				if (fn.equals(elemFields[i].compositeName))
					if (elemFields[i].type.getTypeCode()
							== PType.TYPECODE_OBJINT) {
						return adapter.getOint(resultSet, startIndex + i, null);
					} else
						break;
		} catch (SQLException e) {
			throw new PExceptionIO(e, "Problem while retreiving PNameGetter field from ResultSet: field <" + fn + ">");
		}
		throw new PExceptionTyping("Wrong composite name field " + fn + " in a GenClass.");
	}

	/**
	 * It retrieves the value of a long field.
	 * @param fn	A String containing the name of the field.
	 * @param ctx A context used by PNameGetter to extract the related data.
	 * @return	The long value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 */
	public long pngetLongField(String fn, Object ctx)
			throws PExceptionTyping, PExceptionIO {
		try {
			for (short i = 1; i < elemFields.length; i++)
				if (fn.equals(elemFields[i].compositeName))
					if (elemFields[i].type.getTypeCode() == PType.TYPECODE_LONG)
						return adapter.getLong(resultSet, startIndex + i, -1);
					else
						break;
		} catch (SQLException e) {
			throw new PExceptionIO(e, "Problem while retreiving PNameGetter field from ResultSet: field <" + fn + ">");
		}
		throw new PExceptionTyping("Wrong composite name field " + fn + " in a GenClass.");
	}

	/**
	 * It retrieves the value of a Long field.
	 * @param fn	A String containing the name of the field.
	 * @return	The Long value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	public Long pngetOlongField(String fn, Object ctxt) throws PExceptionTyping, PExceptionIO {
		try {
			for (short i = 1; i < elemFields.length; i++)
				if (fn.equals(elemFields[i].compositeName))
					if (elemFields[i].type.getTypeCode()
							== PType.TYPECODE_OBJLONG) {
						return adapter.getOlong(resultSet, startIndex + i, null);
					} else
						break;
		} catch (SQLException e) {
			throw new PExceptionIO(e, "Problem while retreiving PNameGetter field from ResultSet: field <" + fn + ">");
		}
		throw new PExceptionTyping("Wrong composite name field " + fn + " in a GenClass.");
	}

	/**
	 * It retrieves the value of a string field.
	 * @param fn	A String containing the name of the field.
	 * @param ctx A context used by PNameGetter to extract the related data.
	 * @return	The java.lang.String value of the field as stored within the
	 * 			composite name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 */
	public String pngetStringField(String fn, Object ctx)
			throws PExceptionTyping, PExceptionIO {
		try {
			for (short i = 1; i < elemFields.length; i++)
				if (fn.equals(elemFields[i].compositeName))
					if (elemFields[i].type.getTypeCode() == PType.TYPECODE_STRING) {
						return adapter.getString(resultSet, startIndex + i, null);
					} else
						break;
		} catch (SQLException e) {
			throw new PExceptionIO(e, "Problem while retreiving PNameGetter field from ResultSet: field <" + fn + ">");
		}
		throw new PExceptionTyping("Wrong composite name field " + fn + " in a GenClass.");
	}

	/**
	 * It retrieves the value of a string field.
	 * @param fn	A String containing the name of the field.
	 * @param ctx A context used by PNameGetter to extract the related data.
	 * @return	The java.lang.String value of the field as stored within the
	 * 			composite name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 */
	public byte[] pngetByteArrayField(String fn, Object ctx)
			throws PExceptionTyping, PExceptionIO {
		try {
			for (short i = 1; i < elemFields.length; i++)
				if (fn.equals(elemFields[i].compositeName))
					if (elemFields[i].type.getTypeCode() == PType.TYPECODE_BYTEARRAY) {
						return adapter.getByteArray(resultSet, startIndex + i, null);
					} else
						break;
		} catch (SQLException e) {
			throw new PExceptionIO(e, "Problem while retreiving PNameGetter field from ResultSet: field <" + fn + ">");
		}
		throw new PExceptionTyping("Wrong composite name field " + fn + " in a GenClass.");
	}

	/**
	 * It retrieves the value of a string field.
	 * @param fn	A String containing the name of the field.
	 * @param ctx A context used by PNameGetter to extract the related data.
	 * @return	The java.lang.String value of the field as stored within the
	 * 			composite name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 */
	public char[] pngetCharArrayField(String fn, Object ctx)
			throws PExceptionTyping, PExceptionIO {
		try {
			for (short i = 1; i < elemFields.length; i++)
				if (fn.equals(elemFields[i].compositeName))
					if (elemFields[i].type.getTypeCode() == PType.TYPECODE_CHARARRAY) {
						return adapter.getCharArray(resultSet, startIndex + i, null);
					} else
						break;
		} catch (Exception e) {
			throw new PExceptionIO(e, "Problem while retreiving PNameGetter field from ResultSet: field <" + fn + ">");
		}
		throw new PExceptionTyping("Wrong composite name field " + fn + " in a GenClass.");
	}

	/**
	 * It retrieves the value of a date field.
	 * @param fn	A String containing the name of the field.
	 * @return	The java.util.Date value of the field as stored within the
	 * 			composite name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	public Date pngetDateField(String fn, Object ctxt) throws PExceptionTyping, PExceptionIO {
		try {
			for (short i = 1; i < elemFields.length; i++)
				if (fn.equals(elemFields[i].compositeName))
					if (elemFields[i].type.getTypeCode()
							== PType.TYPECODE_DATE) {
						return adapter.getDate(resultSet, startIndex + i, null);
					} else
						break;
		} catch (SQLException e) {
			throw new PExceptionIO(e, "Problem while retreiving PNameGetter field from ResultSet: field <" + fn + ">");
		}
		throw new PExceptionTyping("Wrong composite name field " + fn + " in a GenClass.");
	}

	/**
	 * It retrieves the value of a BigInteger field.
	 * @param fn	A String containing the name of the field.
	 * @param ctxt	A context used by PNameGetter to extract the related data.
	 * @return	The java.math.BigInteger value of the field as stored within the
	 * 			composite name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	public BigInteger pngetBigIntegerField(String fn, Object ctxt) throws PException {
		try {
			for (short i = 1; i < elemFields.length; i++)
				if (fn.equals(elemFields[i].compositeName))
					if (elemFields[i].type.getTypeCode()
							== PType.TYPECODE_BIGINTEGER) {
						return adapter.getBigInteger(resultSet, startIndex + i, null);
					} else
						break;
		} catch (SQLException e) {
			throw new PExceptionIO(e, "Problem while retreiving PNameGetter field from ResultSet: field <" + fn + ">");
		}
		throw new PExceptionTyping("Wrong composite name field " + fn + " in a GenClass.");
	}

	/**
	 * It retrieves the value of a BigDecimal field.
	 * @param fn	A String containing the name of the field.
	 * @param ctxt	A context used by PNameGetter to extract the related data.
	 * @return	The java.math.BigDecimal value of the field as stored within the
	 * 			composite name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	public BigDecimal pngetBigDecimalField(String fn, Object ctxt) throws PException {
		try {
			for (short i = 1; i < elemFields.length; i++)
				if (fn.equals(elemFields[i].compositeName))
					if (elemFields[i].type.getTypeCode()
							== PType.TYPECODE_BIGDECIMAL) {
						return adapter.getBigDecimal(resultSet, startIndex + i, null);
					} else
						break;
		} catch (SQLException e) {
			throw new PExceptionIO(e, "Problem while retreiving PNameGetter field from ResultSet: field <" + fn + ">");
		}
		throw new PExceptionTyping("Wrong composite name field " + fn + " in a GenClass.");
	}
}
