/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.rdb.generator;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.generator.lib.CommonHelper;
import org.objectweb.jorm.mapper.rdb.adapter.RdbAdapterFactory;
import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapter;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbClassMapping;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbClassMultiMapping;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbExternalTable;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbJoin;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbMapping;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbPrimitiveElementMapping;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbTable;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ClassMapping;
import org.objectweb.jorm.metainfo.api.Mapping;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.Package;
import org.objectweb.jorm.metainfo.api.ParentClassMapping;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.api.Reference;
import org.objectweb.jorm.metainfo.api.ReferenceMapping;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

/**
 * This helper contains all common tool methods shared by the RdbBindingMOP and
 * the RdbMappingMOP.
 * @author Sebastien Chassande-Barrioz
 */
public class RdbCommonHelper extends CommonHelper {

	private Map rdbAdapters = new HashMap();

	public RdbCommonHelper() {
	}
	public RdbCommonHelper(Logger logger) {
		setLogger(logger);
	}

	/**
	 * Retrieves the adapter associated with the given mapper.
	 * @param mappername    The name of the corresponding mapper.
	 * @return  The RDB adapter.
	 * @throws org.objectweb.jorm.api.PException
	 */
	protected RdbAdapter getRdbAdapter(String mappername) throws PException {
		RdbAdapter res = (RdbAdapter) rdbAdapters.get(mappername);
		if (res != null) {
			return res;
		}
		String dbn;
		if (mappername.indexOf(".") == -1)
			dbn = RdbAdapterFactory.DATABASE_NAME_JDBC;
		else
			dbn = mappername.substring(mappername.indexOf(".") + 1);
		try {
			res = RdbAdapterFactory.getTypeConverter(dbn);
		} catch (org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapterException e) {
			throw new PException(e, "");
		}
		rdbAdapters.put(mappername, res);
		return res;
	}

	public String getSqlType(RdbPrimitiveElementMapping pem,
							 String adapter,
							 boolean usedInPk) throws PException {
		String userSQLType = pem.getType();
		if (userSQLType == null || userSQLType.length() == 0) {
			StringBuffer sb = new StringBuffer(adapter);
			sb.append(".getSqlType(");
			PrimitiveElement pe = (PrimitiveElement) pem.getLinkedMO();
			sb.append(pe.getType().getProgName());
			sb.append(".getTypeCode(), ");
			sb.append(usedInPk);
			sb.append(", ");
			if (pe.getSize() == PType.NOSIZE) {
				sb.append("RdbAdapter.NOSIZE");
			} else {
				sb.append(pe.getSize());
			}
			sb.append(", ");
			if (pe.getScale() == PType.NOSIZE) {
				sb.append("RdbAdapter.NOSIZE");
			} else {
				sb.append(pe.getScale());
			}
			sb.append(")");
			return sb.toString();
		} else {
			return "\"" + userSQLType + "\"";
		}
	}
	/**
	 * Creates a generation context that can be shared between binding and
	 * mapping generation for a particular class.
	 *
	 * If the current class has a parent class and if a vertical mapping
	 * has been defined: the main table of the current class then references
	 *
	 *
	 * cm.getRdbTable() returns the main table of one of the parent classes (P class).
	 * rcm.getRdbAllExternalTables() : returns all the external tables defines for
	 * the current class and for all the parent classes up to the P class.
	 *
	 */
	RdbGenInfos getRdbGenInfos(Class clazz, String projectName, String mapperName)
	throws PException {
			Mapping mapping = clazz.getClassProject(projectName).getMapping(mapperName);
			return createRdbGenInfosContext((RdbMapping) mapping,
				(Class) mapping.getClassMapping().getJormClass());
	}

	RdbGenInfos getRdbGenInfos(Mapping mapping)
			throws PException {
		return createRdbGenInfosContext((RdbMapping) mapping,
			(Class) mapping.getClassMapping().getLinkedMO());
	}

	RdbGenInfos getRdbGenInfos(Mapping mapping, Map ctx)
		throws PException {
	    return createRdbGenInfosContext((RdbMapping) mapping,
	            (Class) mapping.getClassMapping().getLinkedMO(), ctx);
	}

	RdbGenInfos createRdbGenInfosContext(RdbMapping mapping, Class clazz) throws PException {
	    return createRdbGenInfosContext(mapping, clazz, new HashMap());
	}
	RdbGenInfos createRdbGenInfosContext(RdbMapping mapping, Class clazz, Map ctx)
			throws PException {
	    //mapping2geninfos.clear();
		debug = logger != null && logger.isLoggable(BasicLevel.DEBUG);

        // create the gen info and register it
		RdbGenInfos rgi = new RdbGenInfos();

        // set mapping, logger & debug
		rgi.mapping = mapping;
		rgi.logger = logger;
		rgi.debug = debug;



        // get the main table 
		RdbClassMultiMapping classMapping = (RdbClassMultiMapping)
                mapping.getClassMapping();
		if (debug) {
			logger.log(BasicLevel.DEBUG, "mapping.getClassMapping() " + classMapping);
			logger.log(BasicLevel.DEBUG, "Creating the generation MI for the class = "
					+ ((Class) classMapping.getLinkedMO()).getFQName());
		}


        // create a RdbGenTable from the RdbTable of the class
        RdbTable mainRdbTable = classMapping.getMainRdbTable();
        if (mainRdbTable == null) {
            logger.log(BasicLevel.DEBUG, "No main table defined for this " +
                    "class (implicit inheritance ?). Abort Gen infos creation");
            return rgi;
        }
		if (debug) {
			logger.log(BasicLevel.DEBUG, "Main table: "
                    + mainRdbTable.getName());
		}

        RdbGenTable mainGenTable = new RdbGenTable(mainRdbTable, clazz, rgi);

		rgi.mainTable = mainGenTable;


		// Retrieves inherited mapping definition
		getInheritedData(classMapping,
				mainGenTable.inheritedColumns,
				rgi.readableColumns,
				rgi.genRefs,
				rgi.tables,
				clazz,
				ctx);
        if (debug) {
            logger.log(BasicLevel.DEBUG, "Inherited columns:");
            printColumns(rgi.readableColumns);
        }

        // Filtered mapping and concrete parent classes

		rgi.tables.put(mainGenTable.tableName, mainGenTable);
		addColumnUnique(mainGenTable.columns, rgi.readableColumns, false);


		// Get the external tables and defined columns
        for (Iterator it = classMapping.getExternalTables().iterator();
             it.hasNext();) {
            RdbExternalTable extTable = (RdbExternalTable) it.next();

            // translation: RdbTable => RdbGenTable
            RdbGenTable genTable = new RdbGenTable(extTable, mainGenTable,
                    clazz, rgi);

            // cross reference: genInfos <->> genTable
            rgi.tables.put(genTable.tableName, genTable);


            if (debug) {
                logger.log(BasicLevel.DEBUG,
                        "ext table -- name = " + genTable.getTableName());
            }

            // add columns that do not participate in the join
            // with main table (OL: why ??)
            for (int i = 0; i < genTable.columns.size(); i++) {
                RdbGenColumn genColumn = (RdbGenColumn) genTable.columns.get(i);
                if (genColumn.joinCol == null) {
                    rgi.readableColumns.add(genColumn);
                }
            }
        }

        // define GenRef of the class identifier
		//System.out.println("get the namedef");
		NameDef nd = classMapping.getIdentifierMapping().getNameDef();
		Integer ndIdInteger = (Integer) ctx.get("ndid");
		
		int ndId = (ndIdInteger == null ? 0 : ndIdInteger.intValue());
		
		rgi.genId = defineGenRef(nd, null, ndId++, classMapping, rgi);

        if (debug) {
            for (Iterator iter = clazz.getAllFields().iterator();
                 iter.hasNext();) {
                TypedElement typedElem = (TypedElement) iter.next();
                if (typedElem instanceof Reference) {
                    logger.log(BasicLevel.DEBUG, "current class: " + clazz.getName() + "current reference field " + typedElem.getName());
                    logger.log(BasicLevel.DEBUG, "classMapping.getReferenceMapping(typedElem.getName()) " + classMapping.getReferenceMapping(typedElem.getName()));
                    NameDef namedef = (NameDef) classMapping.getReferenceMapping(typedElem.getName()).getLinkedMO();
                    logger.log(BasicLevel.DEBUG, "current reference field " + typedElem.getName() + " namedef " + namedef);
                }
            }
        }

		// define GenRef for each reference field of the current class
		for (Iterator it = classMapping.getReferenceMappings().iterator(); it.hasNext();) {
			nd = (NameDef) ((ReferenceMapping) it.next()).getLinkedMO();
			String fn = ((Reference) nd.getParent()).getName();
			rgi.genRefs.put(fn, defineGenRef(nd, fn, ndId++, classMapping, rgi));
		}
		ctx.put("ndid", new Integer(ndId));

        // fix the mustGenerateUpdate attribute for this table
		for (Iterator it = rgi.tables.values().iterator(); it.hasNext();) {
			RdbGenTable rgt = ((RdbGenTable) it.next());
			rgt.fixMustGenerateUpdate();
			if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
				logger.log(BasicLevel.DEBUG, "GenerateUpdate for table " + rgt.tableName +
						": " + rgt.mustGenerateUpdate);
				if (!rgt.mustGenerateUpdate) {
					logger.log(BasicLevel.DEBUG, "columns=" + rgt.columns);
				}
			}
		}

		Collections.sort(rgi.readableColumns, new RGCComparator());
		if (debug) {
			logger.log(BasicLevel.DEBUG, "ReadableColumn: ");
		    for(int i=0; i<rgi.readableColumns.size(); i++) {
		        RdbGenColumn col = (RdbGenColumn) rgi.readableColumns.get(i);
		        if (col.joins == null) {
		            logger.log(BasicLevel.DEBUG, "\tCOL=" + col.columnName 
					        + " / TABLE=" + col.table.tableName
					        + " / COL_IDX=" + rgi.getColumnPosition(col)
					        + " rgc=" + col);
		        } else {
		            logger.log(BasicLevel.DEBUG, "\tCOL=" + col.columnName 
					        + " / TABLE=" + col.table.tableName
					        + " rgc=" + col);
		            for(int j=0; j<col.joins.size(); j++) {
		                RdbGenJoin join = (RdbGenJoin) col.joins.get(j);
		                logger.log(BasicLevel.DEBUG, "\t\tJOIN_IDX=" + join.getJoinIdx()
						        + " / COL_IDX=" + rgi.getColumnPosition(col, join));
		            }
		        }
		    }
			logger.log(BasicLevel.DEBUG, "End of createClassGenerationContext for class " + clazz.getFQName());
		}

        // filter
        rgi.filterExpr = rgi.getFilterExpression();
		return rgi;
	}
	// PRIVATE METHODS

	private class RGCComparator implements Comparator {
		public int compare(Object o1, Object o2) {
			RdbGenColumn rgc1 = (RdbGenColumn) o1;
			RdbGenColumn rgc2 = (RdbGenColumn) o2;
			return rgc1.columnName.compareTo(rgc2.columnName);
		}
	}

    /**
     * Adds all elements of @source to @dest excep those already in @dest.
     * Insertion begings at position @pos in dest
     * @param source the source list of element to insert
     * @param dest the list into wich inserting
     * @param addInHead if true add in head else add in tail
     */
	private void addColumnUnique(List source, List dest, boolean addInHead) {
        int pos = (addInHead) ? 0 : dest.size();
		for(int i=0; i<source.size(); i++) {
			boolean found = false;
			RdbGenColumn rgcToAdd = (RdbGenColumn) source.get(i);
			for(int j=0; j<dest.size() && !found; j++) {
				RdbGenColumn col = (RdbGenColumn) dest.get(j);
				found = col.table.tableName.equals(rgcToAdd.table.tableName)
					&& col.columnName.equals(rgcToAdd.columnName);
			}
			if (!found) {
				dest.add(pos++, rgcToAdd);
			}
		}
	}

	/**
	 * Defines the RdbGenDef associated either with the identifier of this class
	 * mapping, or with one of its references.
	 * @param nd    The NameDef associated with the identifier or a reference.
     * @param fn    ??
	 * @param cnid  The composite name identifier allocated for code generation
     * @param cm    ??
     * @param rgi   ??
	 *              of references.
	 * @return  The created RdbGenRef describing the identifier or a reference.
	 */
	private RdbGenRef defineGenRef(NameDef nd,
                                   String fn,
                                   int cnid,
								   RdbClassMapping cm,
								   RdbGenInfos rgi)
			throws PException {

		RdbGenRef rgr = new RdbGenRef();
		rgr.nd = nd;
		rgr.fieldName = fn;
		if (nd.isFieldName()) {
            // get the primitive element mapping of the field (searching in super classes)
            RdbPrimitiveElementMapping pem = (RdbPrimitiveElementMapping)
                    cm.getPrimitiveElementMapping(nd.getFieldName(), true);
			if (pem == null) {
				throw new PException(
						"Unable to find the mapping of the field "
						+ nd.getFieldName());
			}
			RdbTable table = (RdbTable) pem.getParent();
			if (debug) {
				logger.log(BasicLevel.DEBUG, "nd(" + fn
						+ ") FieldName, field=" + nd.getFieldName()
						+ " / column name=" + pem.getName()
						+ " in the table: " + table.getName());
			}
			RdbGenTable rgt = (RdbGenTable) rgi.tables.get(table.getName());
			if (rgt == null) {
				throw new NullPointerException("No table " + table.getName() + " found among: " + rgi.tables);
			}
			rgr.refColumn = rgt.getColumn(pem.getName());
			if (rgt.joins != null) {
				getJoin(rgt, pem, nd.getFieldName(), rgr);
			}
		} else { // this is a NameRef
			if (debug) {
				logger.log(BasicLevel.DEBUG, "nd(" + fn + ") CommpositeName, cnId=" + cnid);
			}
			rgr.cnId = cnid;
			rgr.cnName = nd.getNameRef().getCompositeName().getName();
			rgr.cnPackage = ((Package) nd.getNameRef().getCompositeName().getParent()).getName();
			rgr.cnFieldColumns = new TreeMap();
			Iterator itp = nd.getNameRef().getProjection().entrySet().iterator();
			RdbPrimitiveElementMapping pem = null;
			while (itp.hasNext()) {
				Map.Entry me = (Map.Entry) itp.next();
				String cnfield = (String) me.getKey();
				String clafn = (String) me.getValue();
                // get the primitive element mapping of the field (searching in super classes)
                pem = (RdbPrimitiveElementMapping)
                        cm.getPrimitiveElementMapping(clafn, true);
                if (pem == null) {
                    throw new PException(
                            "Unable to find the mapping of the field "
                            + nd.getFieldName());
                }
                RdbTable table = (RdbTable) pem.getParent();
                if (debug) {
                    logger.log(BasicLevel.DEBUG, "nd(" + fn
                            + ") composite field name=" + cnfield
                            + " /  class field name=" + clafn
                            + " / column name=" + pem.getName()
                            + " in the table: " + table.getName());
                }
                RdbGenTable rgt = (RdbGenTable) rgi.tables.get(table.getName());
                if (rgt == null) {
                    throw new PException("Cannot find gen table");
                }
                RdbGenColumn c = rgt.getColumn(pem.getName());
                rgr.cnFieldColumns.put(cnfield, c);
                if (rgr.join == null && rgt.joins != null) {
                    getJoin(rgt, pem, clafn, rgr);
                }
            }
		}
		return rgr;
	}

	private void getJoin(RdbGenTable rgt,
						 RdbPrimitiveElementMapping pem,
						 String fn,
						 RdbGenRef rgr) {
		if (debug) {
			logger.log(BasicLevel.DEBUG, "getJoin: fn: " + rgr.fieldName
					+ " / col: " + pem.getName()
					+ " / rgt: " + rgt.tableName);
		}
		if (rgt.joins.size() == 1) {
			rgr.join = (RdbGenJoin) rgt.joins.get(0);
			if (debug) {
				logger.log(BasicLevel.DEBUG, "return: fn: " + rgr.fieldName
						+ " / col: " + pem.getName()
						+ " / join: " + rgr.join.joinIdx);
			}
		}
		Iterator it = pem.getPrimitiveElementByRdbJoin().entrySet().iterator();
		while (it.hasNext()) {
			Map.Entry me2 = (Map.Entry) it.next();
			PrimitiveElement pe = (PrimitiveElement) me2.getValue();
			if (pe.getName().equals(fn)) {
				rgr.join = (RdbGenJoin)
						rgt.jn2join.get(((RdbJoin) me2.getKey()).getName());
				if (debug) {
					logger.log(BasicLevel.DEBUG, "return fn: " + rgr.fieldName
							+ " / col: " + pem.getName()
							+ " / joinName: " + ((RdbJoin) me2.getKey()).getName()
							+ " / join: " + rgr.join.joinIdx);
				}
			}
		}
	}

    /**
     * Gets the inherited data
     * @param cm the rdb class mapping of the class
     * @param inheritedColumns list in which to return the inherited
     * @param inheritedReadableColumns
     * @param inheritedGenRefs
     * @param inheritedExtTables
     * @param constantHolderClass
     * @throws PException
     */
	private void getInheritedData(RdbClassMapping cm,
								  List inheritedColumns,
								  List inheritedReadableColumns,
								  Map inheritedGenRefs,
								  Map inheritedExtTables,
								  Class constantHolderClass, 
								  Map ctx)
			throws PException {
        // loop on super classes
		for (Iterator pcmit = cm.getParentClassMappings().iterator();
			 pcmit.hasNext();) {
			ParentClassMapping pcm = (ParentClassMapping) pcmit.next();
            // only consider 'added' or 'extended' inheritance rules
			if (cm.inheritsStructures(pcm)) {
				Class superClass = pcm.getMOClass();
				if (debug) {
					logger.log(BasicLevel.DEBUG,
							"rule-name = " + pcm.getRuleName() + "=<" + superClass.getFQName() + ">"
							+ " abstract =<" + superClass.isAbstract() + ">");
				}

				RdbClassMapping superClassMapping = (RdbClassMapping)
						superClass.getClassMapping(
								cm.getProjectName(),
								cm.getMapperName());
				
                if (superClassMapping.getMainRdbTable() != null) {
                    RdbGenInfos superGenInfos =
                            getRdbGenInfos((Mapping) superClassMapping.getParent(), ctx);
                    addColumnUnique(superGenInfos.getReferenceTable().columns,
                            inheritedColumns, true);
                    addColumnUnique(superGenInfos.readableColumns,
                            inheritedReadableColumns, true);
                    inheritedGenRefs.putAll(superGenInfos.genRefs);
                    inheritedExtTables.putAll(superGenInfos.tables);
                }
               
                // recursively gets the infos inherited by the super class
				getInheritedData(superClassMapping,
						inheritedColumns,
						inheritedReadableColumns,
						inheritedGenRefs,
						inheritedExtTables,
						constantHolderClass,
						ctx);
			}
		}
	}

	private void printColumns(List columns) {
		RdbGenColumn rgc = null;
		for (int i = 0; i < columns.size(); i++) {
			rgc = (RdbGenColumn) columns.get(i);
			logger.log(BasicLevel.DEBUG, "\tcolumn : " + rgc.columnName
					+ " isHiddenField: " + rgc.isHiddenField()
					+ " isConstantField: " + (rgc.constant != null));
		}
	}
}
