/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.rdb.generator;

import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapter;

import java.util.List;

/**
 * This class contains all required information related to a column to which
 * some field of a JORM class is mapped.
 * @author  P. Dechamboux
 */
public class RdbGenColumn {
	/**
	 * The name of the field that is mapped to this column.
	 */
	public String fieldName = null;
	/**
	 * Tells is this field is an hidden one.
	 */
	public boolean hiddenField = true;

    /**
     * Contains the constant as a String in the case is this field is a constant
	 * field, otherwise (not a constant field) null.
     */
    public String constant;

	/**
	 * The name of the RDB column.
	 */
	public String columnName = null;
	/**
	 * The size of the RDB column in case of string, serialized,
	 * [byte|char]array...
	 */
	public int columnSize = RdbAdapter.NOSIZE;
	public int columnScale = RdbAdapter.NOSIZE;
	/**
	 * The JORM type associated with this mapped field.
	 */
	public PType columnType = null;
	/**
	 * Tells if this column may store null values.
	 */
	public boolean columnNotNull = false;
	/**
	 * Tells if this column belongs to the table primary key.
	 */
	public boolean columnInPK = false;
	/**
	 * The table descrption which this column belongs to.
	 */
	public RdbGenTable table = null;

	/**
	 * the sql type of the column. A null value means to use the value of the
	 * choosen by the adapter.
	 */
	public String columnSqlType = null;

	/**
	 * references a column in the main table in case of this column participates
	 * in the join with the main table. In other case this field is null.
	 */
	public RdbGenColumn joinCol = null;

	/**
	 * contains the association between a RdbGenJoin and a primitive element.
	 * This field is null is the current column is a column of the main table,
	 * because no join is needed of course.
	 */
	public List joins = null;
	public List pes = null;

	public RdbGenColumn getJoinCol() {
		return joinCol;
	}

	public String getColumnSqlType() {
		return columnSqlType;
	}

	public String getFieldName() {
		return fieldName;
	}

	public String getFieldName(RdbGenJoin rgj) throws PException {
		if (rgj != null && joins != null) {
			int idx = joins.indexOf(rgj);
			if (idx == -1) {
				throw new PException("No field found for the column "
									 + columnName + " with the join " + rgj);
			}
			return ((PrimitiveElement) pes.get(idx)).getName();
		}
		return fieldName;
	}

	public String getConstant() {
		return constant;
	}

	public List getJoins() {
		return joins;
	}

	public boolean isHiddenField() {
		return hiddenField;
	}

	public String getColumnName() {
		return columnName;
	}

	public PType getColumnType() {
		return columnType;
	}

	public boolean isColumnNotNull() {
		return columnNotNull;
	}

	public boolean isColumnInPK() {
		return columnInPK;
	}

	public RdbGenTable getTable() {
		return table;
	}

	public String getColumnNotNull() {
		return "" + columnNotNull;
	}

	public String getColumnInPK() {
		return "" + columnInPK();
	}

	public boolean columnInPK() {
        if (table.isMainTable) {
			return table.genInfos.colInGenId(this);
		} else if (table.colocatedTable) {
			return false;
		} else {
			return this.joinCol != null;
		}
	}

	/**
	 * Gets the SQL type associated with this column into the database for
	 * which we define the mapping.
	 * @return  The string containing the name of the relevant SQL type.
	 * @throws PException
	 */
	public String getSqlType(String adapter) throws PException {
		if (columnSqlType != null) {
			return "\"" + columnSqlType + "\"";
		} else {
			StringBuffer sb = new StringBuffer(adapter);
			sb.append(".getSqlType(");
			sb.append(columnType.getProgName());
			sb.append(".getTypeCode(), ");
			sb.append(columnInPK());
			sb.append(", ");
			sb.append(columnSize);
			sb.append(", ");
			sb.append(columnScale);
			sb.append(")");
			return sb.toString();
		}
	}

	/**
	 * Constructs the string that defines the way to access to the relevant
	 * result part into a particular ResultSet.
	 * @param adapter The name of the RdbAdaptervariable used to set the
	 *              given query parameter.
	 * @param resultset The name of the ResultSet variable used to get the query
	 *                  result.
	 * @param pos       The position of the result to get.
	 * @return  The string defining the access to the relevant result part into
	 *          a particular ResultSet.
	 * @throws PException
	 *
	 * example: "adapter.getBoolean(pstmt, 4);
	 * example: "adapter.getDate(pstmt, 4, "TIME");
	 */
	public String getSqlGet(String adapter,
							String resultset,
							String pos,
							String nullValue) throws PException {
		StringBuffer sb = new StringBuffer(adapter);
		sb.append(".");
		sb.append(getAdapterGet(columnType));
		sb.append("(");
		sb.append(resultset);
		sb.append(", ");
		sb.append(pos);
		switch (columnType.getTypeCode()) {
		case PType.TYPECODE_DATE:
			sb.append(", ");
			if (columnSqlType != null && columnSqlType.length() > 0) {
				sb.append("\"");
			}
			sb.append(columnSqlType);
			if (columnSqlType != null && columnSqlType.length() > 0) {
				sb.append("\"");
			}
			break;
		}
		sb.append(", ");
		if (nullValue != null && nullValue.length()>0) {
			sb.append(nullValue);
		} else {
			sb.append(getNullValue());
		}
		sb.append(")");
		return sb.toString();
	}

	public String getNullValue() {
		switch (columnType.getTypeCode()) {
		case PType.TYPECODE_BOOLEAN:
			return "false";
		case PType.TYPECODE_CHAR:
			return "((" + columnType.getJavaName() + ") 0)";
		case PType.TYPECODE_BYTE:
		case PType.TYPECODE_SHORT:
		case PType.TYPECODE_INT:
		case PType.TYPECODE_LONG:
		case PType.TYPECODE_FLOAT:
		case PType.TYPECODE_DOUBLE:
			return "((" + columnType.getJavaName() + ") -1)";
		default:
			return "null";
		}
	}


	/**
	 * Constructs the string that defines the way to assign to the relevant
	 * query parameter into a particular prepared statement through the adapter.
	 * @param adapter The name of the RdbAdaptervariable used to set the
	 *              given query parameter.
	 * @param pstmt The name of the PreparedStatement variable used to set the
	 *              given query parameter.
	 * @param pos   The position of the query parameter to be assigned.
	 * @param value is the String value which must be assigned into the
	 * 			prepared statment.
	 * @return  The string defining the assignement of the relevant query
	 *          parameter into a particular PreparedStatement throught the
	 * 			rdb adapter.
	 * @throws PException
	 *
	 * example: "adapter.setBoolean(pstmt, 4, true);
	 * example: "adapter.setDate(pstmt, 4, ((java.util.Date) fieldVal), "TIME");
	 */
	public String getSqlSet(String adapter,
							String pstmt,
							String value,
							int pos) throws PException {
		StringBuffer sb = new StringBuffer(adapter);
		sb.append(".");
		sb.append(getAdapterSet(columnType));
		sb.append("(");
		sb.append(pstmt);
		sb.append(", ");
		sb.append(pos);
		sb.append(", ");
		sb.append(value);
		switch (columnType.getTypeCode()) {
		case PType.TYPECODE_DATE:
			sb.append(", ");
			if (columnSqlType != null && columnSqlType.length() > 0) {
				sb.append("\"");
			}
			sb.append(columnSqlType);
			if (columnSqlType != null && columnSqlType.length() > 0) {
				sb.append("\"");
			}
			break;
		}
		sb.append(")");
		return sb.toString();
	}

	public String getAdapterGet(PType type) {
		return "get" + getAdapterSuffix(type);
	}
	public String getAdapterSet(PType type) {
		return "set" + getAdapterSuffix(type);
	}

	private String getAdapterSuffix(PType type) {
		switch (type.getTypeCode()) {
		case PType.TYPECODE_BYTEARRAY:
			return "ByteArray";
		case PType.TYPECODE_BOOLEAN:
			return "Boolean";
		case PType.TYPECODE_OBJBOOLEAN:
			return "Oboolean";
		case PType.TYPECODE_FLOAT:
			return "Float";
		case PType.TYPECODE_OBJFLOAT:
			return "Ofloat";
		case PType.TYPECODE_DOUBLE:
			return "Double";
		case PType.TYPECODE_OBJDOUBLE:
			return "Odouble";
		case PType.TYPECODE_SERIALIZED:
			return "Serialized";
		default:
			return type.getCodingName();
		}
	}

	public String getSqlTypeCode(PType pt) throws PException {
		StringBuffer sb = new StringBuffer("adapter.getSqlTypeCode(");
		sb.append(pt.getProgName());
		sb.append(".getTypeCode(), ");
		if (columnSqlType != null && columnSqlType.length() > 0) {
			sb.append("\"");
		}
		sb.append(columnSqlType);
		if (columnSqlType != null && columnSqlType.length() > 0) {
			sb.append("\"");
		}
		sb.append(")");
		return sb.toString();
	}

	public String getAccessorValue(RdbGenInfos genInfos,
								   boolean isSpecific) throws Exception {
		return getAccessorValue(genInfos, isSpecific, null);
	}

	public String getAccessorValue(RdbGenInfos genInfos,
								   boolean isSpecific, RdbGenJoin rgj) throws Exception {

		if (hiddenField) {
			if (genInfos.colInGenId(this)) {
				throw new PException("unmanaged case");
			} else {
				RdbGenRef gen_ref = genInfos.getGenRefOfColumn(this, rgj);
				if (isSpecific) {
					return "paccessor.paGet"
							+ RdbGenInfos.commonHelper.upperFL(gen_ref.fieldName)
							+ "(conn)";
				} else {
					return "paccessorGen.paGetRefField("
							+ gen_ref.fieldName + "_NAME, conn)";
				}
			}
		} else {
			if (isSpecific) {
				return "paccessor.paGet"
						+ RdbGenInfos.commonHelper.upperFL(getFieldName(rgj))
						+ "()";
			} else {
				return "paccessorGen." +
						RdbGenInfos.commonHelper.getPaGetMethod(columnType)
						+ "(" + getFieldName(rgj) + "_NAME)";
			}
		}
	}

	public String getRefValue(RdbGenInfos genInfos,
							  boolean isSpecific) throws Exception {
		return getRefValue(genInfos, isSpecific, null);
	}

	public String getRefValue(RdbGenInfos genInfos,
							  boolean isSpecific, RdbGenJoin rgj) throws Exception {
		try {
			if (!hiddenField) {
                //A visible field is fetch through the PAccessor
				throw new PException("unmanaged case");
			}
			RdbGenRef gen_ref = null;
			if (genInfos.colInGenId(this)) {
                //The column is included into the identifier
				gen_ref = genInfos.genId;
				if (gen_ref.refColumn != null) {
                    //the identifier is composed by a single column
					return "pname." + RdbGenInfos.commonHelper
						.getPNameEncodeFunction(columnType) + "()";
				} else {
                    // composite identifier, the PNameGetter of the identifier is computed before
                    //TODO: pass as parameter, the variable name of the PNameGetter 
					return "_pngId.pnGet" +
							RdbGenInfos.commonHelper.upperFL(gen_ref.getFieldName(this))
							+ "(null)";
				}
			} else {
                //The column is not part of the identifier (however it is a hidden field)
                // get the corresponding reference field of the column (there is obligatory at least one)
				gen_ref = genInfos.getGenRefOfColumn(this, rgj);
   				if (gen_ref == null) {
                    System.out.println("Pb in getRefValue of column " + getColumnName());
					return "JORM_GENERATION_ERROR";
				}
                // check if the name def of the ref is a composite ref 
				if (gen_ref.refColumn != null) {
                    // not composite, compute the name of the PNC variable name corresponding 
                    // to the reference field, then call the encode method corresp?nding to 
                    // the type of the column, pass in parameter the Pname provided by the PAccessor
					return gen_ref.fieldName + "Pnc." + RdbGenInfos.commonHelper
							.getPNameEncodeFunction(columnType)
							+ "(" + getAccessorValue(genInfos, isSpecific, rgj) + ")";
				} else {
                    // composite case : compute the name of the variable correponding to the pname 
                    // getter of the reference, call the pnGetXXX of the right type, pass null in parameter                     
					return "png" + gen_ref.cnId + ".pnGet"
							+ RdbGenInfos.commonHelper.upperFL(
									gen_ref.getFieldName(this))
							+ "(null)";
				}

			}
		} catch (Exception e) {
			e.printStackTrace();
			System.exit(-1);
			throw e;
		}
	}

	public String getColumnName(String tableSuffix) {
		return table.tableName
				+ (tableSuffix == null ? "" : tableSuffix)
				+ "." + columnName;
	}

	public boolean isInJoin(RdbGenJoin rgj) {
		return (joins != null && joins.contains(rgj))
				|| rgj.getJoinColumnsInExt().contains(this);

	}
	
	public boolean equals(RdbGenColumn rgc) {
		if (rgc == null)
			return false;
		return ( (columnName == null ? rgc.columnName == null : columnName.equals(rgc.columnName))
				&& (table == null ? rgc.table == null : table.tableName.equals(rgc.table.tableName))
				&& (columnSqlType == null ? rgc.columnSqlType == null : columnSqlType.equals(rgc.columnSqlType))
				&& (columnType == rgc.columnType)
				&& (fieldName == null ? rgc.fieldName == null : fieldName.equals(rgc.fieldName))
				&& (hiddenField == rgc.hiddenField)
				&& (joinCol == null ? rgc.joinCol == null : joinCol.equals(rgc.joinCol))
				&& (columnInPK == rgc.columnInPK)
				&& (columnNotNull == rgc.columnNotNull)
				&& (columnScale == rgc.columnScale)
				&& (constant == null ? rgc.constant == null : constant.equals(rgc.constant)));
	}
}
