/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2004 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.rdb.generator;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.generator.lib.CommonHelper;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbClassMapping;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbFilter;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbMapping;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbPrimitiveElementMapping;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ClassMapping;
import org.objectweb.jorm.metainfo.api.ParentClassMapping;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.naming.api.PNameGetter;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.api.VariableOperand;
import org.objectweb.medor.expression.converter.rdb.Expression2WhereClauseImpl;
import org.objectweb.medor.expression.lib.ExpressionPrinter;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * This is the root class that contains all the required information for
 * generating Java code specific to the RDB mapping. It allows the support of
 * all elaborated mapping rules.
 * @author  P. Dechamboux, O. Lobry
 */
public class RdbGenInfos {
	private static final String COLUMN = ", ";
	/**
	 * The set of names all tables used to store the generated class. It also contains
	 * the reference table that is supposed to hold the object identifier (see
	 * mainTable). Indexed by their names.  < String -> RdbGenTable >
	 */
	public Map tables = new HashMap();
	/**
	 * The reference table used in this mapping. It entirely stores the
	 * identifier of the related JORM object, as specified by variable genId.
	 */
	public RdbGenTable mainTable = null;
	/**
	 * The set of all columns used to store the generated class. It defines the
	 * order with which the columns will be used within all kinds of queries.
	 * This set is the union of all columns specified with each table involved
	 * in this mapping.
	 */
	public List readableColumns = new ArrayList();
	/**
	 * The structure of the identifier of the related JORM object.
	 */
	public RdbGenRef genId = null;
	/**
	 * The associated Mapping
	 */
	public RdbMapping mapping = null;
	/**
	 * The set of references defined by the generated class.
	 * There is one RdbGenRef object registered for each reference field,
	 * in association with its field name into this map.
	 */
	public Map genRefs = new HashMap();


    String filterExpr = null;

	public Logger logger;
	public boolean debug = false;

	public static CommonHelper commonHelper = new CommonHelper();


	public RdbGenInfos() {
	}

	public Map getTables() {
		return tables;
	}

	public RdbGenTable getReferenceTable() {
		return mainTable;
	}

	public List getReadableColumns() {
		return readableColumns;
	}

	public RdbGenRef getGenId() {
		return genId;
	}

	public Map getGenRefs() {
		return genRefs;
	}

	/**
	 * Computes the list of references (i.e., their definition) stored into this
	 * particular table.
	 * @param rgt   The table for which to find the reference definitions.
	 * @return  The computed list of definitions.
	 */
	public List getTableCompositeGenRefs(RdbGenTable rgt) {
		return getTableCompositeGenRefs(rgt, null);
	}

	public List getTableCompositeGenRefs(RdbGenTable rgt, RdbGenJoin rgj) {
		ArrayList res = new ArrayList();
		Iterator it = genRefs.values().iterator();
		while (it.hasNext()) {
			RdbGenRef rgr = (RdbGenRef) it.next();
			if (rgr.refColumn == null && rgr.join == rgj && !res.contains(rgr)) {
				res.add(rgr);
			}
		}
		return res;
	}

	/**
	 * Gets the position (within [1..n] of the given column into the overall
	 * list defined into this RdbGenInfos.
	 * @param irgc  The column for which to search the position.
	 * @return  Its position into the genColumns list. -1 if not found (BUG:
	 *          should never appear).
	 */
	public int getColumnPosition(RdbGenColumn irgc) {
		return getColumnPosition(irgc, null);
	}
	
	public int getColumnPosition(RdbGenColumn irgc, RdbGenJoin rgj) {
	    int idx = 1;
	    if (debug) {
	        logger.log(BasicLevel.DEBUG, "getColumnPosition(" + irgc.columnName + ", rgj=" + rgj +")");
	    }
		for (int i = 0; i < readableColumns.size(); i++) {
			RdbGenColumn rgc = (RdbGenColumn) readableColumns.get(i);
		    if (debug) {
		        logger.log(BasicLevel.DEBUG, "\tcurrent column: " + rgc.columnName);
		        logger.log(BasicLevel.DEBUG, "\tpirgc=" + irgc.toString() + " - current=" + rgc.toString());
		    }
			if (rgc.equals(irgc)) {
		        if (rgj != null) {
				    if (debug) {
				        logger.log(BasicLevel.DEBUG, "\t\tadd join idx: " + rgj.joinIdx);
				    }
			        idx += rgj.joinIdx;
			    }
			    if (debug) {
			        logger.log(BasicLevel.DEBUG, "\treturn " + idx);
			    }
				return idx;
			} else if (rgc.joins != null) {
			    if (debug) {
			        logger.log(BasicLevel.DEBUG, "\t\tcurrent has joins: " + rgc.joins.size());
			    }
				idx += rgc.joins.size();
			} else {
			    if (debug) {
			        logger.log(BasicLevel.DEBUG, "\t\tidx++");
			    }
				idx++;
			}
		    if (debug) {
		        logger.log(BasicLevel.DEBUG, "\t\tidx=" + idx);
		    }
		}
		return -1;
	}

	/**
	 * Constructs the string representing the SQL query used for testing the
	 * existence of a JORM object.
	 * @return  The query string.
	 */
	public String getExistStatement() {
		StringBuffer sb = new StringBuffer("SELECT ");
		String sep = "";
		Iterator it;
		String tabname = null;
		if (genId.refColumn != null) {
			sb.append(sep);
			appendFQColName(sb, genId.refColumn);
			tabname = genId.refColumn.table.tableName;
		} else /* genCNId must not be null */ {
			it = genId.cnFieldColumns.keySet().iterator();
			while (it.hasNext()) {
				String cnfn = (String) it.next();
				RdbGenColumn rgc = (RdbGenColumn) genId.cnFieldColumns.get(cnfn);
				sb.append(sep);
				sep = COLUMN;
				appendFQColName(sb, rgc);
				if (tabname == null) {
					tabname = rgc.table.tableName;
				}
			}
		}
		sb.append(" FROM ");
		sb.append(tabname);
		sb.append(" WHERE ");
		sep = "";
		if (genId.refColumn != null) {
			sb.append(sep);
			appendFQColName(sb, genId.refColumn);
			sb.append(" = ?");
		} else /* genCNId must not be null */ {
			it = genId.cnFieldColumns.keySet().iterator();
			while (it.hasNext()) {
				String cnfn = (String) it.next();
				RdbGenColumn rgc = (RdbGenColumn) genId.cnFieldColumns.get(cnfn);
				sb.append(sep);
				sep = " AND ";
				appendFQColName(sb, rgc);
				sb.append(" = ?");
			}
		}
		return sb.toString();
	}
    public String getReadSelectFields() {
        return getReadSelectFields(false);
    }

	/**
	 * Build the select clause that access all fields of the class.
	 */
	public String getReadSelectFields(boolean aliasingWithPE) {
		StringBuffer select = new StringBuffer();
		String sep = "";
		for (int i = 0; i < readableColumns.size(); i++) {
			RdbGenColumn rgc = (RdbGenColumn) readableColumns.get(i);
			if (rgc.joins != null && rgc.joins.size() > 1) {
				String[] s = new String[rgc.joins.size()];
				for (int j = 0; j < rgc.joins.size(); j++) {
					RdbGenJoin rgj = (RdbGenJoin) rgc.joins.get(j);
					s[rgj.joinIdx] = getFQColName(rgc, "_J" + rgj.joinIdx);
				}
				for (int j = 0; j < s.length; j++) {
					select.append(sep);
					sep = COLUMN;
					select.append(s[j]);
                    if (aliasingWithPE) {
                        select.append("\"+ adapter.getColumnAliasExpr(\"");
                        select.append(((PrimitiveElement) rgc.pes.get(i)).getName());
                        select.append("\") + \"");
                    }
				}
			} else if (rgc.constant == null) { //excludes constant fields
				select.append(sep);
				sep = COLUMN;
				appendFQColName(select, rgc);
                if (aliasingWithPE) {
                    select.append("\"+ adapter.getColumnAliasExpr(\"");
                    select.append(rgc.fieldName);
                    select.append("\") + \"");
                }
			}
		}
		return select.toString();
	}

	public String getReadSelectFieldsAsColArray() {
		StringBuffer select = new StringBuffer("new String[]{");
		String sep = "";
		for (int i = 0; i < readableColumns.size(); i++) {
			RdbGenColumn rgc = (RdbGenColumn) readableColumns.get(i);
			if (rgc.joins != null && rgc.joins.size() > 1) {
				String[] s = new String[rgc.joins.size()];
				for (int j = 0; j < rgc.joins.size(); j++) {
					RdbGenJoin rgj = (RdbGenJoin) rgc.joins.get(j);
					s[rgj.joinIdx] = getFQColName(rgc, "_J" + rgj.joinIdx);
				}
				for (int j = 0; j < s.length; j++) {
					select.append(sep);
					sep = COLUMN;
					select.append("\"").append(s[j]).append("\"");
				}
			} else if (rgc.constant == null) { //excludes constant fields
				select.append(sep);
				sep = COLUMN;
				select.append("\"");
				appendFQColName(select, rgc);
				select.append("\"");
			}
		}
		select.append("}");
		return select.toString();
	}

	
	
	/**
	 * @return a list of RdbGenColumn corresponding to constant fields
	 */
	public List getConstantVisibleColumns() {
		List constantCols = new ArrayList();
		for (int i = 0; i < readableColumns.size(); i++) {
			RdbGenColumn rgc = (RdbGenColumn) readableColumns.get(i);
			if (rgc.constant != null && !rgc.hiddenField)  {
				constantCols.add(rgc);
			}
		}
		return constantCols;
	}

	/**
	 * @return the number of fields accessed by the select clause of the class.
	 */
	public int getReadSelectNbFields() {
		int res = 0;
		for (int i = 0; i < readableColumns.size(); i++) {
			RdbGenColumn rgc = (RdbGenColumn) readableColumns.get(i);
			if (rgc.joins != null && rgc.joins.size() > 1) {
				String[] s = new String[rgc.joins.size()];
				for (int j = 0; j < rgc.joins.size(); j++) {
					RdbGenJoin rgj = (RdbGenJoin) rgc.joins.get(j);
					s[rgj.joinIdx] = getFQColName(rgc, "_J" + rgj.joinIdx);
				}
				for (int j = 0; j < s.length; j++) {
					res++;
				}
			} else if (rgc.constant == null) { //excludes constant fields
				res++;
			}
		}
		return res;
	}

	/**
	 * Build the data structure representing the from clause and the joins for
	 * accessing all fields of the class.
	 * @return
	 */
	public List getReadFromTables() {
		ArrayList al = new ArrayList();
		al.add("JoinedTable jt = new JoinedTable(\""
				+ mainTable.tableName + "\");");
		al.add("JoinedTable.Join jtj = null;");
		Iterator it = tables.values().iterator();
		while (it.hasNext()) {
			RdbGenTable rgt = (RdbGenTable) it.next();
			if (rgt.joins != null) {
				boolean multiJoin = rgt.joins.size() > 1;
				for (int i = 0; i < rgt.joins.size(); i++) {
					RdbGenJoin join = (RdbGenJoin) rgt.joins.get(i);
					String tableSuffix = null;
					if (multiJoin) {
						tableSuffix = "_J" + ((RdbGenJoin) rgt.joins.get(i)).joinIdx;
						al.add("jtj = jt.createChildren(\""
								+ rgt.tableName + "\", \""
								+ rgt.tableName + tableSuffix + "\");");
					} else {
						al.add("jtj = jt.createChildren(\""
								+ rgt.tableName + "\");");
					}
					for (int j = 0; j < join.joinColumnsInMain.size(); j++) {
						al.add("jtj.addJoinColumn(\""
								+ ((RdbGenColumn) join.joinColumnsInMain.get(j)).columnName
								+ "\", \""
								+ ((RdbGenColumn) join.joinColumnsInExt.get(j)).columnName
								+ "\");");
					}
				}
			}
		}
		return al;
	}


	public String getReadWhereClause() {
		// build the where clause with only the identifier matching
		StringBuffer where = new StringBuffer();
		String sep = "";
		// apply the identifier matching
		if (genId.refColumn != null) {
			where.append(sep);
			appendFQColName(where, genId.refColumn);
			where.append(" = ?");
		} else {
			Iterator it = genId.cnFieldColumns.keySet().iterator();
			while (it.hasNext()) {
				String cnfn = (String) it.next();
				RdbGenColumn rgc = (RdbGenColumn) genId.cnFieldColumns.get(cnfn);
				where.append(sep);
				sep = " AND ";
				appendFQColName(where, rgc);
				where.append(" = ?");
			}
		}
		return where.toString();
	}

	/**
	 * Constructs the string representing the SQL query used to get the extent
	 * of a JORM class (the ID PName of all objects of this class).
	 * @return  The query string.
	 */
	public String getPNameIteratorSelectClause() {
	    try {
		StringBuffer select = new StringBuffer();
		if (genId.getCnFieldColumns() == null) {
			appendFQColName(select, genId.getRefColumn());
		} else {
			String sep = "";
			Iterator it = genId.getCnFieldColumns().keySet().iterator();
			while (it.hasNext()) {
				RdbGenColumn rgc = (RdbGenColumn) genId.getCnFieldColumns().get(it.next());
				if (!colInGenId(rgc)) {
				continue;
				}
				select.append(sep);
				sep = COLUMN;
				appendFQColName(select, rgc);
			}
		}
		return select.toString();
	    } catch(RuntimeException e) {
	        e.printStackTrace();
	        throw e;
	    }
	}


	public String getPNameIteratorSelectClauseWithPrefetch() {
		StringBuffer select = new StringBuffer();
		// build the select clause
		select.append(getReadSelectFields(true));
		if (genId.getCnFieldColumns() == null) {
			select.append(COLUMN);
			appendFQColName(select, genId.getRefColumn());
			select.append("\" + adapter.getColumnAliasExpr(\"pk0\") + \"");
		} else {
			Iterator it = genId.getCnFieldColumns().keySet().iterator();
			int i = 0;
			while (it.hasNext()) {
				RdbGenColumn rgc = (RdbGenColumn) genId.getCnFieldColumns().get(it.next());
				if (!colInGenId(rgc)) {
					continue;
				}
				select.append(COLUMN);
				appendFQColName(select, rgc);
				select.append("\" + adapter.getColumnAliasExpr(\"pk");
				select.append(i);
				select.append("\") + \"");
			}
		}
		return select.toString();
	}


	/**
	 * Return the filter as a medor expression.
	 */
	public Expression getFilterMedorExpression() {
		Expression exp = null;
        try {
            // check if there is an explicit mapping filter
               RdbFilter rdbFilter = ((RdbClassMapping) mapping.getClassMapping())
                    .getRdbFilter();
            if (rdbFilter != null) { //there is an explicit mapping filter
                exp = rdbFilter.getExpression();
                if (debug) {
                    logger.log(BasicLevel.DEBUG, "Rdb filter specified for the class "
                        + mapping.getClassMapping().getJormClass().getFQName()
                        + ", expression=" + ExpressionPrinter.e2str(exp));
                }
            } else {
                // try to compute it from the naming filter (only in case of
                // extended or added inheritance rules)
                if (((RdbClassMapping) mapping.getClassMapping()).inheritsStructures()
                        || !mapping.getClassMapping().getJormClass().getSubClasses().isEmpty()) {
                    // if there is no explicit filter but the
                    exp = ((RdbClassMapping) mapping.getClassMapping())
                            .getMappingFilterFromNamingFilter();
                    if (debug) {
                        logger.log(BasicLevel.DEBUG, "Class "
                                + mapping.getClassMapping().getJormClass().getFQName()
                                + " has a filtered mapping and the filter is computed from the naming filter:" 
                                +  ExpressionPrinter.e2str(exp));
                    }
                } else if (debug) {
                    logger.log(BasicLevel.DEBUG, "Class "
                            + mapping.getClassMapping().getJormClass().getFQName()
                            + " has no filter");
                }
            }
            return exp;
        } catch (ExpressionException e) {
            if (logger != null) {
                logger.log(BasicLevel.ERROR, "Cannot compute filter expression: " 
                    + e.getMessage(), e);
            } else {
                e.printStackTrace(System.err);
            }
            throw new InternalError("Cannot compute filter expression: " + e);
        }
	}
	
	/**
	 * Return the filter as a string. It is a where clause. 
	 */
    public String getFilterExpression() {
        if (filterExpr != null) {
            return filterExpr;
        }
        try {
        	Expression exp = getFilterMedorExpression();
            if (exp == null) {
                filterExpr = "";
            } else {
                StringBuffer sb = new StringBuffer();
                new JormExpression2WhereClause()
                        .convertExp2WhereClauseBuilder(exp, "adapter", sb, new HashMap());
                filterExpr = sb.toString();
            }
            return filterExpr;
        } catch (ExpressionException e) {
            throw new InternalError("Cannot compute filter expression: " + e);
        }
    }
    
    String pnameIteratorWhereClause = null;

	public String getPNameIteratorWhereClause() throws ExpressionException {
		if (pnameIteratorWhereClause == null) {
            pnameIteratorWhereClause = getFilterExpression();
		}
		if (pnameIteratorWhereClause != null
			&& pnameIteratorWhereClause.length() > 0) {
			return pnameIteratorWhereClause;
		} else {
			return null;
		}
	}




	/**
	 * Constructs the string representing the SQL query used to update the part
	 * of a JORM object stored into the given table.
	 * @param rgt   The description of the table to be modified.
	 * @return  The query string.
	 */
	public String getUpdateStatement(RdbGenTable rgt) throws PException {
		StringBuffer sb = new StringBuffer("UPDATE ");
		sb.append(rgt.tableName);
		sb.append(" SET ");
		String sep = "";
		for (int i = 0; i < rgt.inheritedColumns.size(); i++) {
			if (updatable((RdbGenColumn) rgt.inheritedColumns.get(i))) {
				sb.append(sep);
				sep = COLUMN;
				sb.append(((RdbGenColumn) rgt.inheritedColumns.get(i)).columnName);
				sb.append(" = ?");
			}
		}
		for (int i = 0; i < rgt.columns.size(); i++) {
			if (updatable((RdbGenColumn) rgt.columns.get(i))) {
				sb.append(sep);
				sep = COLUMN;
				sb.append(((RdbGenColumn) rgt.columns.get(i)).columnName);
				sb.append(" = ?");
			}
		}
		appendModificationWhereClause(sb, rgt);
		return sb.toString();
	}

	public String getUpdateStatement(RdbGenJoin rgj) throws PException {
		StringBuffer sb = new StringBuffer("UPDATE ");
		sb.append(rgj.table.tableName);
		sb.append(" SET ");
		String sep = "";
		if (rgj.table.colocatedTable) {
			for (int i = 0; i < rgj.joinColumnsInExt.size(); i++) {
				sb.append(sep);
				sep = COLUMN;
				sb.append(((RdbGenColumn) rgj.joinColumnsInExt.get(i)).columnName);
				sb.append(" = ?");
			}

		} else {
			for (int i = 0; i < rgj.table.columns.size(); i++) {
				RdbGenColumn rgc = (RdbGenColumn) rgj.table.columns.get(i);
				if (rgc.joinCol == null) {
					sb.append(sep);
					sep = COLUMN;
					sb.append(rgc.columnName);
					sb.append(" = ?");
				}
			}
		}
		appendModificationWhereClause(sb, rgj, rgj.table.colocatedTable);
		return sb.toString();
	}

	public String getUpdateNullStatement(RdbGenTable rgt) throws PException {
		if (rgt.joins != null && rgt.joins.size() > 0) {
			throw new PException("umanaged case, must use " +
								 "getUpdateNullStatement(RdbGenJoin) for each join.");
		}
		StringBuffer sb = new StringBuffer("UPDATE ");
		sb.append(rgt.tableName);
		sb.append(" SET ");
		String sep = "";
		for (int i = 0; i < rgt.columns.size(); i++) {
			RdbGenColumn rgc = (RdbGenColumn) rgt.columns.get(i);
			if (updatable(rgc)) {
				if (rgt.colocatedColumn(rgc)) {
					continue;
				}
				sb.append(sep);
				sep = COLUMN;
				sb.append(rgc.columnName);
				sb.append(" = null");
			}
		}
		appendModificationWhereClause(sb, rgt);
		return sb.toString();
	}

	public String getUpdateNullStatement(RdbGenJoin rgj) throws PException {
		StringBuffer sb = new StringBuffer("UPDATE ");
		sb.append(rgj.table.tableName);
		sb.append(" SET ");
		String sep = "";
		for (int i = 0; i < rgj.joinColumnsInExt.size(); i++) {
			sb.append(sep);
			sep = COLUMN;
			sb.append(((RdbGenColumn) rgj.joinColumnsInExt.get(i)).columnName);
			sb.append(" = null");
		}
		appendModificationWhereClause(sb, rgj, false);
		return sb.toString();
	}

	/**
	 * Constructs the string representing the SQL query used to insert the part
	 * of a JORM object stored into the given table.
	 * @param rgt   The description of the table to be modified.
	 * @return  The query string.
	 */
	public String getInsertStatement(RdbGenTable rgt) throws PException {
		String sep = "";
		StringBuffer sb = new StringBuffer("INSERT INTO ");
		sb.append(rgt.tableName);
		sb.append(" (");
		ArrayList columns = new ArrayList();
		columns.addAll(rgt.inheritedColumns);
		columns.addAll(rgt.columns);
		for (int i = 0; i < columns.size(); i++) {
			sb.append(sep);
			sep = COLUMN;
			sb.append(((RdbGenColumn) columns.get(i)).columnName);
		}
		sb.append(") VALUES (");
		sep = "";
		for (int i = 0; i < columns.size(); i++) {
			sb.append(sep);
			sep = COLUMN;
			sb.append("?");
		}
		sb.append(")");
		return sb.toString();
	}

	public String getInsertStatement(RdbGenJoin rgj) {
		RdbGenTable rgt = rgj.getTable();
		int i = 0;
		try {
			String sep = "";
			StringBuffer sb = new StringBuffer("INSERT INTO ");
			sb.append(rgt.tableName);
			sb.append(" (");
			for (i = 0; i < rgt.columns.size(); i++) {
				RdbGenColumn col = (RdbGenColumn) rgt.columns.get(i);
				if (col.isInJoin(rgj)) {
					sb.append(sep);
					sep = COLUMN;
					sb.append(((RdbGenColumn) rgt.columns.get(i)).columnName);
				}
			}
			sb.append(") VALUES (");
			sep = "";
			for (i = 0; i < rgt.columns.size(); i++) {
				RdbGenColumn col = (RdbGenColumn) rgt.columns.get(i);
				if (col.isInJoin(rgj)) {
					sb.append(sep);
					sep = COLUMN;
					sb.append("?");
				}
			}
			sb.append(")");
			return sb.toString();
		} catch (RuntimeException e) {
			System.out.println(rgt.tableName);
			System.out.println(rgt.columns.get(i));
			System.out.println(((RdbGenColumn) rgt.columns.get(i)).joins);
			e.printStackTrace();
			throw e;
		}
	}

	/**
	 * Constructs the string representing the SQL query used to delete the part
	 * of a JORM object stored into the given table.
	 * @param rgt   The description of the table to be modified.
	 * @return  The query string.
	 */
	public String getDeleteStatement(RdbGenTable rgt) throws PException {
		StringBuffer sb = new StringBuffer("DELETE FROM ");
		sb.append(rgt.tableName);
		appendModificationWhereClause(sb, rgt);
		return sb.toString();
	}

	public String getDeleteStatement(RdbGenJoin rgj) throws PException {
		return getDeleteStatement(rgj.getTable());
	}

	/**
	 * Constructs the string representing the SQL query used to delete the part
	 * of a JORM object stored into the given table.
	 * @param rgj   The description of the join between the external table and
	 * the main table
	 * @return  The query string.
	 */
	public String getExtDeleteStatement(RdbGenJoin rgj) throws PException {
		StringBuffer sb = new StringBuffer("DELETE FROM ");
		sb.append(rgj.table.tableName);
		appendModificationWhereClause(sb, rgj, rgj.table.colocatedTable);
		return sb.toString();
	}

	public String setIdPnParam(String cnfn,
							   int idx,
							   String adapter,
							   String pstmt,
							   boolean isSpecific) throws PException {
		RdbGenColumn rgc = (RdbGenColumn) genId.cnFieldColumns.get(cnfn);
		return rgc.getSqlSet(adapter, pstmt, setIdPNParameterOperand( cnfn, isSpecific), idx);
	}

	/**
	 * If isSpecific is true returns a String like this:
	 * ((org.objectweb.speedo.pobjects.detach.VehicleIdPNG) _pngId).pnGetName(null)
	 * else, it returns a String like that:
	 * ((PNameGetter) _pngId).pngetStringField("name", null)
	 */
	public String setIdPNParameterOperand(String cnfn, boolean isSpecific) throws PException {
		RdbGenColumn rgc = (RdbGenColumn) genId.cnFieldColumns.get(cnfn);
		StringBuffer value = new StringBuffer();
		if (isSpecific) {
			value.append("((");
			value.append(genId.getCnFQName());
			value.append("PNG) _pngId).pnGet");
			value.append(commonHelper.upperFL(cnfn));
			value.append("(null)");
		} else {
			value.append("((PNameGetter) _pngId).");
			value.append(commonHelper.getPNameGetterGetFunction(rgc.columnType));
			value.append("(\"");
			value.append(cnfn);
			value.append("\", null)");
		}
		return value.toString();
	}
	
	/**
	 * Searches the RdbGenRef object, from the genRefs list, which the given
	 * column belongs to.
	 * @param rgc   The column for which to find a RdbGenRef.
	 * @return  The relevant RdbGenRef or null if none found.
	 */
	public RdbGenRef getGenRefOfColumn(RdbGenColumn rgc) {
		return getGenRefOfColumn(rgc, null);
	}

	public RdbGenRef getGenRefOfColumn(RdbGenColumn rgc, RdbGenJoin rgj) {
		if (rgc.joinCol != null) {
			return getGenRefOfColumn(rgc.joinCol, rgj);
		}
		PrimitiveElement pe = null;
		if (rgj != null && rgc.joins != null) {
			int idx = rgc.joins.indexOf(rgj);
			if (idx != -1) {
				pe = (PrimitiveElement) rgc.pes.get(idx);
			}
		}
		Iterator it = genRefs.values().iterator();
		while (it.hasNext()) {
			RdbGenRef rgr = (RdbGenRef) it.next();
			if (rgr.refColumn != null) {
				if (rgr.refColumn == rgc && (pe == null
						|| rgr.nd.getFieldName().equals(pe.getName()))) {
					//if (rgj !=null) System.out.println("getGenRefOfColumn(" + rgc.columnName + COLUMN + rgj.joinIdx + ") return " + rgr.fieldName);
					return rgr;
				}
			} else {
				Iterator it2 = rgr.cnFieldColumns.values().iterator();
				Collection classfields = rgr.nd.getNameRef().getProjection().values();
				while (it2.hasNext()) {
					RdbGenColumn currentRgc = (RdbGenColumn) it2.next();
					if (currentRgc == rgc && (pe == null
							|| classfields.contains(pe.getName()))) {
						//if (rgj !=null) System.out.println("getGenRefOfColumn(" + rgc.columnName + COLUMN + rgj.joinIdx + ") return " + rgr.fieldName);
						return rgr;
					}
				}
			}
		}
		if (rgj !=null) System.out.println("getGenRefOfColumn(" + rgc.columnName + COLUMN + rgj.joinIdx + ") return null");
		return null;
	}

	/**
	 * Verifies if the given column is used in the identifier of this class,
	 * which is defined by genId.
	 * @param genCol   The column for which to verify its belonging to the ID.
	 * @return  true if it belongs to genId.
	 */
	public boolean colInGenId(RdbGenColumn genCol) {
		if (genCol.joinCol != null) {
            // the column belongs to an external table and is part 
            // of the join with the main table
            // then we check if the corresponding column 
            // in the main table is part of the id
			return colInGenId(genCol.joinCol);
		}
        if (genId.refColumn != null) {
			if (genId.refColumn == genCol) {
				return true;
			}
		} else {
			Iterator itCol = genId.cnFieldColumns.values().iterator();
			while (itCol.hasNext()) {
                RdbGenColumn genColOfComposite = (RdbGenColumn) itCol.next();
				if (genColOfComposite == genCol) {
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * Searches the name of the field which defines the given reference.
	 * @param gr    The reference definition.
	 * @return  The relevant field name.
	 */
	public String getGenRefFieldName(RdbGenRef gr) {
		Iterator it = genRefs.keySet().iterator();
		while (it.hasNext()) {
			String res = (String) it.next();
			if (genRefs.get(res) == gr) {
				return res;
			}
		}
		return null;
	}

	/**
	 * Searches the name of the field in the composite name associated with the
	 * identifier, which is mapped to the given column.
	 * @param rgc   The column to search.
	 * @return  The name of the mapped field.
	 */
	public String getPnFieldName(RdbGenColumn rgc) {
		Iterator it = genId.cnFieldColumns.keySet().iterator();
		while (it.hasNext()) {
			String res = (String) it.next();
			if (rgc == genId.cnFieldColumns.get(res)) {
				return res;
			}
		}
		return null;
	}

	// PRIVATE METHODS

	/**
	 * Appends the definition of a fully qualified name of a field of a table
	 * (example: field F of table T => appends "T.F").
	 * @param sb
	 * @param rgc
	 */
	private void appendFQColName(StringBuffer sb, RdbGenColumn rgc) {
		sb.append(rgc.table.tableName);
		sb.append(".");
		sb.append(rgc.columnName);
	}

	private String getFQColName(RdbGenColumn rgc, String tableSuffix) {
		return rgc.table.tableName + tableSuffix + "." + rgc.columnName;
	}

	/**
	 * Appends the where clause for a given table for the prupose of a
	 * modification of this table (i.e., an update or a delete).
	 * @param sb    The string buffer into which to produce the clause.
	 * @param rgt   The involved table.
	 */
	private void appendModificationWhereClause(StringBuffer sb,
											   RdbGenTable rgt)
			throws PException {
		String sep = "";
		sb.append(" WHERE ");
		if (rgt != mainTable) {
			throw new PException("umanaged case, the table "
								 + rgt.tableName + " is an external table");
		}
		if (genId.refColumn != null) {
			sb.append(genId.refColumn.columnName);
			sb.append(" = ?");
		} else {
			Iterator it = genId.cnFieldColumns.values().iterator();
			while (it.hasNext()) {
				RdbGenColumn cur = (RdbGenColumn) it.next();
				sb.append(sep);
				sep = " AND ";
				sb.append(cur.columnName);
				sb.append(" = ?");
			}
		}
	}

	/**
	 * Appends the where clause for a given table for the prupose of a
	 * modification of this table (i.e., an update or a delete).
	 * @param sb    The string buffer into which to produce the clause.
	 * @param rgj   The involved join.
	 */
	private void appendModificationWhereClause(StringBuffer sb,
											   RdbGenJoin rgj,
											   boolean isNotJoinColumn) {
		String sep = "";
		sb.append(" WHERE ");
		sep = "";
		if (isNotJoinColumn) {
			for (int i = 0; i < rgj.table.columns.size(); i++) {
				RdbGenColumn rgc = (RdbGenColumn) rgj.table.columns.get(i);
				if (rgc.joinCol == null) {
					sb.append(sep);
					sep = " AND ";
					sb.append(rgc.columnName);
					sb.append(" = ?");
				}
			}
		} else {
			for (int i = 0; i < rgj.joinColumnsInExt.size(); i++) {
				sb.append(sep);
				sep = " AND ";
				sb.append(((RdbGenColumn) rgj.joinColumnsInExt.get(i)).columnName);
				sb.append(" = ?");
			}
		}
	}

	public boolean updatable(RdbGenColumn rgc) {
		return (!rgc.hiddenField || rgc.joinCol != null || !colInGenId(rgc))
				&& rgc.constant == null;
	}

	public boolean updatable(RdbGenColumn rgc, RdbGenJoin rgj) {
		if (rgj == null) {
			return updatable(rgc);
		}
		if (rgj.table.colocatedTable) {
			return rgc.joinCol != null && rgc.isInJoin(rgj);
		} else {
			return rgc.joinCol == null;
		}
	}

    public boolean hasFilter() {
        return filterExpr!= null && !"".equals(filterExpr);
    }



	class JormExpression2WhereClause extends Expression2WhereClauseImpl {

		public void convertExp2WhereClauseBuilder(Expression expression,
												  String rdbAdapterVarName,
												  StringBuffer sb, Map ctx) throws ExpressionException {
			if (expression instanceof ParameterOperand) {
				// A parameter operand corresponds to a PEM mapped into the
				// main table
				String fieldName = ((ParameterOperand) expression).getName();
				RdbPrimitiveElementMapping rpem =
					getRPEM(fieldName, mapping.getClassMapping());
				if (rpem == null) {
					throw new ExpressionException(
							"No mapping found for the field '" + fieldName
							+ "' into the mapping filter");
				}
				RdbGenColumn rgc = mainTable.getColumn(rpem.getName());
				if (rgc == null) {
					throw new ExpressionException(
							"No column found for the field '" + fieldName
							+ "' into the mainb table '"
							+ mainTable.tableName
							+ "' in the mapping filter");
				}
				appendFQColName(sb, rgc);
			} else if (expression instanceof VariableOperand) {
				sb.append("\" + ");
				sb.append(rdbAdapterVarName);
				sb.append(".getValueAsSQLString(getNamingFilterKey(), ");
				sb.append(expression.getType().getProgName());
				sb.append(".getTypeCode()) + \"");
			} else {
				super.convertExp2WhereClauseBuilder(expression, rdbAdapterVarName, sb, ctx);
			}
		}
		private RdbPrimitiveElementMapping getRPEM(String fieldName,
												   ClassMapping cm)
			throws ExpressionException {
			if (debug) {
				logger.log(BasicLevel.DEBUG, "Try to find the field " + fieldName
					+ " on the class mapping of the class "
					+ ((Class) cm.getLinkedMO()).getFQName());
			}
			RdbPrimitiveElementMapping rpem = (RdbPrimitiveElementMapping)
					cm.getPrimitiveElementMapping(fieldName);
			if (rpem == null) {
				for(Iterator it = cm.getParentClassMappings().iterator();
					rpem == null && it.hasNext();) {
					ParentClassMapping pcm = (ParentClassMapping) it.next();
					ClassMapping _cm = pcm.getMOClass()
						.getClassProject(pcm.getProjectName())
						.getMapping(pcm.getMapperName()).getClassMapping();
					rpem = getRPEM(fieldName, _cm);
				}

			}
			return rpem;
		}
	}

}
