/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.rdb.generator;

import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.api.PException;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

/**
 * Defines the structure of a reference of a JORM class. This structure is
 * either a simple column or a composite name.
 * In the case of a simple column, only refColumn is defined (not null).
 * In the case of a composite name, all variables are defined except refColumn
 * that is null.
 * @author  P. Dechamboux
 */
public class RdbGenRef {
	/**
	 * The name of the composite name used, if it is a composite name.
	 */
	public String cnName = null;
	/**
	 * An index in [1..n] is associated with each composite name reference.
	 */
	public int cnId = -1;
	/**
	 * The package name of the composite name.
	 */
	public String cnPackage = null;
	/**
	 * It associates a projection column for each field of the composite name.
     * < name of the field in the composite name (String) -> RdbGenColumn > 
	 */
	public Map cnFieldColumns = null;
	/**
	 * The column associated with this reference (non composite name).
     * Null if the namedef is composite, otherwise directly references 
     * the column encoding the name 
	 */
	public RdbGenColumn refColumn = null;

	public NameDef nd = null;

    /**
     * name of the reference field. If null the object 
     * describes the mapping of the Id
     */
	public String fieldName = null;

	public RdbGenJoin join = null;

	public String getFieldName() {
		return fieldName;
	}

	public NameDef getNd() {
	    return nd;
	}
	public String getCnName() {
		return cnName;
	}

	public int getCnId() {
		return cnId;
	}

	public String getCnPackage() {
		return cnPackage;
	}

	public String getCnFQName() {
		return (cnPackage == null ? cnName : cnPackage + "." + cnName);
	}

	public Map getCnFieldColumns() {
		return cnFieldColumns;
	}

	public RdbGenColumn getRefColumn() {
		return refColumn;
	}

	/**
	 * Gets the composite name field that is mapped to the given column.
	 * @param rgc   The involved column.
	 * @return  The field name of the composite name that is mapped to this
	 *          column.
	 */
	public String getFieldName(RdbGenColumn rgc) {
		if (rgc.joinCol != null) {
			return getFieldName(rgc.joinCol);
		}
		if (refColumn != null) {
			return null;
		}
		Iterator it = cnFieldColumns.keySet().iterator();
		while (it.hasNext()) {
			String res = (String) it.next();
			if (cnFieldColumns.get(res) == rgc) {
				return res;
			}
		}
		return null;
	}

	public String getClassFieldName(String compositeFieldName) {
		return (String) nd.getNameRef().getProjection().get(compositeFieldName);
	}

	public String getAccessorValue(boolean isSpecific) {
		if (isSpecific) {
			return "paccessor.paGet"
					+ RdbGenInfos.commonHelper.upperFL(fieldName) + "(conn)";
		} else {
			return "paccessorGen.paGetRefField("
					+ fieldName + "_NAME, conn)";
		}

	}

	public RdbGenJoin getJoin(RdbGenColumn rgc) throws PException {
		return getJoin(rgc, nd.getFieldName());
	}

	public RdbGenJoin getJoin(RdbGenColumn rgc,
							  String fn) throws PException {
		try {
			if (rgc.pes == null) {
				throw new PException("No join needed to reach the field " + fn
									 + " over the " + rgc.columnName + " column.");
			}
			int i = 0;
			while (i < rgc.pes.size()
					&& !fn.equals(((PrimitiveElement) rgc.pes.get(i)).getName())) {
				i++;
			}
			if (i < rgc.pes.size()) {
			    RdbGenJoin rgj = (RdbGenJoin) rgc.joins.get(i);
			    if (rgj == null) {
					new PException("No RdbGenJoin found for the field '" 
					        + fn + "' and the column '" 
					        + rgc.columnName + "'.").printStackTrace();
			    }
			    return rgj;
			} else {
				throw new PException("Field " + fn + " is not mapped over the "
									 + rgc.columnName + " column.");
			}
		} catch (RuntimeException e) {
			e.printStackTrace();
			throw e;
		}
	}
	public Set getcnField2ColumnName() {
	    Map p = new HashMap();
	    try {
	    if (nd.isFieldName()) {
            //No alias is need because the table containing identifier is the 
            //main table.
            p.put("", refColumn.table.tableName + "." + refColumn.columnName);
	    } else if (nd.isNameRef()) {
	        
	        for (Iterator iter = cnFieldColumns.entrySet().iterator(); iter.hasNext();) {
	            Map.Entry me = (Map.Entry) iter.next();
	            RdbGenColumn rgc = (RdbGenColumn) me.getValue();
	            //No alias is need because the table containing identifier is the 
	            //main table.
	            p.put(me.getKey(), rgc.table.tableName + "." + rgc.columnName);
	        }
        }
	    }catch(RuntimeException e){
	        System.out.println(this);
	        e.printStackTrace();throw e;};
	    return p.entrySet();
	    
	}
	
	public String toString() {
	    StringBuffer sb = new StringBuffer(super.toString());
	    sb.append(" / cnId=").append(cnId);
	    sb.append(" / cnPackage=").append(cnPackage);
	    sb.append(" / cnFieldColumns=").append(cnFieldColumns);
	    sb.append(" / refColumn=").append(refColumn);
	    sb.append(" / nd=").append(nd);
	    if (nd != null) {
	        if (nd.isNameRef()) {
	            sb.append(" / nd.nr.cn=").append(nd.getNameRef().getCompositeName());
	            sb.append(" / nd.nr.cn=").append(nd.getNameRef().getProjection());
	        }
	    }
	    sb.append(" / fieldName=").append(fieldName);
	    sb.append(" / join=").append(join);	     
	    return sb.toString();
	}
}
