/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.rdb.generator;

import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.api.ScalarField;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.mapper.rdb.metainfo.*;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.HashMap;
import java.util.Map;

/**
 * Defines one of the table that is used to mapped a particular JORM class.
 * In case of the main table (see RdbGenInfos), no join information is
 * defined.
 * @author  P. Dechamboux
 */
public class RdbGenTable {
	public RdbGenInfos genInfos = null;
	/**
	 * The name of the table involved in this mapping.
	 */
	public String tableName = null;
	/**
	 * The list of columns of this table used in this mapping.
	 */
	public List columns = new ArrayList();

	/**
	 * The list of inherited columns of this table used in this mapping.
	 */
	public List inheritedColumns = new ArrayList();

	/**
	 * The filter.
	 */
	// public RdbFilter filter = null;

	/**
	 * The list of columns of this table that are shared with a colocated
	 * object (i.e., an object stored into the same table along with this
	 * object.
	 */
	public List colocatedColumns = null;
	/**
	 * Tells that this table stores colocated objects.
	 */
	public boolean colocatedTable = false;

	public List joins = null;
	public Map jn2join = null;

    /**
     * indicates if the table is a read-only table
     */
	public boolean readOnly = false;

    /**
     * indicates if code must be generated to update the table
     * when updating an instance
     */
	public boolean mustGenerateUpdate = true;

    /**
     * indicates if the table is the main table of a the class
     */
	public boolean isMainTable;

	public RdbGenTable() {
	}

    /**
     * Creates an RdbGenTable from an external table.
     * @param extTable the external table
     * @param mainTable the associated main table
     * @param clazz the associated table
     * @throws PException
     */
	public RdbGenTable(RdbExternalTable extTable,
					   RdbGenTable mainTable,
					   Class clazz,
                       RdbGenInfos genInfos) throws PException {

		constructFromTable(extTable, clazz, genInfos);
		isMainTable = false;
		readOnly = extTable.isReadOnly();
		joins = new ArrayList();
		jn2join = new HashMap();
		int joinIdx = 0;
        for (Iterator itj = extTable.getRdbJoins().iterator(); itj.hasNext();) {
            RdbJoin j = (RdbJoin) itj.next();
            RdbGenJoin rgj = new RdbGenJoin(mainTable, this, j, joinIdx);
            joinIdx++;
            jn2join.put(j.getName(), rgj);
            joins.add(rgj);
        }
        if ( extTable.getPrimitiveElementMappings() != null) {
		    for (Iterator it = extTable.getPrimitiveElementMappings().iterator(); it.hasNext();) {
			    RdbPrimitiveElementMapping rpem = (RdbPrimitiveElementMapping) it.next();
			    Iterator entryIt = rpem.getPrimitiveElementByRdbJoin().entrySet().iterator();
			    if (!entryIt.hasNext()) {
				    continue;
			    }
			    if (rpem.getName() == null) {
				    throw new PException("Primitive element "
									    + ((PrimitiveElement) rpem.getLinkedMO()).getName()
									    + " has a null column name");
			    }
			    RdbGenColumn rgc = getColumn(rpem.getName());
			    rgc.pes = new ArrayList();
			    rgc.joins = new ArrayList();
			    PrimitiveElement pe = null;
			    while (entryIt.hasNext()) {
				    Map.Entry me = (Map.Entry) entryIt.next();
				    RdbJoin j = (RdbJoin) me.getKey();
				    pe = (PrimitiveElement) me.getValue();
				    rgc.pes.add(pe);
				    rgc.joins.add(jn2join.get(j.getName()));
			    }
			    rgc.hiddenField = pe instanceof ScalarField;
			    rgc.columnType = pe.getType();
		    }
        }
	}

    /**
     * Creates an RdbGenTable from a main table.
     * @param mainTable the associated main table
     * @param clazz the associated table
     * @throws PException
     */
	public RdbGenTable(RdbTable mainTable, Class clazz, RdbGenInfos genInfos) throws PException {
        constructFromTable(mainTable, clazz, genInfos);
        isMainTable = true;
	}


    private void constructFromTable(RdbTable table, Class clazz,
                                    RdbGenInfos genInfos) throws PException {
        this.genInfos = genInfos;
        tableName = table.getName();
        colocatedTable = table.isColocated() && !table.isColocatedMaster();

        for (Iterator it = table.getPrimitiveElementMappings().iterator(); it.hasNext();) {
            RdbPrimitiveElementMapping rpem = (RdbPrimitiveElementMapping) it.next();
            RdbGenColumn rgc = new RdbGenColumn();
            rgc.columnName = rpem.getName();
            if (rgc.columnName == null) {
                throw new PException("Primitive element "
                                     + ((PrimitiveElement) rpem.getLinkedMO()).getName()
                                     + " has a null column name");
            }
            rgc.columnNotNull = rpem.isNotNull();
            rgc.table = this;
            PrimitiveElement pe = (PrimitiveElement) rpem.getLinkedMO();
            if (pe != null) {
                rgc.columnType = pe.getType();
                rgc.columnSize = pe.getSize();
                rgc.columnScale = pe.getScale();
                rgc.hiddenField = pe instanceof ScalarField;
                if (pe.isConstant()) {
                    //fetch the constant value on the current class
                    rgc.constant = clazz.getConstantValue(pe.getName());
                }
                rgc.fieldName = pe.getName();
            }
            rgc.columnSqlType = rpem.getType();
            if (rgc.columnSqlType != null && rgc.columnSqlType.length() == 0) {
                rgc.columnSqlType = null;
            }
            columns.add(rgc);
        }

    }
    /**
     * Fix the mustGenerateUpdate attribute.
     * One must generate code to update the table when an instance of the jorm
     * class (to which the genTable is associated) is updated.
     * The table is concerned only if there are columns which do not participate
     * in the join with the main table nor in the identifier
     */
	public void fixMustGenerateUpdate() {
		mustGenerateUpdate = false;
        // loop on columns
		for (int i = 0; i < columns.size() && !mustGenerateUpdate; i++) {
			RdbGenColumn rgc = (RdbGenColumn) columns.get(i);
			mustGenerateUpdate |= isUpdatableColumn(rgc);
		}
		//loop on inherited columns
		for (int i = 0; i < inheritedColumns.size() && !mustGenerateUpdate; i++) {
			RdbGenColumn rgc = (RdbGenColumn) inheritedColumns.get(i);
			mustGenerateUpdate |= isUpdatableColumn(rgc);
		}
	}

	private boolean isUpdatableColumn(RdbGenColumn rgc) {
		//column must not participate in the join with the main table
		if (rgc.joinCol != null) {
			return false;
		}

        // column must not participate in the identifier
		if (joins == null && genInfos.colInGenId(rgc)) {
			return false;
		}
        // otherwise we must generate some code that update the table
		return true;
	}
	
	
	public boolean getMustGenerateUpdate() {
		return mustGenerateUpdate;
	}

	public boolean getReadOnly() {
		return readOnly;
	}

	public List getJoins() {
		return joins;
	}

	public RdbGenInfos getGenInfos() {
		return genInfos;
	}

	public String getTableName() {
		return tableName;
	}

	public String getTableNameNoDot() {
		return tableName.replace('.', '_');
	}

	public List getColumns() {
		return columns;
	}

	public List getInheritedColumns() {
		return inheritedColumns;
	}

	public List getColocatedColumns() {
		return colocatedColumns;
	}

	public boolean getColocatedTable() {
		return colocatedTable;
	}

	public boolean isMainTable() {
		return isMainTable;
	}

	/**
	 * Specifies if only prepared statement can be used with this table mapping.
	 */
	private Boolean psMand = null;

	/**
	 * Tells if only PreparedStatement can be used to store data associated
	 * with this table.
	 * @return  true if the use of PreparedStatement is mandatory.
	 */
	public boolean preparedStatementMandatory() {
		if (psMand == null) {
			for (int i = 0; i < columns.size(); i++) {
				if (((RdbGenColumn) columns.get(i)).columnType == PTypeSpace.SERIALIZED) {
					psMand = new Boolean(true);
					return true;
				}
			}
			psMand = new Boolean(false);
		}
		return psMand.booleanValue();
	}

	/**
	 * Tells if the given column belongs to a colocated ones, which means that
	 * it is shared with the colocated object and must not be touched (e.g.
	 * should not be nullified).
	 * @param rgc   The involved column.
	 * @return  true if it is a colocated column.
	 */
	boolean colocatedColumn(RdbGenColumn rgc) {
		if (colocatedColumns != null) {
			for (int i = 0; i < colocatedColumns.size(); i++) {
				if (colocatedColumns.get(i) == rgc) {
					return true;
				}
			}
		}
		return false;
	}
    /**
     * get the genColumn corresponding to the given column columnName. If
     * not found in this table definition, search in the table definition of the
     * super classes
     * @param columnName the name of a column in the table
     * @return the genColumn corresponding to the column name
     */
	public RdbGenColumn getColumn(String columnName) {
	    RdbGenColumn rgc = null;
	    for (int i = 0; i < columns.size(); i++) {
	        rgc = (RdbGenColumn) columns.get(i);
	        if (rgc.columnName.equals(columnName)) {
	            return rgc;
	        }
	    }
	    for (int i = 0; i < inheritedColumns.size(); i++) {
	        rgc = (RdbGenColumn) inheritedColumns.get(i);
	        if (rgc.columnName.equals(columnName)) {
	            return rgc;
	        }
	    }
	    
	    return null;
	}
}
