/**
 * Copyright (C) 2003-2004
 * - France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors: Olivier Lobry (olivier.lobry@francetelecom.com)
 * Date: 14 sept. 2004
 * Time: 11:15:44
 */

package org.objectweb.jorm.mapper.rdb.lib;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbClassMapping;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbClassMultiMapping;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbJoin;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbPrimitiveElementMapping;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbTable;
import org.objectweb.jorm.mapper.rdb.util.ColumnAliasing;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.ParentClassMapping;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

public class RdbExtentGenInfos {
	
    public Logger logger;
    public boolean debug = false;

    public RdbExtentGenInfos(Logger logger) {
        this.logger = logger;
        debug = logger != null && logger.isLoggable(BasicLevel.DEBUG);
    }

    /**
     * Calculates the query representing the extent of the class.
     * The query will be made of an union (in case of horizontal mappings)
     * of filtered (in case of filtered and vertical mappings) tables.
     * Elements of the extent may come from the tables on which the class is mapped
     * or from tables on which its sub(sub...)classes are mapped.
     * The main problem is to retrieve those tables and in those tables the right
     * columns mapping fields defined in the mother class knowing that:
     *   - different inheritance rules may be encountered (filtered, horizontal,
     *     vertical). Some will need union, some others will need selection
     *   - classes may be mapped on multiple tables, requiring joins to retrieve the
     *     needed columns
     *   - only tables columns mapping fields of the mother class should be included
     *     in the expression of the query
     *   - columns mapping the same field in different class mapping may have
     *     different names, thus requiring renaming for realizing the required unions
     *
     * How the query is calculated depends on whether the class is at the top
     * of the hierarchy tree.
     * 1) If is at the top, its extent is made of all the instances coming from its
     *    tables and those of its children (if they are mapped to other tables).
     * 2) If it is in the middle of the tree and its inheritance rule is 'to-new-structures'
     *    then, none of its instances comes from its parents' tables, so we can apply the
     *    same algorithm
     * 3) If it is in the middle and its inheritance rule is 'to-extended-structures', then its
     *    filter and those of its children must be used not to get instances of its
     *    parent class.
     *
     * In the first two cases, the query is created with the following formula:
     *   Q(A) = T(A) U (U (PROJECT(ATTR(A)) T))
     *               |
     *      T in T(SUB(A)) and T != T(A)
     * where Q(A) = query for A's extent, ATTR(A) = attributes of A,
     *       T(X) = main table for class X, SUB(A) = all sub(sub(...))classes of A
     *
     * In the third case, the query is created with the following formula:
     *   Q(A) = (FILTER(A) AND (AND FILTER(X))) T(A) U (U (PROJECT(ATTR(A)) T(X)))
     *                           |                      |
     *     X in SUB_F(A) and T(X)=T(A)   T in T(SUB(A) - SUB_F(A)) and T != T(A)
     * where Q(A) = query for A's extent, ATTR(A) = attributes of A,
     *       T(X) = main table for class X,
     *       SUB_F(A) = all sub(sub(...))classes that is 'filter chained'* w/ A
     *    (*) 'filter chained' means that all inheritance link from A to X
     *         is a filter inheritance rule
     *
     * Actually, this is a simplification since a single class may be mapped to
     * multiple tables and the extent be calculated by joining these tables.
     * However, since the filters express conditions on the key and the key is
     * entirely contained in a table, this does not change the principles
     * discussed above but simply complexifies the expression of the formulae.
     * Instead of T(A) (resp. T(X)) we will have the joins between A's
     * (resp. X's) tables.
     *
     * Also, the case of abstract classes must be taken into account. If class A
     * its extents is the union of the extents of its subclasses. This is true
     * whether it defines or extends an inherited mapping or not. So there is
     * no need to construct a "select from where" at its level.
     *
     * @return a collection of ExtentMappingInfos that must be unioned to get the extent
     */
    //TODO: remove this method when integrating the horizontal and verticla mapping in speedo
    //not forget to change the method call in the velocity template with the next method
    public Collection getPolymorphicClassExtent(Class classA, String projectName,
                                               String mapperName) throws PException {
        //System.out.println("debug=" + debug + "logger=" + logger
        //    + (logger == null ? "" : ", logger.name=" + logger.getName()));
        if (debug) logger.log(BasicLevel.INFO, "Calculating extent for class"
                    + classA.getFQName() + " project: " + projectName +
                ", mapperName: " + mapperName);
        RdbClassMapping classMappingOfA = (RdbClassMapping)
                classA.getClassMapping(projectName, mapperName);


        // The term "class A" is used for the name of the class of which we are calculating
        // the extent and "class X" is used as a sub(sub...)class of A

        // get all the primitive elements of class A (non-hidden)
        ArrayList primElsOfA =  new ArrayList();
        if (debug) logger.log(BasicLevel.DEBUG, "Primitive elements are: ");
        for (Iterator itTE = classA.getAllFields().iterator(); itTE.hasNext();) {
            Object te = itTE.next();
            if (te instanceof PrimitiveElement) {
                primElsOfA.add(te);
                if (debug) logger.log(BasicLevel.DEBUG, "   " +
                        ((PrimitiveElement) te).getName());
            }
        }
        // add hidden fields
        for (Iterator itTE = classA.getAllHiddenFields().iterator(); itTE.hasNext();) {
            Object te = itTE.next();
            if (te instanceof PrimitiveElement) {
                primElsOfA.add(te);
                if (debug) logger.log(BasicLevel.DEBUG, "   " +
                        ((PrimitiveElement) te).getName() + " (hidden)");
            }
        }

        HashMap extMappingInfosMap = new HashMap(); // <Class->ExtentMappingInfos>
        HashSet extMappingInfos = new HashSet(); // <ExtentMappingInfos>

        // Remaining classes to inspect. The algorithm follows a top-down
        // horizontal traversal of the tree, classes being added progressively
        // during the traversal
        ArrayList remClasses = new ArrayList();
        remClasses.add(classA);
        while (!remClasses.isEmpty()) {
            // remove the first class in list to treat it
            Class classX = (Class) remClasses.remove(0);
            // and add its children
            Collection xSubClasses = classX.getSubClasses();
            if (xSubClasses != null) {
                remClasses.addAll(xSubClasses);
            }

            // get its class mapping in the current project
            RdbClassMultiMapping classMappingOfX = (RdbClassMultiMapping)
                    classX.getClassMapping(projectName, mapperName);
            if (classMappingOfX == null) {
                throw new PException("Could not find a class mapping for class "
                        + classA.getName());
            }


            // will reference the extent mapping information relative to class X
            // EMIs can be shared by classes that map fields to the same structures
            RdbExtentMappingInfos emi;


            // sub-class inspection
            if (classX != classA) {
                if (debug) logger.log(BasicLevel.DEBUG, "Inspecting subclass: "
                            + classX.getName());
                Class[] parent = new Class[1];
                RdbExtentMappingInfos emiParent = getParentEMI(classX, extMappingInfosMap, parent);
                // NB: the ext mapping info may be attached to a ancestor of the parent
                // so do not use emiParent.clazz to retrieve the pcm but the real parent instead
                ParentClassMapping pcmOfX = classMappingOfX.
                        getParentClassMapping(parent[0].getFQName());

                // sanity check
                if (pcmOfX == null) {
                    throw new PException("Could not find a parent class mapping" +
                            " for class " + classX.getName() + " to super class" +
                            classA.getName());
                }

                // gets the inheritance rule
                String ruleName = pcmOfX.getRuleName();
                if (debug) logger.log(BasicLevel.DEBUG, "  inherits mapping from : " +
                        ((Class) pcmOfX.getLinkedMO()).getFQName() +
                        " with rule=" + ruleName);


                // in the case of an 'extended' or 'added' mapping,
                // add the filter if it exists (and if required) to the
                // emi of the parent
                if ((RdbClassMapping.MAP_NEW_FIELDS_TO_EXTENDED_STRUCTURES
                        .equalsIgnoreCase(ruleName)) ||
                        (RdbClassMapping.MAP_NEW_FIELDS_TO_ADDED_STRUCTURES
                        .equalsIgnoreCase(ruleName))) {
                    // if the parent has unmapped fields, then we need to
                    // construct the union of its children's extents
                    // Otherwise, we can generate a single select/from/where
                    // clause with approriate filters
                    if (emiParent.hasUnmappedFields) {
                        if (debug) logger.log(BasicLevel.DEBUG,
                                "Parent class has unmapped fields, create new EMI");
                        emi = new RdbExtentMappingInfos(classX, classX.getName(), primElsOfA.size());
                        extMappingInfos.add(emi);
                        emi.addFilters = emiParent.addFilters;
                        emi.hasUnmappedFields = classMappingOfX.
                                hasUnmappedPrimitiveElements(primElsOfA);
                    } else {
                        if (debug) logger.log(BasicLevel.DEBUG,
                                "Parent class has all its fields mapped, complement its EMI");
                        emi = emiParent;
                    }

                    // add filter if required and if some exist (there is none
                    // if the class is abstract)
                    if (emi.addFilters && (!classX.isAbstract())) {
                        emi.filters.add(classX.getFQName());
                    }
                } else if (RdbClassMapping.REMAP_FIELDS_TO_NEW_STRUCTURES
                        .equalsIgnoreCase(ruleName)) {
                    // mapping completely redefined on new structures,
                    // so create a new extent mapping info to generate a seperate
                    // select/from/where clause
                    emi = new RdbExtentMappingInfos(classX, classX.getName(), primElsOfA.size());
                    // TODO: should optimize this. The necessity to add filters depends on the inheritance
                    // relationship with the super classes AND the sub-classes
                    emi.addFilters = true;
                    extMappingInfos.add(emi);
                    emi.hasUnmappedFields = classMappingOfX.hasUnmappedPrimitiveElements(primElsOfA);
                } else {
                    throw new InternalError("Unexpected inheritance rule: "
                            + ruleName);
                }
                // register the extent mapping info for that class
                extMappingInfosMap.put(classX, emi);

                // if class A is abstract it has no instances and has
                // (or at least should have) no filters
                // so we can stop here
                if (classX.isAbstract()) {
                    if (debug) logger.log(BasicLevel.DEBUG, "Class is abstract, next one");
                    continue;
                }

            } else { // class X == class A
                emi = new RdbExtentMappingInfos(classX, classX.getName(), primElsOfA.size());
                extMappingInfos.add(emi);
                extMappingInfosMap.put(classX, emi);

                // in case of class A, determines the 'add filters' condition
                // that will be checked for its subclasses
                // The necessity to add filters depends on the inheritance
                // relationship with the super classes AND the sub-classes
                emi.addFilters = classMappingOfX.inheritsStructures();
                emi.hasUnmappedFields = classMappingOfX.hasUnmappedPrimitiveElements(primElsOfA);
                if (debug) logger.log(BasicLevel.DEBUG, "  should add filters ? "
                        + (emi.addFilters ? "yes" : "no"));

                // if class A is abstract it has no instances and has
                // (or at least should have) no filters
                // so we can stop here
                if (classX.isAbstract()) {
                    if (debug) logger.log(BasicLevel.DEBUG, "Class is abstract, next one");
                    continue;
                }
                if (emi.addFilters) {
                    emi.filters.add(classX.getFQName());
                }
            }



            // finds how primitive elements defined in class A are mapped
            // in class X and construct select/from/where clauses to fetch
            // the corresponding tuples from X's tables
            int pos = 0;
            for (Iterator it = primElsOfA.iterator(); it.hasNext(); pos++) {
                PrimitiveElement pe = (PrimitiveElement) it.next();
                if (debug) logger.log(BasicLevel.DEBUG, "     find mapping for "
                        + "primitive element " + pe.getName());
                if (emi.mappingDone(pos)) {
                    if (debug) logger.log(BasicLevel.DEBUG, "    already done");
                    continue;
                }
                // search the PEM, looking in super classes if necessary
                RdbPrimitiveElementMapping mappingOfpeInX = (RdbPrimitiveElementMapping)
                        classMappingOfX.getPrimitiveElementMapping(pe.getName(), true);
                // for classes mapped to several tables, the primitive element
                // may be mapped on an external table, requiring a join
                RdbJoin join = mappingOfpeInX.getJoinByPrimitiveElement(pe);
                String tableName;
                if (join != null) {
                    if (debug) logger.log(BasicLevel.DEBUG, "     mapped to external table");
                    tableName = join.getExternalTable().getName();
                    emi.addJoin(join);
                } else {
                    if (debug) logger.log(BasicLevel.DEBUG, "     mapped to main table");
                    emi.mainTable = (RdbTable) mappingOfpeInX.getParent();
                    tableName = emi.mainTable.getName();
                }
                
                emi.addProjection(tableName, mappingOfpeInX.getName(),
                        ColumnAliasing.getAliasFromColumn(pe.getName()), classX.getFQName(), pos);
            }
        } // while !remClasses.isEmpty()

        // remove empty requests
        for (Iterator emiIt = extMappingInfos.iterator(); emiIt.hasNext();) {
            RdbExtentMappingInfos emi = (RdbExtentMappingInfos) emiIt.next();
            if (emi.isEmpty()) {
                emiIt.remove();
            }
        }

    return extMappingInfos;
    }
    
    /**
     * This method is used to create the list of extent mapping infos for the class classA.
     * @param classA: the JORM class
     * @param projectName: the name of the project
     * @param mapperName: the name of the mapper
     * @param prefetch: true or false
     */
    public Collection getPolymorphicClassExtent(Class classA, String projectName, String mapperName, boolean prefetch) throws PException {
		if (debug)
			logger.log(BasicLevel.INFO, "Calculating extent for class"
					+ classA.getFQName() + " project: " + projectName
					+ ", mapperName: " + mapperName);
		RdbClassMapping classMappingOfA = (RdbClassMapping) classA
				.getClassMapping(projectName, mapperName);

		// The term "class A" is used for the name of the class of which we are
		// calculating
		// the extent and "class X" is used as a sub(sub...)class of A

		//the list of primitive elements for classA
		ArrayList primElsOfA = new ArrayList();
		
		// get all the primitive elements of class A (including non-hidden fields) (and its subclasses if prefetch is true) 
		ArrayList primEls = getPrimitiveElements(prefetch, classA, projectName, primElsOfA);

		HashMap extMappingInfosMap = new HashMap(); // <Class->ExtentMappingInfos>
		HashSet extMappingInfos = new HashSet(); // <ExtentMappingInfos>

		// Remaining classes to inspect. The algorithm follows a top-down
		// horizontal traversal of the tree, classes being added progressively
		// during the traversal
		ArrayList remClasses = new ArrayList();
		remClasses.add(classA);
		while (!remClasses.isEmpty()) {
			// remove the first class in list to treat it
			Class classX = (Class) remClasses.remove(0);
			// and add its children
			Collection xSubClasses = classX.getSubClasses();
			if (xSubClasses != null) {
				remClasses.addAll(xSubClasses);
			}

			// get its class mapping in the current project
			RdbClassMultiMapping classMappingOfX = (RdbClassMultiMapping) classX
					.getClassMapping(projectName, mapperName);
			if (classMappingOfX == null) {
				throw new PException(
						"Could not find a class mapping for class "
								+ classA.getName());
			}

			// will reference the extent mapping information relative to class X
			// EMIs can be shared by classes that map fields to the same
			// structures
			RdbExtentMappingInfos emi;

			// sub-class inspection
			if (classX != classA) {
				if (debug)
					logger.log(BasicLevel.DEBUG, "Inspecting subclass: "
							+ classX.getName());
				Class[] parent = new Class[1];
				RdbExtentMappingInfos emiParent = getParentEMI(classX,
						extMappingInfosMap, parent);
				// NB: the ext mapping info may be attached to a ancestor of the
				// parent
				// so do not use emiParent.clazz to retrieve the pcm but the
				// real parent instead
				ParentClassMapping pcmOfX = classMappingOfX
						.getParentClassMapping(parent[0].getFQName());

				// sanity check
				if (pcmOfX == null) {
					throw new PException(
							"Could not find a parent class mapping"
									+ " for class " + classX.getName()
									+ " to super class" + classA.getName());
				}

				// gets the inheritance rule
				String ruleName = pcmOfX.getRuleName();
				if (debug)
					logger.log(BasicLevel.DEBUG, "  inherits mapping from : "
							+ ((Class) pcmOfX.getLinkedMO()).getFQName()
							+ " with rule=" + ruleName);

				// in the case of an 'extended' or 'added' mapping,
				// add the filter if it exists (and if required) to the
				// emi of the parent
				if ((RdbClassMapping.MAP_NEW_FIELDS_TO_EXTENDED_STRUCTURES.equalsIgnoreCase(ruleName))
					||
					((RdbClassMapping.MAP_NEW_FIELDS_TO_ADDED_STRUCTURES.equalsIgnoreCase(ruleName))  && !prefetch)  ) {
					// if the parent has unmapped fields, then we need to
					// construct the union of its children's extents
					// Otherwise, we can generate a single select/from/where
					// clause with approriate filters
					if (emiParent.hasUnmappedFields) {
						if (debug)
							logger
									.log(BasicLevel.DEBUG,
											"Parent class has unmapped fields, create new EMI");
						emi = new RdbExtentMappingInfos(classX, classX.getName(),
								primEls.size());
						extMappingInfos.add(emi);
						emi.addFilters = emiParent.addFilters;
						emi.hasUnmappedFields = classMappingOfX
								.hasUnmappedPrimitiveElements(primElsOfA);
					} else {
						if (debug)
							logger
									.log(BasicLevel.DEBUG,
											"Parent class has all its fields mapped, complement its EMI");
						emi = emiParent;
					}

					// add filter if required and if some exist (there is none
					// if the class is abstract)
					if (emi.addFilters && (!classX.isAbstract())) {
						emi.filters.add(classX.getFQName());
					}
				} else if (RdbClassMapping.REMAP_FIELDS_TO_NEW_STRUCTURES.equalsIgnoreCase(ruleName) 
							|| 
							((RdbClassMapping.MAP_NEW_FIELDS_TO_ADDED_STRUCTURES.equalsIgnoreCase(ruleName))  && prefetch)) {
					// mapping completely redefined on new structures,
					// so create a new extent mapping info to generate a
					// seperate
					// select/from/where clause
					emi = new RdbExtentMappingInfos(classX, classX.getName(),
							primEls.size());
					// The necessity to add filters
					// depends on the inheritance
					// relationship with the super classes AND the sub-classes
					emi.addFilters = classMappingOfX.inheritsStructures();
					extMappingInfos.add(emi);
					emi.hasUnmappedFields = classMappingOfX
							.hasUnmappedPrimitiveElements(primElsOfA);
				} else {
					throw new InternalError("Unexpected inheritance rule: "
							+ ruleName);
				}
				// register the extent mapping info for that class
				extMappingInfosMap.put(classX, emi);

				// if class A is abstract it has no instances and has
				// (or at least should have) no filters
				// so we can stop here
				if (classX.isAbstract()) {
					if (debug)
						logger.log(BasicLevel.DEBUG,
								"Class is abstract, next one");
					continue;
				}

			} else { // class X == class A
				emi = new RdbExtentMappingInfos(classX, classX.getName(),
						primEls.size());
				extMappingInfos.add(emi);
				extMappingInfosMap.put(classX, emi);

				// in case of class A, determines the 'add filters' condition
				// that will be checked for its subclasses
				// The necessity to add filters
				// depends on the inheritance
				// relationship with the super classes AND the sub-classes
				emi.addFilters = classMappingOfX.inheritsStructures();
				emi.hasUnmappedFields = classMappingOfX
						.hasUnmappedPrimitiveElements(primElsOfA);
				if (debug)
					logger.log(BasicLevel.DEBUG, "  should add filters ? "
							+ (emi.addFilters ? "yes" : "no"));

				// if class A is abstract it has no instances and has
				// (or at least should have) no filters
				// so we can stop here
				if (classX.isAbstract()) {
					if (debug)
						logger.log(BasicLevel.DEBUG,
								"Class is abstract, next one");
					continue;
				}
				if (emi.addFilters) {
					emi.filters.add(classX.getFQName());
				}
			}

			// finds how primitive elements defined in class A (and subclasses if prefetch is true) are mapped
			// in class X and construct select/from/where clauses to fetch
			// the corresponding tuples from X's tables
			int pos = 0;
			for (Iterator it = primEls.iterator(); it.hasNext(); pos++) {
				PrimitiveElement pe = (PrimitiveElement) it.next();
				if (debug)
					logger.log(BasicLevel.DEBUG, "     find mapping for "
							+ "primitive element " + pe.getName());
				//if the mapping is already done and it is not the NULL mapping (a NULL mappping for the c column is: select t.a, t.b, NULL, t.d from t)
				//do nothing
				if (emi.mappingDone(pos) && !emi.mappingNull(pos)) {
					if (debug)
						logger.log(BasicLevel.DEBUG, "    already done");
					continue;
				}
				// search the PEM, looking in super classes if necessary
				RdbPrimitiveElementMapping mappingOfpeInX = (RdbPrimitiveElementMapping) classMappingOfX
						.getPrimitiveElementMapping(pe.getName(), true);
				String tableName = "";
				if(prefetch && mappingOfpeInX == null){
					//if prefetch is true and the pe is not in the list of primitive elements for the class classX, 
					//put NULL for the name of the column in the select statement and nothing for the tableName: 
					//select a, b, NULL, d from...
					//is the class associated the current class (classX) or the class of the pe???
					emi.addProjection(tableName, RdbExtentMappingInfos.NULL_COLUMN, ColumnAliasing.getAliasFromColumn(pe.getName()), ((Class) pe.getParent()).getFQName()/*classX.getFQName()*/, pos);
				}
				else{
					// 	for classes mapped to several tables, the primitive element
					// may be mapped on an external table, requiring a join
					RdbJoin join = mappingOfpeInX.getJoinByPrimitiveElement(pe);
					if (join != null) {
						if (debug)
							logger.log(BasicLevel.DEBUG,"     mapped to external table");
						tableName = join.getExternalTable().getName();
						emi.addJoin(join);
					} else {
						if (debug)
							logger.log(BasicLevel.DEBUG,"     mapped to main table");
						emi.setMainTable((RdbTable) mappingOfpeInX.getParent());
						tableName = emi.getMainTable().getName();
					}
					emi.addProjection(tableName, mappingOfpeInX.getName(), ColumnAliasing.getAliasFromColumn(pe.getName()), ((Class) pe.getParent()).getFQName()/*classX.getFQName()*/, pos);
				}	
			}
		} // while !remClasses.isEmpty()

		// remove empty requests
		for (Iterator emiIt = extMappingInfos.iterator(); emiIt.hasNext();) {
			RdbExtentMappingInfos emi = (RdbExtentMappingInfos) emiIt.next();
			if (emi.isEmpty()) {
				emiIt.remove();
			}
		}

		return extMappingInfos;
	}
    
    /**
     * Returns the list of the primitive elements (hidden and non-hidden) for:
     * 	- the class classA , 
     * 	- all its parents and parents of parents...
     *  - all its subclasses and subclasses of subclasses... 
     */
    public ArrayList getAllPrimitiveElementsInGraph(Class classA, String projectName){
    	//get all the ancestor(s) of classA
    	Collection ancestors = classA.getAllAncestors();
    	if (ancestors == Collections.EMPTY_LIST)
    		ancestors = new ArrayList(1);
    	//add classA to this list
    	ancestors.add(classA);
    	//the list of primitive elements to return
    	ArrayList primEls = new ArrayList();
    	Iterator it = ancestors.iterator();
    	//for each class
    	while (it.hasNext()) {
    		Class currentClass = (Class) it.next();
    		boolean withSubclasses = false;
    		if (currentClass == classA)
    			withSubclasses = true;
    		//gets the list of pe for the class currentClass
    		// plus the list of pe for subclasses just in the case of the classA class
    		ArrayList primElsOfCurrent = getPrimitiveElements(withSubclasses, currentClass, projectName, null);
    		Iterator peIt = primElsOfCurrent.iterator();
    		while (peIt.hasNext()) {
    			PrimitiveElement pe = (PrimitiveElement) peIt.next();
    			if (!primEls.contains(pe)) {
    				//adds the pe if not already in the list
    				primEls.add(pe);
    			}
    		}
    	}
    	
    	return primEls;
    }
    
    /**
     * Returns the list of the primitive elements (hidden and non-hidden) for:
     * 	- the class classA if withSubtypes is false
     *  - the class classA and all its subclasses and subclasses of subclasses... if withSubtypes is true
     * This method also compute the list of primitive elements, only for classA and put it in the primElsOfA list if primElsOfA is not null
     */
    private ArrayList getPrimitiveElements(boolean withSubtypes, Class classA, String projectName, ArrayList primElsOfA){
    	//the list of primitive elements
    	ArrayList primEls = new ArrayList();
    	//the list of class to process
    	ArrayList remClasses = new ArrayList();
    	//add the first class classA
    	remClasses.add(classA);
    	
    	while(!remClasses.isEmpty()){
    		Class currentClass = (Class) remClasses.remove(0);
    		
    		//if prefetch is true, add the subclasses as classes to process
    		//else, only process classA
    		if(withSubtypes)
    			remClasses.addAll(currentClass.getSubClasses());
    		
    		NameDef ndef = currentClass.getNameDef(projectName);
    		if (debug)
    			logger.log(BasicLevel.DEBUG, "Primitive elements are: ");
    		for (Iterator itTE = currentClass.getAllFields().iterator(); itTE.hasNext();) {
    			Object te = itTE.next();
    			if (te instanceof PrimitiveElement) {
    				//add it only if it is not already in the list
    				if(!primEls.contains(te))
    					primEls.add(te);
    				if (debug)
    					logger.log(BasicLevel.DEBUG, "   " + ((PrimitiveElement) te).getName());
    			}
    		}
    		// 	add hidden fields
    		for (Iterator itTE = currentClass.getAllHiddenFields().iterator(); itTE.hasNext();) {
    			Object te = itTE.next();
    			if (te instanceof PrimitiveElement) {
    				//add it only if it is not already in the list
    				if(!primEls.contains(te))
    					primEls.add(te);
    				if (debug)
    					logger.log(BasicLevel.DEBUG, "   "	+ ((PrimitiveElement) te).getName() + " (hidden)");
    			}
    		}
    		
    		if(primElsOfA != null && currentClass == classA)
    			primElsOfA.addAll(primEls);
    	}
		return primEls;
    }
  
    /**
     * finds the parent's extent mapping info of a class in the inheritance hierarchy
     * @param clazz the clazz from wich searching for parents
     * @param EMIMap map of < clazz,extent mapping infos > into wich searching
     * @return
     */
    private RdbExtentMappingInfos getParentEMI(Class clazz, Map EMIMap, Class[] parent) {
        logger.log(BasicLevel.DEBUG, "Get parent's EMI");
        if (parent != null) {
            parent[0] = null;
        }
        RdbExtentMappingInfos emiOfParent = null;
        Collection parentsOfX = clazz.getSuperClasses();
        for (Iterator itP = parentsOfX.iterator(); itP.hasNext();) {
            Class superClass = (Class) itP.next();
            emiOfParent = (RdbExtentMappingInfos)
                    EMIMap.get(superClass);
            if (emiOfParent != null) {
                if (parent != null) {
                    logger.log(BasicLevel.DEBUG, "Return parent=" + superClass.getFQName());
                    parent[0] = superClass;
                }
                return emiOfParent;
            }
        }
        logger.log(BasicLevel.DEBUG, "Found none");
        return null;
    }
   
}