/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.rdb.lib;

import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.api.PMappingStructuresManager;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.api.PMapCluster;
import org.objectweb.util.monolog.api.Logger;

import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Collection;

/**
 * Defines the way to manage relational schema for the RDB mappers.
 * @author	P. Dechamboux
 */
public class RdbPMappingStructuresManager implements PMappingStructuresManager {
	private PMapper pMapper = null;
	private ArrayList clusterList = new ArrayList();
	protected Logger logger = null;
	
	/**
	 * Closes a mapper connection.
	 * @param conn	The mapper connection.
	 * @throws PException
	 */
	void closeConnection(Object conn) throws PException {
		pMapper.closeConnection(conn);
	}

	/**
	 * Gets a connection from the mapper.
	 * @return The allocated connection.
	 * @throws PException
	 */
	Object getConnection() throws PException {
		Object conn = pMapper.getConnection();
		if (conn == null)
			throw new PExceptionProtocol("Impossible to initialize the mapping structures without connection");
		return conn;
	}

	/**
	 * Retrieves or creates an RdbPMapCluster associated with this JORM class.
	 * @param jcname	The name of this JORM class.
	 * @param tname		A table name within which the class is mapped.
	 * @return The associated RdbPMapCluster.
	 */
	private RdbPMapCluster getAlwaysPMapCluster(String jcname,
												String tname)
			throws PException {
		RdbPMapCluster cl = null;
		Iterator it = clusterList.iterator();
		while (it.hasNext()) {
			cl = (RdbPMapCluster) it.next();
			if (cl.containTable(tname)) {
				break;
			}
			cl = null;
		}
		if (cl == null) {
			// no map cluster use tname
			cl = (RdbPMapCluster) getPMapCluster(jcname);
			if (cl == null) {
				// no map cluster for jcname : creates a new one
				cl = new RdbPMapCluster(jcname, tname, this);
				clusterList.add(cl);
			}
			// there is already a map cluster for jcname
		} else {
			// there is already a map cluster for tname
			RdbPMapCluster cl2 = (RdbPMapCluster) getPMapCluster(jcname);
			if ((cl != cl2) && (cl2 != null)) {
				// there is also a map cluster for jcname : merge both into cl
				cl.merge(cl2);
				clusterList.remove(cl2);
			} else {
				// no map cluster for jcname : declare jcname into cl
				cl.classDefined(jcname);
			}
		}
		return cl;
	}

	/**
	 * Adds a table column into these RDB mapping structures.
	 * @param jcname    The JORM class name.
	 * @param tname     The table name.
	 * @param cname     The column name.
	 * @param notnull   true if column value should never be null.
	 * @throws PException
	 */
	synchronized public void addTableColumn(String jcname,
											String tname,
											String cname,
											String type,
											boolean notnull,
											boolean ispkcol,
											boolean isMaster) throws PException {
		RdbPMapCluster cl = getAlwaysPMapCluster(jcname, tname);
		cl.addTableColumn(tname, cname, type, notnull, ispkcol, isMaster);
	}


	public synchronized void declareClass(String jcname) {
		Iterator it = clusterList.iterator();
		RdbPMapCluster cl;
		while (it.hasNext()) {
			cl = (RdbPMapCluster) it.next();
			if (cl.containClass(jcname)) {
				return;
			}
		}
		cl = new RdbPMapCluster(jcname, this);
		clusterList.add(cl);
	}

    public PMapper getPMapper() {
        return pMapper;
    }

	// IMPLEMENTATION OF METHODS FROM THE PMappingStructuresManager INTERFACE

	/**
	 * Asks for the map cluster associated with a JORM class.
	 * @return The relevant map cluster.
	 * @throws PException   This class has not been mapped.
	 */
	synchronized public PMapCluster getPMapCluster(String jcname) throws PException {
		Iterator it = clusterList.iterator();
		while (it.hasNext()) {
			RdbPMapCluster cl = (RdbPMapCluster) it.next();
			if (cl.containClass(jcname)
					|| cl.getUnResolvedDependencies().contains(jcname)) {
				return cl;
			}
		}
		return null;
	}

	/**
	 * Asks for all the map clusters defined within this mapper. The returned
	 * Collection is not thread safe.
	 * @return The Iterator over the collection of map clusters.
	 */
	public Collection getPMapClusters() {
		return clusterList;
	}

	synchronized public void addDependency(String jcname1, String jcname2)
			throws PException {
		RdbPMapCluster cl1 = (RdbPMapCluster) getPMapCluster(jcname1);
		if (cl1 == null) {
			throw new PException("You must define the class " + jcname1
								 + " before adding its dependcies (no cluster found).");
		}
		RdbPMapCluster cl2 = (RdbPMapCluster) getPMapCluster(jcname2);
		if (cl2 == null) {
			cl1.addDependency(jcname2);
		} else if (cl2 != cl1) {
			cl1.merge(cl2);
			clusterList.remove(cl2);
		}
	}

	public void classDefined(String jcname) throws PException {
		RdbPMapCluster cl = (RdbPMapCluster) getPMapCluster(jcname);
		if (cl == null) {
			throw new PException("No cluster found for the class " + jcname);
		}
		cl.classDefined(jcname);
	}

	/**
	 * Assigns a mapper to this schema manager.
	 * @param pm	The concerned mapper.
	 */
	public void setPMapper(PMapper pm) {
		pMapper = pm;
	}

	/**
	 * Assigns a logger to this schema manager.
	 * @param l		The concerned logger.
	 */
	public void setLogger(Logger l) {
		logger = l;
	}
	
	public void clear() throws PException {
		clusterList.clear();
	}
}
