/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.rdb.lib;

import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.expression.api.Operand;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.medor.eval.prefetch.api.IntermediaryPrefetchBuffer;
import org.objectweb.medor.eval.prefetch.api.PrefetchBuffer;
import org.objectweb.medor.eval.prefetch.api.PrefetchBufferFactory;
import org.objectweb.medor.eval.prefetch.api.PrefetchBufferHolder;
import org.objectweb.medor.eval.prefetch.lib.IntermediaryPrefetchBufferImpl;
import org.objectweb.medor.eval.prefetch.lib.PrefetchBufferFactoryImpl;
import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapter;
import org.objectweb.jorm.mapper.rdb.lib.PMapperRdb;
import org.objectweb.jorm.mapper.rdb.lib.PClassMappingDecoder;
import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PPrefetchTupleCollection;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.sql.ResultSet;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.Date;

/**
 * Is a TupleCollection encaspulating a prefetched buffer.
 *
 * @author	P. Dechamboux
 */
public class RdbTupleCollection
		implements PPrefetchTupleCollection, Tuple, PrefetchBufferHolder {

	/**
	 * The inner SQL resultset containg the PName and the other field values
	 */
	private ResultSet resultSet;

	/**
	 * the PreparedStatement to close when the collection will be no more used
	 */
	private PreparedStatement preparedStatement;

	/**
	 * The index of the PName in the current TupleCollection;
	 */
	private int pnameIndex;

	/**
	 * The RdbAdapter to access the resultset.
	 */
	private RdbAdapter rdbAdapter;

	/**
	 * The inner prefetch buffer to fill when the TupleCollection is read
	 */
	private PrefetchBuffer prefetchBuffer;

	/**
	 * The PClassMapping of the persistent class which is prefetched.
	 */
	private PClassMapping pclassMapping;

	/**
	 * The PNameGetter able to build the PName of the persistent class which is
	 * prefetched.
	 */
	private Object pnameGetter;

	/**
	 * A factory of prefetched buffer
	 */
	private static PrefetchBufferFactory pbf = new PrefetchBufferFactoryImpl();

	/**
	 * The row index in the ResultSet / TupleCollection
	 */
	private int pos = 0;

	private Logger logger;

	public RdbTupleCollection(Object txctx,
							  ResultSet rs,
							  PreparedStatement ps,
							  int pi,
							  Logger l,
							  PClassMapping pcm,
							  boolean withSubType,
							  Object png)
			throws MedorException, SQLException {
		logger = l;
		logger.log(BasicLevel.DEBUG, "TupleCollection creation");
		resultSet = rs;
		preparedStatement = ps;
		pnameIndex = pi;
		rdbAdapter = ((PMapperRdb) pcm.getPMapper()).getRdbAdapter();
		pclassMapping = pcm;
		pnameGetter = png;
		boolean isPolymorphic = (pcm instanceof RdbPPolymorphicClass) && withSubType;
		prefetchBuffer = pcm.getPMapper().getPrefetchCache().createPrefetchBuffer(
				pbf, pcm, txctx, pi, !isPolymorphic);
		prefetchBuffer.setTupleCollection(this);
		//if the class is polymorphic, create intermediary prefetch buffer
		if (isPolymorphic) {
			IntermediaryPrefetchBuffer ipb = new IntermediaryPrefetchBufferImpl(
										prefetchBuffer,
										pcm.getIndexesTable(pcm));
			if (!pcm.getPMapper().getPrefetchCache().registerPrefetchBuffer(
					ipb, pcm, txctx)) {
				throw new MedorException("No prefetch buffer registered for the pcm " + pcm.getClassName() + ".");
			}
			try{
				//get the list of all the sub pclassmappings
				PClassMapping[] subPCMs = pcm.getSubPCMs();
				if(subPCMs != null){
					for (int i = 0; i < subPCMs.length; i++) {
						//create an intermediary prefetchBuffer that has the prefetchBuffer
						//of the superclass as delegate prefetch buffer
						//and the association table for the pcm of the superclass
						IntermediaryPrefetchBuffer intermediaryPb = new IntermediaryPrefetchBufferImpl(
																prefetchBuffer,
																subPCMs[i].getIndexesTable(pcm));
						if (!subPCMs[i].getPMapper().getPrefetchCache().registerPrefetchBuffer(
								intermediaryPb, subPCMs[i], txctx)) {
							throw new MedorException("No prefetch buffer registered for the pcm " + subPCMs[i].getClassName() + ".");
						}
					}
				}	
			}
			catch(Exception e){
				throw new MedorException("Error while trying to register the prefetchBuffer with the subclasses. ", e);
			}
		}
	}

	public PrefetchBuffer getPrefetchBuffer() {
		return prefetchBuffer;
	}

	public int getPNameIndex() {
		return pnameIndex;
	}

	public TupleStructure getMetaData() throws MedorException {
		return null;
	}

	public boolean isLast() throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		logger.log(BasicLevel.DEBUG, "");
		try {
			return resultSet.isLast();
		} catch (SQLException sqlex) {
			throw new MedorException(sqlex);
		}
	}

	public void close() throws MedorException {
		if (resultSet != null) {
			try {
				resultSet.close();
				preparedStatement.close();
			} catch (SQLException e) {
				throw new MedorException("Impossible to close the result set: ", e);
			} finally {
				resultSet = null;
				preparedStatement = null;
			}
		}
	}

	/**
	 * Moves the cursor down one row from its
	 * current position to the next tuple
	 */
	public boolean next() throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		logger.log(BasicLevel.DEBUG, "");
		try {
			if (resultSet.next()) {
				pos++;
				logger.log(BasicLevel.DEBUG, "resultset.next()=true");
				return true;
			} else {
			    pos = resultSet.getRow();
				logger.log(BasicLevel.DEBUG, "resultset.next()=false");
				return false;
			}
		} catch (SQLException sqlex) {
			throw new MedorException(sqlex);
		}
	}

	/**
	 * Moves the cursor to the first Tuple of this TupleCollection
	 */
	public void first() throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		logger.log(BasicLevel.DEBUG, "");
		try {
			if (resultSet.first()) {
				pos = 1;
				logger.log(BasicLevel.DEBUG, "resultset.previous()=true");
			} else
				throw new MedorException("can't move the cursor to the first row");
		} catch (SQLException ex) {
			throw new MedorException(ex);
		}
	}

	/**
	 * Retrieves the current row Number of this TupleCollection
	 */
	public int getRow() throws MedorException {
		return pos;
	}

	public Tuple getTuple() throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		logger.log(BasicLevel.DEBUG, "");
		return this;
	}

	public Tuple getTuple(int row) throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		row(row);
		return this;
	}

	public boolean isEmpty() throws MedorException {
		throw new MedorException("No need to support it here!!");
	}

	public boolean row(int i) throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		try {
			logger.log(BasicLevel.DEBUG, "resultSet.absolute(" + i + ")");
			if (pos == i) {
				return true;
			}
			if (i == 1) {
				first();
				return true;
			}
			if (pos == (i - 1)) {
				return next();
			}
			if (resultSet.absolute(i)) {
				pos = i;
				return true;
			} else
				return false;
		} catch (SQLException sqlex) {
			throw new MedorException(sqlex);
		}
	}

	// IMPLEMENTATION OF METHODS OF THE Tuple INTERFACE

	/**
	 * Give the number of the attribute in this Tuple object.
	 * @return an int
	 */
	public int getSize() {
		try {
			return resultSet.getMetaData().getColumnCount() + 1;
		} catch (SQLException e) {
			return -1;
		}
	}

	/**
	 * Give the array of Operand object representation of the Tuple.
	 * @return an array of Operand type.
	 */
	public Operand[] toOperandArray() {
		return null;
	}

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple
	 * as a Boolean of java programming language
	 */
	public boolean getBoolean(int i) throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		logger.log(BasicLevel.DEBUG, "numColumn=" + i);
		try {
			return rdbAdapter.getBoolean(resultSet, (i >= pnameIndex) ? i - 1 : i, false);
		} catch (SQLException e) {
			throw new MedorException(e);
		}
	}

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple as a Java byte
	 */
	public byte getByte(int i) throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		logger.log(BasicLevel.DEBUG, "numColumn=" + i);
		try {
			return rdbAdapter.getByte(resultSet, (i >= pnameIndex) ? i - 1 : i, (byte) 0);
		} catch (SQLException e) {
			throw new MedorException(e);
		}
	}

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple as a Java byte
	 */
	public byte[] getByteArray(int i) throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		logger.log(BasicLevel.DEBUG, "numColumn=" + i);
		try {
			return rdbAdapter.getByteArray(resultSet, (i >= pnameIndex) ? i - 1 : i, null);
		} catch (SQLException e) {
			throw new MedorException(e);
		}
	}

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple as a Java char
	 */
	public char getChar(int i) throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		logger.log(BasicLevel.DEBUG, "numColumn=" + i);
		try {
			return rdbAdapter.getChar(resultSet, (i >= pnameIndex) ? i - 1 : i, (char) 0);
		} catch (SQLException e) {
			throw new MedorException(e);
		}
	}

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple as a Java char[]
	 */
	public char[] getCharArray(int i) throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		logger.log(BasicLevel.DEBUG, "numColumn=" + i);
		try {
			return rdbAdapter.getCharArray(resultSet, (i >= pnameIndex) ? i - 1 : i, null);
		} catch (SQLException e) {
			throw new MedorException(e);
		}
	}

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple as a java.sql.Data
	 */
	public Date getDate(int i) throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		logger.log(BasicLevel.DEBUG, "numColumn=" + i);
		try {
			return rdbAdapter.getDate(resultSet, (i >= pnameIndex) ? i - 1 : i, null);
		} catch (SQLException e) {
			throw new MedorException(e);
		}
	}

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple as a Java double
	 */
	public double getDouble(int i) throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		logger.log(BasicLevel.DEBUG, "numColumn=" + i);
		try {
			return rdbAdapter.getDouble(resultSet, (i >= pnameIndex) ? i - 1 : i, ((double) 0.0));
		} catch (SQLException e) {
			throw new MedorException(e);
		}
	}

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple as a Java float
	 */
	public float getFloat(int i) throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		logger.log(BasicLevel.DEBUG, "numColumn=" + i);
		try {
			return rdbAdapter.getFloat(resultSet, (i >= pnameIndex) ? i - 1 : i, ((float) 0.0));
		} catch (SQLException e) {
			throw new MedorException(e);
		}
	}

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple as a java int
	 */
	public int getInt(int i) throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		logger.log(BasicLevel.DEBUG, "numColumn=" + i);
		try {
			return rdbAdapter.getInt(resultSet, (i >= pnameIndex) ? i - 1 : i, -1);
		} catch (SQLException e) {
			throw new MedorException(e);
		}
	}

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple
	 * as an org.objectweb.medor.filter.Operand Object
	 */
	public Operand getLikeOperand(int i) throws MedorException {
		return null;
	}

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple as a Java long
	 */
	public long getLong(int i) throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		logger.log(BasicLevel.DEBUG, "numColumn=" + i);
		try {
			return rdbAdapter.getLong(resultSet, (i >= pnameIndex) ? i - 1 : i, ((long) -1));
		} catch (SQLException e) {
			throw new MedorException(e);
		}
	}

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple
	 * as an Object in the java programming language
	 */
	public Object getObject(int i) throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		logger.log(BasicLevel.DEBUG, "numColumn=" + i);
		try {
			if (i == pnameIndex) {
				// should return the PName
				return ((PClassMappingDecoder) pclassMapping).decode(resultSet, pnameIndex, pnameGetter);
			} else {
				Object o = resultSet.getObject((i >= pnameIndex) ? i - 1 : i);
				return (resultSet.wasNull() ? null : o);
			}
		} catch (SQLException sqlex) {
			throw new MedorException(sqlex);
		} catch (Exception e) {
			throw new MedorException("Invalid field index " + i, e);
		}
	}

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple
	 * as a Short in the java programming language
	 */
	public short getShort(int i) throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		logger.log(BasicLevel.DEBUG, "numColumn=" + i);
		try {
			return rdbAdapter.getShort(resultSet, (i >= pnameIndex) ? i - 1 : i, (short) -1);
		} catch (SQLException e) {
			throw new MedorException(e);
		}
	}

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple
	 * as a BigDecimal in the java programming language
	 */
	public BigDecimal getBigDecimal(int i) throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		logger.log(BasicLevel.DEBUG, "numColumn=" + i);
		try {
			return rdbAdapter.getBigDecimal(resultSet, (i >= pnameIndex) ? i - 1 : i, null);
		} catch (SQLException e) {
			throw new MedorException(e);
		}
	}

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple
	 * as a BigDecimal in the java programming language
	 */
	public BigInteger getBigInteger(int i) throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		logger.log(BasicLevel.DEBUG, "numColumn=" + i);
		try {
			BigDecimal bd = rdbAdapter.getBigDecimal(resultSet, (i >= pnameIndex) ? i - 1 : i, null);
			return (bd == null) ? null : bd.toBigInteger();
		} catch (SQLException e) {
			throw new MedorException(e);
		}
	}

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple
	 * as a String in java programming language
	 */
	public String getString(int i) throws MedorException {
		if (resultSet == null) {
			throw new MedorException("Impossible to use a closed TupleCollection");
		}
		logger.log(BasicLevel.DEBUG, "numColumn=" + i);
		try {
			return rdbAdapter.getString(resultSet, (i >= pnameIndex) ? i - 1 : i, null);
		} catch (SQLException e) {
			throw new MedorException(e);
		}
	}

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple
	 * as a org.objectweb.medor object
	 */
	public TupleCollection getTupleCollection(int i) throws MedorException {
		return this;
	}

	public boolean isDefined(int i) {
		try {
			return !resultSet.wasNull();
		} catch (SQLException e) {
			return false;
		}
	}

	// IMPLEMENTATION OF METHODS OF THE PrefetchBufferHolder INTERFACE

	public void invalidatePrefetchBuffer() throws MedorException {
		logger.log(BasicLevel.DEBUG, "Invalidating PrefetchBuffer " + prefetchBuffer + " for " + this);
		if (prefetchBuffer != null) {
			this.close();
			prefetchBuffer = null;
		} else {
			logger.log(BasicLevel.DEBUG, "The tupleCollection was closed earlier: closing it for real");
		}
	}
}
