/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2004 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm@objectweb.org
 *
 */
package org.objectweb.jorm.mapper.rdb.metainfo;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.lib.BasicMetaObject;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.Operand;
import org.objectweb.medor.expression.api.Operator;
import org.objectweb.medor.expression.lib.BasicOperand;
import org.objectweb.medor.expression.lib.BasicParameterOperand;
import org.objectweb.medor.expression.lib.Equal;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.HashMap;

public class RdbFilter extends BasicMetaObject {

	/**
	 * This map associates a list of Predicate objects to a filter column.
	 * key: the name of the filter column
	 * value: a list of Predicate objects
	 */
	//HashMap filtercol2predicates = null;

	/**
	 * This list stores the names of PrimitiveElements which take part in the
	 * filter.
	 */
	HashMap filteringElements = null;

	/**
	 * The initial string parsed to construct the Expression
	 */
	private String stringExp = null;

	/**
	 * The MEDOR Expression constructed using the input String
	 */
	private Expression exp = null;

	/**
	 * Builds a new RdbFilter object.
	 *
	 * This object encapsulates information about filter columns.
	 * The parent object is a RdbClassMapping object.
	 *
	 * @param parent	the parent of the current object.
	 */
	public RdbFilter(MetaObject parent) {
		super(parent);
		filteringElements = new HashMap();
	}

	/**
	 * Indicates if a PrimitiveElement is part of the filter.
	 * @param pe   a PrimitiveElement.
	 * @return true, if pe is a part of the filter, else false.
	 */
	public boolean isFilter(PrimitiveElement pe) {
		logger.log(BasicLevel.DEBUG, "Looking whether PrimitiveElement " + pe.getName() + " is in the filter");
		return filteringElements.containsKey(pe.getName());
	}

	/**
	 * Returns the predicate value associated with the filter PrimitiveElement
	 * name.
	 * If the predicate does not correspond to an equal operator, a PException is raised.
	 * @param elementName   a filter column name.
	 * @return the predicate value associated with the filter PrimitiveElement.
	 */
	public String getEqualPredicateValue(String elementName) throws PException {
		String predValue = (String) filteringElements.get(elementName);
		if (predValue == null) {
			throw new PException("No fixed value can be given for this PrimitiveElement " +
					elementName + " although it is a hidden filter field!");
		}
		return predValue;
	}

	public void setExpression(Expression e) {
		exp = e;
		/* parse the expression to add the PrimitiveElement to the list of
		  filtering elements
		*/
		addElementsToFilter(e);
	}

	private void addElementsToFilter(Expression e) {
		if (e instanceof BasicParameterOperand) {
			filteringElements.put(((BasicParameterOperand) e).getName(), null);
		} else if (e instanceof Operand) {
			//do nothing
		} else if (e instanceof Equal) {
			Equal eq = (Equal) e;
			Expression eq0 = eq.getExpression(0);
			Expression eq1 = eq.getExpression(1);
			if (eq0 instanceof BasicParameterOperand) {
				if (eq1 instanceof BasicOperand) {
					filteringElements.put(
							((BasicParameterOperand) eq0).getName(),
							((BasicOperand) eq1).getObject()
					);
				} else {
					filteringElements.put(((BasicParameterOperand) eq0).getName(), null);
					addElementsToFilter(eq1);
				}
			} else if (eq1 instanceof BasicParameterOperand) {
				if (eq0 instanceof BasicOperand) {
					filteringElements.put(
							((BasicParameterOperand) eq1).getName(),
							((BasicOperand) eq0).getObject()
					);
				} else {
					filteringElements.put(((BasicParameterOperand) eq1).getName(), null);
					addElementsToFilter(eq0);
				}
			} else {
				addElementsToFilter(eq.getExpression(0));
				addElementsToFilter(eq.getExpression(1));
			}
		} else if (e instanceof Operator) {
			Operator op = (Operator) e;
			for (int i = 0; i < op.getOperandNumber(); i++) {
				addElementsToFilter(op.getExpression(i));
			}
		}
	}

	public Expression getExpression() {
		return exp;
	}

	public void setStringExpression(String e) {
		stringExp = e;
	}

	public String getStringExpression() {
		return stringExp;
	}
}
