/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.rdb.metainfo;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;

import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.mapper.rdb.adapter.api.JoinedTable;
import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapter;
import org.objectweb.jorm.mapper.rdb.lib.PMapperRdb;
import org.objectweb.jorm.mapper.rdb.lib.RdbExtentGenInfos;
import org.objectweb.jorm.mapper.rdb.lib.RdbExtentMappingInfos;
import org.objectweb.jorm.mapper.rdb.lib.RdbPPolymorphicClass;
import org.objectweb.jorm.mapper.rdb.util.ColumnAliasing;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ClassProject;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.api.PrimitiveElementMapping;
import org.objectweb.jorm.metainfo.lib.BasicMetaObject;

/**
 * This class stores two queries in String format:
 * 	a query with prefetch
 * 	a query without prefetch
 * These queries retrieve the Extent of a class in case of inheritance.
 * The list of fields to select is also stored as a collection of String. 
 * @author Y.Bersihand
 */
public class RdbInheritanceQuery extends BasicMetaObject {

	/**
     * 
     */
    private static final long serialVersionUID = 1L;

    //index for the pk fields: shared by all the RdbInheritanceQuery instances
	private static int pkIndex = 0;
	
	public static final String PK = "pk";
	
	private RdbClassMapping rdbClassMapping;
	//the query with prefetch
	private String queryPrefetch;
	//the query with the PK only
	private String queryPKOnly;
	//the query with the PK + the fields of the class and its super classes
	private String queryPKAndFields;
	//the list of fields to select
	private Collection fieldsPrefetch;
	private Collection fieldsPKOnly;
	private Collection fieldsPKAndFields;
	//the list if class associated to the fields in the fieldsPrefetch list
	private Collection prefetchClasses;
	
	public RdbInheritanceQuery(MetaObject parent) {
		super(parent);
		rdbClassMapping = (RdbClassMapping) parent;
		queryPrefetch = "";
		queryPKOnly = "";
		queryPKAndFields = "";
		fieldsPrefetch = new ArrayList();
		fieldsPKOnly = new ArrayList();
		fieldsPKAndFields = new ArrayList();
		prefetchClasses = new ArrayList();
	}
	
	/**
	 * Returns the query  of the extent as a String. 
	 */
	public String getExtentQuery(PMapper mapper, boolean prefetch, boolean pkOnly, String primaryKey, PClassMapping pcm){
		if (prefetch) { 
			return getQueryPrefetch(mapper, primaryKey, pcm);
		} else {
			if (pkOnly) {
				return getQueryPKOnly(mapper, primaryKey, pcm);
			} else {
				return getQueryPKAndFields(mapper, primaryKey, pcm);
			}
		}
	}
	
	public Collection getPKfield() {
	    return fieldsPKOnly;
	}
	
	/**
	 * Returns the list of fields to be selected. 
	 */
	public Collection getExtentFieldsNames(boolean prefetch, boolean pkOnly){
		if (prefetch) {
			return fieldsPrefetch;
		} else {
			if (pkOnly) {
				return fieldsPKOnly;
			} else {
				return fieldsPKAndFields;
			}
		}
	}
	
	/**
	 * Adds all the fields of the collection c to the fieldsPKOnly list.
	 * If one the fields is already present in the list, it is not added a second time.
	 */
	public void addFieldPKOnly(Collection c){
		Iterator itC = c.iterator();
		while (itC.hasNext()) {
			String s = (String) itC.next();
			if (!fieldsPKOnly.contains(s))
				fieldsPKOnly.add(s);
		}
	}
	
	/**
	 * Adds all the fields of the collection fields to the fieldsPrefetch list.
	 * If one the fields is already present in the list, it is not added a second time.
	 * In the same time, classes associated to these added fields are added to the prefetchClasses list.
	 */
	public void addFieldPrefetch(Collection fields, Collection classes){
		Iterator itF = fields.iterator();
		Iterator itC = classes.iterator();
		while (itF.hasNext()) {
			String field = (String) itF.next();
			String className = (String) itC.next();
			if (!fieldsPrefetch.contains(field)) {
				//add the field
				fieldsPrefetch.add(field);
				//add the class associated
				prefetchClasses.add(className);
			}
		}
	}
	
	/**
	 * Adds all the fields of the collection c to the fieldsPKAndFields list.
	 * If one the fields is already present in the list, it is not added a second time.
	 */
	public void addFieldPKAndFields(Collection c){
		Iterator itC = c.iterator();
		while (itC.hasNext()) {
			String s = (String) itC.next();
			if (!fieldsPKAndFields.contains(s))
				fieldsPKAndFields.add(s);
		}
	}
	
	/**
	 * 
	 * @param mapper
	 * @param primaryKey: a string representing the list of columns composing the primary key
	 * @return
	 */
	private String getQueryPKOnly(PMapper mapper, String primaryKey, PClassMapping pcm){
		if(queryPKOnly.equals("")){
			//compute query
			RdbExtentGenInfos extentGenInfos = new RdbExtentGenInfos(logger);
			
			RdbAdapter adapter;
			try {
				adapter = ((PMapperRdb) mapper).getRdbAdapter();
				if (adapter == null) {
					throw new PException("No rdb adapter found on the mapper: " + mapper);
				}
				//Calculate the extent structure (a set of ExtentMappingInfos)
				//each one corresponding to a select/from/where statement plus filters
				Collection emis = extentGenInfos.getPolymorphicClassExtent((Class)rdbClassMapping.getLinkedMO(), rdbClassMapping.getProjectName(), rdbClassMapping.getMapperName(), false);
				//loop on extent mapping info objects
	        	JoinedTable jt = null;
	        	JoinedTable.Join joinCond = null;
	        	Iterator itEmis = emis.iterator();
	        	while (itEmis.hasNext()) {
	        		RdbExtentMappingInfos emi = (RdbExtentMappingInfos) itEmis.next();
	        		//add the aliases to the list of fields (belonging to the primary key only) to select 
	        		addFieldPKOnly(emi.getColumnAliases(false, primaryKey, null));
	        		//create the JoinedTable object to pass to the getQuery
	        		jt = new JoinedTable(emi.getMainTable().getName());
	        		HashSet hsJoins = emi.getJoins();
	        		Iterator itJoins = hsJoins.iterator();
	        		while (itJoins.hasNext()) {
	        			RdbJoin rdbJoin = (RdbJoin) itJoins.next();
	        			joinCond = jt.createChildren(rdbJoin.getExternalTable().getName());
	        			int pos = 0;
	        			Iterator itCName = rdbJoin.getPTJoinColumnNames().iterator();
	        			while (itCName.hasNext()) {
	        				String cName = (String) itCName.next(); 
	        				joinCond.addJoinColumn(cName, (String) rdbJoin.getETJoinColumnNames().get(pos));
	        				pos++;
	        			}
	        		}
	        		//select x1 as y1, x2 as y2, x3 as y3 from T
	        		queryPKOnly += adapter.getQuery(emi.getSelectPKOnly(adapter, primaryKey), Collections.singletonList(jt), null, false, false);
	        		//where x1='filter'  
	        		queryPKOnly += emi.getWhereParameter(pcm);
	        		if(itEmis.hasNext())
	        			queryPKOnly += " UNION ";
	        	}
	        	//System.out.println("fields pk only=" + fieldsPKOnly.toString());
			} catch (Exception e) {
				e.printStackTrace();
	        }
		}
		return queryPKOnly;
	}
	
	/**
	 * 
	 * @param mapper
	 * @param primaryKey: a string representing the list of columns composing the primary key
	 * @return
	 */
	private String getQueryPKAndFields(PMapper mapper, String primaryKey, PClassMapping pcm){
		if(queryPKAndFields.equals("")){
			//compute query
			RdbExtentGenInfos extentGenInfos = new RdbExtentGenInfos(logger);
			
			RdbAdapter adapter;
			try {
				adapter = ((PMapperRdb) mapper).getRdbAdapter();
				if (adapter == null) {
					throw new PException("No rdb adapter found on the mapper: " + mapper);
				}
				//Calculate the extent structure (a set of ExtentMappingInfos)
				//each one corresponding to a select/from/where statement plus filters
				Collection emis = extentGenInfos.getPolymorphicClassExtent((Class)rdbClassMapping.getLinkedMO(), rdbClassMapping.getProjectName(), rdbClassMapping.getMapperName(), false);
				//loop on extent mapping info objects
	        	JoinedTable jt = null;
	        	JoinedTable.Join joinCond = null;
	        	Iterator itEmis = emis.iterator();
	        	while (itEmis.hasNext()) {
	        		RdbExtentMappingInfos emi = (RdbExtentMappingInfos) itEmis.next();
	        		//add the aliases to the list of fields (belonging to the primary key only) to select 
	        		addFieldPKAndFields(emi.getColumnAliases(false, primaryKey, (Class)rdbClassMapping.getLinkedMO()));
	        		//create the JoinedTable object to pass to the getQuery
	        		jt = new JoinedTable(emi.getMainTable().getName());
	        		HashSet hsJoins = emi.getJoins();
	        		Iterator itJoins = hsJoins.iterator();
	        		while (itJoins.hasNext()) {
	        			RdbJoin rdbJoin = (RdbJoin) itJoins.next();
	        			joinCond = jt.createChildren(rdbJoin.getExternalTable().getName());
	        			int pos = 0;
	        			Iterator itCName = rdbJoin.getPTJoinColumnNames().iterator();
	        			while(itCName.hasNext()){
	        				String cName = (String) itCName.next(); 
	        				joinCond.addJoinColumn(cName, (String) rdbJoin.getETJoinColumnNames().get(pos));
	        				pos++;
	        			}
	        		}
	        		//select x1 as y1, x2 as y2, x3 as y3 from T
	        		queryPKAndFields += adapter.getQuery(emi.getSelectPKAndFields(adapter, (Class)rdbClassMapping.getLinkedMO()), Collections.singletonList(jt), null, false, false);
	        		//where x1='filter'  
	        		queryPKAndFields += emi.getWhereParameter(pcm);
	        		if (itEmis.hasNext())
	        			queryPKAndFields += " UNION ";
	        	}
	        	//System.out.println("fields pk and fields=" + fieldsPKAndFields.toString());
			} catch (Exception e) {
				e.printStackTrace();
	        }
		}
		return queryPKAndFields;
	}
	
	/**
	 * 
	 * @param mapper
	 * @param primaryKey : a string representing the list of columns composing the primary key
	 * @return
	 */
	private String getQueryPrefetch(PMapper mapper, String primaryKey, PClassMapping pcm){
		if(queryPrefetch.equals("")){
			//compute query
			RdbExtentGenInfos extentGenInfos = new RdbExtentGenInfos(logger);
			
			RdbAdapter adapter;
			try {
				adapter = ((PMapperRdb) mapper).getRdbAdapter();
				if (adapter == null) {
					throw new PException("No rdb adapter found on the mapper: " + mapper);
				}
				//Calculate the extent structure (a set of ExtentMappingInfos)
				//each one corresponding to a select/from/where statement plus filters
				Collection emis = extentGenInfos.getPolymorphicClassExtent((Class)rdbClassMapping.getLinkedMO(), rdbClassMapping.getProjectName(), rdbClassMapping.getMapperName(), true);
				//loop on extent mapping info objects
	        	JoinedTable jt = null;
	        	JoinedTable.Join joinCond = null;
	        	Iterator itEmis = emis.iterator();
	        	while (itEmis.hasNext()) {
	        		RdbExtentMappingInfos emi = (RdbExtentMappingInfos) itEmis.next();
	        		//add the aliases to the list of fields to select and the classes associated to the list prefetchClasses 
	        		addFieldPrefetch(emi.getColumnAliases(true, "", null), emi.getColumnClasses());
	        		//add the pk aliases to the list of pk fields 
	        		addFieldPKOnly(emi.getColumnAliases(false, primaryKey, null));
	        		//create the JoinedTable object to pass to the getQuery
	        		jt = new JoinedTable(emi.getMainTable().getName());
	        		HashSet hsJoins = emi.getJoins();
	        		Iterator itJoins = hsJoins.iterator();
	        		while (itJoins.hasNext()) {
	        			RdbJoin rdbJoin = (RdbJoin) itJoins.next();
	        			joinCond = jt.createChildren(rdbJoin.getExternalTable().getName());
	        			int pos = 0;
	        			Iterator itCName = rdbJoin.getPTJoinColumnNames().iterator();
	        			while (itCName.hasNext()) {
	        				String cName = (String) itCName.next(); 
	        				joinCond.addJoinColumn(cName, (String) rdbJoin.getETJoinColumnNames().get(pos));
	        				pos++;
	        			}
	        		}
	        		//select x1 as y1, x2 as y2, x3 as y3 from T
	        		String currentQuery = adapter.getQuery(
	        		        emi.getSelectPrefetch(adapter, rdbClassMapping.getJormClass()), 
	        		        Collections.singletonList(jt), null, false, false);
	        		//add the PName fields in end the of the columns to select,
	        		//that is to say between the last column and the FROM key word
	        		int fromIndex = currentQuery.toUpperCase().indexOf(" FROM ");
	        		String tmpQuery = currentQuery.substring(0, fromIndex);
                    //get the pk field names from PClassMapping corresponding
                    // the RdbExtentMappingInfos
                    PClassMapping currentPCM = mapper.lookup(emi.getClazz().getFQName());
                    String pkString;
                    if (currentPCM instanceof RdbPPolymorphicClass) {
                        pkString = ((RdbPPolymorphicClass) currentPCM).getPNameFields();
                    } else  if (currentPCM == pcm) {
                        pkString = primaryKey;
                    } else {
                        throw new RuntimeException("Impossible to compute the inheritance query with prefetch of the persistent class " + pcm.getClassName());
                    }
                    
	        		tmpQuery += ", " + aliasPK(pkString, adapter) + " ";
	        		tmpQuery += currentQuery.substring(fromIndex);
                    currentQuery = tmpQuery;
	        		//either WHERE x1='filter' or AND x1='filter' 
                    currentQuery += emi.getWhereParameter(pcm);
                    queryPrefetch += currentQuery;
	        		if (itEmis.hasNext())
	        			queryPrefetch += " UNION ";
	        	}
	        	//System.out.println("fields with prefetch=" + fieldsPrefetch.toString());
			} catch (Exception e) {
				e.printStackTrace();
	        }
			//compute the association table of field indexes for each subclass
			setSubTreeFieldAssociationTable(mapper, pcm);
		}
		return queryPrefetch;
	}
	
	/**
	 * For each subclass of the class linked to the pcm (including itself), 
	 * compute the association table of field indexes
	 * and add the element [pcm -> table] in the map of the subclass PClassMapping 
	 */
	public void setSubTreeFieldAssociationTable(PMapper mapper, PClassMapping pcm){
		try{
			ArrayList pcms = new ArrayList();
			pcms.add(pcm);
			PClassMapping[] subPCMs = pcm.getSubPCMs();
			//add children (if any)
			if (subPCMs != null) {
				for (int indexPCM = 0; indexPCM < subPCMs.length; indexPCM++) {
					pcms.add(subPCMs[indexPCM]);
				}
			}	
			String[] classes = new String[prefetchClasses.size()];
			Iterator it = prefetchClasses.iterator();
			int i = 0;
			while (it.hasNext()) {
				classes[i++] = (String) it.next();
			}
			//get the mapper name
			String mapperName = mapper.getMapperName();
			if (mapperName.indexOf(".") != -1)
				mapperName = mapperName.substring(0, mapperName.indexOf("."));
			//	for each pcm
			Iterator itPCM = pcms.iterator();
			while (itPCM.hasNext()) {
				PClassMapping currentPCM = (PClassMapping) itPCM.next();
				ArrayList indexes = new ArrayList(0);
				//get the jorm class associated to the childPCM	
				org.objectweb.jorm.metainfo.api.Class clazz = mapper.getMetaInfoManager().getClass(currentPCM.getClassName());
				ClassProject cp = clazz.getClassProject(currentPCM.getProjectName());
				RdbMapping rdbm = (RdbMapping) cp.getMapping(mapperName);
				RdbClassMapping rdbcm = (RdbClassMapping) rdbm.getClassMapping();
				Class cl = rdbcm.getJormClass(); 
				if (cl == null) {
					return;
				}
				//get all the superclasses
				Collection superClasses = getSuperClasses(cl);
	            
				//for each field in the select clause, 
				//decide wether it is part of the association table or not
				for (int indexField = 0; indexField < prefetchClasses.size(); indexField++) {
					if (superClasses.contains(classes[indexField])) {
						//get the field name
					    //Be careful that the prefetchFieldName is the alias one, we need the column one
						String fieldName = ((String) ((ArrayList)fieldsPrefetch).get(indexField));
						fieldName = ColumnAliasing.getColumnFromAlias(fieldName);
						PrimitiveElementMapping pem = rdbcm.getPrimitiveElementMapping(fieldName, true);
						RdbJoin rdbJoin = ((RdbPrimitiveElementMapping) pem).getJoinByPrimitiveElement((PrimitiveElement)cl.getTypedElement(fieldName));
						//create the indexed pem: if no join, then put the void string
						IndexedPEM indexedPem = new IndexedPEM(pem, (rdbJoin != null ? rdbJoin.getName():""), indexField);
						//and add it in the list
						indexes.add(indexedPem);
					}
				}
				//sort the list of indexed pem: elements are sorted according to
				//the name of sql column then according to the name of joins
				Collections.sort(indexes);
				//fill in the int[] table using the sorted indexes list
				int[] table = new int[indexes.size()];
				for (int j = 0; j < table.length; j++) { 
					table[j] = ((IndexedPEM) indexes.get(j)).index;
				}
				//	add the association [initial PCM -> table] to the sub pcm
				currentPCM.addAssociation(pcm, table);
			}
		} catch(PException e){
			e.printStackTrace();
		}
	}
	
	/**
	 * Returns the list of super class names including the name of the class cl
	 */
	private Collection getSuperClasses(Class cl){
		Collection superClasses = new ArrayList(1);
        //add the current class name to the list of super classes
        superClasses.add(cl.getFQName());
        //get all the ancestors of the class cl
        Iterator it = cl.getAllAncestors().iterator();
        //and add the names of these classes to the superClasses list
        while (it.hasNext()) {
        	Class currentClass = (Class) it.next();
        	superClasses.add(currentClass.getFQName());
        }
        return superClasses;
	}
	
	
	/**
	 * Assuming the primaryKey String is of type "x1,x2,x3", 
	 * returns "x1 as pk0, x2 as pk1, x3 as pk2"
	 * @param primaryKey
	 * @return
	 */
	private String aliasPK(String primaryKey, RdbAdapter adapter){
		String tmp = primaryKey;
		String aliasName = PK;
		String separator = ",";
		int separatorIndex = tmp.indexOf(separator); 
        int localPkIndex = 0;
		while (separatorIndex != -1) {
			String alias = adapter.getColumnAliasExpr(aliasName + localPkIndex++);
			String beginning = primaryKey.substring(0, separatorIndex);
			String end = primaryKey.substring(separatorIndex);
			primaryKey = beginning + alias + end; 
			separatorIndex = primaryKey.indexOf(separator, separatorIndex + alias.length() + 1);			
		}
		primaryKey += adapter.getColumnAliasExpr(aliasName + localPkIndex++);
		return primaryKey;
	}
	
	/**
	 * This class is only used to sort the elements while building the association table
	 * different pcms in an inheritance graph.
	 * See the setSubTreeFieldAssociationTable method.
	 */
	private class IndexedPEM implements Comparable {
		protected PrimitiveElementMapping pem;
		protected int index;
		protected String joinName;
		
		public IndexedPEM(PrimitiveElementMapping pem, String joinName, int index) {
			this.pem = pem;
			this.joinName = joinName;
			this.index = index;
		}

		//compare pem sql column names first
		//and then compare join names  
		public int compareTo(Object o) {
			IndexedPEM indexedPem = (IndexedPEM) o;
			int result = ((RdbPrimitiveElementMapping)pem).compareTo(indexedPem.pem);
			if (result == 0)
				result = joinName.compareTo(indexedPem.joinName);
			return result;
		}
		
	}
}
