/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */
package org.objectweb.jorm.mapper.rdb.metainfo;

import org.objectweb.jorm.metainfo.lib.BasicMetaObject;

import java.util.ArrayList;
import java.util.List;

/**
 * Implementation of the RdbJoin interface. It represents the structure of a join
 * between an external table and its corresponding main table.
 * The join relationship requires two columns represented by
 * primaryTableColumnName and externalTableColumnName.
 * As a meta-object the parent is the external table. The main table is retrieved
 * through the external table.
 * @see	    RdbClassMultiMapping
 */
public class RdbJoin extends BasicMetaObject implements Comparable {
    /**
     * Column names of the primary table involved in the join.
     */
    private ArrayList ptJoinColumnNames;

    /**
     * Column names of the external table involved in the join.
     */
    private ArrayList etJoinColumnNames;

    private String name;
    /**
     * Builds a new BasicRdbJoin object between an external table and its
     * asociated main table. The list of joined columns is given afterwhile by
     * calling the addJoinColumnNames.
     *
     * @param externalTable the external table of the join.
     * @param name the name given to the join meta-object
     */
    public RdbJoin(RdbExternalTable externalTable, String name) {
        super(externalTable);
        this.name = name;
        this.ptJoinColumnNames = new ArrayList();
        this.etJoinColumnNames = new ArrayList();
    }

    /**
     * Returns the external table involved in the join.
     * @return the external table involved in the join.
     */

    public RdbExternalTable getExternalTable() {
        return (RdbExternalTable) parent;
    }

    /**
     * Returns the main table involved in the join. Retrieved through the external
     * table.
     * @return the external table involved in the join.
     */
    public RdbTable getMainTable() {
        return getExternalTable().getMainTable();
    }
    /**
     * Compares the join to another join. Join meta-object are equal if their names
     * are equal.
     * @param o the join meta-object to which comparing this join meta-object
     * @return true is join mata-objects have the same names
     */
    public int compareTo(Object o) {
        return name.compareTo(((RdbJoin) o).getName());
    }

	/**
	 * Returns the list of the column names of the primary (main) table
     * involved in the join. If main T1 and ext T2 joined through
     * T1.c11 == T2.c21 ... T1.c1n == T2.c2n,
     * then this method returns (c11, ..., c1n)
	 * @return the list of the column names of the primary (main) table
     * involved in the join.
	 */
    public List getPTJoinColumnNames() {
        return ptJoinColumnNames;
    }

    /**
     * Returns the list of the column names of the primary (main) table
     * involved in the join. If main T1 and ext T2 joined through
     * T1.c11 == T2.c21 ... T1.c1n == T2.c2n,
     * then this method returns (c21, ..., c2n)
     * @return the list of the column names of the primary (main) table
     * involved in the join.
     */
    public List getETJoinColumnNames() {
        return etJoinColumnNames;
    }

	/**
	 * Adds two join column names 
	 */
    public void addJoinColumnNames(String ptJoinColumnName,
                                   String etJoinColumnName) {
        if (! ptJoinColumnNames.contains( ptJoinColumnName)) {
            if (! etJoinColumnNames.contains( etJoinColumnName)) {
                ptJoinColumnNames.add(ptJoinColumnName);
                etJoinColumnNames.add(etJoinColumnName);
            }
        }
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }
}
