/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.rdb.metainfo;

import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ClassMapping;
import org.objectweb.jorm.metainfo.api.ClassProject;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.Mapping;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.ParentClassMapping;
import org.objectweb.jorm.metainfo.lib.BasicMapping;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.Collection;
import java.util.Iterator;



/**
 * Description
 */
public class RdbMapping extends BasicMapping {
	/**
	 * Builds a new BasicRdbMapping object.
	 * This object contains the mapping structures of a class.
     * The parent object is a Project object.
	 * @param mapperName   the mapper name,
     *        parent	    the parent of the current object.
	 */
	public RdbMapping(String mapperName, MetaObject parent) {
		super(mapperName, parent);
	}

    ///////////////////////////////////////////////////////////////////
	//                      from RdbMapping interface
	///////////////////////////////////////////////////////////////////

	/**
     * NB: OL: Always create multi mapping since only multimapping are expected
     * when parsing the meta-info
	 * Creates a new BasicRdbClassMapping object.
	 * @param   ruleName   the name of the rule used to map the class,
     *          linkedMO   the Class object referenced by the current object,
     *          parent     the parent object of the current object.
	 * @return	an RdbClassMapping object.
	 */
	public ClassMapping createClassMapping(String ruleName) {
        return createClassMultiMapping(ruleName);
    }

	/**
	 * Creates a new BasicRdbClassMultiMapping object.
	 * @param   ruleName   the name of the rule used to map the class,
     *          linkedMO   the Class object referenced by the current object,
     *          parent     the parent object of the current object.
	 * @return	an RdbClassMultiMapping object.
	 */
	public RdbClassMultiMapping createClassMultiMapping(String ruleName) {
        RdbClassMultiMapping res = (RdbClassMultiMapping) getClassMapping();
        if (res == null) {
            res = new RdbClassMultiMapping(
                    ruleName, getParent().getParent(), this);
            setClassMapping(res);
            res.setLogger(logger);
        } else {
            res.setRuleName(ruleName);
        }
        return res;
    }

    /**
	 * Creates a new BasicRdbGenClassMapping object.
	 * @param   ruleName   the name of the rule used to map the class,
     *          linkedMO   the GenClassRef object referenced by the current
     *                     object,
     *          parent     the parent object of the current object.
	 * @return	an RdbGenClassMapping object.
	 */
	public RdbGenClassMapping createGenClassMapping(String ruleName,
                                                    MetaObject linkedMO) {
        String gcid = ((GenClassRef) linkedMO).getGenClassId();
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG,
                    "(gcid " + gcid + ")");
        }
        RdbGenClassMapping res = (RdbGenClassMapping) getGenClassMapping(gcid);
        if (res == null) {
            res = new RdbGenClassMapping(ruleName, linkedMO, this);
            res.setLogger(logger);
            addGenClassMapping(gcid, res);
        } else {
            res.setRuleName(ruleName);
            res.setLinkedMO(linkedMO);
        }
        return res;
    }


	public boolean isFilteredMapping() {
		//lookup if the current class a filtered parent class mapping
		boolean isFilteredMapping = isFilteredMapping(this);
		//if the class has children with filtered parent class mapping
		if (!isFilteredMapping) {
			Class clazz = (Class) getClassMapping().getLinkedMO();
			Collection cs = clazz.getSubClasses();
			if (cs.isEmpty()) {
				return false;
			}
			String proj = ((ClassProject) getParent()).getProjectName();
			for (Iterator it = cs.iterator(); it.hasNext() && !isFilteredMapping;) {
				Mapping m = ((Class) it.next()).getClassProject(proj)
						.getMapping(getMapperName());
				isFilteredMapping = isFilteredMapping(m);
			}
		}
		return isFilteredMapping;
	}

	private boolean isFilteredMapping(Mapping m) {
		Collection pcms = m.getClassMapping().getParentClassMappings();
		if (pcms.isEmpty()) {
			return false;
		}
		boolean isFilteredMapping = false;
		for(Iterator it=pcms.iterator(); it.hasNext() && !isFilteredMapping;) {
			ParentClassMapping pcm = (ParentClassMapping) it.next();
			isFilteredMapping =
				RdbClassMapping.MAP_NEW_FIELDS_TO_EXTENDED_STRUCTURES
				.equalsIgnoreCase(pcm.getRuleName());
		}
		return isFilteredMapping;
	}

}
