/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2004 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.rdb.metainfo;

import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.api.PrimitiveElementMapping;
import org.objectweb.jorm.metainfo.lib.BasicMappingStructure;
import org.objectweb.jorm.api.PException;

import java.util.Map;
import java.util.Iterator;
import java.util.TreeMap;

/**
 * Description
 */
public class RdbPrimitiveElementMapping extends BasicMappingStructure
		implements PrimitiveElementMapping, Comparable {
	/**
	 * The name of the column
	 */
	private String name;

	/**
	 * The sql type of the column
	 */
	private String type;

	/**
	 * Indicates if the column can be null or not
	 */
	private boolean notNull;

	private Map join2pe;

	/**
	 * Builds a new BasicRdbPrimitiveElementMapping.
	 * @param name  the name of the column
	 * @param type  the sql type of the column
	 * @param notNull   true, if the column can not be null, else false.
	 * @param parent    the parent of the current object
	 */
	public RdbPrimitiveElementMapping(String name,
										   String type,
										   boolean notNull,
										   MetaObject linkedMO,
										   MetaObject parent) {
		super(parent, linkedMO);
		this.name = name;
		this.type = type;
		this.notNull = notNull;
		this.join2pe = new TreeMap();
	}

	/**
	 * Builds a new BasicRdbPrimitiveElementMapping.
	 * @param name  the name of the column
	 * @param type  the sql type of the column
	 * @param notNull   true, if the column can not be null, else false.
	 * @param parent    the parent of the current object
	 */
	public RdbPrimitiveElementMapping(String name,
										   String type,
										   boolean notNull,
										   MetaObject linkedMO,
										   MetaObject parent,
										   RdbJoin join) {
		this(name, type, notNull, linkedMO, parent);
		join2pe.put(join, linkedMO);
	}

	public int compareTo(Object o) {
		return name.compareTo(((RdbPrimitiveElementMapping) o).getName());
	}


	///////////////////////////////////////////////////////////////////
	//                      from RdbPrimitiveElementMapping interface
	///////////////////////////////////////////////////////////////////

	/**
	 * Returns the name of the current column.
	 * @return	the string representation of column name
	 */
	public String getName() {
		return name;
	}

	/**
	 * Returns the SQL type of the current object.
	 * @return	the string representation of the sql type
	 */
	public String getType() {
		return type;
	}

	/**
	 * Returns an RdbFilter object.
	 * @return	an RdbFilter object.
	 */
	public RdbFilter getRdbFilter() {
        RdbFilter filter = ((RdbClassMapping) getParent().getParent()).getRdbFilter();
		return filter;
	}

	/**
	 * Indicates if the column is used as a filter column
	 * @return	true, if the column is a filter column, else false
	 */
	public boolean isFilter() {
        RdbFilter filter = getRdbFilter();
        if (filter == null) {
            return false;
        }
		return filter.isFilter((PrimitiveElement) linkedMO);
	}

 	/**
	 * Returns the predicate value associated with the column.
     * If the predicate does not correspond to an equal operator, a PException is raised.
	 * @return the predicate value associated with the column.
	 */
    public String getEqualPredicateValue() throws PException {
        RdbFilter filter = getRdbFilter();
		return filter.getEqualPredicateValue(((PrimitiveElement)linkedMO).getName());
     }

	/**
	 * Allows to know if the column is null or not.
	 * @return	true, if the column is null, else false
	 */
	public boolean isNotNull() {
		return notNull;
	}

	public void setName(String columnName) {
		this.name = columnName;
	}

	public void setType(String sqlType) {
		this.type = sqlType;
	}

	public void setIsNotNull(boolean isNotNull) {
		this.notNull = isNotNull;
	}

	public void bindPrimitiveElement(RdbJoin join, PrimitiveElement pe) {
		join2pe.put(join, pe);
	}

	public PrimitiveElement lookupPrimitiveElement(RdbJoin join) {
		return (PrimitiveElement) join2pe.get(join);
	}

	public Map getPrimitiveElementByRdbJoin() {
		return join2pe;
	}

	public RdbJoin getJoinByPrimitiveElement(PrimitiveElement pe) {
		RdbJoin join = null;
		if (!(join2pe.keySet().isEmpty())) {
			Iterator joinIterator = join2pe.keySet().iterator();
			while (joinIterator.hasNext()) {
				RdbJoin currentJoin = (RdbJoin) joinIterator.next();
				if (join2pe.get(currentJoin) == pe) {
					join = currentJoin;
					return join;
				}
			}
		}
		return join;
	}

}
