/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.rdb.metainfo;

import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.api.PrimitiveElementMapping;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.CommonClassMapping;
import org.objectweb.jorm.metainfo.lib.BasicMappingStructure;
import org.objectweb.jorm.api.PException;

import java.util.Collection;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;

/**
 * Define the mapping of primitive elements onto the columns of a table.
 * Note that the mapping definition of the fields of a class onto the columns
 * of a table may be sparsed between several objects of this class. This can
 * espacially happen with inheritance with the 'extended' or 'added' rules.
 *
 *
 * @author S.Chassande-Barrioz
 */
public class RdbTable extends BasicMappingStructure {
    /**
     * Name of the table in the relational schema
     */
	String name = null;
    /**
     * TODO
     */
	boolean colocated = false;
    /**
     * TODO
     */
	boolean colocatedMaster = false;

    /**
     * Map use to register the primitive elements mapping indexed with their
     * column name. < String,PrimitiveElementMapping >
     */
	Map colName2pem = null;

    /**
     * Indicates if the table is read-only
     */
	boolean readOnly;

	public RdbTable(MetaObject parent, MetaObject linkedMO, String name) {
		super(parent, linkedMO);
		colName2pem = new HashMap();
		this.name = name;
	}



	public boolean isReadOnly() {
		return readOnly;
	}

	public void setReadOnly(boolean readOnly) {
		this.readOnly = readOnly;
	}

	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	public Collection getPrimitiveElementMappings() {
		return colName2pem.values();
	}

    /**
     * Creates and registers a primitive element mapping, that is how
     * a primitive element is mapped onto a column of the table. It implictly
     * defines a new column name in the table, the type and the 'not-null'
     * property being passed as parameters.
     * @param pe the primitive element
     * @param columnName the name of the column onto the primitive element is mapped
     * @param sqlType the SQL type of the column
     * @param notNull if true the column cannot have null values
     * @return a new primitive element registered with this table mapping
     * @throws PException if a primitive element is already mapped to the column
     */
	public RdbPrimitiveElementMapping createPrimitiveElementMapping(
			PrimitiveElement pe,
			String columnName,
			String sqlType,
			boolean notNull) throws PException {

        // checks for existence of the mapping
		RdbPrimitiveElementMapping pem = (RdbPrimitiveElementMapping)
				colName2pem.get(columnName);
		if (pem != null) {
			throw new PException("Column " + columnName
								 + " already defined in the table " + name);
		}
        // creates a new primitive element mapping
		pem = new RdbPrimitiveElementMapping(
				columnName, sqlType, notNull, pe, this);
		((CommonClassMapping) getParent()).addPrimitiveElementMapping(pem);
		pem.setLoggerFactory(getLoggerFactory());
        // and registers it
		colName2pem.put(columnName, pem);
		return pem;
	}

    /**
     * Creates and registers a primitive element mapping, that is how
     * a primitive element is mapped onto a column of the table. It implictly
     * defines a new column name in the table, the type being undefined
     * and the 'not-null' property being set to false.
     * @param pe the primitive element
     * @param columnName the name of the column onto the primitive element is mapped
     * @return a new primitive element registered with this table mapping
     * @throws PException if a primitive element is already mapped to the column
     */
	public RdbPrimitiveElementMapping createPrimitiveElementMapping(
			PrimitiveElement pe,
			String columnName) throws PException {
		return createPrimitiveElementMapping(pe, columnName, null, false);
	}

    /**
     * Removes a primitive element mapping from the table mapping definition
     * @param pem the primitive element mapping to remove
     * @return null if the pem was not part of the table mapping definition, the
     * pem itself otherwise
     */
	public PrimitiveElementMapping removePrimitiveElementMapping(PrimitiveElementMapping pem) {
		return (PrimitiveElementMapping) colName2pem.remove(pem);
	}

    /**
     * Gets the primitive element mapping by its column name
     * @param columnName the name of the column
     * @return the primitive element mapping definition on this column
     */
	public PrimitiveElementMapping getPrimitiveElementMappingByCol(String columnName) {
		return (PrimitiveElementMapping) colName2pem.get(columnName);
	}

    /**
     * Gets the primitive element mapping by its field name
     * @param fieldName the name of the field
     * @return the primitive element mapping definition of this field
     */
    public PrimitiveElementMapping getPrimitiveElementMappingByField(String fieldName) {
        for (Iterator it = colName2pem.values().iterator(); it.hasNext();) {
            RdbPrimitiveElementMapping pem =
                    (RdbPrimitiveElementMapping) it.next();
            if (((PrimitiveElement) pem.getLinkedMO()).getName().equals(fieldName))
                return pem;
        }
        return null;
    }

    /**
     * Gets the set of column names of the table definition
     * @return the set of column names of the table definition (String)
     */
    public Set getColumns() {
        return colName2pem.keySet();
    }

    /**
     * Gets the collocated attribute
     * @return the collocated attribute
     */
	public boolean isColocated() {
		return colocated;
	}

    /**
     * Sets the collocated attribute
     * @param c true if the table is collocated
     */
	public void setColocated(boolean c) {
		colocated = c;
	}

	public boolean isColocatedMaster() {
		return colocatedMaster;
	}

	public void setColocatedMaster(boolean cm) {
		colocatedMaster = cm;
	}
}
