/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.metainfo.api;

import java.util.Iterator;
import java.util.Collection;

/**
 * This interface gathers methods to the ClassMapping and GenClassMapping
 * interfaces. It gives the notion of reference
 * @author	X. Spengler
 */
public interface ClassMapping extends CommonClassMapping {

	/**
	 * 
	 * @return the JORM Class linked to the ClassMapping.
	 * This JORM class is also the linkedMO.
	 * The getLinkedMO() method returns a MetaObject and not a Class object.
	 */
	public Class getJormClass();
	
	/**
	 * Adds a ReferenceMapping object.
	 *
	 * @param  refMapping  the ReferenceMapping object to add to the list.
	 */
	void addReferenceMapping(ReferenceMapping refMapping);

	/**
	 * Retrieves the ReferenceMapping instance corresponding to the name of
	 * a reference field.
	 * @param referenceName is the name of a reference field
	 * @return a ReferenceMapping instance or a null value if no mapping is
	 * defined for the reference name
	 */
	ReferenceMapping getReferenceMapping(String referenceName);

	/**
	 * Returns an Iterator over ReferenceMapping objects.
	 *
	 * @return  an Iterator.
	 */
	Iterator iterateReferenceMappings();

	/**
	 * Returns a collection of ReferenceMapping objects.
	 *
	 * @return  a collection.
	 */
	Collection getReferenceMappings();

	Collection getParentClassMappings();


    ParentClassMapping getParentClassMapping(String className);

    /**
     * Creates a implicit parent class mapping. This method is called to create
     * parent class mapping to super classes for which there is no parent class
     * mapping defined, which means that an implicit inheritance rule of mapping
     * should be applied. For example, with the rdb mapper, this implicit rule
     * is to inherit all the mappings defined in the super class.
     *
     * @param superClass the super class that will be referenced by this
     * parent class mapping
     * @return a parent class mapping with in implicit rule to the super class
     */
    ParentClassMapping createImplicitParentClassMapping(Class superClass);

    /**
     * Get the primitive element mapping corresponding to a field name by searching
     * recursively in the super classes until it is found (if requested)
     * @param fieldName the name of the field of which the primitive element
     * mapping must be returned
     * @param searchInSuper if true search recursively in super classes
     * @return the primitive element mapping if found, null otherwise
     */
    PrimitiveElementMapping getPrimitiveElementMapping(String fieldName,
                                                       boolean searchInSuper);
    /**
     * Add implicit dependencies of mapping structure creation to mapping structure
     * definition. For example, in the case of relational mappings, the creation
     * of tables cannot be known until all the columns of the table is known and
     * those columns may not be known until the mapping of super/sub classes
     * is known.
     */
    void addImplicitDependencies();
}
