/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.metainfo.api;

import org.objectweb.jorm.type.api.PType;

import java.util.Collection;

/**
 * A GenClassRef is an interface which describes method to define
 * a class field which can be a reference to a generic class.
 * A GenClassRef object is composed by a PrimitiveElement (a primitive type),
 * or a ClassRef (a reference to a Class), or an another GenClassRef (a
 * reference to a generic class.
 * A recursion is defined when a GenClassRef object is composed by an other
 * GenClassRef object. A final object of the recursion is a PrimitiveElement
 * or a ClassRef objects.
 * @author	X. Spengler
 */
public interface GenClassRef extends Reference {
	/**
	 * Creates a new PrimitiveElement object, and attach it to the current GenClassRef
	 * This method must be used when the recursion is finished (we are on
	 * final object which is a PrimitiveElement).
	 * @param type is the PType of the primitive element which will be created
	 * @return PrimitiveElement	the primitive element which defines the current
	 *                          GenClassRef object. This object is the final
	 *                          object of the recursion.
	 */
	PrimitiveElement createPrimitiveElement(PType type, int size, int scale);

	/**
	 * Creates a new ClassRef object, and attach it to the current GenClassRef
	 * This method must be used when the recursion is finished (we are on
	 * final object which is a ClassRef).
	 * @param clazz	the reference to the Class object which defines the
	 *                  current GenClassRef object. This object is the final
	 *                  object of the recursion.
	 * @return ClassRef a new ClassRef object
	 */
	ClassRef createClassRef(Class clazz);

	/**
	 * Creates a new GenClassRef object.
	 * This method must be used when the recursion is not finished (we are on
	 * not final object).
	 * @param genClassRefName	the name of the generic class used to create a
	 *                          new GenClassRef object
	 * @return	a new GenClassRef object
	 */
	GenClassRef createGenClassRef(String genClassRefName);

	/**
	 * Returns the PrimitiveElement object which defines the current GenClassRef
	 * object.
	 * This method is used after a call to isPrimitive() method, to determine
	 * if the current object is a "basic" type.
	 * @return	the current type of the field. This method always returns a
	 *          PType object, and can not return null value because of the
	 *          previous test.
	 */
	PrimitiveElement getPrimitiveElement();

	/**
	 * Removes the primitive element.
	 * @param fieldName is the name of the field to be removed
	 * @return the removed primitive element.
	 */
	TypedElement removeTypedElement(String fieldName);
	
	/**
	 * Returns the ClassRef object which defines the current GenClassRef object.
	 * This method is used after a call to isClassRef() method, to determine if
	 * the current object is defined by a final class or not.
	 * @return	the final Class which describes the type of the object which
	 *          defines the GenClassRef object. This method always returns a
	 *          Class, and can not return null value because of the previous
	 *          test.
	 */
	ClassRef getClassRef();

	/**
	 * Returns the GenClassRef object which defines the current GenClassRef
	 * object.
	 * This object exists when it is in a recursion. This method is used after a
	 * call to isGenClassRef() method, to determine if the current object is a
	 * reference to a generic class or not.
	 * @return	the underlying reference to the generic class. This method
	 *          always returns a GenClassRef, and can not return null value
	 *          because of the previous test.
	 */
	GenClassRef getGenClassRef();

	/**
	 * Allows to know if the current object is primitive or not (final and
	 * "basic").
	 * @return	true, if the current object is primitive, else false, if the
	 *          object is not primitive (either a reference to a Class, or
	 *		    a reference to a generic class).
	 */
	boolean isPrimitive();

	/**
	 * Allows to know if the current object is a reference to a Class or not.
	 * @return	true, if the object is a reference to a Class, else false, if
	 *          the object is not a reference to a Class (either a primitive
	 *          object, or a reference to a generic class).
	 */
	boolean isClassRef();

	/**
	 * Allows to know if the current object is a reference to a generic class
	 * or not.
	 * @return	true, if the object is a reference to a generic class, else
	 *	      	false, if the object is not a reference to a generic class
	 *		    (either a primitive object, or a reference to a class).
	 */
	boolean isGenClassRef();

	/**
	 * Creates a new NameDef object for the new GenClassRef object.
	 * @return	a new object used to describe the PName projection for
	 *	    	the current Class
	 */
	NameDef createIdNameDef();

	/**
	 * Returns an iterator on existing NameDef of GenClassRef object.
	 * If no NameDef exists, an empty iterator is returned.
	 * @return	an iterator on NameDef object
	 */
	Collection getIdNameDef();

	/**
	 * Returns a NameDef object.
	 * If the namedef does not exist, null is returned.
	 *
	 * @param projectName   the name of the project.
	 * @return  an existing namedef object,else null is returned.
	 */
	NameDef getIdNameDef(String projectName);

	/**
	 * Returns the name of the GenClass object. ("Set", ..)
	 * @return the name of the GenClassRef object
	 */
	String getGenClassName();

	/**
	 * Returns the id of the GenClass object. ("Set/List/String", ..)
	 * @return the id of the GenClassRef object
	 */
	String getGenClassId();

	/**
	 * Create a new hidden field for the refgenclass.
	 * @return the ScalarField
	 */
	ScalarField createHiddenField(String fieldName, PType type, int size, int scale);

	/**
	 * return the collection of the hiddenfield of the refgenclass definition.
	 * @return the collection containing the hiddenfield (ScalarField).
	 */
	Collection getHiddenFields();

	/**
	 * retrieve an hiddenfield from its name.
	 * @return the hiddenfield.
	 */
	ScalarField getHiddenField(String fieldName);

	/**
	 * @param indexFieldName	the name of the index field
	 */
	void addIndexField(String indexFieldName);

	/**
	 * Allows to know all the existing index field for the current generic
	 * class.
	 * This iterator contains PrimitiveElement objects. If no index field
	 * exists, an empty iterator is returned.
	 *
	 * @return	an Iterator on index field (PrimitiveElement). If there is no
	 *          index field to return, an empty iterator is returned.
	 */
	Collection getIndexFields();
}
