/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.metainfo.lib;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.metainfo.api.*;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.Package;
import org.objectweb.jorm.metainfo.lib.BasicMetaObject;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.MalformedExpressionException;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.api.Operator;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.parser.string.ExpressionParser;
import org.objectweb.medor.expression.parser.api.ParameterTypeProvider;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.HashMap;
import java.util.Map;
import java.util.Collections;
import java.util.List;

/**
 * BasicClass is an implementation of the Class interface defined
 * in the meta information system. This object is defined by a list
 * of fields, a name, a list of mapping.
 * @author	X. Spengler
 */
public class BasicClass extends BasicMetaObject implements Class {
	/**
	 * abstractClass boolean defines if the current class is an abstract
	 * class or not.
	 */
	private boolean abstractClass;
	private String fileName;

	/**
	 * name is the name of the current class object.
	 */
	private String name;

	/**
	 * nameDefs defines a list of NameDef that the user can potentially
	 * used for the mapping.
	 * key: the name of the NameDef object
	 * value: a NameDef object
	 */

	private ArrayList nameDefs;
	private HashMap nd2Exp;
	private HashMap nd2Key;

	/**
	 * classRefs defines the list of the fields reprensented by
	 * a reference to a Class.
	 * key: the name of the ClassRef object
	 * value: the ClassRef object
	 */
	private Map classRefs;

	/**
	 * genClassRefs defines the list of the fields which are defined
	 * by a Generic Class.
	 * key: the name of the GenClassRef object
	 * value: the GenClassRef object
	 */
	private Map genClassRefs;

	/**
	 * primitiveElements defines the list of the fields with a primitive type.
	 * key: the name of the PrimitiveElement object
	 * value: the PrimitiveElement object
	 */
	private Map primitiveElements;

	/**
	 * hiddenFields defines the list of the hiddenfields.
	 * These fields are scalar fields.
	 * key: the name of the hiddenField object
	 * value: the scalarfield object
	 */
	private Map hiddenFields;

	/**
	 * pe2Value defines the constant values of PrimitiveElements which are
	 * constant.
	 */
	private Map pe2Value;
	/**
	 * derivedClasses represents the list of the classes derived from
	 * the current class object.
	 * element: a Class MetaObject object
	 */
	private ArrayList derivedClasses;

	/**
	 * This hashmap associates a full qualified class name with
     * an Class object.
	 */
    private Map superClasses;

	/**
	 * This HashMap associates a project name with a vector of ClassProject
	 * objects.
	 * The notion of project allows the developer to define multiple mappings
	 * for a class per mapper type.
	 */
	private HashMap classProjects;

	/**
	 * Builds a new Class object.
	 * This object is mainly defined by its name, if the class is abstract or
	 * not, and its parent (a schema).
	 *
	 * @param className     the name of the current class
	 * @param parent    the parent of the current object
	 */
	public BasicClass(String className,
					  MetaObject parent) {
		super(parent);
		this.name = className;
		this.abstractClass = false;
		classRefs = new HashMap();
		genClassRefs = new HashMap();
		nameDefs = new ArrayList();
		primitiveElements = new HashMap();
		hiddenFields = new HashMap();
		derivedClasses = new ArrayList();
        superClasses = new HashMap();
		classProjects = new HashMap();
		this.fileName = getFQName().replace('.',
		    System.getProperty("file.separator").charAt(0)) + ".pd";
	}

	protected Collection getChildren() {
		ArrayList res = new ArrayList();
		res.addAll(nameDefs);
		res.addAll(classProjects.values());
		res.addAll(superClasses.values());
		res.addAll(getAllFields());
		res.addAll(getAllHiddenFields());
		return res;
	}

	///////////////////////////////////////////////////////////////////
	//                      from Class interface
	///////////////////////////////////////////////////////////////////

	public Package getPackage() {
	    return (Package) parent;
	}
	
	/**
	 * Allows to retrieve the file name of this class MO.
	 * @return  The file name.
	 */
	public String getFileName() {
		return fileName;
	}

	/**
	 * Allows to assign the file name of this class MO.
	 * @param filename  The file name.
	 */
	public void setFileName(String filename) {
		this.fileName = filename;
	}

	/**
	 * Allows to know if the current class is an abstract class or not.
	 *
	 * @return	true, if the current class is an abstract class, and false,
	 *          if the current class is not abstract
	 */
	public boolean isAbstract() {
		return abstractClass;
	}

	public void setAbstract(boolean isAbstract) {
		abstractClass = isAbstract;
	}

	/**
	 * Allows to know the name of the current class object.
	 * @return	the string representation of the class name
	 */
	public String getName() {
		return name;
	}

	public String getFQName() {
		String packName = null;
		MetaObject parent = getParent();
		if (parent instanceof Package)
			packName = ((Package) parent).getName();
		if (packName == null || packName.length() == 0)
			return name;
		else
			return packName + "." + name;
	}

	/**
	 * Allows to obtain an existing Field.
	 * A TypedElement is a class field, and this method finds a field with its
	 * name (fieldName). This method is a generic method to return a
	 * TypedElement object. The field can be hidden or not.
	 *
	 * @param fieldName	the name of the class field
	 * @return	a field object. If the field does not exist, null is returned.
	 */
	public TypedElement getTypedElement(String fieldName) {
		if (debug)
			logger.log(BasicLevel.DEBUG, "try to return field (" + fieldName + ")");

		TypedElement res = null;
		// this method finds in the 3 structures where is the field
		// fieldName could be a PrimitiveElement
		res = (TypedElement) primitiveElements.get(fieldName);
		if (res != null) {
			if (debug)
				logger.log(BasicLevel.DEBUG, "the field is a PrimitiveElement object");
			return res;
		}
		res = (TypedElement) hiddenFields.get(fieldName);
		if (res != null) {
			if (debug)
				logger.log(BasicLevel.DEBUG, "the field is an hiddenfield object");
			return res;
		}
		res = (TypedElement) classRefs.get(fieldName);
		// fieldName could be a reference to a class
		if (classRefs.get(fieldName) != null) {
			if (debug)
				logger.log(BasicLevel.DEBUG, "the field is a ClassRef object");
			return res;
		}
		res = (TypedElement) genClassRefs.get(fieldName);
		// fieldName could be a reference to a generic class
		if (res != null) {
			if (debug)
				logger.log(BasicLevel.DEBUG, "the field is a GenClassRef object");
			return (GenClassRef) genClassRefs.get(fieldName);
		}
		Iterator it = getSuperClasses().iterator();
		while (it.hasNext() && res == null) {
			res = ((Class) it.next()).getTypedElement(fieldName);
		}
		// attention, the fieldname does not exist. where is it ?
		if (res == null && debug)
			logger.log(BasicLevel.DEBUG, "the field (" + fieldName
			    + ") is not defined in the current class");
		return res;
	}

	/**
	 * Removes a typed element.
	 * @param fieldName is the name of the field to be removed
	 * @return the removed element or null if the field has not been found.
	 */
	public TypedElement removeTypedElement(String fieldName) {
		TypedElement res = (TypedElement) primitiveElements.remove(fieldName);
		if (res != null) {
			return res;
		}
		res = (TypedElement) hiddenFields.remove(fieldName);
		if (res != null) {
			return res;
		}
		res = (TypedElement) classRefs.remove(fieldName);
		// fieldName could be a reference to a class
		if (classRefs.get(fieldName) != null) {
			return res;
		}
		res = (TypedElement) genClassRefs.remove(fieldName);
		return res;
	}

	/**
	 * Creates a new NameDef object for the current Class object.
	 * Here we can define if the namedef is defined outside jorm or not.
	 * In the case where the namedef is already defined for the current class,
	 * the old one is used and returned.
	 *
	 * @return	a new object used to describe the PName projection for
	 *          the current Class
	 */
	public NameDef createNameDef() {
		if (debug)
			logger.log(BasicLevel.DEBUG,
					   "create a new NameDef () for the current Class '" + name + "'");
		NameDef newNameDef = new BasicNameDef(this);
		setLoggingOnChild(newNameDef);
		nameDefs.add(newNameDef);
		return newNameDef;
	}


	/**
	 * Returns a NameDef object.
	 * If the namedef does not exist, null is returned.
	 *
	 * @param name   the name of a NameDef object.
	 * @return  a namedef object, else null is returned.
	 */
	public NameDef getNameDef(String name) {
		Iterator iter = nameDefs.iterator();
		while (iter.hasNext()) {
			NameDef nameDef = (NameDef) iter.next();
			if (nameDef.getName().equals(name)) {
				return nameDef;
			}
		}
		return null;
	}

	public Collection getNameDefs() {
		return nameDefs;
	}

	public synchronized Expression getInheritanceFilter(NameDef nd) throws ExpressionException {
		Object filter =  (nd2Exp == null ? null : nd2Exp.get(nd));
		if (filter instanceof String) {
			filter = parseInheritanceFilter((String) filter, nd);
			nd2Exp.put(nd, filter);
		}
		//if filter is null, try the parents
		if (filter == null && !superClasses.isEmpty()) {
			Iterator it = getSuperClasses().iterator();
			while (it.hasNext() && filter == null) {
				Class cl = (Class) it.next();
				filter = cl.getInheritanceFilter(nd);
			}
		}
		return (Expression) filter;
	}

	public Object getInheritanceNamingKey(NameDef nd) {
		return  (nd2Key == null ? null : nd2Key.get(nd));
	}

	public void setInheritanceFilter(NameDef nd, Expression e) {
		if (nd == null || e == null) {
			return;
		}
		if (nd2Exp == null) {
			nd2Exp = new HashMap(5);
		}
		nd2Exp.put(nd, e);
	}

	public void setInheritanceFilter(NameDef nd, String filter) {
		if (nd == null || filter == null) {
			return;
		}
		if (nd2Exp == null) {
			nd2Exp = new HashMap(5);
		}
		nd2Exp.put(nd, filter);
	}

	public void setInheritanceNamingKey(NameDef nd, Object key) {
		if (nd == null) {
			return;
		}
		if (nd2Key == null) {
			nd2Key = new HashMap(5);
		}
		nd2Key.put(nd, key);
	}

	/**
	 * Allows to obtain a new PrimitiveElement from a class field name an its
	 * type.
	 * A PrimitiveElement is a Class field. If the class field already exists,
	 * it is returned.
	 *
	 * @param fieldName	the name of the primitive field
	 * @param type      the PType of the primitive field
	 * @return	a new PrimitiveElement object, or the existing one if already
	 *          defined
	 */
	public PrimitiveElement createPrimitiveElement(String fieldName,
												   PType type,
												   int size, int scale) {
		if (debug)
			logger.log(BasicLevel.DEBUG, "create a new PrimitiveElement (" +
										 fieldName + ") for the current Class (" + name + ")");

		// try to get the fieldname from the current structure
		PrimitiveElement result = (PrimitiveElement)
				primitiveElements.get(fieldName);
		if (result != null) {
			// be careful, the primitive element already exists, we return it.
			if (logger != null) {
				logger.log(BasicLevel.WARN,
					   "Try to create an existing PrimitiveElement (" +
					   fieldName + "), return existing one");
			}
		} else {
			// oh, the primitive element does not exist, we must create it.
			result = new BasicPrimitiveElement(fieldName, type, size, scale, this);
			setLoggingOnChild(result);
			primitiveElements.put(fieldName, result);
		}
		return result;
	}

	public void setConstantValue(String fieldName, String cv) {
		if (pe2Value == null) {
			pe2Value = new HashMap();
		}
		pe2Value.put(fieldName, cv);
	}

	public String getConstantValue(String fieldName) {
		return (pe2Value == null ? null : (String) pe2Value.get(fieldName));
	}

	/**
	 * Creates a new scalar field and adds it to the list of hidden fields.
	 * If the field already exists, it is returned, else a new one is built.
	 *
	 * @param fieldName	the name of the field to create
	 * @param type	    the type of the field to create
	 * @return	a new ScalarField object if this object does not already exist
	 */
	public ScalarField createHiddenField(String fieldName, PType type, int size, int scale) {
		if (debug)
			logger.log(BasicLevel.DEBUG,
					   "create a new HiddenField (" + fieldName + ") for the current class (" + name + ")");

		// verify if the field is not yet in the database
		ScalarField result = (ScalarField) hiddenFields.get(fieldName);
		if (result != null) {
			if (logger != null) {
				logger.log(BasicLevel.WARN,
					   "Try to create an existing HiddenField (" +
					   fieldName + "), return existing one");
			}
		} else {
			// else create a new one
			result = new BasicScalarField(fieldName, type, size, scale, this);
			setLoggingOnChild(result);
			hiddenFields.put(fieldName, result);
		}
		return result;
	}

	/**
	 * Allows to obtain a new ClassRef from a field name an its class.
	 * A ClassRef object is a field which is represented by a reference to a
	 * Class. If the field already exists, it is returned.
	 *
	 * @param fieldName	the name of the field
	 * @param class_	the string representation of the class which represents
	 *                  the reference of the field
	 * @return	a new ClassRef object, or the existing one if already defined
	 */
	public ClassRef createClassRef(String fieldName, Class class_) {
		if (debug)
			logger.log(BasicLevel.DEBUG, "create a new ClassRef (" + fieldName +
										 ") for the current Class (" + name + ")");

		// just verify if the fielname does not exist in the current structure.
		ClassRef result = (ClassRef) classRefs.get(fieldName);
		if (result != null) {
			// be careful, the object exists, we return it
			if (logger != null) {
				logger.log(BasicLevel.WARN,
					   "attention, try to create an existing ClassRef (" +
					   fieldName + "), return existing one");
			}
		} else {
			// we have to create a new object
			result = new BasicClassRef(fieldName, class_, this);
			setLoggingOnChild(result);
			classRefs.put(fieldName, result);
		}
		return result;
	}

	/**
	 * Allows to obtain a new GenClassRef from a field name an its generic
	 * class.
	 * A GenClassRef object is a field which is represented by a
	 * reference to a generic class. If the field already exists,
	 * it is returned.
	 *
	 * @param fieldName	the name of the field
	 * @param genName	the GenClass object which represents the reference of
	 *                  the field
	 * @return	a new GenClassRef object, or the existing one if already
	 *          defined
	 */
	public GenClassRef createGenClassRef(String fieldName, String genName) {
		if (debug)
			logger.log(BasicLevel.DEBUG, "create a new GenClassRef (" +
										 genName + ") from field (" + fieldName + ") from Class (" +
										 name + ")");

		// verify if the fieldname already exists
		GenClassRef result = (GenClassRef) genClassRefs.get(fieldName);
		if (result != null) {
			// be careful, the object already exists, we return it
			if (logger != null) {
				logger.log(BasicLevel.WARN,
					   "attention, try to create an existing GenClassRef (" + fieldName +
					   "), return existing one");
			}
		} else {
			// we have to create a new object and return it
			result = new BasicGenClassRef(genName, fieldName, this);
			setLoggingOnChild(result);
			genClassRefs.put(fieldName, result);
		}
		return result;
	}

	/**
	 * returns an ArrayList which contains the current fields.
	 * @return  a list of current fields
	 */
	protected List listField() {
		return listField(new ArrayList());
	}
	protected List listField(List allFields) {
		allFields.addAll(primitiveElements.values());
		allFields.addAll(classRefs.values());
		allFields.addAll(genClassRefs.values());
		Collections.sort(allFields, FieldComparator.instance);
		return allFields;
	}
	protected List listAllFields(List allFields) {
		listField(allFields);
		// and add the fields from the inherited classes
		Iterator iter = getSuperClasses().iterator();
		while (iter.hasNext())
			((BasicClass) iter.next()).listAllFields(allFields);
		return allFields;
	}

	public Collection getFields() {
		List fields = listField(new ArrayList());
		Collections.sort(fields, FieldComparator.instance);
		return fields;
	}

	public Collection getAllFields() {
		List allFields = listAllFields(new ArrayList());
		Collections.sort(allFields, FieldComparator.instance);
		return allFields;
	}

	protected List listHiddenFields(List allFields) {
		allFields.addAll(hiddenFields.values());
		return allFields;
	}

	protected List listAllHiddenField(List allFields) {
		listHiddenFields(allFields);
		Iterator iter = getSuperClasses().iterator();
		while (iter.hasNext())
			((BasicClass) iter.next()).listAllHiddenField(allFields);
		return allFields;
	}

	public Collection getHiddenFields() {
		List hiddenFields = listHiddenFields(new ArrayList());
		Collections.sort(hiddenFields, FieldComparator.instance);
		return hiddenFields;
	}

	public Collection getAllHiddenFields() {
		List allFields = listAllHiddenField(new ArrayList());
		Collections.sort(allFields, FieldComparator.instance);
		return allFields;
	}

	/**
	 * retrieve an hiddenfield from its name.
	 *
	 * @return the hiddenfield or null if the field does not exist.
	 */
	public ScalarField getHiddenField(String fieldName) {
		if (debug)
			logger.log(BasicLevel.DEBUG,
					   "try to return an hidden field (" + fieldName + ")");
		// this method finds in the 3 structures where is the field
		return (ScalarField) hiddenFields.get(fieldName);
	}

	public Collection getSubClasses() {
		return derivedClasses;
	}

	/**
	 * Returns a collection of Class objects.
	 * Returns only the direct super classes of this class.
	 * In the following graph,
	 * 		A
	 * 	---------
	 * |		|
	 * B		C
	 * 			-----
	 * 				|
	 * 				D
	 * superclass of D is C.
	 * The s in superclasses comes from the fact that in JORM a class can have many superclasses.
	 * 
	 * @return	 a collection.
	 */
	public Collection getSuperClasses() {
		return superClasses.values();
	}

	public Class getSuperClass(String fqcn) {
		return (Class) superClasses.get(fqcn);
	}

	/**
	 * Adds a new inherited class to the list of the inherited classes for the
	 * current class object.
	 * If the class already exists, nothing is done.
	 *
	 * @param class_	the inherited class to add to inherited classes list
     *
	 */
	public void addSuperClass(Class class_) {
		if ((class_ != null) &&
            (!superClasses.containsKey(class_.getFQName()))) {
            superClasses.put(class_.getFQName(), class_);
			class_.addSubClass(this);
		}
	}

	public void addSubClass(Class class_) {
		if ((class_ != null) &&
				(!derivedClasses.contains(class_))) {
			derivedClasses.add(class_);
			class_.addSuperClass(this);
		}
	}

	/**
	 * Returns the number of inherited classes for the current Class object.
	 *
	 * @return  the number of inherited classes
	 */
	public int getInheritedClassNumber() {
		return superClasses.size();
	}

	/**
	 * Returns all the ancestors of the class, i.e.:
	 *  	A
	 * 	---------
	 * |		|
	 * B		C
	 * 			-----
	 * 				|
	 * 				D
	 * all ancestors of D are C and A.
	 *
	 */
	public List getAllAncestors() {
		if (superClasses.isEmpty()) {
			return Collections.EMPTY_LIST;
		}
		List ancestors = new ArrayList();
		ancestors.add(this); //init the algorithm
		int i=0;
		while(i < ancestors.size()) {
			Class clazz = (Class) ancestors.get(i);
			Collection c = clazz.getSuperClasses();
			// Add super classes which have not already been listed
			for(Iterator it = c.iterator(); it.hasNext();) {
				Class superClass = (Class) it.next();
				if (!ancestors.contains(superClass)) {
					ancestors.add(superClass);
				}
			}
			i++; // treat the next element in the list
		}
		ancestors.remove(0);
		return ancestors;
	}

	/**
	 * Return the list (most often with only one element) of ancestors.
	 * In the following graph,
	 * 		A
	 * 	---------
	 * |		|
	 * B		C
	 * 			-----
	 * 				|
	 * 				D
	 * ancestor of D is A.
	 * The s of ancestors comes from the fact that with JORM a class can have many superclasses.
	 */
	public List getAncestors() {
		if (superClasses.isEmpty()) {
			return Collections.EMPTY_LIST;
		}
		List ancestors = new ArrayList();
		List roots = new ArrayList();
		ancestors.add(this); //init the algorithm
		while(!ancestors.isEmpty()) {
			Class clazz = (Class) ancestors.remove(0);
			Collection c = clazz.getSuperClasses();
			if (c.isEmpty()) {
				if (!roots.contains(clazz)) {
					roots.add(clazz);
				}
			} else {
				// Add super classes which have not already listed
				for(Iterator it = c.iterator(); it.hasNext();) {
					Class superClass = (Class) it.next();
					if (!ancestors.contains(superClass)) {
						ancestors.add(superClass);
					}
				}
			}
		}
		return roots;
	}

    public boolean isPolymorphic() {
        return (!getSubClasses().isEmpty() || !getSuperClasses().isEmpty());
    }

    public PType getPType() {
		try {
			PType type = getManager().getPTypeSpace().getPType(getFQName());
			if (type != null) {
				return type;
			}			
			if (superClasses.isEmpty()) {
				return getManager().getPTypeSpace().createPType(getFQName());
			} else {
				List ancestors = getAllAncestors();
				List couples = new ArrayList(ancestors.size());
				for(int i=0; i<ancestors.size(); i++) {
					Class clazz = (Class) ancestors.get(i);
					Collection superClasses = clazz.getSuperClasses();
					for(Iterator it = superClasses.iterator(); it.hasNext();) {
						Class superClass = (Class) it.next();
						couples.add(new String[] {clazz.getFQName(), superClass.getFQName()});
					}
				}
				return getManager().getPTypeSpace().createPType(
							getFQName(),
							(String[][]) couples.toArray(new String[couples.size()][]));
			}
		} catch(Exception e) {
			logger.log(BasicLevel.ERROR,
				"impossible to compute the PType of the class "
				+ getFQName(), e);
			return null;
		}
	}

	/**
	 * Returns the number of fields which are references to classes.
	 *
	 * @return  the number of fieldRef
	 */
	public int getFieldRefNumber() {
		return classRefs.size();
	}

	/**
	 * Returns a ClassProject object.
	 *
	 * @param  projectName       a project name.
	 * @return a ClassProject object.
	 */
	public ClassProject getClassProject(String projectName) {
		return (ClassProject) classProjects.get(projectName);
	}

	/**
	 * Returns a ClassMapping object.
	 *
	 * @param  projectName      a project name,
     *         mappername       a mapper name.
	 * @return a ClassMapping object.
	 */
	public ClassMapping getClassMapping(String projectName, String mapperName) {
        ClassProject cp = (ClassProject) classProjects.get(projectName);
        if (cp == null) return null;
        Mapping mapping = cp.getMapping(mapperName);
        if (mapping == null) return null;
        return mapping.getClassMapping();
	}

	/**
	 * Returns a collection of ClassProject objects.
	 * @return	 a collection.
	 */
	public Collection getClassProjects() {
		return classProjects.values();
	}

	/**
	 * Adds a ClassProject object.
	 *
	 * @param  projectName       a project name,
	 *         classProject	    a ClassProject object.
	 */
	public void addClassProject(String projectName, ClassProject classProject) {
		if (!classProjects.containsKey(projectName)) {
			classProjects.put(projectName, classProject);
		}
	}

	/**
	 * Creates a new ClassProject object.
	 *
	 * @param projectName	the name of the project.
	 * @return a new ClassProject object, or an existing one.
	 */
	public ClassProject createClassProject(String projectName) {
		if (debug)
			logger.log(BasicLevel.DEBUG,
					   "create a new ClassProject for the current Class (" + name +
					   ") projectName=<" + projectName + ">");

		// verifies whether the classProject already exists
		ClassProject cp = (ClassProject) classProjects.get(projectName);
		if (cp == null) {
			cp = new BasicClassProject(projectName, this);
			classProjects.put(projectName, cp);
			return cp;
		} else {
			if (debug) {
				logger.log(BasicLevel.DEBUG,
					   "return the existing ClassProject " + projectName);
			}
			return cp;
		}
	}

	/**
	 * Removes a ClassProject instance.
	 * @param projectname is the name of removed ClassProject
	 * @return The removed ClassProject
	 */
	public ClassProject removeClassProject(String projectname) {
		return (ClassProject) classProjects.remove(projectname);
	}

	public Expression parseInheritanceFilter(String filter, NameDef nd)
		throws ExpressionException {
		return new ExpressionParser()
			.parse(filter, new BasicParameterTypeProvider());
	}

	/**
	 * Determine if we need to generate the kfpnc.
	 */
	public boolean generateKFPNC() throws ExpressionException {
		Expression exp = null;
		//for each namedef
		for (int i = 0; i < nameDefs.size(); i++) {
			NameDef nd = (NameDef) nameDefs.get(i);
			exp = getInheritanceFilter(nd);
			//if only one is false, then we need to generate the kfpnc
			if(!detectFilterElementNotInPK(exp, nd))
				return true;
		}
		return false;
	}
	
	public boolean detectFilterElementNotInPK(Expression exp, NameDef nd) throws ExpressionException {
		if (exp instanceof Operator) {
			Operator operator = (Operator) exp;
			int nb = operator.getOperandNumber();
			if (nb == 1) {
				return detectFilterElementNotInPK(operator.getExpression(0), nd);
			} else if (nb == 2) {
				return (detectFilterElementNotInPK(operator.getExpression(0), nd) 
						|| 	detectFilterElementNotInPK(operator.getExpression(1), nd));
			} else {
				throw new ExpressionException("Operators with more than 2 operands are not supported: " + exp);
			}
		} else if (exp instanceof ParameterOperand) {
			ParameterOperand po = (ParameterOperand) exp;
			if (nd.isNameRef()) {
				//get the list of fields of the composite namedef
				Collection fields = nd.getNameRef().getProjection().values();
				return !(fields.contains(po.getName()));
			} else {
				return !(po.getName().equals(nd.getFieldName()));
			}
		} 
		return false;
	}
	
	private class BasicParameterTypeProvider implements ParameterTypeProvider {
		public PType getParameterPType(String s) throws MalformedExpressionException {
			TypedElement te = getTypedElement(s);
			return (te == null ? null : te.getType());
		}
	}

}

