/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.metainfo.lib;

import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ClassRef;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.api.ScalarField;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.type.api.PExceptionTyping;
import org.objectweb.jorm.lib.JormPathHelper;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.List;

/**
 * BasicGenClassRef defines a reference to a generic class which is class field.
 * This object is created by the GenClass object to define its fields.
 * @author	X. Spengler
 */
public class BasicGenClassRef
		extends BasicMetaObject
		implements GenClassRef, Comparable {

	private static final String PRIMITIVE_ELEMENT_NAME = "element";
	/**
	 *  the name of the reference field
	 */
	private String name;

	/**
	 * the name of the generic class
	 */
	private String genClassName;
	private String genClassId;

	/**
	 * indexField is an index field for the generic class
	 */
	private ScalarField indexField;

	/**
	 * the list of the name definition declared by the user in the
	 * current reference to the generic class.
	 * key: the name of the NameDef
	 * value: a NameDef object
	 */
	private ArrayList refNameDefs;

	/**
	 * the list of the name definition declared by the user in the
	 * current reference to the generic class.
	 * key: the name of the NameDef
	 * value: a NameDef object
	 */
	private ArrayList idNameDefs;

	/**
	 * hiddenFields defines the list of the hiddenfields.
	 * These fields are scalar fields.
	 * key: the name of the hiddenField object
	 * value: the scalarfield object
	 */
	private Map hiddenFields;

	/**
	 * indexFields is a list of index fields for the generic class.
	 * key: the name of the index field
	 * value: a PrimitiveElement
	 */
	private List indexFields;

	/**
	 * if true, the current GenClassRef is a primitive field
	 */
	private boolean primitive;

	/**
	 * if true, the current GenClassRef is a reference to a class
	 */
	private boolean classRef;

	/**
	 * if true, the current GenClassRef is a reference to a generic class
	 */
	private boolean genClassRef;

	/**
	 * currentGenClassRef is the reference to the generic class which defines
	 * the type of the current field
	 */
	private GenClassRef currentGenClassRef;

	/**
	 * currentClassRef is the reference to the class which defines the type
	 * of the current field
	 */
	private ClassRef currentClassRef;

	/**
	 * primitiveType defines the type of the current field
	 */
	private PrimitiveElement primitiveElement;

	private PType type;

	/**
	 * Builds a new BasicGenClassRef object.
	 * This object is defined by its name, the reference to the generic class
	 * and its parent.
	 *
	 * @param genClassName  the name of the generic class reference
	 * @param name	        the string representation of the generic class
	 *                      field name
	 * @param parent	the parent of the current object
	 */
	public BasicGenClassRef(String genClassName,
							String name,
							MetaObject parent) {
		super(parent);
		this.name = name;
		this.genClassName = genClassName;
		indexField = null;
		hiddenFields = new HashMap();
		refNameDefs = new ArrayList();
		idNameDefs = new ArrayList();
		indexFields = new ArrayList();
		genClassId = JormPathHelper.getPath(this, false);
	}

	public int compareTo(Object o) {
		return name.compareTo(((TypedElement) o).getName());
	}

	///////////////////////////////////////////////////////////////////
	//                      from GenClassRef interface
	///////////////////////////////////////////////////////////////////
	/**
	 * Creates a new PrimitiveElement object, and attach it to the current GenClassRef
	 * This method must be used when the recursion is finished (we are on
	 * final object which is a PrimitiveElement).
	 *
	 * @param type is the PType of the primitive element which will be created
	 * @return PrimitiveElement	the primitive element which defines the current
	 *                          GenClassRef object. This object is the final
	 *                          object of the recursion.
	 */
	public PrimitiveElement createPrimitiveElement(PType type, int size, int scale) {
		primitiveElement = new BasicPrimitiveElement(
				PRIMITIVE_ELEMENT_NAME + '_' + type.getJormName(), type, size, scale, this);
		setLoggingOnChild(primitiveElement);

		// set the flags, primitive to true, the others to false
		primitive = true;
		classRef = false;
		genClassRef = false;
		return primitiveElement;
	}

	/**
	 * Returns a new ScalarField object.
	 *
	 * @param indexFieldName	the name of the index field
	 * @param type	            the type of the index field
	 * @return	a new ScalarField object for the index field
	 */
	public ScalarField createIndexField(String indexFieldName,
										PType type, int size, int scale) {
		if (debug)
			logger.log(BasicLevel.DEBUG,
					   "create a new PrimitiveElement (index field (" +
					   indexFieldName + ")) for the current " + "GenClass (" +
					   this.genClassName + ")");
		ScalarField sf = new BasicScalarField(indexFieldName, type, size, scale, this);
		setLoggingOnChild(sf);
		return sf;
	}

	/**
	 * Returns the ScalarField object corresponding to the index field.
	 * @return	a ScalarField object.
	 */
	public ScalarField getIndexField() {
		return indexField;
	}

	/**
	 * Creates a new ClassRef object, and attach it to the current GenClassRef
	 * This method must be used when the recursion is finished (we are on
	 * final object which is a ClassRef).
	 *
	 * @param clazz	the reference to the Class object which defines the
	 *                  current GenClassRef object. This object is the final
	 *                  object of the recursion.
	 * @return ClassRef a new ClassRef object
	 */
	public ClassRef createClassRef(Class clazz) {
		String classname = clazz.getName();
		int idx = classname.lastIndexOf('.');
		if (idx != -1) {
			classname = classname.substring(idx + 1, classname.length());
		}
		currentClassRef = new BasicClassRef(classname, clazz, this);
		setLoggingOnChild(currentClassRef);
		// set the flags, classRef to true, the others to false
		classRef = true;
		genClassRef = false;
		primitive = false;

		return currentClassRef;
	}

	/**
	 * Creates a new GenClassRef object.
	 * This method must be used when the recursion is not finished (we are on
	 * not final object).
	 *
	 * @param genClassRefName	the name of the generic class used to create a
	 *                          new GenClassRef object
	 * @return	a new GenClassRef object
	 */
	public GenClassRef createGenClassRef(String genClassRefName) {
		// set the flags, genClassRef to true, the others to false
		classRef = false;
		genClassRef = true;
		primitive = false;

		if (debug)
			logger.log(BasicLevel.DEBUG, "create a new GenClassRef ("
										 + genClassRefName + ") for the current GenClassRef ("
										 + name + ")");

		if (currentGenClassRef == null) {
			currentGenClassRef = new BasicGenClassRef(genClassRefName, null, this);
			setLoggingOnChild(currentGenClassRef);
		}
		return currentGenClassRef;
	}

	/**
	 * Returns the PrimitiveElement object which defines the current GenClassRef
	 * object.
	 * This method is used after a call to isPrimitive() method, to determine
	 * if the current object is a "basic" type.
	 *
	 * @return	the current type of the field. This method always returns a
	 *          PType object, and can not return null value because of the
	 *          previous test.
	 */
	public PrimitiveElement getPrimitiveElement() {
		return primitiveElement;
	}

	public TypedElement removeTypedElement(String fieldName){
		TypedElement res = (TypedElement) hiddenFields.remove(fieldName);
		return res;
	}
	
	/**
	 * Returns the ClassRef object which defines the current GenClassRef object.
	 * This method is used after a call to isClassRef() method, to determine if
	 * the current object is defined by a final class or not.
	 *
	 * @return	the final Class which describes the type of the object which
	 *          defines the GenClassRef object. This method always returns a
	 *          Class, and can not return null value because of the previous
	 *          test.
	 */
	public ClassRef getClassRef() {
		return currentClassRef;
	}

	/**
	 * Returns the GenClassRef object which defines the current GenClassRef
	 * object.
	 * This object exists when it is in a recursion. This method is used after a
	 * call to isGenClassRef() method, to determine if the current object is a
	 * reference to a generic class or not.
	 *
	 * @return	the underlying reference to the generic class. This method
	 *          always returns a GenClassRef, and can not return null value
	 *          because of the previous test.
	 */
	public GenClassRef getGenClassRef() {
		return currentGenClassRef;
	}

	/**
	 * Allows to know if the current object is primitive or not (final and
	 * "basic").
	 *
	 * @return	true, if the current object is primitive, else false, if the
	 *          object is not primitive (either a reference to a Class, or
	 *		    a reference to a generic class).
	 */
	public boolean isPrimitive() {
		return primitive;
	}

	/**
	 * Allows to know if the current object is a reference to a Class or not.
	 *
	 * @return	true, if the object is a reference to a Class, else false, if
	 *          the object is not a reference to a Class (either a primitive
	 *          object, or a reference to a generic class).
	 */
	public boolean isClassRef() {
		return classRef;
	}

	/**
	 * Allows to know if the current object is a reference to a generic class
	 * or not.
	 *
	 * @return	true, if the object is a reference to a generic class, else
	 *	      	false, if the object is not a reference to a generic class
	 *		    (either a primitive object, or a reference to a class).
	 */
	public boolean isGenClassRef() {
		return genClassRef;
	}

	/**
	 * Creates a new NameDef object for the new GenClassRef object.
	 * If the specified namedef object already exists, it is returned, else
	 * a new NameDef object is created.
	 *
	 * @return	a new object used to describe the PName projection for
	 *	    	the current Class, or an existing object if already defined
	 */

	public NameDef createIdNameDef() {
		if (debug)
			logger.log(BasicLevel.DEBUG,
					   "create a new ID NameDef () for the current GenClassRef (" + name + ")");
		// in the case it does not exist, we create it.
		NameDef newNameDef = new BasicNameDef(this);
		setLoggingOnChild(newNameDef);
		idNameDefs.add(newNameDef);
		return newNameDef;
	}


	/**
	 * Returns an iterator on existing NameDef of GenClassRef object.
	 * If no NameDef exists, an empty iterator is returned.
	 *
	 * @return	an iterator on NameDef object
	 */
	public Collection getIdNameDef() {
		return idNameDefs;
	}


	/**
	 * Create a new hidden field for the refgenclass.
	 *
	 * @return the ScalarField
	 */
	public ScalarField createHiddenField(String fieldName, PType type, int size, int scale) {
		if (debug)
			logger.log(BasicLevel.DEBUG,
					   "create a new HiddenField (" + fieldName + ") for the current refgenclass (" + name + ")");

		// verify if the field is not yet in the database
		if (hiddenFields.get(fieldName) != null) {
			if (logger != null) {
				logger.log(BasicLevel.WARN,
						   "attention, try to create an existing HiddenField (" +
						   fieldName + "), return existing one");
			}
			return (ScalarField) hiddenFields.get(fieldName);
		}

		// else create a new one
		ScalarField newHiddenField = new BasicScalarField(fieldName, type, size, scale, this);
		setLoggingOnChild(newHiddenField);
		hiddenFields.put(fieldName, newHiddenField);
		return newHiddenField;
	}

	/**
	 * return an iterator over the hiddenfield of the refgenclass definition.
	 *
	 * @return the iterator containing the hiddenfield.
	 */
	public Collection getHiddenFields() {
		return hiddenFields.values();
	}


	/**
	 * retrieve an hiddenfield from its name.
	 *
	 * @return the hiddenfield.
	 */
	public ScalarField getHiddenField(String fieldName) {
		if (debug) {
			logger.log(BasicLevel.DEBUG,
					   "try to return an hidden field (" + fieldName + ")");
		}
		return (ScalarField) hiddenFields.get(fieldName);
	}

	/**
	 * Returns the name of the GenClass object.
	 *
	 * @return the name of the GenClassRef object
	 */
	public String getGenClassName() {
		return genClassName;
	}

	public String getGenClassId() {
		return genClassId;
	}

	/**
	 * Creates a new NameDef object for the current reference class object.
	 * If this NameDef already exists for the current object, it is returned,
	 * in other case, a new NameDef object is created and returned.
	 *
	 * @return	a new object used to describe the name projection for
	 *	      	the current reference class, or an existing one if it
	 *          already exists
	 */

	public NameDef createRefNameDef() {
		if (debug)
			logger.log(BasicLevel.DEBUG,
					   "create a new REF NameDef () for the current Class (" + name + ")");
		// in the case it does not exist, we create it.
		NameDef newNameDef = new BasicNameDef(this);
		setLoggingOnChild(newNameDef);
		refNameDefs.add(newNameDef);
		return newNameDef;
	}

	/**
	 * Returns a NameDef object.
	 * If the namedef does not exist, null is returned.
	 *
	 * @param name   the name of the namedef.
	 * @return  an existing namedef object,else null is returned.
	 */
	public NameDef getRefNameDef(String name) {
		Iterator iter = refNameDefs.iterator();
		while (iter.hasNext()) {
			NameDef nameDef = (NameDef) iter.next();
			if (nameDef.getName().equals(name)) {
				return nameDef;
			}
		}
		return null;
	}

	/**
	 * Returns a NameDef object.
	 * If the namedef does not exist, null is returned.
	 *
	 * @param name   the name of the namedef.
	 * @return  an existing namedef object,else null is returned.
	 */
	public NameDef getIdNameDef(String name) {
		Iterator iter = idNameDefs.iterator();
		while (iter.hasNext()) {
			NameDef nameDef = (NameDef) iter.next();
			if (nameDef.getName().equals(name)) {
				return nameDef;
			}
		}
		return null;
	}

	/**
	 * Returns an iterator on existing NameDef for the current Class.
	 * This method returns an empty iterator if no namedef exists in this Class.
	 *
	 * @return	an iterator on NameDef, or an empty iterator if no namedef is
	 *          defined
	 */

	public Collection getRefNameDef() {
		return refNameDefs;
	}

	public void addIndexField(String indexFieldName) {
		indexFields.add(hiddenFields.get(indexFieldName));
	}

	/**
	 * Allows to know all the existing index field for the current generic
	 * class.
	 * This iterator contains PrimitiveElement objects. If no index field
	 * exists, an empty iterator is returned.
	 *
	 * @return	an Iterator on index field (PrimitiveElement). If there is no
	 *          index field to return, an empty iterator is returned.
	 */
	public Collection getIndexFields() {
		return indexFields;
	}

	protected Collection getChildren() {
		ArrayList res = new ArrayList();
		res.addAll(hiddenFields.values());
		res.addAll(refNameDefs);
		res.addAll(idNameDefs);
		return res;
	}

	///////////////////////////////////////////////////////////////////
	//                      from TypedElement interface
	///////////////////////////////////////////////////////////////////


	/**
	 * Returns the name of the current typed element.
	 *
	 * @return	the string representation of the typed element
	 */
	public String getName() {
		return name;
	}

	/**
	 * Allows to know the type of the field.
	 *
	 * @return	a PType object which defines the type of the current field
	 */
	public PType getType() {
		if (type != null)
			return type;
		try {
			if (primitive) {
				type = getManager().getPTypeSpace().createPType(
						new String[]{genClassName}, primitiveElement.getType());
			} else if (classRef) {
				type = getManager().getPTypeSpace().createPType(
						new String[]{genClassName}, currentClassRef.getType());
			} else { //genClassRef
				ArrayList al = new ArrayList();
				GenClassRef gcr = this;
				al.add(gcr.getGenClassName());
				while (gcr.isGenClassRef()) {
					al.add(gcr.getGenClassName());
					gcr = gcr.getGenClassRef();
				}
				PType lasttype = null;
				if (gcr.isClassRef())
					lasttype = gcr.getClassRef().getType();
				else
					lasttype = gcr.getPrimitiveElement().getType();
				type = getManager().getPTypeSpace().createPType(
						(String[]) al.toArray(new String[0]), lasttype);
			}
		} catch (PExceptionTyping e) {
			if (logger != null) {
				logger.log(BasicLevel.ERROR,
						   "Impossible to calculate the type of the GenClassRef field "
						   + genClassId, e);
			}
		}
		return type;
	}
}
