/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */
package org.objectweb.jorm.metainfo.lib;

import org.objectweb.jorm.metainfo.api.CompositeName;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.NameRef;
import org.objectweb.jorm.metainfo.api.Package;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.jorm.metainfo.api.Manager;

import java.util.Map;
import java.util.HashMap;

/**
 * The BasicNameRef is used to reference a composite name.
 * nameref are used in namedef to define pname with multiple fields.
 * @author	N. De Palma
 */
public class BasicNameRef extends BasicMetaObject implements NameRef {

	/**
	 * the name of the class reference
	 */
	private String name;

	/**
	 * name of the schema for the compositename referenced by this nameref
	 */
	private String cnSchemaName = null;

	/**
	 * the meta object describing the composite name referenced by this nameref.
	 */
	private CompositeName compositeNameClass;

	/**
	 * the list of the projection definition declared by the user in the
	 * current reference to the class.
	 * key: the name of the nameField
	 * value: the name of the classfield
	 */
	private HashMap projections;

	/**
	 * Builds a new BasicNameRef object.
	 * @param cn_Class	the composite name referenced by the nameref
	 * @param parent    the parent meta-object of the current nameref (this is a namedef).
	 */
	public BasicNameRef(CompositeName cn_Class, MetaObject parent) {
		super(parent);
		this.name = cn_Class.getName();
		this.compositeNameClass = cn_Class;
		projections = new HashMap();
	}

	/**
	 * Add a projection between a field in the composite name structure and a field in the class.
	 * @param namefield the name of the field in the composite name
	 * @param classfield the name of the field in the class
	 */
	public void addProjection(String namefield, String classfield) {
		projections.put(namefield, classfield);
	}

	/**
	 * Retrieve the list of projections between the fields of the composite name and the field of the class.
	 * @return a map object describing the projection.
	 */
	public Map getProjection() {
		return projections;
	}

	/**
	 * Return the class field name associated with the composite name field name.
	 * @param fieldName the name of a field in the composite name.
	 * @return a String object containing a class field name.
	 */
	public String getClassFieldName(String fieldName) {
		return (String) projections.get(fieldName);
	}

	/**
	 * Retrieve the name of the nameref. This name is equal to the name of the referenced composite name.
	 * @return the nameof the nameref used to describe the current pname.
	 */
	public String getName() {
		return name;
	}

	/**
	 * Retrieve the metaobject describing the composite name referenced by this nameref.
	 * @return the composite name.
	 */
	public CompositeName getCompositeName() {
		return compositeNameClass;
	}

	public boolean equals(Object nr) {
		if (nr instanceof NameRef)
			return ((NameRef) nr).getName().equals(this.name);
		else
			return false;
	}


	public String getCNPackageName() {
		if (cnSchemaName == null) {
			MetaObject mo = compositeNameClass;
			while (!(mo instanceof Package)) {
				mo = mo.getParent();
				if (mo instanceof Manager) return null;
			}
			cnSchemaName = ((Package) mo).getName();
		}
		return cnSchemaName;
	}

	/**
	 * test if the type of the compositename referenced in this nameref is equals to the type
	 * referenced by the nameref passed as parameter.
	 * @obj the nameref to compare to
	 * @return true if the schema name ans the compositename name are equal.
	 */
	public boolean equalCNType(Object obj) {
		if (!(obj instanceof NameRef)) return false;
		return ((name.equals(((NameRef) obj).getName())) &&
				(getCNPackageName().equals(((NameRef) obj).getCNPackageName())));
	}
}
