/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.metainfo.lib;

import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.ScalarField;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.api.PException;
import org.objectweb.util.monolog.api.BasicLevel;

import java.text.SimpleDateFormat;
import java.text.ParseException;
import java.math.BigDecimal;
import java.math.BigInteger;

/**
 * A ScalarField describes corresponds to a hidden field of a ClassObject or
 * a Generic Class. For Classes, it is typically used to represent PNames. For
 * Generic Class, it can also be used for the index attributes of the
 * Generic Class.
 * @author	X. Spengler
 */
public class BasicScalarField extends BasicPrimitiveElement implements ScalarField {

	/**
	 * This string represents the null value for the field.
	 */
	private String nullvalueStr = null;
	private Object nullValue = null;
	private boolean hasNullValue = false;;

	/**
	 * Builds a new BasicScalarField object.
	 *
	 * @param name  the name of the scalar field
	 * @param type  the PType of the field
	 * @param parent the parent meta-object of the current field
	 */
	public BasicScalarField(String name, PType type, int size, int scale, MetaObject parent) {
		super(name, type, size, scale, parent);
	}

	/**
	 * Set the null value for this field.
	 * @param value a string representation of the nullvalue
	 *
	 */
	public void setNullValue(String value) {
		nullvalueStr = value;
		hasNullValue = true;
		try {
			nullValue = parseValue(getType(), value);
		} catch (Exception e) {
			hasNullValue = false;
			if (logger != null) {
				logger.log(BasicLevel.WARN,
					"The null value specified for the field '" + getName()
					+ "' is malformed:", e);
			}
		}
	}

	/**
	 * get the null value for this field.
	 * @return the string representation of the null value or null if there is no null value for this field.
	 */
	public String getNullValue() {
		return nullvalueStr;
	}

	public Object getNullValueObject() {
		return nullValue;
	}

	/**
	 * Test if the null value exists for this field.
	 * @return true if the null value exist for this field.
	 */
	public boolean hasNullValue() {
		return hasNullValue;
	}

	public static Object parseValue(PType type, String value)
		throws NumberFormatException, ParseException, PException {
		if (type != null) {
			if (value == null) {
				return null;
			} else {
				switch (type.getTypeCode()) {
				case PType.TYPECODE_BYTE:
				case PType.TYPECODE_OBJBYTE:
					return Byte.valueOf(value);
				case PType.TYPECODE_CHAR:
				case PType.TYPECODE_OBJCHAR:
					return new Character(value.charAt(0));
				case PType.TYPECODE_SHORT:
				case PType.TYPECODE_OBJSHORT:
					return Short.valueOf(value);
				case PType.TYPECODE_INT:
				case PType.TYPECODE_OBJINT:
					return Integer.valueOf(value);
				case PType.TYPECODE_LONG:
				case PType.TYPECODE_OBJLONG:
					return Long.valueOf(value);
				case PType.TYPECODE_DATE:
					return new SimpleDateFormat().parse(value);
				case PType.TYPECODE_CHARARRAY:
					return value.toCharArray();
				case PType.TYPECODE_BYTEARRAY:
					return value.getBytes();
				case PType.TYPECODE_STRING:
					return value;
				case PType.TYPECODE_BIGDECIMAL:
					return new BigDecimal(value);
				case PType.TYPECODE_BIGINTEGER:
					return new BigInteger(value);
				default:
					throw new PException("Impossible to parse with unkknwon type: "
						+ type.getJormName());
				}
			}
		} else {
			throw new PException("Impossible to parse without type: "
				+ type.getJormName());
		}
	}

}
