/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.mi2xml.lib;

import org.apache.xerces.dom.DocumentImpl;
import org.objectweb.jorm.metainfo.api.*;
import org.objectweb.jorm.metainfo.api.Package;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.mi2xml.api.MappingDomtreeBuilder;
import org.objectweb.jorm.mi2xml.api.DomtreeBuilder;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.compiler.api.PExceptionCompiler;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.util.monolog.api.BasicLevel;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.util.*;

/**
 * BasicDomtreeBuilder implements the DomtreeBuilder interface.
 */
public class BasicDomtreeBuilder implements DomtreeBuilder, Loggable {
	//TODO: declare this in a class
	public final static String ID = "ID";
	public final static String REF = "REF";
	public final static String IDVALUE_SEP = ".";
	public final static String JORM = "jorm";
	public final static String PACKAGE = "package";
	public final static String CLASS = "class";
	public final static String MAPPING = "mapping";
	public final static String COMPOSITE_NAME = "composite-name";
	public final static String FIELD = "field";
	public final static String SCALAR_FIELD = "scalar-field";
	public final static String EXTENSION = "extension";
	public final static String NAME_DEF = "name-def";
	public final static String PRIMITIVE_TYPE = "primitive-type";
	public final static String CLASS_REF = "class-ref";
	public final static String GEN_CLASS_REF = "gen-class-ref";
	public final static String FIELD_REF = "field-ref";
	public final static String SCALAR_TYPE = "scalar-type";
	public final static String NULL_VALUE = "null-value";
	public final static String GEN_CLASS = "gen-class";
	public final static String INDEX = "index";
	public final static String SYSTEM = "system";
	public final static String COMPOSITE_NAME_REF = "composite-name-ref";
	public final static String COMPOSITE_NAME_FIELD_PROJECTION =
			"composite-name-field-projection";

	/**
	 * This hashmap associates a mapper name with a MappingDomtreeBuilder object.
	 */
	private Map mappingDomtreeBuilders;

	/**
	 * a simple logger to log
	 */
	private Logger logger;

	/**
	 * a logger factory to create other loggers if needed
	 */
	private LoggerFactory loggerFactory;

	/**
	 * A Map of GenClassRef objects.
	 * key = a GenClassRef object,
	 * value = a String object that represents the associated id value.
	 */
	private Map metaobject2idvalue;

	/**
	 * Main constructor for BasicDomtreeBuilder.
	 */
	public BasicDomtreeBuilder() {
		metaobject2idvalue = new Hashtable();
		mappingDomtreeBuilders = new HashMap();
	}

	/**
	 * Adds a MappingDomtreeBuilder object.
	 * @param  mapperName    a mapper name,
	 *         mdb           a MappingDomtreeBuilder object.
	 */
	public void addMappingDomtreeBuilder(String mapperName, MappingDomtreeBuilder mdb) {
		if (getLogger().isLoggable(BasicLevel.DEBUG)) {
			getLogger().log(BasicLevel.DEBUG,
							"Add a MappingDomtreeBuilder (" + mapperName +
							") for the current builder");
		}
		if (!mappingDomtreeBuilders.containsKey(mapperName)) {
			mappingDomtreeBuilders.put(mapperName, mdb);
		}
	}

	/**
	 * Returns a MappingDomtreeBuilder object.
	 * @param  mapperName    a mapper name,
	 * @return a MappingDomtreeBuilder object.
	 */
	private MappingDomtreeBuilder getMappingDomtreeBuilder(String mapperName) throws PException {
		MappingDomtreeBuilder mdb = (MappingDomtreeBuilder) mappingDomtreeBuilders.get(mapperName);
		if (mdb == null) {
			throw new PExceptionCompiler("<" + mapperName +
										 ">DomtreeBuilder has not been created.");
		}
		return mdb;
	}

	/**
	 * Generates the org.w3c.dom.Document object from a <tt>Class</tt>
	 * or a <tt>CompositeName</tt> object.
	 * @param   metaObject    a <tt>Class</tt> or a <tt>CompositeName</tt> object.
	 * @return the  corresponding <tt>org.w3c.dom.Document</tt> object.
	 *
	 * @see org.w3c.dom.Document
	 */
	public Document build(MetaObject metaObject) throws PException {
		Document document = new DocumentImpl();
		Element rootElement =
				document.createElement(JORM);
		// Creates an Element object for the Package object
		String packageName = ((Package) metaObject.getParent()).getName();
		Element packageElement = document.createElement(PACKAGE);
		packageElement.appendChild(document.createTextNode(packageName));
		// Adds packageElement to rootElement.
		rootElement.appendChild(packageElement);

		// Creates an Element object for a Class or a CompositeName object
		if (metaObject instanceof Class) {
			Class clazz = (Class) metaObject;
			String className = clazz.getName();
			Element classElement = document.createElement(CLASS);
			// Adds attributes to classElement.
			classElement.setAttribute("name", className);
			classElement.setAttribute("abstract",
									  ("" + clazz.isAbstract()).toUpperCase());
			// Adds classElement to rootElement
			rootElement.appendChild(classElement);
			// Builds the domtree enrooted at classElement
			processClass(document, classElement, clazz);
			if (!clazz.getClassProjects().isEmpty()) {
				processMappings(document, rootElement, clazz);
			}
		} else {
			CompositeName compositeName = (CompositeName) metaObject;
			Element compositeNameElement =
					document.createElement(COMPOSITE_NAME);
			// Add attributes to compositeName Element
			compositeNameElement.setAttribute("name", compositeName.getName());
			// Adds compositenameElement to rootElement
			rootElement.appendChild(compositeNameElement);
			// builds the domtree enrooted at classElement
			processCompositeName(document, compositeNameElement, compositeName);
		}
		document.appendChild(rootElement);
		return document;
	}


	/**
	 * Adds field elements, scalar field elements, inherited class elements, name defintions elements
	 * to a class element.
	 * @param   document        an org.w3c.dom.Document object,
	 *          classElement    a class element,
	 *          clazz           a <tt>Class</tt> object.
	 */
	private void processClass(Document document, Element classElement, Class clazz) {
		Iterator fieldIterator = clazz.getFields().iterator();
		// Processes the class fields.
		while (fieldIterator.hasNext()) {
			TypedElement typedElement = (TypedElement) fieldIterator.next();
			String fieldName = typedElement.getName();
			String idValue = fieldName;
			Element fieldElement = document.createElement(FIELD);
			fieldElement.setAttribute("name", fieldName);
			classElement.appendChild(fieldElement);
			// Processes PrimitiveElement objects if any.
			if (typedElement instanceof PrimitiveElement) {
				Element ptElement = document.createElement(PRIMITIVE_TYPE);
				//ptElement.setAttribute("id", idValue);
				ptElement.setAttribute("type", typedElement.getType().getJormName());
				fieldElement.appendChild(ptElement);
				//metaobject2idvalue.put(typedElement, idValue);
			}
			// Processes ClassRef objects if any.
			//TODO : add getFQName() to ClassRef
			else if (typedElement instanceof ClassRef) {
				Element crElement = document.createElement(CLASS_REF);
				crElement.setAttribute("class-name", ((ClassRef) typedElement).getMOClass().getFQName());
				fieldElement.appendChild(crElement);
				processClassRef(document, crElement, (ClassRef) typedElement, idValue);
			}
			// Processes GenClassRef objects if any.
			else if (typedElement instanceof GenClassRef) {
				Element gcrElement = document.createElement(GEN_CLASS_REF);
				//gcrElement.setAttribute("shared", ...);
				fieldElement.appendChild(gcrElement);
				processGenClassRef(document, gcrElement, (GenClassRef) typedElement, idValue);
			}
		}
		// Processes the ScalarField objects.
		Iterator sfIterator = clazz.getHiddenFields().iterator();
		while (sfIterator.hasNext()) {
			ScalarField scalarField = (ScalarField) sfIterator.next();
			String name = scalarField.getName();
			String idValue = name;
			Element sfElement = document.createElement(SCALAR_FIELD);
			sfElement.setAttribute("name", name);
			classElement.appendChild(sfElement);
			processScalarField(document, sfElement, scalarField, idValue);
		}
		// Processes the super classes.
		Iterator superIterator = clazz.getSuperClasses().iterator();
		while (superIterator.hasNext()) {
			Class superClazz = (Class) superIterator.next();
			Element extElement = document.createElement(EXTENSION);
            extElement.setAttribute("name", superClazz.getFQName());
			classElement.appendChild(extElement);
		}
		// Processes the NameDef objects.
		Iterator ndIterator = clazz.getNameDefs().iterator();
		while (ndIterator.hasNext()) {
			NameDef nameDef = (NameDef) ndIterator.next();
			/* String name = nameDef.getName();
			String idValue = ID + IDVALUE_SEP + nameDef.getName();
			Element ndElement = document.createElement(NAME_DEF);
			ndElement.setAttribute("id", idValue);
			*/
			Element ndElement = document.createElement(NAME_DEF);
			ndElement.setAttribute("name", nameDef.getName());
			classElement.appendChild(ndElement);
			//metaobject2idvalue.put(nameDef, idValue);
			processNameDef(document, ndElement, nameDef);
		}
	}

	/**
	 * Adds name definition elements to a class reference element.
	 * @param   document        an org.w3c.dom.Document object,
	 *          crElement       a class reference element,
	 *          classRef        a <tt>ClassRef</tt> object,
	 *          parentIdValue   a String value identifying the class reference element.
	 */
	private void processClassRef(Document document, Element crElement,
								 ClassRef classRef, String parentIdValue) {
		// Processes the NameDef objects.
		Iterator ndIterator = classRef.getRefNameDef().iterator();
		while (ndIterator.hasNext()) {
			NameDef nameDef = (NameDef) ndIterator.next();
			/* String name = nameDef.getName();
			String idValue = parentIdValue + IDVALUE_SEP + REF + IDVALUE_SEP + name;
			Element ndElement = document.createElement(NAME_DEF);
			ndElement.setAttribute("id", idValue);
			*/
			Element ndElement = document.createElement(NAME_DEF);
			ndElement.setAttribute("name", nameDef.getName());
			crElement.appendChild(ndElement);
			//metaobject2idvalue.put(nameDef, idValue);
			processNameDef(document, ndElement, nameDef);
		}
	}

	/**
	 * Adds a generic class element and name definition elements to a
	 * generic class reference element.
	 * @param   document        an org.w3c.dom.Document object,
	 *          gcrElement      a generic class reference element,
	 *          genClassRef     a <tt>GenClassRef</tt> object,
	 *          parentIdValue   a String value identifying the generic class reference element.
	 */
	private void processGenClassRef(Document document, Element gcrElement,
									GenClassRef genClassRef, String parentIdValue) {
		// Processes the generic class.
		/* String genClassName = genClassRef.getGenClassName();
		String genclassIdValue = parentIdValue + IDVALUE_SEP + genClassName;
		*/
		Element gcElement = document.createElement(GEN_CLASS);
		//gcElement.setAttribute("id", genclassIdValue);
		gcElement.setAttribute("gen-class-name", genClassRef.getGenClassName());
		gcrElement.appendChild(gcElement);
		metaobject2idvalue.put(genClassRef, genClassRef.getGenClassId());
		processGenClass(document, gcElement, genClassRef, "");
		// Processes the NameDef objects.
		Iterator ndIterator = genClassRef.getRefNameDef().iterator();
		while (ndIterator.hasNext()) {
			NameDef nameDef = (NameDef) ndIterator.next();
			/* String name = nameDef.getName();
			String nameDefIdValue = parentIdValue + IDVALUE_SEP + REF + IDVALUE_SEP + name;
			Element ndElement = document.createElement(NAME_DEF);
			ndElement.setAttribute("id", nameDefIdValue);
			*/
			Element ndElement = document.createElement(NAME_DEF);
			ndElement.setAttribute("name", nameDef.getName());
			gcrElement.appendChild(ndElement);
			//metaobject2idvalue.put(nameDef, nameDefIdValue);
			processNameDef(document, ndElement, nameDef);
		}
	}

	/**
	 * Adds a scalar type element and possibly a null value element to
	 * a scalar field element.
	 * @param   document    an org.w3c.dom.Document object,
	 *          sfElement   a name definiton element,
	 *          nameDef    a <tt>NameDef</tt> object,
	 *          idValue       a String value identifying the scalar field element.
	 */
	private void processScalarField(Document document, Element sfElement,
									ScalarField scalarField, String idValue) {
		// Processes the scalar type.
		Element stElement = document.createElement(SCALAR_TYPE);
		//stElement.setAttribute("id", idValue);
		stElement.setAttribute("type", scalarField.getType().getJormName());
		sfElement.appendChild(stElement);
		//metaobject2idvalue.put(scalarField, idValue);
		// Processes the null value.
		if (scalarField.getNullValue() != null) {
			Element nvElement = document.createElement(NULL_VALUE);
			nvElement.setAttribute("value", scalarField.getNullValue());
			stElement.appendChild(nvElement);
		}
	}

	/**
	 * Adds a system or a field reference or a composite name reference element to
	 * a name definition element.
	 * @param   document    an org.w3c.dom.Document object,
	 *          ndElement   a name definiton element,
	 *          nameDef     a <tt>NameDef</tt> object.
	 */
	private void processNameDef(Document document, Element ndElement, NameDef nameDef) {
		// Processes a system or a field reference, or a composite name reference.
		if (nameDef.isSystem()) {
			Element systemElement = document.createElement(SYSTEM);
			ndElement.appendChild(systemElement);
		} else if (nameDef.getFieldName() != null) {
			Element frElement = document.createElement(FIELD_REF);
			frElement.setAttribute("field-name", nameDef.getFieldName());
			ndElement.appendChild(frElement);
		} else if (nameDef.isNameRef()) {
			NameRef nameRef = nameDef.getNameRef();
			Element cnrElement = document.createElement(COMPOSITE_NAME_REF);
			cnrElement.setAttribute("composite-name-name",
									nameRef.getCNPackageName() + "." + nameRef.getName());
			ndElement.appendChild(cnrElement);
			processCompositeNameRef(document, cnrElement, nameRef);
		}
	}

	/**
	 * Adds composite name field projection elements to a composite name reference element.
	 * @param   document    an org.w3c.dom.Document object,
	 *          cnrElement   a composite name reference element,
	 *          nameRef     a <tt>NameRef</tt> object.
	 */
	private void processCompositeNameRef(Document document, Element cnrElement, NameRef nameRef) {
		// Processes the composite name field projection(s).
		Iterator projIterator = nameRef.getProjection().keySet().iterator();
		while (projIterator.hasNext()) {
			String fieldName = (String) projIterator.next();
			String classFieldName = nameRef.getClassFieldName(fieldName);
			Element cnfpElement = document.createElement(COMPOSITE_NAME_FIELD_PROJECTION);
			cnfpElement.setAttribute("composite-name-field-name", fieldName);
			cnfpElement.setAttribute("class-field-name", classFieldName);
			cnrElement.appendChild(cnfpElement);
		}
	}

	/**
	 * Adds a generic class element and name definition elements to a
	 * generic class reference element.
	 * @param   document      an org.w3c.dom.Document object,
	 *          gcrElement    a generic class reference element,
	 *          genClassRef   a <tt>GenClassRef</tt> object,
	 *          parentIdValue a String value identifying the generic class element.
	 */
	private void processGenClass(Document document, Element gcElement,
								 GenClassRef genClassRef, String parentIdValue) {
		// Processes the ScalarField objects.
		Iterator sfIterator = genClassRef.getHiddenFields().iterator();
		while (sfIterator.hasNext()) {
			ScalarField scalarField = (ScalarField) sfIterator.next();
			//String name = scalarField.getName();
			//String idValue =  parentIdValue + IDVALUE_SEP + name;
			Element sfElement = document.createElement(SCALAR_FIELD);
			sfElement.setAttribute("name", scalarField.getName());
			gcElement.appendChild(sfElement);
			//metaobject2idvalue.put(scalarField, idValue);
			processScalarField(document, sfElement, scalarField, "");
		}
		// Processes the NameDef objects.
		Iterator ndIterator = genClassRef.getIdNameDef().iterator();
		while (ndIterator.hasNext()) {
			NameDef nameDef = (NameDef) ndIterator.next();
			/*
			String name = nameDef.getName();
			String idValue = parentIdValue + IDVALUE_SEP + ID + IDVALUE_SEP + name;
			Element ndElement = document.createElement(NAME_DEF);
			ndElement.setAttribute("id", idValue);
			*/
			Element ndElement = document.createElement(NAME_DEF);
			ndElement.setAttribute("name", nameDef.getName());
			gcElement.appendChild(ndElement);
			//metaobject2idvalue.put(nameDef, idValue);
			processNameDef(document, ndElement, nameDef);
		}
		// Processes the index.
		Iterator frIterator = genClassRef.getIndexFields().iterator();
		if (frIterator.hasNext()) {
			Element indexElement = document.createElement(SCALAR_FIELD);
			gcElement.appendChild(indexElement);
			while (frIterator.hasNext()) {
				ScalarField scalarField = (ScalarField) frIterator.next();
				String name = scalarField.getName();
				Element frElement = document.createElement(FIELD_REF);
				frElement.setAttribute("name", name);
				indexElement.appendChild(frElement);
			}
		}
		// Processes PrimitiveElement objects if any.
		if (genClassRef.isPrimitive()) {
			PrimitiveElement primitiveElement = genClassRef.getPrimitiveElement();
			String fieldName = primitiveElement.getName();
			/*
			String idValue = parentIdValue + IDVALUE_SEP + fieldName;
			Element ptElement = document.createElement(PRIMITIVE_TYPE);
			ptElement.setAttribute("id", idValue);
			*/
			Element ptElement = document.createElement(PRIMITIVE_TYPE);
			ptElement.setAttribute("type", primitiveElement.getType().getJormName());
			//metaobject2idvalue.put(primitiveElement, idValue);
			gcElement.appendChild(ptElement);

		}
		// Processes ClassRef objects if any.
		else if (genClassRef.isClassRef()) {
			ClassRef classRef = genClassRef.getClassRef();
			Element crElement = document.createElement(CLASS_REF);
			crElement.setAttribute("name", classRef.getClassName());
			gcElement.appendChild(crElement);
			processClassRef(document, crElement, classRef, parentIdValue);
		}
		// Processes GenClassRef objects if any.
		else if (genClassRef.isGenClassRef()) {
			GenClassRef subGenClassRef = genClassRef.getGenClassRef();
			Element gcrElement = document.createElement(GEN_CLASS_REF);
			gcElement.appendChild(gcrElement);
			processGenClassRef(document, gcrElement, subGenClassRef, parentIdValue);
		}
	}

    //TODO : extension
	/**
	 * Adds a composite name element to a jorm element (the root).
	 * @param   document      an org.w3c.dom.Document object,
	 *          indexElement  an index element,
	 *          scalarField   a <tt>ScalarField</tt> object.
	 */
	private void processCompositeName(Document document, Element cnElement,
									  CompositeName compositeName) {
		// Processes the ScalarField objects.
		Iterator sfIterator = compositeName.getFields().iterator();
		while (sfIterator.hasNext()) {
			ScalarField scalarField = (ScalarField) sfIterator.next();
			String name = scalarField.getName();
			Element sfElement = document.createElement(SCALAR_FIELD);
			sfElement.setAttribute("name", name);
			cnElement.appendChild(sfElement);
			processScalarField(document, sfElement, scalarField, name);
		}
		// Processes the inherited CompositeName objects.
		Iterator icnIterator = compositeName.getSuperCompositeNames().iterator();
		while (icnIterator.hasNext()) {
			CompositeName aCompositeName = (CompositeName) icnIterator.next();
			Element extensionElement = document.createElement(EXTENSION);
			extensionElement.appendChild(document.createTextNode(aCompositeName.getName()));
			cnElement.appendChild(extensionElement);
		}
	}

	/**
	 * Adds mapping elements to a jorm element (the root).
	 * @param   document        an org.w3c.dom.Document object,
	 *          rootElement     the jorm element,
	 *          clazz           the current  <tt>Class</tt> object.
	 */
	private void processMappings(Document document, Element rootElement, Class clazz)
			throws PException {
		MappingDomtreeBuilder mdb = null;
		// Processes the ClassProject objects.
		Iterator cpIterator = clazz.getClassProjects().iterator();
		while (cpIterator.hasNext()) {
			ClassProject classProject = (ClassProject) cpIterator.next();
			String projectName = classProject.getProjectName();
			Element mappingElement = document.createElement(MAPPING);
			mappingElement.setAttribute("project-name", projectName);
			rootElement.appendChild(mappingElement);
			// Processes the Mapping objects.
			Iterator mappingIterator = classProject.getMappings().iterator();
			while (mappingIterator.hasNext()) {
				Mapping mapping = (Mapping) mappingIterator.next();
				String mapperName = mapping.getMapperName();
				// Gets the right builder according to the current mapper type
				// ("rdb", "fos", etc.)
				mdb = getMappingDomtreeBuilder(mapperName);
				((Loggable) mdb).setLogger(logger);
				mdb.setCurrentClass(clazz);
				mdb.setmetaobject2idvalue(metaobject2idvalue);
				ClassMapping classMapping = mapping.getClassMapping();
				if (logger.isLoggable(BasicLevel.DEBUG)) {
					logger.log(BasicLevel.DEBUG,
							   "current class =<" + clazz.getName() + ">");
					logger.log(BasicLevel.DEBUG,
							   "current project =<" + projectName + ">");
					logger.log(BasicLevel.DEBUG,
							   "current mapping =<" + mapperName + ">");
				}
				mdb.processClassMapping(document, mappingElement, classMapping);
				Iterator gcmIterator = mapping.getGenClassMappings().iterator();
				while (gcmIterator.hasNext()) {
					GenClassMapping genClassMapping = (GenClassMapping) gcmIterator.next();
					mdb.processGenClassMapping(document, mappingElement, genClassMapping);
				}
			}
		}
	}
	///////////////////////////////////////////////////////////////////
	//                      from Loggable interface
	///////////////////////////////////////////////////////////////////

	/**
	 * Defines a logger object.
	 *
	 * @param logger	the logger object
	 */
	public void setLogger(Logger logger) {
		this.logger = logger;
	}

	/**
	 * Defines the logger factory to obtain new loggers.
	 *
	 * @param loggerfactory	The LoggerFactory object to obtain a logger object
	 */
	public void setLoggerFactory(LoggerFactory loggerfactory) {
		this.loggerFactory = loggerfactory;
		if (logger == null && loggerFactory != null) {
			logger = loggerFactory.getLogger("org.objectweb.jorm.mi2xml.rdb");
		}
	}

	public Logger getLogger() {
		return logger;
	}

	public LoggerFactory getLoggerFactory() {
		return loggerFactory;
	}
}
