/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.naming.api;

import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PBindingCtrl;
import org.objectweb.jorm.api.PBinding;
import org.objectweb.perseus.cache.api.CacheManager;

import java.util.Iterator;

/**
 * This interface defines a binder as a particular PNameManager that manages
 * the association of bindings with PName. Each JORM class managed by a mapper
 * have an associated PBinder. * Binders provide binding objects specific to the persistent class.
 * @author	R. Basset, P. D?chamboux
 */
public interface PBinder extends PNameManager {
	/**
	 * It looks for a PBinding object with the given PName within a PBinder.
	 * @param pn    The PName to associate to the PBinding. This PName must be
	 * 				valid within this binder just like with a naming context.
	 * @return	The PBinding bound with the given PName.
	 * @exception PExceptionNaming	The given PName is not valid in this
	 * 								PBinder or this PBinding cannot be managed
	 * 								by this PBinder.
	 */
	PBinding lookup(PName pn) throws PException;

	/**
	 * It associates a PBinding object with a PName within a PBinder. This
	 * sets/resets the link between a PBinding and a DSI.
	 * @param pn    The PName to associate to the PBinding. This PName must be
	 * 				valid within this binder just like with a naming context.
	 * @param pb	The PBinding to bind with the given PName; this PBinding
	 * 				must have been created or initialised by this PBinder.
	 * @exception PExceptionNaming	The given PName is not valid in this
	 * 								PBinder or this PBinding cannot be managed
	 * 								by this PBinder.
	 */
	void bind(PName pn, PBindingCtrl pb) throws PException;

	/**
	 * It returns the PClassMapping for which it is the binder.
	 * @return	The PClassMapping to which it is associated.
	 */
	PClassMapping getBinderClassMapping();

	/**
	 * It set the PClassMapping which use this binder slave.
	 * @param pcm The PClassMapping.
	 */
	void setPClassMapping(PClassMapping pcm);

	/**
	 * It releases the link between the PBinding passed as parameter and its
	 * associated DSI represented by the PName referenced by this PBinding.
	 * @param pb	The PBinding to unbind.
	 * @exception PExceptionNaming	This PBinding cannot be managed by this
	 * 								PBinder.
	 */
	void unbind(PBindingCtrl pb) throws PException;

	/**
	 * Returns the cache manager associated with this binder.
	 * @return  The cache manager.
	 */
	CacheManager getCacheManager();

	/**
	 * Assigns the cache manager associated with this binder.
	 * @param cm    The cache manager.
	 * @throws PException
	 */
	void setCacheManager(CacheManager cm) throws PException;
}
