/**
 * Copyright (C) 2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.jorm.naming.lib;

import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PBinder;
import org.objectweb.jorm.naming.api.PNameCoder;
import org.objectweb.jorm.naming.api.FilteredPNamingContext;
import org.objectweb.jorm.naming.api.PNameGetterConverter;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.api.PException;

import java.util.Date;
import java.util.Arrays;
import java.math.BigInteger;
import java.math.BigDecimal;

/**
 * This class is abstract implementation of the FilteredPNamingContext interface.
 * The aim is to manage a tree of naming contexts. At each node a filter is able
 * to say if a pname structure matches to the type represented by the
 * PNamingContext. To use this class it is required
 * - to override the right match and decodeXXX methods in according to the
 * PARENT coding type,
 * - to override the right encodeXXX methods in according to the coding type,
 * - to implement the encodeString and decodeString method
 *
 * Example of overriding of a decodeXXX method:
 * ============================================
 * public PName decodeAbstract(Object en) throws PExceptionNaming {
 *      //check coding type asked with the parent coding type
 *      if (parentCodingType != CTCOMPOSITE) {
 *          throw new PExceptionNaming("Decoding a composite name is not supported: " + en);
 *      }
 *
 *      //obtain a PName with the 'en' parameter
 *      PName pn;
 *      if (en instanceof PName) {
 *          pn = (PName) en
 *      } else if (en instanceof ID1ID2PNG) {
 *          pn = new ID1ID2PName(this,
 *              ((ID1ID2PNG) en).pnGetId1(),
 *              ((ID1ID2PNG) en).pnGetId1());
 *      } else if (en instanceof PNameGetter) {
 *          pn = new ID1ID2PName(this,
 *              ((PNameGetter) en).pnGetString("id1"),
 *              ((PNameGetter) en).pnGetLong("id2") );
 *      } else {
 *          throw new PExceptionNaming("Impossible to decode this parameter: " + en);
 *      }
 *
 *      // resolve the PName
 *      try {
 *          return resolve(null, pn);
 *      } catch (PException e) {
 *          throw new PExceptionNaming(e, "Impossible to resolve " + en);
 *      }
 * }
 *
 * Example of overriding of a match method:
 * ========================================
 * public boolean match(Object en) throws PExceptionNaming {
 *      if (parentCodingType != CTCOMPOSITE) {
 *          throw new PExceptionNaming("Decoding a composite name is not supported: " + en);
 *      }
 *      String id1 = ((FooPNG) en).getId1();
 *      long id2 = ((FooPNG) en).getId2();
 *      return id2 > 0 && "kangaroo".equals(id1)
 * }
 *
 *
 * Example of overriding of a encodeXXX method:
 * ============================================
 * public Object encodeAbstract(PName pn) throws PExceptionNaming {
 *      return pn; //The PName is a PNameGetter
 * }
 *
 *
 * @author S.Chassande-Barrioz
 * @author P.Dechamboux
 */
public abstract class AbstractInheritFilteredPNamingContext
    extends BasicPNamingContext
    implements FilteredPNamingContext {

	/**
	 * The list of sub PNameCoderSelector
	 * element type: PNameCoderSelector
	 */
	private FilteredPNamingContext[] fpncs = new FilteredPNamingContext[0];

	/**
	 * The Binder associated to the current type. A null value means the type is
	 * abstract.
	 */
	protected PBinder binder;

	/**
	 * The codingType of the current PNamingContext
	 */
	protected int codingType;

	/**
	 * the null PName
	 */
	private PName nullPName;

	protected PNameGetterConverter converter;

	public AbstractInheritFilteredPNamingContext() {
	}

	public AbstractInheritFilteredPNamingContext(int codingType) {
		this.codingType = codingType;
	}

	public AbstractInheritFilteredPNamingContext(int codingType, PNameGetterConverter converter) {
		this.codingType = codingType;
		this.converter = converter;
	}

	public PBinder getBinder() {
		return binder;
	}

	public void setBinder(PBinder binder) {
		this.binder = binder;
	}

	public void setCodingType(int codingType) {
		this.codingType = codingType;
	}

	public PNameGetterConverter getConverter() {
		return converter;
	}

	public void setConverter(PNameGetterConverter converter) {
		this.converter = converter;
	}

	public synchronized void bindSubFPNC(FilteredPNamingContext fpnc) {
		FilteredPNamingContext[] newfpncs = new FilteredPNamingContext[fpncs.length + 1];
		System.arraycopy(fpncs, 0, newfpncs, 0, fpncs.length);
		newfpncs[fpncs.length] = fpnc;
		fpncs = newfpncs;
	}

	public synchronized void unbindSubFPNC(FilteredPNamingContext fpnc) {
		int j = Arrays.binarySearch(fpncs, fpnc);
		if (j == -1) {
			//The PNC is not bound
			return;
		}
		FilteredPNamingContext[] newfpncs = new FilteredPNamingContext[fpncs.length - 1];
		if (j > 0) {
			//Copy the begin
			System.arraycopy(fpncs, 0, newfpncs, 0, j);
		}
		if (j < newfpncs.length) {
			//Copy the end
			System.arraycopy(fpncs, j + 1, newfpncs, j, fpncs.length - j);
		}
		fpncs = newfpncs;
	}

	public boolean codingSupported(int codingtype) {
		return codingtype == codingType || codingtype == PNameCoder.CTSTRING;
	}

	public PName getNull() {
		return nullPName;
	}

	public void setNullPName(Object o) throws PException {
		nullPName = (PName) o;
	}

	public PName resolve(Object conn, PName pn) throws PException {
		int i = fpncs.length - 1;
		switch (codingType) {
		case CTCOMPOSITE: {
			Object o = pn.encodeAbstract();
			while(i>=0 && !fpncs[i].match(o)) {
				i--;
			}
			if (i>=0) {
				return fpncs[i].decodeAbstract(o, null);
			} else if (binder != null) {
				return binder.decodeAbstract(o, null);
			} else {
				return pn;
			}
		}
		case CTCHAR:{
			char o = pn.encodeChar();
			while(i>=0 && !fpncs[i].match(o)) {
				i--;
			}
			if (i>=0) {
				return fpncs[i].decodeChar(o);
			} else if (binder != null) {
				return binder.decodeChar(o);
			} else {
				return pn;
			}
		}

		case CTOCHAR: {
			Character o = pn.encodeOchar();
			while(i>=0 && !fpncs[i].match(o)) {
				i--;
			}
			if (i>=0) {
				return fpncs[i].decodeOchar(o);
			} else if (binder != null) {
				return binder.decodeOchar(o);
			} else {
				return pn;
			}
		}

		case CTBYTE: {
			byte o = pn.encodeByte();
			while(i>=0 && !fpncs[i].match(o)) {
				i--;
			}
			if (i>=0) {
				return fpncs[i].decodeByte(o);
			} else if (binder != null) {
				return binder.decodeByte(o);
			} else {
				return pn;
			}
		}

		case CTOBYTE: {
			Byte o = pn.encodeObyte();
			while(i>=0 && !fpncs[i].match(o)) {
				i--;
			}
			if (i>=0) {
				return fpncs[i].decodeObyte(o);
			} else if (binder != null) {
				return binder.decodeObyte(o);
			} else {
				return pn;
			}
		}

		case CTSHORT: {
			short o = pn.encodeShort();
			while(i>=0 && !fpncs[i].match(o)) {
				i--;
			}
			if (i>=0) {
				return fpncs[i].decodeShort(o);
			} else if (binder != null) {
				return binder.decodeShort(o);
			} else {
				return pn;
			}
		}

		case CTOSHORT: {
			Short o = pn.encodeOshort();
			while(i>=0 && !fpncs[i].match(o)) {
				i--;
			}
			if (i>=0) {
				return fpncs[i].decodeOshort(o);
			} else if (binder != null) {
				return binder.decodeOshort(o);
			} else {
				return pn;
			}
		}

		case CTINT: {
			int o = pn.encodeInt();
			while(i>=0 && !fpncs[i].match(o)) {
				i--;
			}
			if (i>=0) {
				return fpncs[i].decodeInt(o);
			} else if (binder != null) {
				return binder.decodeInt(o);
			} else {
				return pn;
			}
		}

		case CTOINT: {
			Integer o = pn.encodeOint();
			while(i>=0 && !fpncs[i].match(o)) {
				i--;
			}
			if (i>=0) {
				return fpncs[i].decodeOint(o);
			} else if (binder != null) {
				return binder.decodeOint(o);
			} else {
				return pn;
			}
		}

		case CTLONG: {
			long o = pn.encodeLong();
			while(i>=0 && !fpncs[i].match(o)) {
				i--;
			}
			if (i>=0) {
				return fpncs[i].decodeLong(o);
			} else if (binder != null) {
				return binder.decodeLong(o);
			} else {
				return pn;
			}
		}

		case CTOLONG: {
			Long o = pn.encodeOlong();
			while(i>=0 && !fpncs[i].match(o)) {
				i--;
			}
			if (i>=0) {
				return fpncs[i].decodeOlong(o);
			} else if (binder != null) {
				return binder.decodeOlong(o);
			} else {
				return pn;
			}
		}

		case CTSTRING: {
			String o = pn.encodeString();
			while(i>=0 && !fpncs[i].match(o)) {
				i--;
			}
			if (i>=0) {
				return fpncs[i].decodeString(o);
			} else if (binder != null) {
				return binder.decodeString(o);
			} else {
				return pn;
			}
		}

		case CTBYTEARRAY: {
			byte[] o = pn.encode();
			while(i>=0 && !fpncs[i].match(o)) {
				i--;
			}
			if (i>=0) {
				return fpncs[i].decode(o);
			} else if (binder != null) {
				return binder.decode(o);
			} else {
				return pn;
			}
		}

		case CTCHARARRAY: {
			char[] o = pn.encodeCharArray();
			while(i>=0 && !fpncs[i].match(o)) {
				i--;
			}
			if (i>=0) {
				return fpncs[i].decodeCharArray(o);
			} else if (binder != null) {
				return binder.decodeCharArray(o);
			} else {
				return pn;
			}
		}

		case CTDATE: {
			Date o = pn.encodeDate();
			while(i>=0 && !fpncs[i].match(o)) {
				i--;
			}
			if (i>=0) {
				return fpncs[i].decodeDate(o);
			} else if (binder != null) {
				return binder.decodeDate(o);
			} else {
				return pn;
			}
		}

		case CTBIGDECIMAL: {
			BigDecimal o = pn.encodeBigDecimal();
			while(i>=0 && !fpncs[i].match(o)) {
				i--;
			}
			if (i>=0) {
				return fpncs[i].decodeBigDecimal(o);
			} else if (binder != null) {
				return binder.decodeBigDecimal(o);
			} else {
				return pn;
			}
		}

		case CTBIGINTEGER: {
			BigInteger o = pn.encodeBigInteger();
			while(i>=0 && !fpncs[i].match(o)) {
				i--;
			}
			if (i>=0) {
				return fpncs[i].decodeBigInteger(o);
			} else if (binder != null) {
				return binder.decodeBigInteger(o);
			} else {
				return pn;
			}
		}

		default:
			return getNull();
		}
	}

	public PName export(Object conn, Object infoitem) throws PException {
		return export(conn, infoitem, null);
	}

	public PName export(Object conn, Object infoitem, Object hints)
	    throws PException {
		if (infoitem instanceof PName) {
			return (PName) infoitem;
		} else {
			return decodeAbstract(infoitem, hints);
		}
	}

	public void unexport(Object conn, PName pn) throws PException {
		//nothing to do
	}

	public void unexport(Object conn, PName pn, Object hints) throws PException {
		//nothing to do
	}

	public String encodeString(PName pn) throws PExceptionNaming {
		throw new UnsupportedOperationException("String coding not supported");
	}
}
