/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2005 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Authors: S.Chassande-Barrioz.
 * Created on 28 juin 2005
 *
 */
package org.objectweb.jorm.naming.lib;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Date;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PNameGetter;
import org.objectweb.jorm.naming.api.PolymorphicPName;
import org.objectweb.jorm.type.api.PExceptionTyping;

/**
 * This abstract class helps the implementation of PName for composite name.
 *
 * @author S.Chassande-Barrioz
 */
public abstract class BasicCompositePName 
	extends BasicPolymorphicPName
	implements PNameGetter {

    protected abstract boolean checkPName(Object obj);
    protected abstract boolean valuesEqual(PName obj);
    
    public BasicCompositePName() {
    }

    public boolean isNull() {
        return this == pnc.getNull();
    }

    /**
	 * Overides the equals method.
	 * If none of the 2 pnames is ploymorphic, uses the normal equals method.
	 * If this pname is polymorphic, compare the values.
	 * If the obj pname is polymorphic, call obj.equals(this).
	 */
    public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (!checkPName(obj)) {
			return false;
		}
		BasicCompositePName pn = (BasicCompositePName) obj;
   		if (isPolymorphic()) {
   	   		//this is polymorphic
			//then compares the ptypes and the values
			return (this.getPType().isa(pn.getPType()) 
						|| pn.getPType().isa(this.getPType())) 
					&&  valuesEqual(pn);
		} else  if (pn.isPolymorphic()) {
	  		//pn is polymorphic, then calls the equals method on it
    		return pn.equals(this);
    	} else {
        	//none of them is polymorphic
    	    //then compare pnc
    		if (pn.pnc == this.pnc) {
    		    //same pnc, then compare values
    		    if (valuesEqual(pn)) {
    		        return true;
    		    }
    		}
	        return false;
    	}
    }

    public boolean equal(char[] ba1, char[] ba2) {
        if (ba1 == null) {
            return ba2 == null;
        } else {
            if (ba2 == null || ba1.length != ba2.length)
                return false;
            for(int i= 0; i< ba1.length; i++) {
                if (ba1[i] != ba2[i]) {
                    return false;
                }
            }
        }
        return true;
    }

    public boolean equal(byte[] ba1, byte[] ba2) {
        if (ba1 == null) {
            return ba2 == null;
        } else {
            if (ba2 == null || ba1.length != ba2.length)
                return false;
            for(int i= 0; i< ba1.length; i++) {
                if (ba1[i] != ba2[i]) {
                    return false;
                }
            }
        }
        return true;
    }

    // IMPLEMENTATION OF THE GENERIC PNameGetter INTERFACE //
    //-----------------------------------------------------//

    public byte pngetByteField(String fn, Object ctx) throws PException {
        throw new PExceptionTyping("Unsupported field type: type=byte, fn=" + fn);
    }

    public Byte pngetObyteField(String fn, Object ctx) throws PException {
        throw new PExceptionTyping("Unsupported field type: type=java.lang.Byte, fn=" + fn);
    }

    public char pngetCharField(String fn, Object ctx) throws PException {
        throw new PExceptionTyping("Unsupported field type: type=char, fn=" + fn);
    }

    public Character pngetOcharField(String fn, Object ctx) throws PException {
        throw new PExceptionTyping("Unsupported field type: type=java.lang.Character, fn=" + fn);
    }

    public short pngetShortField(String fn, Object ctx) throws PException {
        throw new PExceptionTyping("Unsupported field type: type=short, fn=" + fn);
    }

    public Short pngetOshortField(String fn, Object ctx) throws PException {
        throw new PExceptionTyping("Unsupported field type: type=java.lang.Short, fn=" + fn);
    }

    public int pngetIntField(String fn, Object ctx) throws PException {
        throw new PExceptionTyping("Unsupported field type: type=int, fn=" + fn);
    }

    public Integer pngetOintField(String fn, Object ctx) throws PException {
        throw new PExceptionTyping("Unsupported field type: type=java.lang.Integer, fn=" + fn);
    }

    public long pngetLongField(String fn, Object ctx) throws PException {
        throw new PExceptionTyping("Unsupported field type: type=long, fn=" + fn);
    }

    public Long pngetOlongField(String fn, Object ctx) throws PException {
        throw new PExceptionTyping("Unsupported field type: type=java.lang.Long, fn=" + fn);
    }

    public String pngetStringField(String fn, Object ctx) throws PException {
        throw new PExceptionTyping("Unsupported field type: type=java.lang.String, fn=" + fn);
    }

    public byte[] pngetByteArrayField(String fn, Object ctx) throws PException {
        throw new PExceptionTyping("Unsupported field type: type=byte[], fn=" + fn);
    }

    public char[] pngetCharArrayField(String fn, Object ctx) throws PException {
        throw new PExceptionTyping("Unsupported field type: type=char[], fn=" + fn);
    }

    public Date pngetDateField(String fn, Object ctx) throws PException {
        throw new PExceptionTyping("Unsupported field type: type=java.util.Date, fn=" + fn);
    }

    public BigInteger pngetBigIntegerField(String fn, Object ctx) throws PException {
        throw new PExceptionTyping("Unsupported field type: type=java.math.BigInteger, fn=" + fn);
    }

    public BigDecimal pngetBigDecimalField(String fn, Object ctx) throws PException {
        throw new PExceptionTyping("Unsupported field type: type=java.math.BigDecimal, fn=" + fn);
    }
}
