/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */
package org.objectweb.jorm.naming.lib;

import org.objectweb.jorm.naming.api.PBinder;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.api.PBinding;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.api.PBindingCtrl;
import org.objectweb.jorm.api.PStateGraph;
import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.perseus.cache.api.CacheEntry;
import org.objectweb.perseus.cache.api.CacheException;
import org.objectweb.perseus.cache.api.FixableCacheEntry;
import org.objectweb.perseus.cache.api.CacheManager;

/**
 * It defines the a basic PBinder which able to be also a PNamingContext.
 *
 * @author S.Chassande-Barrioz
 */
public abstract class BasicPBinder
		extends BasicPNamingContext
		implements PBinder {

	protected CacheManager cache;
	protected PClassMapping pcm;
	protected PName nullPName;

	public BasicPBinder() {
	}
	public BasicPBinder(PClassMapping _pcm) {
	    this.pcm = _pcm;
	}

	public BasicPBinder(PName nullPName) {
		this.nullPName = nullPName;
	}

	public String getClassName() {
		return ptype.getJormName();
	}


	// IMPLEMENTATION OF THE METHOD OF PBinder INTERFACE //
	//---------------------------------------------------//

	public PBinding lookup(PName pn) throws PException {
		if (pn == null) {
			throw new PExceptionNaming("[" + getClassName() + "]: this pname is null");
		}
		if (pn.isNull()) {
			return null;
		}
		if (cache == null) {
			throw new PExceptionProtocol("No internal cache: should be managed externally!");
		}
		CacheEntry ce = cache.lookup(pn);
		return (ce == null ? null : (PBinding) ce.getCeObject());
	}

	public void bind(PName pn, PBindingCtrl pb) throws PException {
		if (pn == null)
			throw new PExceptionNaming("[" + getClassName() + "]: this pname is null");
		if (pb == null)
			throw new PExceptionNaming("[" + getClassName() + "]: this pbinding is null");
		if (!pn.getClass().equals(getNull().getClass()))
			throw new PExceptionNaming("Can only bind " + getNull().getClass() + ", found: " + pn);
		if (!pn.getPNameManager().equals(this))
			throw new PExceptionNaming("[" + getClassName() + "]: this pname is not valid in this binder ("
									   + ", \nthis:" + this
									   + ", \nthis.type:" + this.getPType().getJormName()
									   + ", \nthis.pcm:" + this.getBinderClassMapping()
									   + ", \npn.pnc: " + pn.getPNameManager()
									   + ", \npn.pnc.type: " + pn.getPNameManager().getPType().getJormName()
									   + ", \npn.pnc.pcm: " + ((PBinder) pn.getPNameManager()).getBinderClassMapping()
									   + ", \npn:" + pn + ")");
		if (!pb.getPClassMapping().equals(pcm))
			throw new PExceptionNaming("[" + getClassName() + "]: this pbinding is not valid in this binder");
		if (pn.isNull())
			throw new PExceptionNaming("[" + getClassName() + "]: this PName represent the null value");
		byte nextstate = PStateGraph.nextStatePBinding(pb.getStatus(),
													   PBinding.ACTION_BIND);
		// Performs the transition
		if (cache != null) {
			synchronized (cache) {
				try {
					cache.fix(cache.bind(pn, pb));
				} catch (CacheException e) {
					throw new PException(e, "[" + getClassName() + "]: problem with cache management");
				}
			}
		}
		pb.setPName(pn);
		pb.setStatus(nextstate);
	}

	public PClassMapping getBinderClassMapping() {
		return pcm;
	}

	public void setPClassMapping(PClassMapping pcm) {
		this.pcm = pcm;
	}

	public void unbind(PBindingCtrl pb) throws PException {
		byte nextstate = PStateGraph.nextStatePBinding(pb.getStatus(),
													   PBinding.ACTION_UNBIND);
		// If the state does not change, do nothing
		if (nextstate == pb.getStatus())
			return;
		// Perform the transition
		if (pb == null)
			throw new PExceptionNaming("[" + getClassName() + "]: this pbinding is null");
		PName pn = pb.getPName();
		if (pn == null)
			throw new PExceptionNaming("[" + getClassName() + "]: this PBinding is not bounded with a pname");
		if (!pb.getPClassMapping().equals(pcm))
			throw new PExceptionNaming("[" + getClassName() + "]: this pbinding is not valid in this binder");
		if (cache != null) {
			synchronized (cache) {
				try {
					if (((FixableCacheEntry) cache).getCeFixCount() == 0) {
						// the cache replacement has decided to remove this entry
						pb.setPName(null);
						pb.setStatus(nextstate);
					} else {
						// this entry becomes eligeable for replacement
						CacheEntry ce = cache.lookup(pn);
						if (ce != null) {
							cache.unfix(ce);
						}
					}
				} catch (CacheException e) {
					throw new PException(e, "[" + getClassName() + "]: problem with cache management");
				}
			}
		} else {
			pb.setPName(null);
			pb.setStatus(nextstate);
		}
	}

	public CacheManager getCacheManager() {
		return cache;
	}

	public void setCacheManager(CacheManager cm) throws PException {
		cache = cm;
	}

	public PName getNull() {
		return nullPName;
	}

	public void setNullPName(Object o) throws PException {
	}
}
