/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 */

package org.objectweb.jorm.naming.lib;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Date;

import org.objectweb.jorm.api.PBinding;
import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.genclass.lib.GenClassMapping;
import org.objectweb.jorm.naming.api.PBinder;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PNameManager;
import org.objectweb.jorm.naming.api.PolymorphicPName;
import org.objectweb.jorm.naming.api.PolymorphicPNamingContext;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.perseus.cache.api.CacheEntry;
import org.objectweb.perseus.cache.api.CacheManager;
import org.objectweb.perseus.persistence.api.ConnectionHolder;
import org.objectweb.perseus.persistence.api.WorkingSet;

/**
 * This class is used to encode/decode asymetric PNames.
 * It is linked to a delegated Binder to which all the encode/decode calls are delegated.
 * It is linked to the cache to perform lookup on resolve(PName) and decodeXX().
 * It is liked to a PClassMapping to delegate the resolve call in case the PName is not in the cache.
 * @author Y.Bersihand
 */
public class BasicPolymorphicPNamingContext implements PolymorphicPNamingContext {

	/**
	 * the delegated binder to which all the encode/decode calls are delegated.
	 */
	private PBinder delegatedBinder;
	
	/**
	 * the PClassMapping to delegate the resolve call.
	 */
	private PClassMapping pcm;
	
	/**
	 * the cache to lookup PNames
	 */
	private CacheManager cache;
	
	/**
	 * the delegated PNameManager: the one originally bound to the PName
	 */
	private PNameManager delegatedPNM;
	
	public BasicPolymorphicPNamingContext() {
		delegatedBinder = null;
		pcm = null;
		cache = null;
		delegatedPNM = null;
	}
	
	public BasicPolymorphicPNamingContext(PBinder delegatedBinder, 
											PClassMapping pcm, 
											CacheManager cache, 
											PNameManager delegatedPNM) {
		super();
		this.delegatedBinder = delegatedBinder;
		this.pcm = pcm;
		this.cache = cache;
		this.delegatedPNM = delegatedPNM;
	}
		
	//////////////////////////////////////
	//	AsymetricPNContext INTERFACE 	//
	//////////////////////////////////////
	
	public CacheManager getCache() {
		return cache;
	}
	public void setCache(CacheManager cache) {
		this.cache = cache;
	}
	public PBinder getDelegatedBinder() {
		return delegatedBinder;
	}
	public void setDelegatedBinder(PBinder delegatedBinder) {
		this.delegatedBinder = delegatedBinder;
	}
	public PClassMapping getPcm() {
		return pcm;
	}
	public void setPcm(PClassMapping pcm) {
		this.pcm = pcm;
	}
	
	/////////////////////////////////
	// 		PNContext INTERFACE	   //
	/////////////////////////////////
	
	public PName resolve(Object conn, PName pname) throws PException {
		//if the pname is linked to a PBinder, returns it
		if(pname.getPNameManager() instanceof PBinder) {
			return pname;
		} else {
			try {
				//checks in the cache
				PName pnFound = checkCache(pname);
				//if the pnFound from the cache is not the same than the parameter
				//returns the pnFound
				if (pnFound != pname)
					return pnFound;
				if (!(conn instanceof ConnectionHolder))
					throw new PException("The conn parameter of PolymorphicPNC.resolve(conn, pname) must be an instance of ConnectionHolder");
				ConnectionHolder connHolder = (ConnectionHolder) conn;
				//get the tx to register the prefetch buffer
				WorkingSet ws = (WorkingSet) connHolder.getWorkingSet();
				//check the prefetch cache
				TupleCollection tupleCollection = pcm.getPMapper().getPrefetchCache().getPrefetchTupleCollection(pname, pcm, ws);
				if (tupleCollection != null && tupleCollection.getMetaData() != null) {
					return pcm.getDecodedPName(tupleCollection, pname, false);	
				}
				//else, delegates to the PClassMapping
				return pcm.resolve(conn, pname);
			} catch (Exception e) {
				throw new PException(e, "Exception while trying to resolve the pname in the BasicAsymetricPNC. " + e.getMessage());
			}
		}
	}
	
	public PName export(Object conn, Object infoitem) throws PException {
		//delegates to the binder
		return delegatedBinder.export(conn, infoitem);
	}

	public PName export(Object conn, Object infoitem, Object hints) throws PException {
		//delegates to the binder
		return delegatedBinder.export(conn, infoitem, hints);
	}

	public void unexport(Object conn, PName pn) throws PException {
		//delegates to the binder
		delegatedBinder.unexport(conn, pn);
	}

	public void unexport(Object conn, PName pn, Object hints) throws PException {
		//delegates to the binder
		delegatedBinder.unexport(conn, pn, hints);	
	}
	
	//////////////////////////////
	// 		PNCoder INTERFACE 	//
	//////////////////////////////
	
	public boolean codingSupported(int codingtype) {
		//delegates to the binder
		return delegatedBinder.codingSupported(codingtype);
	}

	public PName decode(byte[] en) throws PExceptionNaming {
		//delegates to the binder
		PName pname = delegatedBinder.decode(en);
		return checkCache(markAsPolymorphic(pname));
	}

	public PName decodeAbstract(Object en, Object context)
			throws PExceptionNaming, UnsupportedOperationException {
		//delegates to the binder
		PName pname = delegatedBinder.decodeAbstract(en, context);
		return checkCache(markAsPolymorphic(pname));
	}

	public PName decodeByte(byte en) throws PExceptionNaming,
			UnsupportedOperationException {
		///delegates to the binder
		PName pname = delegatedBinder.decodeByte(en);
		return checkCache(markAsPolymorphic(pname));
	}

	public PName decodeObyte(Byte en) throws PExceptionNaming,
			UnsupportedOperationException {
		//delegates to the binder
		PName pname = delegatedBinder.decodeObyte(en);
		return checkCache(markAsPolymorphic(pname));
	}

	public PName decodeChar(char en) throws PExceptionNaming,
			UnsupportedOperationException {
		//delegates to the binder
		PName pname = delegatedBinder.decodeChar(en);
		return checkCache(markAsPolymorphic(pname));
	}

	public PName decodeOchar(Character en) throws PExceptionNaming,
			UnsupportedOperationException {
		//delegates to the binder
		PName pname = delegatedBinder.decodeOchar(en);
		return checkCache(markAsPolymorphic(pname));
	}

	public PName decodeInt(int en) throws PExceptionNaming,
			UnsupportedOperationException {
		//delegates to the binder
		PName pname = delegatedBinder.decodeInt(en);
		return checkCache(markAsPolymorphic(pname));
	}

	public PName decodeOint(Integer en) throws PExceptionNaming,
			UnsupportedOperationException {
		//delegates to the binder
		PName pname = delegatedBinder.decodeOint(en);
		return checkCache(markAsPolymorphic(pname));
	}

	public PName decodeLong(long en) throws PExceptionNaming,
			UnsupportedOperationException {
		//delegates to the binder
		PName pname = delegatedBinder.decodeLong(en);
		return checkCache(markAsPolymorphic(pname));
	}

	public PName decodeOlong(Long en) throws PExceptionNaming,
			UnsupportedOperationException {
		//delegates to the binder
		PName pname = delegatedBinder.decodeOlong(en);
		return checkCache(markAsPolymorphic(pname));
	}

	public PName decodeShort(short en) throws PExceptionNaming,
			UnsupportedOperationException {
		//delegates to the binder
		PName pname = delegatedBinder.decodeShort(en);
		return checkCache(markAsPolymorphic(pname));
	}

	public PName decodeOshort(Short en) throws PExceptionNaming,
			UnsupportedOperationException {
		//delegates to the binder
		PName pname = delegatedBinder.decodeOshort(en);
		return checkCache(markAsPolymorphic(pname));
	}

	public PName decodeString(String en) throws PExceptionNaming {
		//delegates to the binder
		PName pname = delegatedBinder.decodeString(en);
		return checkCache(markAsPolymorphic(pname));
	}

	public PName decodeCharArray(char[] en) throws PExceptionNaming {
		//delegates to the binder
		PName pname = delegatedBinder.decodeCharArray(en);
		return checkCache(markAsPolymorphic(pname));
	}

	public PName decodeDate(Date en) throws PExceptionNaming {
		//delegates to the binder
		PName pname = delegatedBinder.decodeDate(en);
		return checkCache(markAsPolymorphic(pname));
	}

	public PName decodeBigInteger(BigInteger en) throws PExceptionNaming {
		//delegates to the binder
		PName pname = delegatedBinder.decodeBigInteger(en);
		return checkCache(markAsPolymorphic(pname));
	}

	public PName decodeBigDecimal(BigDecimal en) throws PExceptionNaming {
		//delegates to the binder
		PName pname = delegatedBinder.decodeBigDecimal(en);
		return checkCache(markAsPolymorphic(pname));
	}

	public byte[] encode(PName pn) throws PExceptionNaming {
		//delegates to the binder
		return delegatedBinder.encode(pn);
	}

	public Object encodeAbstract(PName pn) throws PExceptionNaming,
			UnsupportedOperationException {
		//delegates to the binder
		return delegatedBinder.encodeAbstract(pn);
	}

	public byte encodeByte(PName pn) throws PExceptionNaming,
			UnsupportedOperationException {
		//delegates to the binder
		return delegatedBinder.encodeByte(pn);
	}

	public Byte encodeObyte(PName pn) throws PExceptionNaming,
			UnsupportedOperationException {
		//delegates to the binder
		return delegatedBinder.encodeObyte(pn);
	}

	public char encodeChar(PName pn) throws PExceptionNaming,
			UnsupportedOperationException {
		//delegates to the binder
		return delegatedBinder.encodeChar(pn);
	}

	public Character encodeOchar(PName pn) throws PExceptionNaming,
			UnsupportedOperationException {
		//delegates to the binder
		return delegatedBinder.encodeOchar(pn);
	}

	public int encodeInt(PName pn) throws PExceptionNaming,
			UnsupportedOperationException {
		//delegates to the binder
		return delegatedBinder.encodeInt(pn);
	}

	public Integer encodeOint(PName pn) throws PExceptionNaming,
			UnsupportedOperationException {
		//delegates to the binder
		return delegatedBinder.encodeOint(pn);
	}

	public long encodeLong(PName pn) throws PExceptionNaming,
			UnsupportedOperationException {
		//delegates to the binder
		return delegatedBinder.encodeLong(pn);
	}

	public Long encodeOlong(PName pn) throws PExceptionNaming,
			UnsupportedOperationException {
		//delegates to the binder
		return delegatedBinder.encodeOlong(pn);
	}

	public short encodeShort(PName pn) throws PExceptionNaming,
			UnsupportedOperationException {
		//delegates to the binder
		return delegatedBinder.encodeShort(pn);
	}

	public Short encodeOshort(PName pn) throws PExceptionNaming,
			UnsupportedOperationException {
		//delegates to the binder
		return delegatedBinder.encodeOshort(pn);
	}

	public String encodeString(PName pn) throws PExceptionNaming {
		//delegates to the binder
		return delegatedBinder.encodeString(pn);
	}

	public char[] encodeCharArray(PName pn) throws PExceptionNaming {
		//delegates to the binder
		return delegatedBinder.encodeCharArray(pn);
	}

	public Date encodeDate(PName pn) throws PExceptionNaming {
		//delegates to the binder
		return delegatedBinder.encodeDate(pn);
	}

	public BigInteger encodeBigInteger(PName pn) throws PExceptionNaming {
		//delegates to the binder
		return delegatedBinder.encodeBigInteger(pn);
	}

	public BigDecimal encodeBigDecimal(PName pn) throws PExceptionNaming {
		//delegates to the binder
		return delegatedBinder.encodeBigDecimal(pn);
	}

	public PName getNull() {
		//delegates to the binder
		return delegatedBinder.getNull();
	}

	public void setNullPName(Object o) throws PException {
		//delegates to the binder
		delegatedBinder.setNullPName(o);
	}

	public boolean supportDynamicComposite() {
		//delegates to the binder
		return delegatedBinder.supportDynamicComposite();
	}

	public boolean supportCompositeField(String fn, PType ft) {
		//delegates to the binder
		return delegatedBinder.supportCompositeField(fn,ft);
	}

	public boolean supportStaticComposite() {
		//delegates to the binder
		return delegatedBinder.supportStaticComposite();
	}

	public PType getPType() {
		//delegates to the binder
		PType type = delegatedBinder.getPType();
		if (type == null) {
			type = pcm.getPType();
			delegatedBinder.setPType(type);
		}
		return type;
	}

	public void setPType(PType pt) {
		//delegates to the binder
		delegatedBinder.setPType(pt);
	}

	
	//////////////////////////////////////
	//			PRIVATE METHODS			//
	//////////////////////////////////////
	
	/**
	 * If a "valid" pname is already in the cache, returns this valid pname,
	 * else return the pname parameter.
	 * The "valid" pname is :
	 * 	- present in the cache,
	 *  - not an instance of GenClassMapping
	 *  - linked to a PClassMapping which is part of the inheritance graph
	 */
	private PName checkCache(PName pname) throws PExceptionNaming {
		PolymorphicPName polymorphicPN = (PolymorphicPName) pname;
		CacheEntry ce = cache.lookup(pname);
		if (ce != null) {
			//get the "valid" pname
			PName pnFound = (PName) ce.getCeIdentifier();
			PBinding pbinding = (PBinding) ce.getCeObject();
			//if the cache entry is linked to a gen pcm, it is not the good one
			PClassMapping pcmFound = pbinding.getPClassMapping(); 
			if ( pcmFound instanceof GenClassMapping) {
				return pname;
			} else {
				try {
					PClassMapping[] subPCMs = pcm.getSubPCMs();
					//checks that the pcm of the cache entry is part of the inheritance graph
					//TODO: check for the parents???
					if (subPCMs != null) {
						for (int i= 0; i<subPCMs.length; i++) {
							if (pcmFound == subPCMs[i]) {
								//this pname found in the cache is the good one
								return pnFound;
							}
						}
					}
					//the pname found is linked to a pcm which is not part
					//of the inheritance graph, so returns the pname
					return pname;
				} catch (Exception e) {
					throw new PExceptionNaming(e,
							"Exception on decode, trying to get all the sub PClassMappings of "
							+ pcm.getClassName() + "." + "Exception nested: " + e.getMessage());
				}
			}
		} else {
			//if there is no entry in the cache, returns the pname
			return pname;
		}
	}
	
	private PolymorphicPName markAsPolymorphic(PName pn) {
		PolymorphicPName polymorphicPN = (PolymorphicPName) pn;
		polymorphicPN.setPolymorphic(true);
		polymorphicPN.setPNameManager((PNameManager)this);
	    return polymorphicPN;
	}
}
