/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2005 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Authors: S.Chassande-Barrioz.
 * Created on 24 juin 2005
 *
 */
package org.objectweb.jorm.naming.lib;

import org.objectweb.jorm.api.PAccessor;
import org.objectweb.jorm.api.PBinding;
import org.objectweb.jorm.api.PBindingCtrl;
import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.api.PStateGraph;
import org.objectweb.jorm.naming.api.PExceptionExistingName;
import org.objectweb.jorm.naming.api.PExceptionNameCoding;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.naming.api.PNCStringCoder;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PolymorphicPNamingContext;
import org.objectweb.perseus.cache.api.CacheEntry;
import org.objectweb.perseus.cache.api.CacheException;

/**
 * This class helps the implementation of PBinder for composite name.
 *
 * @author S.Chassande-Barrioz
 */
public abstract class CompositePBinder extends BasicPBinder {

    public CompositePBinder() {
        super();
    }

    public CompositePBinder(PClassMapping _pcm) {
        super(_pcm);
    }

    public CompositePBinder(PName nullPName) {
        super(nullPName);
    }
    
    /**
     * Indicates if the PName can be managed by this Binder
     */
    protected abstract boolean isManagedPName(Object pn);
    
    /**
     * Indicates if value encapsulated into a pname getter (generic or specific)
     * represents a null PName
     * @param png is the pname getter containing values
     * @param ctx is a hints to pass to the pnGet methods
     */
    protected abstract boolean isNull(Object png, Object ctx) throws PException;
    
    protected abstract void fillStringCoder(PNCStringCoder cnsc, PName pn) throws PExceptionNameCoding; 
    
    protected boolean isAllAutoCalculated() {
        return false;
    }

    public void setNullPName(Object o) throws PException {
	    nullPName = (o instanceof PName ? (PName) o : getPNameFrom(o, null, null, false));
	}
	
	/**
	 * Try to build a PName from a java object. The java object can be a 
	 * generic or specific PNameGetter implementation
	 * @param png is a pname getter (generic or specific)
	 * @param ctx is a context to pass on the pnGet method
	 * @param connection is a connection to the un derlying data store in case
	 * of the value has to be fetch from the data store (ex: sequence)
	 * @return a PName corresponding to the PNameGetter, maybe fill with value
	 * from the data store.
	 */
	protected abstract PName getPNameFrom(Object png, 
	        Object ctx, Object connection, boolean computeACF) throws PException;
	
	/**
	 * Allocate a new PName instance with values containing in existing one.
	 * @param oldpn is the existing PName containg value to pick
	 * @return a new PName instance associated to this PBinder
	 */
	protected abstract PName newPName(PName oldpn) throws PException;

    public Object encodeAbstract(PName pn) 
    	throws PExceptionNaming, UnsupportedOperationException {
        if (!isManagedPName(pn)) {
            throw new PExceptionNaming(
                    "This binder cannot decode this pname: " + pn);
        }
        if (!pn.getPNameManager().equals(this) 
                && !(pn.getPNameManager() instanceof PolymorphicPNamingContext)) {
            throw new PExceptionNaming("This pname is not valid in this pnc");
        }
        return pn;
	}

    public PName decodeAbstract(Object png, Object ctx)
            throws PExceptionNaming, UnsupportedOperationException {
        try {
            if (isNull(png, ctx)) {
                return getNull();
            } else {
                return getPNameFrom(png, ctx, null, false);
            }
        } catch (PExceptionNaming e) {
            throw e;
        } catch (PException e) {
            throw new PExceptionNaming(e);
        }
    }

    public boolean supportDynamicComposite() {
		return true;
	}

	public boolean supportStaticComposite() {
		return true;
	}

	/**
	 * It retrieves the information item associated to the given PName (if any
	 * exists). The pname must be a StrName and must support the hashcode method.
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	The PName to be looked up.
	 * @return	The PBinding associated with the given PName within
	 * 		this binder
	 * @exception PExceptionNaming	Raised if the Pname is not registered into this binder.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */

	public PName resolve(Object conn, PName pn) throws PException {
        if (pn == null)
            throw new PExceptionNaming("This pname is null");
        if (!pn.getPNameManager().equals(this))
            throw new PExceptionNaming("This pname is not valid in this binder");
        return pn;
	}

    public PName export(Object conn, Object en) throws PException {
		if (isManagedPName(en)) {
		    if (((PName) en).getPNameManager()==this) {
			    return (PName) en;
			} else {
			    return newPName((PName) en);
			}
	    }
		if (isAllAutoCalculated()) {
		    return export(conn, en, null);
		} else {
		    throw new PExceptionNaming("Not supported. The export method must be called with a ctx parameter which is the accessor");
		}
	}

    public PName export(Object conn, Object en, Object hints)
		throws PException {
        PName pn = null;
        if (en == null) {
			throw new PExceptionNaming("Cannot export null!");
		}
		if (isManagedPName(en)) {
		    PName oldpn = (PName) en;
		    return (oldpn.getPNameManager()== this 
		            ? oldpn 
		            : newPName(oldpn));
	    }
        if (!(en instanceof PBindingCtrl))
            throw new PExceptionNaming("The item cannot be exported by this binder: " + en);
        
		PBindingCtrl pb = (PBindingCtrl) en;
		byte nextstate = PStateGraph.nextStatePBinding(pb.getStatus(),
			PBinding.ACTION_EXPORT);
		if (nextstate == PBinding.LIFECYCLE_ERROR)
			throw new PExceptionProtocol("Unauthorized operation");

		if (isAllAutoCalculated()) {
		    pn = getPNameFrom(hints, null, conn, true);
		} else {
		    if (hints instanceof PName) {
			    pn = newPName((PName) hints);
		    } else if (hints instanceof PAccessor) {
		        pn = getPNameFrom(pb.getPClassMapping(), hints, conn, true);
		    }
	        if (pn == null) {
	        	pn = getPNameFrom(hints, null, conn, true);
	        	if (pn == null) {
		            throw new PExceptionNaming("This item cannot be exported by this binder");
		        }
	        }
        }
        if (cache != null) {
            synchronized (cache) {
                if (cache.lookup(pn) != null) {
                    throw new PExceptionExistingName(
                            "[" + pcm.getClassName() + "]: an object has been already export with the same identifier");
                }
                try {
                    cache.fix(cache.bind(pn, en));
                } catch (CacheException e) {
                    throw new PException(e, "[" + pcm.getClassName() + "]: problem with cache management");
                }
            }
        }
		pb.setPName(pn);
		pb.setStatus(nextstate);
        return pn;
	}
	
    /**
	 * It removes the PName and all the associated information within a naming
	 * context (i.e., the association between the PName and its information
	 * item). in the case of the StrBinder, the unexport operation do nothing more
	 * than the unbind operation.
	 * @param conn	The connection that can be used to access the data store.
	 * @param pn	A PName valid in this naming context.
	 * @exception PExceptionNaming	Raised in case of problem within the
	 * 								naming context.
	 * @see org.objectweb.jorm.naming.api.PNamingContext
	 */
    public void unexport(Object conn, PName pn) throws PException {
        if (pn.isNull()) {
            throw new PExceptionProtocol("[" + pcm.getClassName() + "]: cannot unexport with a null pname.");
        }
        if (cache != null) {
            CacheEntry ce = cache.lookup(pn);
            if (ce == null) {
                throw new PExceptionProtocol("Unauthorized operation: No Pbinding found");
            }
            PBindingCtrl pb = (PBindingCtrl) ce.getCeObject();
            byte nextstate = PStateGraph.nextStatePBinding(
                pb.getStatus(), PBinding.ACTION_UNEXPORT);
            if (nextstate == PBinding.LIFECYCLE_ERROR) {
                throw new PExceptionProtocol("Unauthorized operation: status =" + pb.getStatus());
            }
            pb.setStatus(nextstate);
        } else {
            throw new PExceptionProtocol("Unauthorized operation: No Pbinding can be found without a CacheManager");
        }
    }

    public void unexport(Object conn, PName pn, Object hints) throws PException {
        if (pn.isNull()) {
            throw new PExceptionProtocol("[" + pcm.getClassName() + "]: cannot unexport with a null pname.");
        }
        PBindingCtrl pb = null;
        if (hints instanceof PBindingCtrl) {
            pb = (PBindingCtrl) hints;
        } else if (cache != null) {
            CacheEntry ce = cache.lookup(pn);
            if (ce != null) {
                pb = (PBindingCtrl) ce.getCeObject();
            }
        }
        if (pb == null) {
            throw new PExceptionProtocol("Unauthorized operation: No Pbinding found");
        }
        byte nextstate = PStateGraph.nextStatePBinding(
            pb.getStatus(), PBinding.ACTION_UNEXPORT);
        if (nextstate == PBinding.LIFECYCLE_ERROR) {
            throw new PExceptionProtocol("Unauthorized operation: status =" + pb.getStatus());
        }
        pb.setStatus(nextstate);
    }

	public String encodeString(PName pn) throws PExceptionNaming {
        if (!isManagedPName(pn)) {
            throw new PExceptionNaming("This binder decode only ${compositename.getName()}PName");
		}
        if (!pn.getPNameManager().equals(this)) {
            throw new PExceptionNaming("This pname is not valid in this pnc");
		}
		PNCStringCoder cnsc = new BasicStringCoder();
		fillStringCoder(cnsc, pn);
		return cnsc.getStringCode();
	}
	

    
}
