/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */
package org.objectweb.jorm.util.lib;

import org.apache.tools.ant.taskdefs.MatchingTask;
import org.apache.tools.ant.types.DTDLocation;
import org.apache.tools.ant.types.Path;
import org.apache.tools.ant.types.Reference;
import org.apache.tools.ant.BuildException;
import org.objectweb.jorm.util.io.api.PathExplorer;
import org.objectweb.jorm.util.io.lib.DirJavaExplorer;
import org.objectweb.jorm.compiler.api.JormCompilerParameter;
import org.objectweb.jorm.compiler.lib.JormCompiler;
import org.objectweb.jorm.compiler.lib.JormCompilerParameterImpl;
import org.objectweb.jorm.compiler.lib.JormCompilerConfiguratorImpl;
import org.objectweb.jorm.compiler.api.JormCompilerConfigurator;
import org.objectweb.jorm.api.PException;

import java.io.File;
import java.util.*;

public class AntJormParser extends MatchingTask {
	private String jormConfFile = "jorm.properties";
	private File pdDir = null;

    /**
     * System separator character (e.g '/' under Unix systems)
     */
    private final static char separator = File.separatorChar;

	private boolean antDtdLocationsSet = false;
	private ArrayList antDtdLocations = new ArrayList();
	private boolean classpathSet = false;
	private PathExplorer classpath = new DirJavaExplorer();
	private boolean antLogConfFileSet = false;
	private String antLogConfFile;
	private String antProjectName;
	private boolean antVerboseSet = false;
	private boolean antVerbose;

	private JormCompilerParameter jcp = null;
	private JormCompilerConfigurator jcc = null;
	private JormCompiler jormcompiler = null;

	public AntJormParser() {

        jcp = new JormCompilerParameterImpl();
		jcc = new JormCompilerConfiguratorImpl();
		jormcompiler = new JormCompiler(jcp, jcc);
        //TODO
        //jormcompiler = new JormCompiler();
        //JormCompilerParameter jcp = jormcompiler.getCompilerParameter();
        //JormConfiguratorImpl jcc = jormcompiler.getCompilerConfigurator();
	}

	/**
	 * It assigns the directory name where the include and exclude clause has
	 * been based. (REQUIRED)
	 */
	public void setPdDir(File d) {
		pdDir = d;
		classpath.addPath(d.getAbsolutePath());
                        System.out.println("pdDir " + pdDir.getAbsolutePath());
	}

	/**
	 * It assigns the classpath where jorm can load .pd files which are
	 * needed (the dependecies).
	 * (OPTIONAL: default value is the PdDir)
	 */
	public void setOtherPdFiles(Path cp) {
		String[] dl = cp.list();
		for (int i = 0; i < dl.length; i++) {
			classpath.addPath(dl[i]);
			classpathSet = true;
		}
	}

	public void setOtherPdFilesRef(Reference r) {
		Path p = new Path(getProject());
		p.setRefid(r);
		setOtherPdFiles(p);
	}

	/**
	 * Create a DTD location record. This stores the location of a DTD. The
	 * DTD is identified by its public Id. The location may either be a file
	 * location or a resource location.
	 * @return the DTD location object to be configured by Ant
	 */
	public DTDLocation createDTD() {
		DTDLocation dtdLocation = new DTDLocation();
		antDtdLocations.add(dtdLocation);
		antDtdLocationsSet = true;
		return dtdLocation;
	}

	/**
	 * It assigns the jorm configuration file. (REQUIRED)
	 * @param s the string representation of the configuration file.
	 */
	public void setJormConf(String s) {
		jormConfFile = s;
	}

	/**
	 * It assigns the jorm configuration file. (REQUIRED)
	 * @param s the string representation of the configuration file.
	 */
	public void setLogConf(String s) {
		antLogConfFile = s;
		antLogConfFileSet = true;
	}

	/**
	 * It assigns the project name which must be used.
	 * This parameter can be override for a class if you use a local jorm
	 * configuration file which defines also the property.
	 * (OPTIONAL default value = "" empty string)
	 */
	public void setProjectName(String s) {
		antProjectName = s;
	}

	/**
	 * It assigns the verbose mode to the jorm generator.
	 * This parameter can be override for a class if you use a local jorm
	 * configuration file which defines also the property.
	 * (OPTIONAL the default value is false)
	 */
	public void setVerbose(boolean verbose) {
		antVerbose = verbose;
		antVerboseSet = true;
	}

    public void initJormCompiler() throws PException {
        try {
            if (jormConfFile == null) {
                jcc.configure();
            } // set jcc.mappers from data stored in the jorm.properties file
            else {
              jcc.configure(jormConfFile);
            } // set jcc.mappers from data stored in jormConfFile
            log("jcc.configure() ----------- ok");
            Collection c = jcc.getSubMappers("rdb");
		    log("rdb sub mappers: " + c);
			if (!antLogConfFileSet) {
				jcc.configureLog(jormConfFile);
			} else {
				jcc.configureLog(antLogConfFile);
			}
            log("jcc.configureLog() ----------- ok");
            // TODO : jcc.knownMappers() seems to be no more used in the loadConfFile() method.
		    log("jormcOpts file= " + jcc.getJormcOptsFile());
            jcp.loadConfFile(jcc.getGlobalJormcOptsFile(), jcc.knownMappers());
            log("jcp.loadConfFile ----------- ok");
            log("pdDir= " + pdDir);
            // The following properties can be set in the jormcOpts.properties file or overloaded in the code
			if (antDtdLocationsSet) {
				jcp.setDtdLocations(antDtdLocations);
            }
            if (antProjectName != null) {
                jcp.setProjectName(antProjectName);
            }
            log("jcp.projectName= " + jcp.getProjectName());
            Collection temp = new ArrayList(Arrays.asList(super.getDirectoryScanner(pdDir).getIncludedFiles()));
            String inputFile = null;
            for (Iterator it = temp.iterator(); it.hasNext();) {
                inputFile = pdDir.getAbsolutePath() + separator + (String) it.next();
                jcp.getInputFiles().add(inputFile);
                log("input file= " + inputFile);
            }
            log("jcp.inputFiles ----------- ok");
			if (classpathSet) {
				jcp.setClasspath(classpath);
			} else {
                DirJavaExplorer pe = new DirJavaExplorer();
                pe.setLogger(jcc.getLoggerFactory().getLogger("org.objectweb.jorm.io.pathexplorer"));
                jcp.setClasspath(pe);
            }
            log("jcp.classpath ----------- ok");
            jcp.setGeneratedWithMapperPackage(false);
            // TODO: jcp.logConFile ?
			if (antLogConfFileSet) {
				jcp.setLogConfFile(antLogConfFile);
			}
			if (antVerboseSet) {
				jcp.setVerbose(antVerbose);
			}
        } catch (PException e) {;
            throw new PException(e, "Impossible to configure the Jorm parser");
        }
    }


	/**
	 * execute method of the Ant task
	 */
	public void execute() throws BuildException {
        try {
            initJormCompiler();
            jormcompiler.parseFiles(jcp.getInputFiles());
            log("parseFiles...... done");
		 } catch (PException e) {
			Exception current = e;
			while (current instanceof PException
					&& ((PException) current).getNestedException() != null) {
				current = ((PException) current).getNestedException();
			}
			throw new BuildException(current);
		}
	}
}
