/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2005 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm@objectweb.org
 *
 */

package org.objectweb.jorm.util.lib;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.PushbackReader;
import java.math.BigDecimal;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.taskdefs.MatchingTask;

public class AntJormTestReport extends MatchingTask {
	private File srcdir;

	private File destdir;

	/**
	 * It assigns the directory name where the include and exclude clause has
	 * been based. (REQUIRED)
	 */
	public void setsrcdir(File d) {
		srcdir = d;
	}

	/**
	 * It assigns the directory name where the report files are going to be
	 * produced. (REQUIRED)
	 */
	public void setdestdir(File d) {
		destdir = d;
	}

	/**
	 * execute method of the Ant task
	 */
	public void execute() throws BuildException {
		try {
			String[] in_files = super.getDirectoryScanner(srcdir)
					.getIncludedFiles();
			File destfile = new File(destdir.getPath() + File.separator
					+ "index.html");
			PrintWriter pw = new PrintWriter(new FileOutputStream(destfile));
			TestInfo[] infos = new TestInfo[in_files.length + 1];
			infos[in_files.length] = new TestInfo();
			infos[in_files.length].time = new BigDecimal(0);
			infos[in_files.length].time.setScale(3, BigDecimal.ROUND_FLOOR);
			for (int i = 0; i < in_files.length; i++) {
				infos[i] = extractInfos(in_files[i]);
				infos[in_files.length].nbTests += infos[i].nbTests;
				infos[in_files.length].nbErrors += infos[i].nbErrors;
				infos[in_files.length].nbFailures += infos[i].nbFailures;
				infos[in_files.length].time = infos[in_files.length].time.add(infos[i].time);
			}
			BigDecimal cent = new BigDecimal(100);
			BigDecimal rate = new BigDecimal(infos[in_files.length].nbErrors + infos[in_files.length].nbFailures);
			rate.setScale(2);
			rate = rate.divide(new BigDecimal(infos[in_files.length].nbTests), 2,
					BigDecimal.ROUND_FLOOR);
			rate = cent.multiply(rate);
			rate = cent.subtract(rate);
			printHeader(pw);
			pw.println("<tr valign=\"top\" class=\"Pass\">");
			pw.println("<td>" + infos[in_files.length].nbTests + "</td><td>"
					+ infos[in_files.length].nbFailures + "</td><td>"
					+ infos[in_files.length].nbErrors + "</td><td>" + rate
					+ "%</td><td>" + infos[in_files.length].time.toString() + "</td>");
			pw.println("</tr>");
			printMiddle(pw);
			for (int i = 0; i < in_files.length; i++) {
				pw.println("<tr valign=\"top\" class=\"Pass\">");
				pw.println("<td><a href=\"" + srcdir + File.separator + infos[i].indexhtml + "\">" +
						infos[i].testname +
						"</a></td><td>" + infos[i].nbTests + "</td><td>" +
						infos[i].nbErrors + "</td><td>" +
						infos[i].nbFailures + "</td><td>" +
						infos[i].time.toString() + "</td>");
				pw.println("</tr>");
			}
			printTrailer(pw);
			pw.flush();
			pw.close();
		} catch (FileNotFoundException e) {
			log(e.getMessage(), Project.MSG_ERR);
		}
	}

	private void printHeader(PrintWriter pw) {
		pw.println("<html>");
		pw.println("<head>");
		pw.println("<META http-equiv=\"Content-Type\" content=\"text/html; charset=US-ASCII\">");
		pw.println("<title>Unit Test Results: Summary</title>");
		pw.println("<link title=\"Style\" type=\"text/css\" rel=\"stylesheet\" href=\"stylesheet.css\">");
		pw.println("</head>");
		//pw.println("<body onload=\"open('allclasses-frame.html','classListFrame')\">");
		pw.println("<body>");
		pw.println("<h1>JORM Unit Test Results</h1>");
		pw.println("<table width=\"100%\">");
		pw.println("<tr>");
		pw.println("<td align=\"left\"></td><td align=\"right\">Designed for use with <a href=\"http://www.junit.org/\">JUnit</a> and <a href=\"http://jakarta.apache.org/\">Ant</a>.</td>");
		pw.println("</tr>");
		pw.println("</table>");
		pw.println("<hr size=\"1\">");
		pw.println("<h2>Summary</h2>");
		pw.println("<table width=\"95%\" cellspacing=\"2\" cellpadding=\"5\" border=\"0\" class=\"details\">");
		pw.println("<tr valign=\"top\">");
		pw.println("<th>Tests</th><th>Failures</th><th>Errors</th><th>Success rate</th><th>Time</th>");
		pw.println("</tr>");
	}

	private void printMiddle(PrintWriter pw) {
		pw.println("</table>");
		pw.println("<table width=\"95%\" border=\"0\">");
		pw.println("<tr>");
		pw.println("<td style=\"text-align: justify;\">");
		pw
				.println("Note: <em>failures</em> are anticipated and checked for with assertions while <em>errors</em> are unanticipated.");
		pw.println("</td>");
		pw.println("</tr>");
		pw.println("</table>");
		pw.println("<h2>Tested area</h2>");
		pw
				.println("<table width=\"95%\" cellspacing=\"2\" cellpadding=\"5\" border=\"0\" class=\"details\">");
		pw.println("<tr valign=\"top\">");
		pw
				.println("<th width=\"80%\">Name</th><th>Tests</th><th>Errors</th><th>Failures</th><th nowrap=\"nowrap\">Time(s)</th>");
		pw.println("</tr>");
	}

	private void printTrailer(PrintWriter pw) {
		pw.println("</table>");
		pw.println("</body>");
		pw.println("</html>");
	}
	
	private TestInfo extractInfos(String fn) {
		TestInfo res = new TestInfo();
		int s1 = fn.lastIndexOf(File.separatorChar);
		int s2 = fn.lastIndexOf(File.separatorChar, s1 - 1);
		int s3 = fn.lastIndexOf(File.separatorChar, s2 - 1);
		if (s3 == -1) {
			s3 = 0;
		} else {
			s3++;
		}
		if (s1 == -1) {
		    //no mapper at all
		    res.testname = "Common test";
		} else if (s2 == -1) {
		    //mapper but no submapper
		    res.testname = fn.substring(0, s1);
		} else {
		    //mapper + submapper
		    res.testname = fn.substring(s3, s2) + "." + fn.substring(s2 + 1, s1);
		}
		res.indexhtml = fn.substring(0, s1 + 1) + "index.html";
		try {
			PushbackReader fr = new PushbackReader(new FileReader(srcdir + File.separator + fn), 10);
			locateNextString(fr, "<table");
			locateNextString(fr, "<table");
			locateNextString(fr, "<tr");
			locateNextString(fr, "<tr");
			locateNextString(fr, "<td>"); // now positioned on the number of tests
			res.nbTests = Integer.parseInt(getString2ClosingTag(fr));
			locateNextString(fr, "<td>"); // now positioned on the number of errors
			res.nbFailures = Integer.parseInt(getString2ClosingTag(fr));
			locateNextString(fr, "<td>"); // now positioned on the number of failures
			res.nbErrors = Integer.parseInt(getString2ClosingTag(fr));
			locateNextString(fr, "<td>"); // now positioned on the success rate
			locateNextString(fr, "<td>"); // now positioned on the execution time
			res.time = new BigDecimal(getString2ClosingTag(fr));
			res.time.setScale(3);
		} catch (FileNotFoundException e) {
			log(e.getMessage(), Project.MSG_ERR);
		} catch (IOException e) {
			log(e.getMessage(), Project.MSG_ERR);
		}
		return res;
	}
	
	private void locateNextString(PushbackReader fr, String s) throws IOException {
		int next = fr.read();
		while (next != -1) {
			if (next == s.charAt(0)) {
				char buf[] = new char[s.length() - 1];
				int next2 = fr.read(), pos = 1;
				while ((next2 != -1) && (pos < s.length())) {
					buf[pos - 1] = (char) next2;
					if (next2 == s.charAt(pos)) {
						pos++;
						next2 = fr.read();
					} else {
						fr.unread(buf, 0, pos);
						break;
					}
				}
				if (pos == s.length()) {
					buf[0] = (char) next2;
					fr.unread(buf, 0, 1);
					return;
				}
			}
			next = fr.read();
		}
		throw new IOException("Wrong format in test result for JORM - string not found: " + s);
	}
	
	private String getString2ClosingTag(PushbackReader fr) throws IOException {
		int next = fr.read();
		StringBuffer sb = new StringBuffer();
		while (next != -1) {
			if (next == '<') {
				break;
			}
			sb.append((char) next);
			next = fr.read();
		}
		return sb.toString();
	}
}

class TestInfo {
	public String indexhtml;
	public String testname;
	public int nbTests = 0;
	public int nbErrors = 0;
	public int nbFailures = 0;
	public BigDecimal time;
}
