/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.xml2mi.lib;

import org.objectweb.jorm.metainfo.api.*;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.xml2mi.api.MappingParser;
import org.objectweb.jorm.util.io.api.PathExplorer;
import org.objectweb.jorm.util.io.lib.DirJavaExplorer;
import org.objectweb.jorm.api.PException;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.util.monolog.api.BasicLevel;
import org.w3c.dom.Element;

import java.util.Map;
import java.util.Hashtable;
import java.util.Iterator;

/**
 * ProxyParser is an implementation of the Parser interface. This object defines
 * a parser which is able to dispatch operation to specific mapping parser,
 * and able to build the corresponding meta information.
 * @author	X. Spengler
 */
public abstract class BasicMappingParser 
	extends ParserHelper
	implements MappingParser {

	/**
	 * The meta information manager for managing meta-objects within the
	 * context of a mapper.
	 */
	protected Manager metaInfoManager;

	/**
	 * The current Class object.
	 */
	protected Class currentClass;

	/**
	 * pathExplorer is able to return input or output stream from a given file.
	 */
	protected PathExplorer pathExplorer;

	/**
	 * A Map of String objects.
	 * key = a String object that represents an  id value.
	 * value = the associated GenClassRef object.
	 */
	protected Map idvalue2genclassref;

	static String fileSeparator = System.getProperty("file.separator");

	///////////////////////////////////////////////////////////////////
	//                      from ProxyParserInt interface
	///////////////////////////////////////////////////////////////////

	/**
	 * Assigns a meta-information manager to a parser in order from it to
	 * create the meta-information related to the XML files it parses.
	 *
	 * @param mim	the meta-information Manager to be associated to this
	 * 				parser
	 */
	public void setMetaInfoManager(Manager mim) {
		metaInfoManager = mim;
	}

	/**
	 * Assigns a PathExplorer object for locating files that have to be parsed.
	 *
	 * @param pathExpl	The PathExplorer to be used for file location
	 */
	public void setPathExplorer(PathExplorer pathExpl) {
		pathExplorer = pathExpl;
		if (pathExplorer == null)
			pathExplorer = new DirJavaExplorer();
	}

	/**
	 * Assigns the current Class object to a parser in order from it to
	 * retrieve the meta-objects (for example, PrimitiveElement objects)
	 * that are referenced from XML mapping elements.
	 *
	 * @param currentClass	the Jorm meta-object associated to the parsed class.
	 */
	public void setCurrentClass(Class currentClass) {
		this.currentClass = currentClass;
	}


	/**
	 * Returns a NameDef object given a NameDef name (name), a field or a scalarfield name (fn)
	 * and possibly the full qualified class name where it is defined. This NameDef object is
	 * referenced from a ReferenceMapping object.
	 * First seeks for a TypedElement object given a field or a scalarfield name and possibly the full
	 * qualified class name (fqcn) where it is defined.
	 * @param linkend	    a String object whose structure is defined as follows: /fqcn/fn/name or fn/name,
	 *        tagName       a String object whose value matches "rdb-class-ref-mapping" or
	 *                      "rdb-gen-class-ref-mapping".
	 * @return a NameDef object.
	 */
	protected NameDef getRefNameDef(String linkend, String tagName) {
		TypedElement typedElement = null;
		NameDef nameDef = null;
		String fieldName = null;
		String namedefName = null;
		int idx1 = linkend.indexOf("/");
		if (idx1 == 0) {
			// link-end value: /fqcn/fn/name
			int idx2 = linkend.indexOf("/", idx1 + 1);
			String fqcn = linkend.substring(1, idx2);
			int idx3 = linkend.indexOf("/", idx2 + 1);
			fieldName = linkend.substring(idx2 + 1, idx3);
			namedefName = linkend.substring(idx3 + 1);
			if (logger.isLoggable(BasicLevel.DEBUG)) {
				logger.log(BasicLevel.DEBUG,
                            " fqcn " + fqcn +
						   " fieldname " + fieldName +
						   " namdefname " + namedefName);
            }
			fqcn = fqcn.replace('.', fileSeparator.charAt(0)) + ".pd";
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                if (motable.contains(fqcn)) {
                    logger.log(BasicLevel.DEBUG, "fqcn is there " + fqcn);
                }
                else {
                    logger.log(BasicLevel.DEBUG,"fqcn is not there " + fqcn);
                }
            }
			Class clazz = (Class) motable.get(fqcn);
			typedElement = clazz.getTypedElement(fieldName);

			if (logger.isLoggable(BasicLevel.DEBUG)) {
				logger.log(BasicLevel.DEBUG,
						   "Get the field " + fieldName +
						   "defined in <" + clazz.getFQName() +
						   ">: <" + typedElement + ">");
			}
		} else {
			// link-end value: fn/name
			fieldName = linkend.substring(0, idx1);
			namedefName = linkend.substring(idx1 + 1);
			typedElement = currentClass.getTypedElement(fieldName);
			if (logger.isLoggable(BasicLevel.DEBUG)) {
				logger.log(BasicLevel.DEBUG,
						   "Get the field [" + fieldName +
						   "] in current class <" + currentClass.getFQName() +
						   ">: <" + typedElement + ">");
			}
		}
		if (typedElement instanceof ClassRef) {
			ClassRef classRef = (ClassRef) typedElement;
			if (logger.isLoggable(BasicLevel.DEBUG)) {
				logger.log(BasicLevel.DEBUG,
						   "Get the ref name def of the ClassRef " + fieldName);
			}
			nameDef = classRef.getRefNameDef(namedefName);
		} else if (typedElement instanceof GenClassRef) {
			GenClassRef genClassRef = (GenClassRef) typedElement;
			if (logger.isLoggable(BasicLevel.DEBUG)) {
				logger.log(BasicLevel.DEBUG,
						   "Get the ref name def of the GenClassRef " + fieldName);
			}
			nameDef = genClassRef.getRefNameDef(namedefName);
		}
		if (logger.isLoggable(BasicLevel.DEBUG)) {
			logger.log(BasicLevel.DEBUG, "Returned name def: " + nameDef);
		}
		return nameDef;
	}

    /**
     * Returns a NameDef object given the full qualified class name of an inherited class
     * and possibly a NameDef name.
     * @param linkend	    a String object whose structure is defined as follows:
     *                      /fqcn or /fqcn/name,
     * @return a NameDef object.
     */
    protected NameDef getNameDefFromSuper(String linkend) {
        int idx1 = linkend.indexOf("/");
        if (idx1 == 0) {
            // link-end value: /fqcn...
            int idx2 = linkend.indexOf("/", idx1 + 1);
			String namedefName = null;
			String fqcn = null;
            if (idx2 != -1) {
                // link-end value: /fqcn/name
                fqcn = linkend.substring(1, idx2);
                namedefName = linkend.substring(idx2 + 1);
            } else {
                // link-end value: /fqcn
                fqcn = linkend.substring(1);
                namedefName = "";
            }
            Class superClass = currentClass.getSuperClass(fqcn);
			Iterator ndIt = superClass.getNameDefs().iterator();
			while(ndIt.hasNext()) {
				NameDef nd = (NameDef) ndIt.next();
				if (namedefName.equals(nd.getName())) {
					if (logger.isLoggable(BasicLevel.DEBUG)) {
						logger.log(BasicLevel.DEBUG, "Returned name def: " + nd);
					}
					return nd;
				}

			}
        }
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "No nameDef found for the linkend " + linkend);
        }
        return null;
    }

    protected PrimitiveElement getPrimitiveElement(String linkend) {
        PrimitiveElement pe = null;
        int idx1 = linkend.indexOf("/");
        if (idx1 == 0) {
            // link-end value: /fqcn/fn
            int idx2 = linkend.indexOf("/", idx1 + 1);
            if (idx2 != -1) {
                // link-end value: /fqcn/name
                String fqcn = linkend.substring(1, idx2);
                String fieldname = linkend.substring(idx2 + 1);
                Class superClass = currentClass.getSuperClass(fqcn);
                pe = (PrimitiveElement) superClass.getTypedElement(fieldname);
            }
        } else {
            // link-end value: fn
            pe = (PrimitiveElement) currentClass.getTypedElement(linkend);
        }
        return pe;
    }

	/**
	 * Seeks for a father node of type field from a name-def node and returns
	 * the corresponding field name.
	 * @param nameDef	the name-def element.
	 * @return a field name.
	 */
	protected String getFieldName(Element nameDef) {
		logger.log(BasicLevel.DEBUG, "Fetching field name of the name def element " + nameDef.getNodeName());
		Element currentElem = nameDef;
		while (currentElem != null &&
				!currentElem.getNodeName().equals("field")) {
			logger.log(BasicLevel.DEBUG, "current: " + currentElem.getNodeName());
			currentElem = (Element) currentElem.getParentNode();
		}
		if (currentElem != null) {
			if (logger.isLoggable(BasicLevel.DEBUG)) {
				logger.log(BasicLevel.DEBUG, "last: " + currentElem.getNodeName()
											 + " => name=" + currentElem.getAttribute("name"));
			}
			return currentElem.getAttribute("name");
		} else {
			return null;
		}
	}


	/**
	 * Seeks for a ClassRef object given a class, a field name.
	 * This method returns the NameDef object of the ClassRef object.
	 *
	 * @return the NameDef object of the ClassRef object.
	 */
	protected NameDef getClassRefNameDef(Class clazz,
										 String fieldName, String name) {
		NameDef res = null;
		TypedElement typedElement = clazz.getTypedElement(fieldName);
		logger.log(BasicLevel.DEBUG, "Fetching field <" + clazz.getFQName() + "." + fieldName + ">: <" + typedElement + ">");
		if (typedElement instanceof ClassRef) {
			ClassRef classRef = (ClassRef) typedElement;
			logger.log(BasicLevel.DEBUG, "Fetching name def of the ClassRef " + fieldName);
			res = classRef.getRefNameDef(name);
		} else if (typedElement instanceof GenClassRef) {
			GenClassRef genClassRef = (GenClassRef) typedElement;
			logger.log(BasicLevel.DEBUG, "Fetching name def of the ClassRef inside GenClassRef*: " + fieldName);
			while ((genClassRef != null) && (!genClassRef.isClassRef())) {
				genClassRef = genClassRef.getGenClassRef();
			}
			if (genClassRef != null) {
				res = genClassRef.getRefNameDef(name);
			}
		}
		logger.log(BasicLevel.DEBUG, "Returned name def: " + res);
		return res;
	}

	/**
	 * Seeks for a father node of type field from a gen-class element or a
	 * name-def element. The name-def element represents the pname of a reference
	 * to a generic class.
	 *
	 * This method returns the nesting level of the generic class as well as the
	 * corresponding field name.
	 *
	 * @return the nesting level of a gen-class element.
	 */
	protected int getGenClassNestingLevel(Element fromElem, String fieldName) {
		int nestingLevel = 0;
		Element currentElem = (Element) fromElem.getParentNode();
		while ((currentElem != null) &&
				(!currentElem.getNodeName().equals("field"))) {
			currentElem = (Element) currentElem.getParentNode();
			if ((currentElem.getNodeName().equals("class-ref")) ||
					(currentElem.getNodeName().equals("gen-class-ref"))) {
				nestingLevel++;
			}
		}
		if (currentElem != null) {
			fieldName = currentElem.getAttribute("field-name");
		}
		return nestingLevel;
	}

	/**
	 * Seeks for a GenClassRef object given a class, a field name and a nesting
	 * level. This method returns the NameDef object of the GenClassRef object.
	 *
	 * @return the NameDef object of the GenClassRef object.
	 */
	protected NameDef getGenClassRefNameDef(Class clazz, String fieldName,
											String name, int nestingLevel) {
		NameDef res = null;
		GenClassRef genClassRef =
				getGenClassRef(clazz, fieldName, nestingLevel);
		if (genClassRef != null) {
			res = genClassRef.getRefNameDef(name);
		}
		return res;
	}


	/**
	 * Seeks for a GenClassRef object given a class, a field name and a nesting
	 * level. This method returns the NameDef object of the GenClassRef object.
	 *
	 * @return the NameDef object of the GenClassRef object.
	 */
	protected GenClassRef getGenClassRef(Class clazz, String fieldName,
										 int nestingLevel) {
		GenClassRef res = null;
		TypedElement typedElement = clazz.getTypedElement(fieldName);
		if (typedElement instanceof GenClassRef) {
			res = (GenClassRef) typedElement;
			GenClassRef genClassRef = (GenClassRef) typedElement;
			while ((genClassRef != null) && (nestingLevel != 0)) {
				res = genClassRef.getGenClassRef();
				nestingLevel--;
			}
		}
		return res;
	}

	/**
	 * Parses the mapping information related to a particular mapper and builds
	 * the corresponding Jorm meta-information. It can be class-related or
	 * generic class-related information.
	 * Manager, PathExplorer and currentClass must be set before calling this
	 * method.
	 *
	 * @param mappingElem   the current XML node.
	 * @param mapping       a Mapping object.
	 */
	public abstract void parseMapping(Element mappingElem, Mapping mapping) throws PException;

	/**
	 * Assigns an HashTable object to the MappingParser object.
	 *
	 * @param idvalue2genclassref	an HashTable that associates an id value with a GenClassRef object.
	 */
	public void setidvalue2genclassref(Map idvalue2genclassref) {
		this.idvalue2genclassref = idvalue2genclassref;
	}

	/**
	 * Assigns an HashTable object to the MappingParser object.
	 *
	 * @param motable	an HashTable that associates a pd file to its corresponding MetaObject.
	 */
	public void setmotable(Hashtable motable) {
		this.motable = motable;
	}

	///////////////////////////////////////////////////////////////////
	//                      from Loggable interface
	///////////////////////////////////////////////////////////////////

	/**
	 * Defines a logger object.
	 *
	 * @param logger	the logger object
	 */
	public void setLogger(Logger logger) {
		this.logger = logger;
	}

	/**
	 * Defines the logger factory to obtain new loggers.
	 *
	 * @param loggerfactory	The LoggerFactory object to obtain a logger object
	 */
	public void setLoggerFactory(LoggerFactory loggerfactory) {
		this.loggerFactory = loggerfactory;
		if (logger == null && loggerFactory != null) {
			logger = loggerFactory.getLogger("org.objectweb.jorm.xml2mi.rdb");
		}
	}

	public Logger getLogger() {
		return logger;
	}

	public LoggerFactory getLoggerFactory() {
		return loggerFactory;
	}
}
