/**
 * Copyright (C) 2001-2005 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Authors: S.Chassande-Barrioz.
 * Created on 10 mars 2005
 *
 */
package org.objectweb.jorm.xml2mi.lib;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.w3c.dom.Document;
import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * This class is an helper for the use of SAX Parser. It manages the entity
 * resolving public id, and handling errors.
 *
 * @author S.Chassande-Barrioz
 */
public class SAXParserHelper implements EntityResolver, ErrorHandler {

    /**
     * the entity location for each public id
     */
    private Properties publicid2location = null;
    
	/**
	 * The classLoader used to try to load entities and xml files
	 */
    private ClassLoader classLoader;
    
    /**
     * The list of errors occuring during the parsing
     */
    private List errors;
    
    /**
     * the parser
     */
    private DocumentBuilder parser;
	
	private Logger logger;

	/**
	 * Creates and initializes a SAX parser with the classLoader of the current
	 * class. The DTD checking is done.
	 * @param publicid2location contains the association between public id and
	 * location
	 */
	public SAXParserHelper(Properties publicid2location, 
	        Logger logger) throws SAXException {
		this.publicid2location = publicid2location;
		this.logger = logger;
		this.classLoader = getClass().getClassLoader();
		if (classLoader == null) {
		    classLoader = ClassLoader.getSystemClassLoader();
		}
		init(true);
	}

	/**
	 * Creates and initializes a SAX parser.
	 * @param publicid2location contains the association between public id and
	 * location
	 * @param cl is the classLoader used to try to load entities and xml files
	 * @param dtdverify indicates if the DTD checking must be done
	 */
	public SAXParserHelper(Properties publicid2location, 
	        Logger logger, 
	        ClassLoader cl, 
	        boolean dtdverify) throws SAXException {
		this.publicid2location = publicid2location;
		this.logger = logger;
		this.classLoader = cl;
		if (classLoader == null) {
		    classLoader = getClass().getClassLoader();
			if (classLoader == null) {
			    classLoader = ClassLoader.getSystemClassLoader();
			}
		}
		init(dtdverify);
	}

	/**
	 * Parses a xml file. Firstly this method try to found the xml file from
	 * the file system or the classpath. Then the method delegates the parsing
	 * to the parse(InputStream) method.
	 * 
	 * @param xmlFileName is the path to the xml file. This path can be valid
	 * from the file system or the classpath/classloader. The path can be an 
	 * URL.
	 * @return the Document representing the XML file.
	 * @throws SAXException is thrown if the file cannot be found or if the
	 * file is malformed.
	 */
    public Document parse(String xmlFileName) throws SAXException {
        InputStream is = null;
		try {
			is = new FileInputStream(xmlFileName);
		} catch (FileNotFoundException e) {
			logger.log(BasicLevel.DEBUG, xmlFileName + " is not a local file ");
		}

		if (is == null) {
			URL urldtd = null;
			try {
				urldtd = new URL(xmlFileName);
				is = urldtd.openStream();
			} catch (IOException e) {
				logger.log(BasicLevel.DEBUG, xmlFileName + " is not an URL ");
			}
		}

		// Try in the classpath
		if (is == null) {
			is = classLoader.getResourceAsStream(xmlFileName);
			if (is == null) {
				logger.log(BasicLevel.DEBUG, xmlFileName + " is not availlable in the classpath");
			}
		}

		//the location can be a jar file
		if (is == null) {
			int idx = xmlFileName.indexOf(".jar!");
			if (idx != -1) {
				try {
					JarFile jarfile = new JarFile(xmlFileName.substring(0, idx));
					// the path is the jar file, can continue
					if (jarfile != null) {
						ZipEntry zipentry = jarfile.getEntry(
						        xmlFileName.substring(idx + 5, xmlFileName.length()));
						// if the entry (file) exists in the jar file
						// extracts it and return an input stream from it
						if (zipentry != null) {
							is = jarfile.getInputStream(zipentry);
						}
					}
				} catch (Exception e) {
					logger.log(BasicLevel.DEBUG, xmlFileName + " is not in a jar file");
				}
			} else {
				logger.log(BasicLevel.DEBUG, xmlFileName + " is not in a jar file");
			}
		}
        
        if (is == null) {
            throw new SAXException("No XML file '" + xmlFileName + "' found");
        } else {
            return parse(is);
        }
        
    }
    
    /**
     * Parses a stream as an xml file. Errors are logged.
     * @param is an inputStream contains the XML xml content
	 * @return the Document representing the XML file.
	 * @throws SAXException is thrown if the file cannot be found or if the
	 * file is malformed.
     */
    public Document parse(InputStream is) throws SAXException {
        if (errors != null) {
            //forget errors of a previous parsing
            errors.clear();
        }
        //Parsing
        try {
            Document d = parser.parse(is, ".");
            if (d == null) {
                throw new SAXException("No document");
            }
            return d;
        } catch (IOException e) {
            throw new SAXException(e);
        }
    }

    /**
     * creates the real parser and initializes it.
	 * @param dtdverify indicates if the DTD checking must be done
     * @throws SAXException
     */
	private void init(boolean dtdverify)throws SAXException {
        //Parser initialisation
        DocumentBuilderFactory domFactory = DocumentBuilderFactory.newInstance();
        domFactory.setValidating(dtdverify);
        try {
            parser = domFactory.newDocumentBuilder();
        } catch (ParserConfigurationException e) {
            throw new SAXException(e);
        }
        parser.setEntityResolver(this);
        parser.setErrorHandler(this);
	}
	
    private List getErrorList() {
	    if (errors == null) {
	        errors = new ArrayList();
	    }
	    return errors;
	}
	
	// IMPLEMENTATION OF METHODS FROM THE ErrorHandler INTERFACE //
    //-----------------------------------------------------------//
	
    public void error(SAXParseException e) throws SAXException {
        logger.log(BasicLevel.ERROR, e.getMessage(), e);
        getErrorList().add(e);
    }
    public void fatalError(SAXParseException e) throws SAXException {
        logger.log(BasicLevel.FATAL, e.getMessage(), e);
        getErrorList().add(e);
    }
    public void warning(SAXParseException e) throws SAXException {
        logger.log(BasicLevel.WARN, e.getMessage(), e);
        getErrorList().add(e);
    }
	
    
	// IMPLEMENTATION OF METHODS FROM THE EntityResolver INTERFACE //
    //-------------------------------------------------------------//

    /**
     * Resolves the entity according to the associaition contained in 
     * #publicid2location. The entity are searched in the file system and in
     * the classpath. 
     */
    public InputSource resolveEntity (String publicId, String systemId)
            throws SAXException {
		String publicid = publicId;
		String location = null;
		if (publicid == null) {
			location = systemId;
		} else {
			logger.log(BasicLevel.DEBUG, "Looking for the publicId: " + publicid);
			for(Iterator it = publicid2location.entrySet().iterator(); it.hasNext();) {
			    Map.Entry me = (Map.Entry) it.next();
				String current = (String) me.getKey();
				logger.log(BasicLevel.DEBUG, "Current public id: " + current);
				if (current.equals(publicid)) {
					location = (String) me.getValue();
					break;
				}
			}
		}
		if (location == null) {
			logger.log(BasicLevel.INFO, 
			        "No dtd location found for the publicId: " + publicid);
			if (systemId == null)
				throw new SAXException("Could not resolve publicid: " + publicid);
			logger.log(BasicLevel.INFO, "Try to resolve " + publicid 
			        + " with the default location " + systemId);
			location = systemId;
		}

		InputStream is = null;

		// try in the file system
		try {
			is = new FileInputStream(location);
			logger.log(BasicLevel.DEBUG, "Resolved " + publicid 
			        + " to local file " + location);
		} catch (FileNotFoundException e) {
			logger.log(BasicLevel.DEBUG, location + " is not a local file ");
		}

		if (is == null) {
			URL urldtd = null;
			try {
				urldtd = new URL(location);
				is = urldtd.openStream();
				logger.log(BasicLevel.DEBUG, "Resolved " + publicid 
				        + " to url " + urldtd);
			} catch (IOException e) {
				logger.log(BasicLevel.DEBUG, location + " is not an URL ");
			}
		}

		// Try in the classpath
		if (is == null) {
			is = classLoader.getResourceAsStream(location);
			if (is != null) {
				logger.log(BasicLevel.DEBUG, "Resolved " + publicid 
				        + " to resource " + location);
			} else {
				logger.log(BasicLevel.DEBUG, location 
				        + " is not availlable in the classpath");
			}
		}

		//the location can be a jar file
		if (is == null) {
			int idx = location.indexOf(".jar!");
			if (idx != -1) {
				try {
					JarFile jarfile = new JarFile(location.substring(0, idx));
					// the path is the jar file, can continue
					if (jarfile != null) {
						ZipEntry zipentry = jarfile.getEntry(
								location.substring(idx + 5, location.length()));
						// if the entry (file) exists in the jar file
						// extracts it and return an input stream from it
						if (zipentry != null) {
							is = jarfile.getInputStream(zipentry);
						}
					}
				} catch (Exception e) {
					logger.log(BasicLevel.DEBUG, location + " is not an jar file");
				}
			}
		}

		if (is != null)
			return new InputSource(is);
		else
			throw new SAXException("Could not resolve publicid: ["
								  + publicid + "] location [" + location + "]");
    }
}
