/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.filter.jorm.lib;

import org.objectweb.jorm.naming.api.PNamingContext;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.api.MalformedExpressionException;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.api.TypingException;
import org.objectweb.medor.expression.api.Operand;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.expression.lib.BasicBinaryOperator;
import org.objectweb.medor.expression.lib.BasicVariableOperand;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.type.lib.QTypeTuple;

import java.math.BigDecimal;
import java.math.BigInteger;

/**
 * A SinglePName is an Operator representing the construction of a PName
 * from a NamingContext and a single Field.
 * <p>
 * The NamingContext is provided as a ParameterOperand. The name of the
 * ParameterOperand is defined to be:
 * <ol><li>The fully qualified class name for the PName of the class itself</li>
 * <li>The fully qualified class name, to which the attribute name is
 * concatenated for a reference PName</li>
 * </ol>
 * The Field is provided as a FieldOperand.
 *
 * @author	<A HREF="mailto:alia.mourad@rd.francetelecom.com><b>
 *			Mourad Alia
 *			</b></A>
 *		<A HREF="mailto:alexandre.lefebvre@rd.francetelecom.com><b>
 * 			Alexandre Lefebvre
 </b></A>
 */
public class SinglePName extends BasicBinaryOperator {

	/**
	 * Constructs a SinglePName from a ParameterOperand representing the
	 * naming context, a FieldOperand, and the PType of the corresponding
	 * PName.
	 * @param f is the FieldOperand from which the value is taken.
	 * @param pnc is the Operand representing the naming context.
	 * @param nameType is the PType of the PName created by this operator.
	 */
	public SinglePName(FieldOperand f, Operand pnc, PType nameType) {
		super(nameType, f, pnc);
	}

	/**
	 *
	 */
	public Operand evaluate(ParameterOperand[] pos, Object t)
		throws ExpressionException {
		Tuple o = (Tuple) t;
		//assign the value of the NamingContext
		expressions[1].evaluate(pos, o);
		//gets the naming context
		PNamingContext nc =
			(PNamingContext) ((Operand) expressions[1]).getObject();
		if (result == null)
			result = new BasicVariableOperand(type);
		PType fieldType = ((FieldOperand) expressions[0]).getField().getType();
		int i = ((FieldOperand) expressions[0]).getIndex();
		try {
			switch (fieldType.getTypeCode()) {
			case QTypeTuple.TYPECODE_BYTE:
                if (!o.isDefined(i)) {
                    result.setValue(nc.getNull());
                } else {
				    result.setValue(nc.decodeByte(o.getByte(i)));
                }
				break;
			case QTypeTuple.TYPECODE_CHAR:
                if (!o.isDefined(i)) {
                    result.setValue(nc.getNull());
                } else {
				    result.setValue(nc.decodeChar(o.getChar(i)));
                }
				break;
			case QTypeTuple.TYPECODE_DATE:
				result.setValue(nc.decodeDate(o.getDate(i)));
				break;
			case QTypeTuple.TYPECODE_INT:
                if (!o.isDefined(i)) {
                    result.setValue(nc.getNull());
                } else {
				    result.setValue(nc.decodeInt(o.getInt(i)));
                }
				break;
			case QTypeTuple.TYPECODE_LONG:
                if (!o.isDefined(i)) {
                    result.setValue(nc.getNull());
                } else {
				    result.setValue(nc.decodeLong(o.getLong(i)));
                }
				break;
			case QTypeTuple.TYPECODE_SHORT:
                if (!o.isDefined(i)) {
                    result.setValue(nc.getNull());
                } else {
				    result.setValue(nc.decodeShort(o.getShort(i)));
                }
				break;
			case QTypeTuple.TYPECODE_STRING:
				result.setValue(nc.decodeString(o.getString(i)));
				break;
			case QTypeTuple.TYPECODE_OBJBYTE:
				result.setValue(nc.decodeObyte((Byte) o.getObject(i)));
				break;
			case QTypeTuple.TYPECODE_OBJCHAR:
				result.setValue(nc.decodeOchar((Character) o.getObject(i)));
				break;
			case QTypeTuple.TYPECODE_OBJINT:
				result.setValue(nc.decodeOint((Integer) o.getObject(i)));
				break;
			case QTypeTuple.TYPECODE_OBJLONG:
				result.setValue(nc.decodeOlong((Long) o.getObject(i)));
				break;
			case QTypeTuple.TYPECODE_OBJSHORT:
				result.setValue(nc.decodeOshort((Short) o.getObject(i)));
				break;
			case QTypeTuple.TYPECODE_BIGINTEGER:
				result.setValue(nc.decodeBigInteger((BigInteger) o.getObject(i)));
				break;
			case QTypeTuple.TYPECODE_BIGDECIMAL:
				result.setValue(nc.decodeBigDecimal((BigDecimal) o.getObject(i)));
				break;
			default:
				throw new MedorException("Impossible to construct PName from field"); // This case don't accure if we compile at first
			}
		}
		catch (Exception e) {
			throw new ExpressionException("Error while building SinglePName", e);
		}
		return result;
	}

	/**
	 * Checks the semantic integrity of an expression.
	 * It checks that all types are compatible and prepare the expression to be
	 * evaluable. It also creates buffers where stores the result. Notes that
	 * when evaluating there is no creation of new objects. This method changes
	 * the state of this expression object, it will be evaluable and not
	 * modifiable.
	 * @throws org.objectweb.medor.expression.api.TypingException when incompatible types error
	 * occures.
	 * @throws org.objectweb.medor.expression.api.MalformedExpressionException if syntax error
	 */
	public Operand compileExpression()
		throws TypingException, MalformedExpressionException {
		if ((expressions[0] != null) && (expressions[1] != null)) {
			result = new BasicVariableOperand(type);
			verified = true;
		}
		else
		// Operands not yet assigned
			throw new MalformedExpressionException("null children value");
		return result;
	}

	public FieldOperand getField() {
		return (FieldOperand) expressions[0];
	}

    public ParameterOperand getPNameManagerParameter() {
        return (expressions[1] instanceof ParameterOperand
			? (ParameterOperand) expressions[1]
			: null);
    }

	//Implementation of Operator interface
	/**
	 * There is no Java operator associated to a PName
	 */
	public String getOperatorString() {
		return "singlePName";
	}
}
