/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2005 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */
package org.objectweb.medor.optim.jorm;

import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.UnaryOperator;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.lib.ExpressionPrinter;
import org.objectweb.medor.optim.api.RewriteRule;
import org.objectweb.medor.optim.lib.DropUnusedProjFieldsRule;
import org.objectweb.medor.query.api.CalculatedField;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.lib.Nest;
import org.objectweb.medor.query.lib.QueryTreePrinter;
import org.objectweb.medor.query.rdb.lib.AggregateRdbQueryNode;
import org.objectweb.medor.query.rdb.lib.BasicRdbExpQueryLeaf;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.ArrayList;

/**
 * This rules transforms a QueryTree built on top of JORM Extents into
 * a QueryTree built on top of RdbExpQueryLeaves.
 * <p>It handles cases where there are Nest QueryNodes, and relies on the
 * JormFlatten2Rdb rule for the rest of the QueryTree.</p>
 * 
 * @author A. Lefebvre
 */
public class Jorm2Rdb extends JormRule {

	public Jorm2Rdb() {
		super("Jorm2Rdb");
	}

	private RewriteRule jormRule = new JormFlatten2Rdb();

	public QueryTree rewrite(QueryTree qt, QueryNode parent)
			throws MedorException {
		log.log(BasicLevel.DEBUG, "Jorm2Rdb: Input query tree:");
		QueryTreePrinter.printQueryTree(qt, log, BasicLevel.DEBUG);

		if (qt instanceof Nest) {
			QueryTree qtInter =
					jormRule.rewrite(((QueryNode)qt).getChildren()[0],
							(QueryNode) qt);
			log.log(BasicLevel.DEBUG, "Jorm2Rdb: intermediate query tree:");
			//WARNING the line below creates a side-effect on the SQL query
			//It should be uncommented with care
			QueryTreePrinter.printQueryTree(qtInter, log);
			//Dropping unused projected fields - necessary to eliminate
			//the calculated class PName (if unused) which would otherwise
			//prevent the Nest to be rewritted into an RDB aggregate
			RewriteRule dropProj = new DropUnusedProjFieldsRule();
			QueryTree qt2 =	dropProj.rewrite(qt, parent);
			log.log(BasicLevel.DEBUG, "Jorm2Rdb: intermediate query tree after dropping unused projections:");
			QueryTreePrinter.printQueryTree(qt2, log);
			qt = qt2;
			//case 1: the result has no CalculatedField: collapse
			if ((qtInter instanceof QueryNode) && !(qtInter instanceof Nest)
					&& (noCalculatedField((QueryNode)qtInter)) ) {
				//build the AggregateRdbQueryNode
				Field[] fs = qtInter.getTupleStructure().getFields();
				for (int i = 0; i < fs.length; i++) {
					if (fs[i] instanceof PropagatedField) {
						log.log(BasicLevel.DEBUG, "PropagatedField " + ((PropagatedField) fs[i]).getPreviousFields()[0]);
						QueryTreeField qtf = (QueryTreeField)
								((PropagatedField) fs[i]).getPreviousFields()[0];
						log.log(BasicLevel.DEBUG, "PF name " + qtf.getName());
					}
				}
				//tmpQT and tmpQT2 for casting purposes
				log.log(BasicLevel.DEBUG, "NestedField " +((Nest)qt).getNestedField());
				Field[] tmpQT = ((Nest)qt).getNestedField().getFields();
				ArrayList tmpQT2 = new ArrayList();
				for (int jj = 0; jj < tmpQT.length; jj++) {
					log.log(BasicLevel.DEBUG, "containing " +tmpQT[jj] + " "
					+ tmpQT[jj].getName() + " (" + (QueryTreeField)
							(((PropagatedField)tmpQT[jj]).
								getPreviousFields()[0]) + ") of node " + 
								((QueryTreeField)
								(((PropagatedField)tmpQT[jj]).
									getPreviousFields()[0])).getQueryTree());
					if (!tmpQT2.contains(
							(((PropagatedField)tmpQT[jj]).
								getPreviousFields()[0]))) {
						tmpQT2.add(
							(((PropagatedField)tmpQT[jj]).
								getPreviousFields()[0]));
					}
				}
				BasicRdbExpQueryLeaf theLeaf = (BasicRdbExpQueryLeaf)
								((QueryNode)qtInter).getChildren()[0];
				QueryNode qtNew =
						new AggregateRdbQueryNode(
							(QueryTreeField[]) (tmpQT2.toArray(new QueryTreeField[] {})),
							((Nest)qt).getNestedField().getName(),
							//replace the group by fields of the Nest
							replaceGroupByFields((Nest)qt),
							theLeaf,
							theLeaf.getDataStore(), qt.getName());
				qtNew.setDistinct(qtInter.getDistinct());
				//add other fields (calculatedFields for aggregates)
				Field[] otherFields = qt.getTupleStructure().getFields();
				for (int fl = 0; fl < otherFields.length; fl++) {
				    log.log(BasicLevel.DEBUG, "Handling other field " + otherFields[fl]);
					if (otherFields[fl] instanceof CalculatedField) {
						CalculatedField cf = (CalculatedField) otherFields[fl];
						replaceAggregatedField(cf.getExpression());

						qtNew.addCalculatedField(
							//withdraw the node name from the CalculatedField name
							cf.getName().substring(
									cf.getName().indexOf(qt.getName())
									+ qt.getName().length() + 1),
							cf.getType(),
							cf.getExpression());
						log.log(BasicLevel.DEBUG, "Adding CalculatedField " + cf + " (" + ExpressionPrinter.e2str(cf.getExpression()) + ")");
					} else if (otherFields[fl] instanceof PropagatedField) {
					    //update the propagated field
					    log.log(BasicLevel.DEBUG, "Adding PropagatedField from " + otherFields[fl]);
					    	QueryTreeField previousField = 
					    	    (QueryTreeField)
					    	    ((PropagatedField)
					    	            ((PropagatedField)otherFields[fl]).getPreviousFields()[0])
					    	            .getPreviousFields()[0];
					    log.log(BasicLevel.DEBUG, "Previous field is " + previousField + " of node " + previousField.getQueryTree());
					    qtNew.addPropagatedField(previousField.getName(),
					            previousField.getType(),
					            new QueryTreeField[] {previousField});
					} else {
					    log.log(BasicLevel.DEBUG, "Doing nothing");
					}
				}
				//TODO propagate the changes to the parent
				log.log(BasicLevel.DEBUG, "Jorm2Rdb: final query tree:");
				QueryTreePrinter.printQueryTree(qtNew, log);
				return qtNew;
			} else {
				/* Case 2: In this case, there are some calculated fields
				(PName for example).
				The aggregation cannot be delegated to the database. */
				log.log(BasicLevel.DEBUG, "Jorm2Rdb: some calculated fields: keeping input query tree");
				return qt;
			}
		}
		else return jormRule.rewrite(qt, parent);
	}

	private void replaceAggregatedField(Expression e) {
		UnaryOperator op = (UnaryOperator) e;
		FieldOperand fo = (FieldOperand) op.getExpression(0);
		fo.setField(((PropagatedField) fo.getField()).getPreviousFields()[0]);
	}

	private QueryTreeField[] replaceGroupByFields(Nest nest) {
		log.log(BasicLevel.DEBUG, "Jorm2Rdb: replacing GroupBy fields");
		QueryTreeField[] groupBy = (nest).getNestingFields();
		QueryTreeField[] newGroupBy = new QueryTreeField[groupBy.length];
		/* replace group by fields of Nest, which are propagated from propagated,
		 * with the initial fields (thus 2 levels of previous)
		 */
		for (int i = 0; i < groupBy.length; i++) {
			log.log(BasicLevel.DEBUG, "Jorm2Rdb: examining GroupBy field " + groupBy[i]);
            if (groupBy[i] instanceof PropagatedField) {
                newGroupBy[i] = 
                    (QueryTreeField)
                            ((PropagatedField)groupBy[i]).getPreviousFields()[0];
            } else if (groupBy[i] instanceof CalculatedField) {
                newGroupBy[i] = groupBy[i];
                log.log(BasicLevel.ERROR, "Jorm2Rdb.replaceGroupByFields: Cannot find previous field of a calculated field :" + groupBy[i]);
            }
			log.log(BasicLevel.DEBUG, "new GroupBy field " + i + ": "
					+ newGroupBy[i] + " of node " + newGroupBy[i].getQueryTree());
		}
		return newGroupBy;
	}
	/**
	 * Checkes whether a QueryNode has calculatedFields
	 * @param qn the QueryNode to be tested
	 * @return true if the QueryNode has no CalculatedFields, false otherwise.
	 * @throws MedorException
	 */
	private boolean noCalculatedField(QueryNode qn) throws MedorException {
		boolean noCCF = true;
		TupleStructure ts = qn.getTupleStructure();
		for (int i = 0; i < ts.getSize(); i++) {
			if (ts.getField(i+1) instanceof CalculatedField) {
				log.log(BasicLevel.DEBUG, "Found CalculatedField " + ts.getField(i+1).getName());
				noCCF = false;
			}
		}
		return noCCF;
	}
}
