/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2004 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */
package org.objectweb.medor.optim.jorm;

import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.jorm.lib.CompositePName;
import org.objectweb.medor.filter.jorm.lib.SinglePName;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.optim.api.LeafRewriter;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.jorm.api.JormExtent;
import org.objectweb.medor.query.jorm.lib.PNameField;
import org.objectweb.medor.query.lib.Project;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.Operand;

import java.util.Iterator;
import java.util.Map;

/**
 * @author S. Chassande-Barrioz
 */
public abstract class JormLeafRewriter extends JormRule
	implements LeafRewriter {

	public JormLeafRewriter(String suffix) {
		super(suffix);
	}

	public abstract boolean canRewrite(QueryLeaf ql);

	public QueryTree rewrite(QueryTree qt, QueryNode parent)
			throws MedorException {
		throw new UnsupportedOperationException("A leaf rewriter only rewrites leaves, not nodes.");
	}

	public QueryTree rewrite(QueryLeaf ql)
			throws MedorException {
		if (!canRewrite(ql))
			throw new MedorException(
				"Impossible to rewrite this QueryLeaf: " + ql);
		JormExtent extent = (JormExtent) ql;
		QueryLeaf specificQL = createQueryLeaf(extent);
		QueryNode qt = new Project("");

		Field[] oldFields = ql.getTupleStructure().getFields();
		for (int i = 0; i < oldFields.length; i++) {
			if (oldFields[i] instanceof PNameField) {
                PNameField pnf = (PNameField) oldFields[i];
                NameDef nd = pnf.getNameDef(extent);
                addReference(nd, pnf, qt, specificQL, extent);
            } else {
                PrimitiveElement pe = getPrimitiveElement(
                    extent, getShortFieldName(oldFields[i].getName()));
                QueryTreeField qtf =
                    getField(specificQL, pe, extent);
                //add a propagated field into the intermediate QueryTree
                qt.addPropagatedField(
                    qtf.getName(),
                    qtf.getType(),
                    new QueryTreeField[]{qtf});
            }
		}
		return qt;
	}

	protected abstract QueryLeaf createQueryLeaf(JormExtent extent)
		throws MedorException;

	public abstract QueryTreeField addPrimitiveElement(QueryLeaf neo,
													   PrimitiveElement pe,
													   JormExtent extent)
		throws MedorException;

	public QueryTreeField getField(QueryLeaf neo,
                                   PrimitiveElement pe,
                                   JormExtent extent) throws MedorException {
        if (neo.getTupleStructure().contains(pe.getName())) {
            return (QueryTreeField)
                neo.getTupleStructure().getField(pe.getName());
        } else {
            return addPrimitiveElement(neo, pe, extent);
        }
    }

	protected void addReference(NameDef nd,
								PNameField pnf,
								QueryNode qt,
								QueryLeaf specificQL,
                                JormExtent extent) throws MedorException {
		//System.out.println("NameDef: " + nd);
        Operand o = getPNCOperand(extent, pnf);
        PType type = pnf.getType();
		QueryTreeField qtf = null;
		if (nd.isFieldName()) {
			String ndfn = getFieldName(specificQL.getName(), nd.getFieldName());
			//Check if the field has not already added on the specific query leaf
			PrimitiveElement pe = getPrimitiveElement(extent, ndfn);
            qtf = getField(specificQL, pe, extent);

			//add a CalculatedField into the intermediate QueryTree.
			// This calculatedField has an expression which is a
			// SinglePName.
			Expression e = new SinglePName(new BasicFieldOperand(qtf),o, type);
			qt.addCalculatedField(pnf.getName(), type, e);
		}
		else if (nd.isNameRef()) {
            Map proj = nd.getNameRef().getProjection();
			FieldOperand[] fos = new BasicFieldOperand[proj.size()];
			String[] cofns = new String[proj.size()];
			int counter = 0;
			for (Iterator it = proj.entrySet().iterator(); it.hasNext();) {
                Map.Entry me = (Map.Entry) it.next();
                cofns[counter] = (String) me.getKey();
                String clafn = (String) me.getValue();
				String ndfn = getFieldName(specificQL.getName(), clafn);
                qtf = getField(specificQL, getPrimitiveElement(extent, ndfn), extent);
				fos[counter] = new BasicFieldOperand(qtf);
                counter++;
			}
			//add a CalculatedField into the intermediate QueryTree.
			// This calculatedField has an expression which is a
			// CompositePName.
			Expression e = new CompositePName(fos, cofns, o, type);
			qt.addCalculatedField(pnf.getName(), type, e);
		}
	}
    protected String getFieldName(String prefix, String nodeName) {
        if (prefix==null || prefix.length()==0)
            return nodeName;
        else
            return prefix + "." + nodeName;
    }

    protected String getShortFieldName(String fn) {
        return fn.substring(fn.lastIndexOf(".")+1, fn.length());
    }
}
