/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.query.jorm.lib;

import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.Reference;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.query.api.QueryTreeField;

import java.util.Iterator;
import java.util.Map;

/**
 * This class represents the extent of a JORM class.
 * It contains and implements its associated TupleStructure.
 */
public class ClassExtent extends BasicJormExtent {

	private Class myClass;

	//this parameter is set to false by default
	//it is set to true via the addPrefetchFields method in the JormQueryTreeHelper class
	private boolean prefetch = false;
	
	public ClassExtent() {
	}

	public ClassExtent(String _name, String pnFieldName) {
		super(_name, pnFieldName);
	}

	/**
	 * Constructs an extent for the JORM class, in the form of a QueryLeaf.
	 * Two options are possible, depending on the value of the classPNameOnly
	 * parameter:
	 * <ol><li>either all attributes of the JORM class are present in the
	 * QueryLeaf</li>
	 * <li>or only the PName attribute is present in the QueryLeaf</li>
	 * </ol>
	 * @param jormClass is the JORM full class name.
	 * @param _name is the name of the node (null value => empty string)
	 * @param classPNameOnly indicates whether the extent should contain only
	 * the PName for the class (if true) or all attributes for that class
	 * (if false).
	 *
	 */
	public ClassExtent(Class jormClass, String _name,
					   String pnameFieldName,
					   boolean classPNameOnly) throws MedorException {
		super(_name, pnameFieldName);
		myClass = jormClass;
		
		//First, create a Field for that Class
		identifier = new PNameField(getFieldName(name, pnameFieldName), myClass, this);
		this.fields.add(identifier);
		name2field.put(identifier.getName(), identifier);

		if (!classPNameOnly) {

			//Fields are PrimitiveElement, ClassRef or GenClassRef
			for (Iterator it = jormClass.getAllFields().iterator(); it.hasNext();) {
				addField((TypedElement) it.next());
			}
		}
	}

	/**
	 * Constructs an extent for the JORM class, in the form of a QueryLeaf.
	 * The field names for those fields which are part of the extent are
	 * passed as a parameter. The fact that the PName is part of the extent
	 * is optional.
	 * @param jormClass is the JORM full class name.
	 * @param _name is the name of the node (null value => empty string)
	 * @param fieldNames is an array of JORM field names which should be present
	 * in the extent
	 * @param addPName is a boolean which is true if the PName should be part
	 * of the extent, false otherwise.
	 * @throws MedorException if one of the requested fieldNames does not
	 * correspond to a field of the JORM class.
	 */
	public ClassExtent(Class jormClass, String _name,
					   String[] fieldNames,
					   boolean addPName,
					   String pnameFieldName)
		throws MedorException {
		super(_name, pnameFieldName);
		myClass = jormClass;

		//First, create a Field for that Class if requested
		if (addPName) {
			PNameField nameField = new PNameField(getFieldName(name, pnameFieldName), myClass, this);
			this.fields.add(nameField);
			name2field.put(nameField.getName(), nameField);
		}
		//for each requested fieldName, adds the corresponding
		for (int i = 0; i < fieldNames.length; i++) {
			TypedElement te = jormClass.getTypedElement(fieldNames[i]);
			if (te == null)
				throw new MedorException("No field named " + fieldNames[i] + " in JORM class " + jormClass.getName());
			addField(te);
		}
	}

	public Object clone(Object clone,
						Map obj2clone) throws CloneNotSupportedException {
		clone = super.clone(clone, obj2clone);
		((ClassExtent) clone).myClass = myClass;
		return clone;
	}

	public QueryTreeField addField(String fieldName) throws MedorException {
		if (fieldName == null) {
			throw new java.lang.IllegalArgumentException("Null field name");
		}
		TypedElement te = myClass.getTypedElement(fieldName);
		if (te == null) {
			throw new MedorException("No field '" + fieldName
				+ "' found on the class '" + myClass.getName() +"'.");
		}
		return addField(te);
	}
	/**
	 * It adds a new Field into the tuple structure.
	 * @param te is the jorm Meta object TypedElement which is the source of
	 * the field
	 * pname or the gen class pname.
	 */
	public QueryTreeField addField(TypedElement te) throws MedorException {
		QueryTreeField basicF = null;
		String fn = getFieldName(name, te.getName());
		basicF = (QueryTreeField) name2field.get(fn);
		if (basicF != null)
			throw new MedorException("the field " + te.getName()
				+ " is already defined (" + fn + ")");
		if (te instanceof Reference)
			basicF = new PNameField(fn, myClass, this);
		else
			basicF = new BasicJormField(fn, this, te);
		fields.add(basicF);
		name2field.put(basicF.getName(), basicF);
		return basicF;
	}

	public MetaObject getMetaObject() {
		return myClass;
	}

    public String getJormName() {
        return myClass.getFQName();
    }
    
	public boolean isPrefetch() {
		return prefetch;
	}
	public void setPrefetch(boolean prefetch) {
		this.prefetch = prefetch;
	}
}
