/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2004 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre, S. Chassande-Barrioz
 */

package org.objectweb.medor.query.jorm.lib;

import java.util.ArrayList;
import java.util.Collections;

import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ClassRef;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.Reference;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.MalformedExpressionException;
import org.objectweb.medor.expression.api.Operand;
import org.objectweb.medor.expression.api.TypingException;
import org.objectweb.medor.expression.lib.And;
import org.objectweb.medor.expression.lib.Equal;
import org.objectweb.medor.expression.lib.Not;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.jorm.lib.NavigatorOperator;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.filter.lib.ExpressionPrinter;
import org.objectweb.medor.filter.lib.MemberOf;
import org.objectweb.medor.lib.Log;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.jorm.api.JormExtent;
import org.objectweb.medor.query.lib.JoinProject;
import org.objectweb.medor.query.lib.QueryTreePrinter;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * This class constructs a QueryTree from a NavigatorOperator, which represents
 * a path expression.
 * <p>It is a factory for normal JoinProject query nodes, which implement the
 * semantics of the path traversal represented by the NavigatorOperator.
 */
public class NavigatorNodeFactory {

	public final static String PNAME_FIELD_NAME = "object";
	private static int pnameFieldNameIndex = 0;

	public final static String GENCLASS_ELEMENT_NAME = "element";

	/**
	 * To navigate up to the last field.
	 */
	public final static byte NAVIGATION = 1;
	/**
	 * To navigate up to the element referenced by the last field.
	 * <p>For a GenClassRef, includes the GenClass element. For a ClassRef,
	 * includes all fields of the referenced class object.
	 */
	public final static byte NAVIGATION_INTO_LAST = 2;

	/**
	 * Applies to navigation ending with a GenClassRef only.
	 * <p> Tests whether there are no GenClass elements for the GenClassRef.
	 */
	public final static byte IS_EMPTY = 4;
	/**
	 * Applies to navigation ending with a GenClassRef only.
	 * <p> Tests whether there are GenClass elements for the GenClassRef.
	 */
	public final static byte IS_NOT_EMPTY = 8;

	public static Logger logger = Log.getLoggerFactory()
			.getLogger("org.objectweb.medor.query.jorm.NavigatorNodeFactory");

	/**
	 * Creates a QueryTree corresponding to the path traversal represented by
	 * the NavigatorOperator.
	 * <p>The end of the path expression can either be a single-valued field or
	 * a reference to a GenClass. In the latter case, the elements
	 * of the GenClass are projected together with their index attributes.
	 * @param nav is the NavigatorOperator representing the path expression.
	 * This operator starts with a FieldOperand pointing to a Field called here
	 * FStart. All fields in the node hosting FStart are projected.
	 * @param aliasName is the root name of the Fields reached by the path
	 * expression.
	 * @param operation indicates which operation to perform (NAVIGATION,
	 * NAVIGATION_INTO_LAST, IS_EMPTY or IS_NOT_EMPTY)
	 * @return the QueryTreeField which matches the NavigatorOperator.
	 */
	public static QueryTreeField navigate(NavigatorOperator nav,
										  String aliasName,
										  byte operation)
			throws MedorException {
		//'qtf' is the QueryTreeField to use for a join
		QueryTreeField qtf = null;
		//pnf is the PNameField which must be used to access to the meta info
		PNameField pnf = null;
		//op is the operation type at each step
		byte op = NAVIGATION;
		boolean joinAlreadyConstructed = false;
		String fieldName = null;
		try {
			fieldName = ((Operand) nav.getExpression(1)).getString();
		} catch (TypingException te) {
			throw new MedorException(te);
		}
		logger.log(BasicLevel.DEBUG, "navigate(" + aliasName + ", "
				+ ExpressionPrinter.e2str(nav.getExpression(0)) + ", " + fieldName
				+ ", op:" + op + ")");

		//iterate over the navigation path, starting with the FieldOperand
		ArrayList path = new ArrayList();
		path.add(nav.getFieldOperand());
		path = nav.getPath(path);

		for (int j = 0; j < path.size() - 1; j++) {

			boolean propagQtf;
			//restores the type of operation to the inital one for the end
			//of the navigation
			if (j == path.size() - 2) {
				op = operation;
			}
			Object element = path.get(j);
			if (element instanceof FieldOperand) {
				qtf = (QueryTreeField) ((FieldOperand) element).getField();
				pnf = getPNameField(qtf);
				logger.log(BasicLevel.DEBUG, "Navigation from FieldOperand " + qtf.getName());
				propagQtf = true;
			} else {
				//propagQtf = false;
				//adding even intermediate fields
				propagQtf = true;
				logger.log(BasicLevel.DEBUG, "Further navigation for " + element);
				//The alias name is null because only the first element (the last
				//join) can be have an alias
			}

			fieldName = (String) path.get(j + 1);
			logger.log(BasicLevel.DEBUG, "Navigate to " + fieldName);

			logger.log(BasicLevel.DEBUG, "propagQtf=" + propagQtf);
			logger.log(BasicLevel.DEBUG, "qtf=" + qtf.getName() + " of node " + qtf.getQueryTree());
			if (pnf == null)
				throw new MedorException("Impossible to find a PNameField");

			// 'qt' is the QueryTree of 'qtf"
			QueryTree qt = qtf.getQueryTree();

			if (!pnf.isClassPName()) {
				logger.log(BasicLevel.DEBUG, "Reference from a Class to a Class");
				Reference ref = pnf.getReference();

				Class jormClass = null;
				if (ref instanceof ClassRef)
					jormClass = ((ClassRef) ref).getMOClass();
				else
					throw new MedorException("Impossible to navigate with PName:"
							+ ref.getName() + " " + ref.getParent());

				ClassExtent extent = null;
				String nodeName;
				if (j == path.size() - 2) {
					//this is the end of the navigation
					logger.log(BasicLevel.DEBUG, "Extent at the end of the navigation");
					nodeName = (op != NAVIGATION_INTO_LAST
							? aliasName
							: jormClass.getFQName());
					pnameFieldNameIndex += 1;
					//create the extent corresponding to the reference
					extent = new ClassExtent(
							jormClass,
							nodeName,
							PNAME_FIELD_NAME + pnameFieldNameIndex,
							false);
				} else {
					//in the middle of the navigation: propagate only the required
					//field
					logger.log(BasicLevel.DEBUG, "Extent in the middle of the navigation");
					nodeName = jormClass.getFQName();
					pnameFieldNameIndex += 1;
					//create the extent corresponding to the reference
					extent = new ClassExtent(
							jormClass,
							nodeName,
							new String[]{fieldName},
							true,
							PNAME_FIELD_NAME + pnameFieldNameIndex);
				}
				// pnf references the PNameField of the joined extent
				pnf = (PNameField) extent.getField(extent.getPNameFieldName());

				JoinProject join;
				if (!joinAlreadyConstructed) {
					logger.log(BasicLevel.DEBUG, "Constructing new JoinProject");
					join = new JoinProject("", op != NAVIGATION);
					joinAlreadyConstructed = true;
					//add the join condition: pnf should be equal to the
					//PName of the extent
					join.setQueryFilter(
							new Equal(
									new BasicFieldOperand(qtf),
									new BasicFieldOperand(pnf)
							)
					);
				} else {
					join = (JoinProject) qtf.getQueryTree();
					logger.log(BasicLevel.DEBUG, "Re-using JoinProject " + join);
					propagQtf = false;
					//add the join condition: pnf should be equal to the
					//PName of the extent
					join.setQueryFilter(
							new And(
								new Equal(
									new BasicFieldOperand(
											((PropagatedField)qtf).getPreviousFields()[0]),
									new BasicFieldOperand(pnf)
								),
								join.getQueryFilter()
						)
					);
				}

				logger.log(BasicLevel.DEBUG, "Current qt is:");
				QueryTreePrinter.printQueryTree(qt, logger);
				logger.log(BasicLevel.DEBUG, "Current join is:");
				QueryTreePrinter.printQueryTree(join, logger);

				//project in qt all fields of the QueryTree containing
				//the origin of the path expression
				Field[] fToAdd = qt.getTupleStructure().getFields();
				logger.log(BasicLevel.DEBUG, "Adding fields:");
				logger.log(BasicLevel.DEBUG, "qtf is:" + qtf);
				for (int i = 0; i < fToAdd.length; i++) {
					if (propagQtf || qtf != fToAdd[i]) {
						logger.log(BasicLevel.DEBUG, "\t-" + fToAdd[i].getName());
						//propagate fields only if not already there
						try {
							join.getField(fToAdd[i].getName());
							logger.log(BasicLevel.DEBUG, "Field already present: not adding.");

						} catch (MedorException e) {
							logger.log(BasicLevel.DEBUG, "Field not present: adding.");

							join.addPropagatedField(
								fToAdd[i].getName(),
								fToAdd[i].getType(),
								new QueryTreeField[]{(QueryTreeField) fToAdd[i]});
						}
					}
					//else
					//logger.log(BasicLevel.DEBUG, "\t-Do not propagate "  + fToAdd[i].getName());
				}

				// project in qt the fields corresponding to the end of the path
				// expression
				//TODO add only of end of navigation
				fToAdd = extent.getTupleStructure().getFields();
				for (int i = 0; i < fToAdd.length; i++) {
					logger.log(BasicLevel.DEBUG, "Adding propagated field " + fToAdd[i].getName());
					join.addPropagatedField(
							fToAdd[i].getName(),
							fToAdd[i].getType(),
							new QueryTreeField[]{(QueryTreeField) fToAdd[i]});
				}
				qtf = (QueryTreeField)
						join.getField(extent.getFieldName(nodeName, fieldName));
				//qtf references a propagated field to a field named 'fieldName'

				qt = join;
			} else { //it is a class PName
				logger.log(BasicLevel.DEBUG, "Adding the field if not already present");
				if (pnf.isInGenClass()) {
					throw new MedorException(
							"Impossible to navigate with a multi valued reference");
				}
				logger.log(BasicLevel.DEBUG, "Checking that " + fieldName + " appears in " + qt);
				// The field 'fieldName' must be found on 'qt'
				String fn = contains(qt.getTupleStructure(),
						pnf.getQueryTree().getName() + "." + fieldName);
				if (fn != null) {
					qtf = (QueryTreeField) qt.getTupleStructure().getField(fn);
				} else if (qt instanceof ClassExtent) {
					//In case where qt is a ClassExtent
					ClassExtent extent = (ClassExtent) qt;
					TypedElement te =
							((Class) extent.getMetaObject()).getTypedElement(fieldName);
					if (te == null)
						throw new MedorException("The jorm class "
								+ ((Class) extent.getMetaObject()).getName()
								+ " does not provide a field " + fieldName);
					qtf = extent.addField(te);
				} else
					throw new MedorException(
							"No field " + fieldName + " found on the query tree " + qt);
			}
			logger.log(BasicLevel.DEBUG, "Query tree before the end of treatment:");
			QueryTreePrinter.printQueryTree(qt, logger);
			qtf = endOfNavigation(op, qtf, qt, aliasName);
			pnf = getPNameField(qtf);
		}
		return qtf;
	}


	/**
	 * Manages the end of the navigation in the case of a reference
	 * @param op is the navigation operation type
	 * @param qtf is the QueryTreeField
	 * @param qt the QueryTree
	 * @param aliasName the alias name of the QueryTree
	 * @return the QueryTreeField corresponding to the end of the navigation.
	 * @throws MedorException
	 */
	private static QueryTreeField endOfNavigation(byte op,
												  QueryTreeField qtf,
												  QueryTree qt,
												  String aliasName)
			throws MedorException {
		PNameField pnf = getPNameField(qtf);
		//pnf can be null
		logger.log(BasicLevel.DEBUG, "Computing the end of navigation with pnf=" + pnf + " for operation " + op);

		if (op != NAVIGATION && pnf != null) {
			//it is a reference
			logger.log(BasicLevel.DEBUG, "pnf.ref=" + pnf.getReference());
			if (pnf.getReference() instanceof GenClassRef) {
				// Multivalued reference => create a GenClass node and the join

				GenClassRef gcr = (GenClassRef) pnf.getReference();
				String nodeName = ((Class) gcr.getParent()).getFQName()
						+ "." + gcr.getName();
				pnameFieldNameIndex += 1;
				//use the alias if possible
				GenClassExtent extent = new GenClassExtent(
						gcr,
						nodeName,
						PNAME_FIELD_NAME + pnameFieldNameIndex,
						GENCLASS_ELEMENT_NAME);

				//'pnf' references the PNameField of the GenClass
				pnf = (PNameField) extent.getField(extent.getPNameFieldName());

				JoinProject join;
				boolean newjoin = false;
				if (qt instanceof JoinProject) {
					logger.log(BasicLevel.DEBUG, "Already a join. Propagating right fields only");
					join = (JoinProject) qt;
					propagateFieldsRight(extent, join, op, aliasName);
				} else {
					logger.log(BasicLevel.DEBUG, "Creating a join. Propagating left and right fields");
					join = new JoinProject("" , op != NAVIGATION);
					propagateFieldsLeft(qt, join);
					propagateFieldsRight(extent, join, op, aliasName);
					newjoin = true;
				}

				//add the join condition: pnf should be equal to the
				//PName of the extent
				try {
					switch (op) {
					case NAVIGATION_INTO_LAST:
						if (newjoin) {
							join.setQueryFilter(new Equal(
								new BasicFieldOperand(qtf),
								new BasicFieldOperand(pnf)));
						} else {
							join.setQueryFilter(
								new And(
									join.getQueryFilter(),
									new Equal(
										new BasicFieldOperand(
											((PropagatedField)qtf).getPreviousFields()[0]),
										new BasicFieldOperand(pnf))
										)
								);
						}
						return (QueryTreeField) join.getField((aliasName == null
								? extent.getElementFieldName()
								: aliasName));
					case IS_NOT_EMPTY:
						if (newjoin) {
							join.setQueryFilter(new MemberOf(
								Collections.singletonList(new BasicFieldOperand(qtf)),
								Collections.singletonList(new BasicFieldOperand(pnf))));
						} else {
							join.setQueryFilter(
								new And(
									join.getQueryFilter(),
									new MemberOf(
										Collections.singletonList(new BasicFieldOperand(((PropagatedField)qtf).getPreviousFields()[0])),
										Collections.singletonList(new BasicFieldOperand(pnf))
										)
									)
								);
						}
						return (QueryTreeField) join.getField(qtf.getName());
					case IS_EMPTY:
						if (newjoin) {
							join.setQueryFilter(new Not(new MemberOf(
									Collections.singletonList(new BasicFieldOperand(qtf)),
									Collections.singletonList(new BasicFieldOperand(pnf)))));
						} else {
							join.setQueryFilter(
								new And(
									join.getQueryFilter(),
									new Not(new MemberOf(
											Collections.singletonList(new BasicFieldOperand(((PropagatedField)qtf).getPreviousFields()[0])),
											Collections.singletonList(new BasicFieldOperand(pnf))
										)
									)
								)
							);
						}
						return (QueryTreeField) join.getField(qtf.getName());
					default:
						throw new MedorException("Operation type not supported: " + op);
					}
				} catch (MalformedExpressionException e1) {
					throw new MedorException(e1);
				}
			} else if (op == NAVIGATION_INTO_LAST
					&& pnf.getReference() instanceof ClassRef) {

				ClassRef cr = (ClassRef) pnf.getReference();
				String nodeName = (aliasName == null
						? cr.getMOClass().getFQName()
						: aliasName) + "." + cr.getName();
				pnameFieldNameIndex += 1;
				ClassExtent extent = new ClassExtent(
						cr.getMOClass(),
						nodeName,
						PNAME_FIELD_NAME + pnameFieldNameIndex,
						false);

				//'pnf' references the PNameField of the GenClass
				pnf = (PNameField) extent.getField(extent.getPNameFieldName());

				JoinProject join = new JoinProject("", op != NAVIGATION);
				propagateFieldsLeft(qt, join);
				propagateFieldsRight(extent, join, op, null);
				join.setQueryFilter(new Equal(
						new BasicFieldOperand(qtf),
						new BasicFieldOperand(pnf)));
				qtf = pnf;
			} else {
				throw new MedorException("Operation type (" + op
						+ ") not supported on the field: " + pnf);
			}
		}
		logger.log(BasicLevel.DEBUG, "End of navigation");
		return qtf;
	}

	/**
	 * Propagates all fields from qt and, in the case of NAVIGATION_INTO_LAST,
	 * all fields from the extent into the result qn.
	 */
	private static void propagateFieldsLeft(QueryTree qt,
			QueryNode qn) throws MedorException {
		//project in 'join' all fields of the inherited node
		Field[] fToAdd = qt.getTupleStructure().getFields();
		for (int i = 0; i < fToAdd.length; i++) {
			logger.log(BasicLevel.DEBUG, "Propagate from left: " + fToAdd[i].getName());
			qn.addPropagatedField(
					fToAdd[i].getName(),
					fToAdd[i].getType(),
					new QueryTreeField[]{(QueryTreeField) fToAdd[i]});
		}
	}
	
	private static void propagateFieldsRight(JormExtent extent,
			QueryNode qn,
			byte op,
			String gcElemAlias) throws MedorException {
		if (op == NAVIGATION_INTO_LAST) {
			//project in 'join' all fields of the GenClassExtent
			// and find the QueryTreeField which represents the element
			Field[] fToAdd = extent.getTupleStructure().getFields();
			String pnfn = extent.getPNameFieldName();
			String gcElemName = null;
			if (gcElemAlias != null) {
				gcElemName = ((GenClassExtent) extent).getElementFieldName();
			}
			//System.out.println("Right: " + extent.getName());
			for (int i = 0; i < fToAdd.length; i++) {
				if (!fToAdd[i].getName().equals(pnfn)) {
					String fn = (fToAdd[i].getName().equals(gcElemName)
							? gcElemAlias : fToAdd[i].getName());
					logger.log(BasicLevel.DEBUG, "Propagate from right: "
							+ fToAdd[i].getName() + " to " + fn);
					qn.addPropagatedField(
							fn,
							fToAdd[i].getType(),
							new QueryTreeField[]{(QueryTreeField) fToAdd[i]});
				}
			}
		}
	}


	private static PNameField getPNameField(Field field) throws MedorException {
		if (field instanceof PNameField)
			return (PNameField) field;
		else if (field instanceof PropagatedField) {
			return getPNameField(
					((PropagatedField) field).getPreviousFields()[0]);
		} else
			return null;
	}

	private static String contains(TupleStructure ts, String fn) {
		String res = (ts.contains(fn) ? fn : null);
		int idx = fn.lastIndexOf(".", fn.length());
		while (res == null && idx != -1) {
			res = fn.substring(idx + 1, fn.length());
			res = (ts.contains(res) ? res : null);
			idx = fn.lastIndexOf(".", idx - 1);
		}
		return res;
	}

	public static void resetNameIndexes() {
		pnameFieldNameIndex = 0;
	}

	public static QueryTreeField navigate_old(NavigatorOperator nav,
											  String aliasName,
											  byte op)
			throws MedorException {
		//'qtf' is the QueryTreeField to use for a join
		QueryTreeField qtf;
		//pnf is the PNameField which must be used to access to the meta info
		PNameField pnf;

		boolean joinAlreadyConstructed = false;

		String fieldName = null;
		try {
			fieldName = ((Operand) nav.getExpression(1)).getString();
		} catch (TypingException e) {
			throw new MedorException(e);
		}
		Expression e = nav.getExpression(0);
		logger.log(BasicLevel.DEBUG, "navigate(" + aliasName + ", "
				+ ExpressionPrinter.e2str(e) + ", " + fieldName
				+ ", op:" + op + ")");
		boolean propagQtf;
		if (e instanceof FieldOperand) {
			qtf = (QueryTreeField) ((FieldOperand) e).getField();
			pnf = getPNameField(qtf);
			logger.log(BasicLevel.DEBUG, "Navigation from FieldOperand " + qtf.getName());
			propagQtf = true;
		} else if (e instanceof NavigatorOperator) {
			logger.log(BasicLevel.DEBUG, "Navigation from NavigatorOperator");
			Expression el = ((NavigatorOperator) e).getExpression(0);
			propagQtf = el instanceof FieldOperand
					&& getPNameField(((FieldOperand) el).getField()).isClassPName();
			//The alias name is null because only the first element (the last
			//join) can be have an alias
			qtf = navigate((NavigatorOperator) e, null, NAVIGATION);
			pnf = getPNameField(qtf);
			if (qtf.getQueryTree() instanceof JoinProject) {
				joinAlreadyConstructed = true;
			}
			logger.log(BasicLevel.DEBUG, "Recursive call constructed join ? " + joinAlreadyConstructed);

		} else
			throw new MedorException(
					"Unknown left operand of the specified NavigatorOperand: " + e);

		logger.log(BasicLevel.DEBUG, "propagQtf=" + propagQtf);
		logger.log(BasicLevel.DEBUG, "qtf=" + qtf.getName() + " of node " + qtf.getQueryTree());
		if (pnf == null)
			throw new MedorException("Impossible to find a PNameField");

		// 'qt' is the QueryTree of 'qtf"
		QueryTree qt = qtf.getQueryTree();

		if (!pnf.isClassPName()) {
			logger.log(BasicLevel.DEBUG, "Reference from a Class to a Class");
			Reference ref = pnf.getReference();

			Class jormClass = null;
			if (ref instanceof ClassRef)
				jormClass = ((ClassRef) ref).getMOClass();
			else
				throw new MedorException("Impossible to navigate with PName:"
						+ ref.getName() + " " + ref.getParent());

			ClassExtent extent = null;
			String nodeName;
			if (aliasName != null) {
				//this is the end of the navigation
				nodeName = (op != NAVIGATION_INTO_LAST
						? aliasName
						: jormClass.getFQName());
				pnameFieldNameIndex += 1;
				//create the extent corresponding to the reference
				extent = new ClassExtent(
						jormClass,
						nodeName,
						PNAME_FIELD_NAME + pnameFieldNameIndex,
						false);
			} else {
				//in the middle of the navigation: propagate only the required
				//field
				nodeName = jormClass.getFQName();
				pnameFieldNameIndex += 1;
				//create the extent corresponding to the reference
				extent = new ClassExtent(
						jormClass,
						nodeName,
						new String[]{fieldName},
						true,
						PNAME_FIELD_NAME + pnameFieldNameIndex);
			}
			// pnf references the PNameField of the joined extent
			pnf = (PNameField) extent.getField(extent.getPNameFieldName());

			JoinProject join;
			//if (!joinAlreadyConstructed) {
			join = new JoinProject("", op != NAVIGATION);
			//} else {
			//	join = (JoinProject) qtf.getQueryTree();
			//	propagQtf = false;
			//}

			//add the join condition: pnf should be equal to the
			//PName of the extent
			join.setQueryFilter(
					new Equal(
							new BasicFieldOperand(qtf),
							new BasicFieldOperand(pnf)
					)
			);

			//project in qt all fields of the QueryTree containing
			//the origin of the path expression
			Field[] fToAdd = qt.getTupleStructure().getFields();
			logger.log(BasicLevel.DEBUG, "Adding fields:");

			for (int i = 0; i < fToAdd.length; i++) {
				if (propagQtf || qtf != fToAdd[i]) {
					logger.log(BasicLevel.DEBUG, "\t-" + fToAdd[i].getName());
					join.addPropagatedField(
							fToAdd[i].getName(),
							fToAdd[i].getType(),
							new QueryTreeField[]{(QueryTreeField) fToAdd[i]});
				}
				//else
				//System.out.println("\t-Do not propagate "  + fToAdd[i].getName());
			}

			// project in qt the fields corresponding to the end of the path
			// expression
			fToAdd = extent.getTupleStructure().getFields();
			for (int i = 0; i < fToAdd.length; i++) {
				logger.log(BasicLevel.DEBUG, "Adding propagated field " + fToAdd[i].getName());
				join.addPropagatedField(
						fToAdd[i].getName(),
						fToAdd[i].getType(),
						new QueryTreeField[]{(QueryTreeField) fToAdd[i]});
			}
			qtf = (QueryTreeField)
					join.getField(extent.getFieldName(nodeName, fieldName));
			//qtf references a propagated field to a field named 'fieldName'

			qt = join;
		} else {
			logger.log(BasicLevel.DEBUG, "Adding the field if not already present");
			if (pnf.isInGenClass()) {
				throw new MedorException(
						"Impossible to navigate with a multi valued reference");
			}
			logger.log(BasicLevel.DEBUG, "Checking that " + fieldName + " appears in " + qt);
			// The field 'fieldName' must be found on 'qt'
			String fn = contains(qt.getTupleStructure(),
					pnf.getQueryTree().getName() + "." + fieldName);
			if (fn != null) {
				qtf = (QueryTreeField) qt.getTupleStructure().getField(fn);
			} else if (qt instanceof ClassExtent) {
				//In case where qt is a ClassExtent
				ClassExtent extent = (ClassExtent) qt;
				TypedElement te =
						((Class) extent.getMetaObject()).getTypedElement(fieldName);
				if (te == null)
					throw new MedorException("The jorm class "
							+ ((Class) extent.getMetaObject()).getName()
							+ " does not provide a field " + fieldName);
				qtf = extent.addField(te);
			} else
				throw new MedorException(
						"No field " + fieldName + " found on the query tree " + qt);
		}
		return endOfNavigation(op, qtf, qt, aliasName);
	}


}
