/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.query.jorm.lib;

import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.Reference;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.ClassRef;
import org.objectweb.jorm.metainfo.api.Mapping;
import org.objectweb.jorm.metainfo.api.ClassProject;
import org.objectweb.jorm.metainfo.api.ReferenceMapping;
import org.objectweb.jorm.lib.JormPathHelper;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.jorm.api.JormExtent;
import org.objectweb.medor.api.MedorException;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.Iterator;
import java.util.Map;

/**
 * This class represents Fields corresponding to the PName of a class or a
 * Genclass(the NameDef of the class itself) or the the PName associated to a
 * reference field or to a genclass element.
 * The name of such a Field is a constant.
 *
 * Inherited field te:
 * ===================
 * In case where this PNameField represents the a reference field, this
 * field is not null and references the Jorm Meta object Reference.
 *
 * In case where this PNameField represents GenClass identifier or the
 * genclass element this field is not null and references the Jorm meta
 * object GenClassRef.
 *
 * @author Alexandre Lefebvre, Sebastien Chassande-Barrioz
 */
public class PNameField extends BasicJormField {

    /**
     * In case where this PNameField represents the PName of a Jorm class, this
     * field is not null and references the Jorm Meta object Class.
     */
    protected Class clazz;

    /**
     * If the current PNameField represents a PName of a Jorm class or genclass,
     * this field is equals to 'true'. Otherwise the current PNameField
     * represents a reference field or the genclass element then this field is
     * equals to 'false'.
     */
    protected boolean isClassPName;

    /**
     * This fiedl returns 'true' if the current PNameField has been built to
     * describe the PName identifier of a GenClass or the PName of the GenClass
     * element.
     */
    protected boolean isInGenClass;

	public PNameField() {
	}

    /**
     * This constructor must be used to built a PNameField associated to a
     * GenClass identifier or the genclass element.
     * @param fieldName is the name which must be given to the field.
     * @param gcr is the jorm meta object which describes the genclass.
     * @param isGenClassPName indicates if this PNameField will represent a
     * genclass identifier(value='true'). Otherwise the this PNameField will
     * represent a genclass element(value='false').
     * @param qt is the QueryTree which is associated to this QueryTree field.
     */
    public PNameField(String fieldName,
                      GenClassRef gcr,
                      boolean isGenClassPName,
                      QueryTree qt) throws MedorException {
		super(fieldName, qt, gcr);
        isInGenClass = true;
        isClassPName = isGenClassPName;
    }

    /**
     * This constructor must be used to built a PNameField associated to a
     * classRef.
     * @param cr is the jorm meta object ClassRef associated to this PNameField.
     * @param qt is the QueryTree which is associated to this QueryTree field.
     */
    public PNameField(String fieldName, ClassRef cr, QueryTree qt) throws MedorException {
		super(fieldName, qt, cr);
        isInGenClass = cr.getParent() instanceof GenClassRef;
        isClassPName = false;
    }

    /**
     * This constructor must be used to built a PNameField associated to a Class
     * identifier or a reference field.
     * @param fn is the name which must be given to the field.
     * @param cl is the jorm meta object Class associated to this PNameField. In
     * case of the class contains a field which the name is equals to @fn then
     * this PNameField represents a reference field.
     * @param qt is the QueryTree which is associated to this QueryTree field.
     */
    public PNameField(String fn, Class cl, QueryTree qt) throws MedorException {
		super(fn, qt);
        isInGenClass = false;
        TypedElement te = getTypedElement(cl, fn);
        if (te != null && te instanceof Reference) {
            type = te.getType();
            isClassPName = false;
            this.te = te;
        }
        else {
            clazz = cl;
            isClassPName = true;
			type = clazz.getPType();
        }
    }

	public Object clone(Object clone,
						Map obj2clone) throws CloneNotSupportedException {
		clone = super.clone(clone, obj2clone);
		((PNameField) clone).clazz = clazz;
		((PNameField) clone).isClassPName = isClassPName;
		((PNameField) clone).isInGenClass = isInGenClass;
		return clone;
	}

    /**
     * It returns 'true' when this PNameField represents the identifier of a jorm
     * class or a jorm genclass.
     */
    public boolean isClassPName() {
        return isClassPName;
    }

    /**
     * It returns the jorm meta object Reference in case of this PNameField
     * represents a reference field in a jorm class.
     */
    public Reference getReference() {
        return (Reference) te;
    }

    /**
     * It returns the jorm meta object Class in case of this PNameField
     * represents an identifier of a jorm class.
     */
    public Class getMetaObjectClass() {
        return clazz;
    }

    /**
     * It returns the jorm meta object GenClassRef in case of this PNameField
     * represents a GenClass identifier or a GenClass element.
     */
    public GenClassRef getGenClassRef() {
        return (GenClassRef)te;
    }

    /**
     * It returns 'true' if the current PNameField has been built to
     * describe the PName identifier of a GenClass or the PName of the GenClass
     * element.
     */
    public boolean isInGenClass() {
        return isInGenClass;
    }

    private TypedElement getTypedElement(Class clazz, String fn) {
        int idx = fn.length();
        TypedElement te = null;
        te = clazz.getTypedElement(fn);
        while ((idx = fn.lastIndexOf(".", idx)) != -1 && te == null) {
            te = clazz.getTypedElement(fn.substring(idx + 1, fn.length()));
            idx--;
        }
        return te;
    }

    /**
     * Retrieves the NameDef corresponding to this PNameField
     * @param extent attached to this PNameField
     * @throws MedorException if it is not possible to fetch the Mapping for
     * the project name and the mapper attached to the JormExtent.
     */
    public NameDef getNameDef(JormExtent extent) throws MedorException {
        if (isClassPName) {
            if (isInGenClass) {
                //Pname of the gen class identifier
                return getRefNameDef((GenClassRef) extent.getMetaObject(), extent);
            } else {
                //PName of the class identifier
                return getClassNameDef(extent);
            }
        } else {
            if (isInGenClass) {
                return getElemNameDef((Reference) te, extent);
            } else {
                //PName of the class reference
                return getRefNameDef((Reference) te, extent);
            }
        }
    }

    /**
     * Retrieves the name of the ParameterOperand containing the PNamingContext.
     */
    public String getPNamingContextParameter() {
        if (isClassPName) {
            if (isInGenClass) {
                //Pname of the gen class identifier
                return JormPathHelper.getPath((GenClassRef) te, false);
            } else {
                //PName of the class identifier
                return JormPathHelper.getPath(clazz);
            }
        } else {
			if (te instanceof GenClassRef) {
				//reference to a GenClass from a Class, or a GenClass
				return JormPathHelper.getPath((GenClassRef) te,
						!isInGenClass);
			}
			else
				//reference to a Class from a Class, or a GenClass
				return JormPathHelper.getPath((ClassRef) te);
         }
    }

    /**
     * Retrieves the Jorm meta object of the Class represented by this
     * PNameField. If the parameter is
     * a Class then the same value is returned. If the parameter is a ClassRef
     * or a GenClassRef then referenced Class is returned. Otherwise a null
     * value is returned.
     */
    public MetaObject getTargetedMetaObject() {
        if (isInGenClass) {
            if (isClassPName) {
                return te;
            } else {
                GenClassRef gcr = (GenClassRef) te;
                if (gcr.isClassRef()) {
                    return gcr.getClassRef().getMOClass();
                } else if (gcr.isGenClassRef()) {
                    return gcr.getGenClassRef();
                } else {
                    return null;
                }
            }
        } else {
            if (isClassPName) {
                return clazz;
            } else {
                return te;
            }
        }
    }
    /**
     * Retrieves the NameDef representing the identifier of a GenClass/
     * @param gcr is the jorm meta object representing the GenClass
     * @param extent is the GenClassExtent corresponding to the 'gcr'
     * @throws MedorException if it is not possible to fetch the Mapping for
     * the project name and the mapper attached to the JormExtent.
     */
    private NameDef getRefNameDef(GenClassRef gcr,
                                 JormExtent extent) throws MedorException {
        return (NameDef) getMapping(extent)
                .getGenClassMapping(gcr.getGenClassId())
                .getIdentifierMapping().getLinkedMO();
    }


    /**
     * Retrieves the NameDef representing the element of a GenClass/
     * @param r is the jorm meta object representing the reference element of
     * the GenClass
     * @param extent is the GenClassExtent corresponding to the 'gcr'
     * @throws MedorException if it is not possible to fetch the Mapping for
     * the project name and the mapper attached to the JormExtent.
     */
    private NameDef getElemNameDef(Reference r,
                                 JormExtent extent) throws MedorException {
        return (NameDef) getMapping(extent)
                .getGenClassMapping(((GenClassRef) r.getParent()).getGenClassId())
                .getReferenceMapping().getLinkedMO();
    }

    /**
     * Retrieves the NameDef representing the reference to a Class. This
     * reference is hosted by a Class (not a GenClass see getElemNameDef)
     * @param r is the jorm meta object representing the class reference
     * @param extent is the ClassExtent corresponding to the 'cr'
     * @throws MedorException if it is not possible to fetch the Mapping for
     * the project name and the mapper attached to the JormExtent.
     */
    private NameDef getRefNameDef(Reference r,
                                 JormExtent extent) throws MedorException {
		if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
			logger.log(BasicLevel.DEBUG, "Extent class mapping " + getMapping(extent).getClassMapping());
			Iterator it = getMapping(extent).getClassMapping().getReferenceMappings().iterator();
			while (it.hasNext()) {
				logger.log(BasicLevel.DEBUG, "Reference rule name " + ((ReferenceMapping)it.next()).getRuleName());
			}
			logger.log(BasicLevel.DEBUG, "+ reference mapping " + getMapping(extent).getClassMapping().getReferenceMapping(r.getName()));
			logger.log(BasicLevel.DEBUG, "+ rule name " + getMapping(extent).getClassMapping().getReferenceMapping(r.getName()).getRuleName());
			logger.log(BasicLevel.DEBUG, "+ linkedMO " + getMapping(extent).getClassMapping().getReferenceMapping(r.getName()).getLinkedMO());
		}
		return (NameDef) getMapping(extent).getClassMapping()
                .getReferenceMapping(r.getName()).getLinkedMO();
    }

    /**
     * Retrieves the Jorm meta object representing a Class. If the parameter is
     * a Class then the same value is returned. If the parameter is a ClassRef
     * or a GenClassRef then the hosting Class is returned. Otherwise a null
     * value is returned.
     */
    private Class getClass(MetaObject mo) {
        if (mo instanceof Class) {
            return (Class) mo;
        } else if (mo instanceof Package || mo instanceof Manager) {
            return null;
        } else {
            return getClass(mo.getParent());
        }
    }

    /**
     * It retrieves the Mapping of a JormExtent.
     * @param extent represents a persitent object which the Mapping is expected
     * @throws MedorException if no mapping exists for the project name and the
     * mapper specified by the extent.
     */
    private Mapping getMapping(JormExtent extent) throws MedorException {
        Class c = getClass(extent.getMetaObject());
        ClassProject cp = c.getClassProject(extent.getProjectName());
        if (cp == null) {
            throw new MedorException("No classproject found for the class "
                    + c.getFQName() + " and  the project "
                    + extent.getProjectName());
        }
        String mapperName = extent.getPMapper().getMapperName();
        int idx =  mapperName.indexOf('.');
        Mapping m = cp.getMapping(idx == -1
                ? mapperName
                : mapperName.substring(0, idx));
        if (m == null) {
            throw new MedorException("No mapping found for the class "
                    + c.getFQName() + ",  the project " + extent.getProjectName()
                    + " and  the mapper " + mapperName);
        }
        return m;
    }

    /**
     * Retrieves the NameDef representing the identifier of a Class.
     * @param extent is the ClassExtent representing the class
     * @throws MedorException if it is not possible to fetch the Mapping for
     * the project name and the mapper attached to the JormExtent.
     */
    private NameDef getClassNameDef(JormExtent extent) throws MedorException {
        return (NameDef) getMapping(extent).getClassMapping()
                .getIdentifierMapping().getLinkedMO();
    }

}
