/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2005 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */
package org.objectweb.medor.query.jorm.lib;

import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ClassRef;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.Reference;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.lib.And;
import org.objectweb.medor.expression.lib.Equal;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.lib.Log;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.lib.JoinProject;
import org.objectweb.medor.query.jorm.api.JormExtent;
import org.objectweb.util.monolog.api.Logger;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;

/**
 * QueryBuilder
 *
 * @author Rafael H. Schloming &lt;rhs@mit.edu&gt;
 **/

public class QueryBuilder {

    public static Logger log = Log.getLoggerFactory()
        .getLogger("org.objectweb.medor.query.jorm.lib.QueryBuilder");

    private static final String GENCLASS_ELEMENT_NAME = "genClassElement";

    private static int queryBuilderCounter = 0;
    
    public static synchronized Integer allocateQueryBuilderId() {
        return new Integer(queryBuilderCounter++);
    }
    
    // our parent query if any
    private QueryBuilder m_parent = null;

    // the query we're building
    private JoinProject m_join = new JoinProject();

    // unique id for PName fields
    private int m_id = 0;

    // map from paths we've navigated to the fields resulting from
    // that navigation
    private Map m_fields = new HashMap();

    // track the fields we've projected
    private Map m_projected = new HashMap();

    /*
     * QueryBuilder identifier.
     */
    private Integer queryBuilderId;
    
    /**
     * track the joins made to navigate a reference.
     * key=is a QueryTreeField parent
     * value= if parent represents a GenClassRef then the content is a a QTF or
     * a java.util.List of QTF. If the parent represents a ClassRef
     * then the content is the unique QTF already join to the parent. 
     * To resume this structure holds the unique join with a ClassRef and the 
     * list of joins with a GenClassRef.  
     */
    private Map m_joins = new HashMap();

    /**
     * Construct a new top level query builder instance.
     */

    public QueryBuilder() {
        this(null);
    }

    /**
     * Constructs a new query builder for creating subqueries of
     * <code>parent</code>.
     *
     * @param parent the containing query
     */

    public QueryBuilder(QueryBuilder parent) {
        m_parent = parent;
        queryBuilderId = allocateQueryBuilderId();
    }

    /**
     * Returns the query tree built by calls to this QueryBuilder
     * instance.
     */

    public QueryTree getQueryTree() {
        return m_join;
    }

    /**
     * Returns true if the query tree built by this query navigates to
     * the given path.
     *
     * @param path a dot separated path
     */

    public boolean contains(String path) {
        return m_fields.containsKey(path);
    }

    /**
     * Assigns a name to a field value in the namespace of the query
     * being built.
     *
     * @param name a path within the namespace of the query
     * @param field the value for <code>name</code>
     */

    public void define(String name, QueryTreeField field)
        throws MedorException {
        if (m_fields.containsKey(name)) {
            throw new IllegalArgumentException
                ("duplicate definition: " + name +
                 ", old value: " + m_fields.get(name) +
                 ", new value: " + field);
        }
        m_fields.put(name, field);
    }

    /**
     * Adds the specified field to the set of projected fields in the
     * query being built.
     *
     * @param qtf the field to project
     * @return the propogated field used to project <code>qtf</code>
     */
    public QueryTreeField project(QueryTreeField qtf) throws MedorException {
        return project(qtf.getName(), qtf);
    }

    public QueryTreeField project(String alias, QueryTreeField qtf) throws MedorException {
        
        QueryTreeField result = (QueryTreeField) m_projected.get(qtf);
        if (result == null) {
            result = m_join.addPropagatedField
                (alias, qtf.getType(), new QueryTreeField[] { qtf });
            m_projected.put(qtf, result);
        }
        //sets the QueryTree group of the child query tree to be this QueryBuilder
        qtf.getQueryTree().getAnnotations().put(JormExtent.ANNOTATION_QUERYTREE_GROUP, queryBuilderId);
        return result;
    }

    /**
     * converts from an array representation of a path to a dot
     * separated string representation
     */

    private String name(String[] path, int length) {
        StringBuffer buf = new StringBuffer();
        String sep = "";
        for (int i = 0; i < length; i++) {
            buf.append(sep);
            buf.append(path[i]);
            sep = ".";
        }
        return buf.toString();
    }

    private int id() {
        if (m_parent == null) {
            return m_id++;
        } else {
            return m_parent.id();
        }
    }

    /**
     * Finds the underlying PNameField of a propogated field if it
     * exists.
     */

    private PNameField getPNameField(Field field) {
        if (field instanceof PNameField) {
            return (PNameField) field;
        } else if (field instanceof PropagatedField) {
            return getPNameField
                (((PropagatedField) field).getPreviousFields()[0]);
        } else {
            return null;
        }
    }

    /**
     * This method returns the PNameField that should be used for
     * further navigation starting from the set of instances
     * identified by <code>parent</code>. The method may return
     * <code>parent</code> itself if the field is already suitable for
     * navigation, or it may join another extent into the query if no
     * extent has already been joined for navigating from the
     * <code>parent</code> field.
     *
     * @param parent the starting point for navigation
     * @param name the node name to use when joining
     * @return a pname field ready for navigation
     */

    private PNameField join(QueryTreeField parent, String nodeName)
        throws MedorException {
        PNameField ppnf = getPNameField(parent);
        if (ppnf.isClassPName()) {
            return ppnf;
        }

        Reference ref = ppnf.getReference();
        
        PNameField pnf = null;
        if (!(ref instanceof GenClassRef)) {
            pnf = (PNameField) m_joins.get(parent);
            if (pnf != null) { 
                return pnf; 
            }
        }

        JormExtent ext;
        String name = nodeName == null ? parent.getName() : nodeName;
        String pnfname = "object" + id();
        if (ref instanceof ClassRef) {
            ClassRef cref = (ClassRef) ref;
            ext = new ClassExtent(cref.getMOClass(), name, pnfname, true);
        } else if (ref instanceof GenClassRef) {
            GenClassRef gcref = (GenClassRef) ref;
            ext = new GenClassExtent
                (gcref, name, pnfname, GENCLASS_ELEMENT_NAME);
        } else {
            throw new IllegalStateException
                ("unknown ref type: " + ref.getClass());
        }
        //sets the QueryTree group of the child query tree to be this QueryBuilder
        ext.getAnnotations().put(JormExtent.ANNOTATION_QUERYTREE_GROUP, queryBuilderId);

        pnf = (PNameField) ext.getTupleStructure().getField
            (ext.getPNameFieldName());

        Expression cond = new Equal
            (new BasicFieldOperand(parent), new BasicFieldOperand(pnf));
        Expression filter = m_join.getQueryFilter();
        if (filter == null) {
            filter = cond;
        } else {
            filter = new And(filter, cond);
        }
        m_join.setQueryFilter(filter);
        if (ref instanceof GenClassRef) {
            Object o = m_joins.get(parent);
            if (o == null) {
                m_joins.put(parent, pnf);
            } else {
                if (!(o instanceof List)) {
                    Object other = o;
                    o = new ArrayList();
                    ((List) o).add(other);
                    m_joins.put(parent, o);
                }
                ((List) o).add(pnf);
            }
        } else {
            m_joins.put(parent, pnf);
        }

        return pnf;
    }

    /**
     * Performs a single step in the navigation of a path. This may
     * involve simple accessing an existing query tree field or adding
     * a new field to the extent specified by the {@link
     * #join(QueryTreeField, String)} method.
     *
     * @param parent the starting point for navigation
     * @param field the field to navigate
     * @return a query tree field referring to the result of the
     *         navigation operation
     */

    public QueryTreeField navigate(QueryTreeField parent, String field)
    	throws MedorException {
        return navigate(parent, field, null);
    }
    public QueryTreeField navigate(QueryTreeField parent, String field, String nodeName)
        throws MedorException {
        PNameField pnf = join(parent, nodeName);

        QueryTreeField qtf;
        QueryTree qt = pnf.getQueryTree();
        TupleStructure ts = qt.getTupleStructure();
        String fq = qt.getName() + "." + field;
        if (ts.contains(fq)) {
            qtf = (QueryTreeField) ts.getField(fq);
        } else if (qt instanceof ClassExtent) {
            ClassExtent extent = (ClassExtent) qt;
            Class klass = (Class) extent.getMetaObject();
            TypedElement te = klass.getTypedElement(field);
            if (te == null) {
                StringBuffer buf = new StringBuffer();
                buf.append("No such field '").append(field);
                buf.append("' in '").append(klass.getFQName());
                buf.append("' from ").append(parent);
                for (Iterator it = klass.getAllFields().iterator();
                     it.hasNext(); ) {
                    TypedElement el = (TypedElement) it.next();
                    buf.append(", ").append(el.getName());
                }
                throw new IllegalArgumentException(buf.toString());
            } else {
                qtf = extent.addField(te);
            }
        } else {
            throw new IllegalStateException
                ("wrong extent type: " + qt.getClass() +
                 "\nfor field: " + field);
        }

        return qtf;
    }

    private boolean isGenClassReference(QueryTreeField qtf) {
        PNameField pnf = getPNameField(qtf);
        if (pnf == null) { return false; }
        if (!pnf.isClassPName()) {
            return pnf.getReference() instanceof GenClassRef;
        }
        return false;
    }

    /**
     * Extends the query being built by this QueryBuilder instance
     * along the path specified by <code>path</code> for
     * <code>length</code> segments of that path.
     *
     * @param path the path to navigate
     * @param length the number of segments of the path to navigate
     * @return the query tree field referring to the result of the
     *         navigation operation
     */

    private QueryTreeField navigate(String[] path, int length, String alias)
        throws MedorException {
        String name = name(path, length);
        QueryTreeField qtf = (QueryTreeField) m_fields.get(alias == null ? name : alias);
        // only join once for any field
        if (qtf != null) { return qtf; }

        if (length == 0) {
            throw new IllegalArgumentException
                ("no such path: " + name(path, path.length));
        }

        QueryTreeField parent = navigate(path, length - 1, null);
        qtf = navigate(parent, path[length - 1], alias);
        if (isGenClassReference(qtf)) {
            qtf = navigate(qtf, GENCLASS_ELEMENT_NAME);
        }
        define(alias == null ? name : alias, qtf);

        return qtf;
    }

    /**
     * Extends the query being built by this QueryBuilder instance
     * along the path specified by <code>path</code>.
     *
     * @param path the path to navigate
     * @return the query tree field referring to the result of the
     *         navigation operation
     */
    public QueryTreeField navigate(String[] path) throws MedorException {
        return navigate(path, null);
    }
    public QueryTreeField navigate(String[] path, String alias) throws MedorException {
        return navigate(path, path.length, alias);
    }

    private static final String DOT = "\\.";

    /**
     * Extends the query being built by this QueryBuilder instance
     * along the dot separated path specified by <code>path</code>.
     *
     * @param path the path to navigate
     * @return the query tree field referring to the result of the
     *         navigation operation
     */
    public QueryTreeField navigate(String path) throws MedorException {
        return navigate(path, path);
    }
    public QueryTreeField navigate(String path, String alias) throws MedorException {
        String[] parts = path.split(DOT);
        return navigate(parts, alias);
    }

}
